/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-12H        *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5800             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.agents;

import rca.content.lang.Codec;
import rca.content.lang.sl.SLCodec;
import rca.content.onto.OntologyException;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.core.AID;
import rca.domain.DFService;
import rca.domain.FIPAAgentManagement.DFAgentDescription;
import rca.domain.FIPAAgentManagement.ServiceDescription;
import rca.domain.FIPAException;
import rca.domain.FIPANames;
import rca.lang.acl.ACLMessage;
import rca.util.leap.ArrayList;
import rca.util.leap.List;
import org.jlab.coda.runcontrol.ontologies.CodaPlatform.*;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.*;
import org.jlab.coda.runcontrol.util.coolparser.CoolParser;
import org.jlab.coda.runcontrol.util.db.*;
import org.jlab.coda.runcontrol.util.rcConstants;
import org.jlab.coda.runcontrol.util.rcClientAddress;
import org.jlab.coda.runcontrol.util.rcSubscription;
import org.jlab.coda.cMsg.*;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.concurrent.ConcurrentHashMap;
import java.io.*;


/**
 * Class is responsible parsing the CoolOntology configuration files,
 * and organizing AFECS agent creation and specialization.
 *
 * @author Vardan Gyurjyan
 * Date: Jan 6, 2006
 * Time: 11:32:48 AM

 */
public class Cortex extends CAgent {

    /** Controls agent differentiation process*/
    private volatile boolean STOPDIFF;

    /** The name of the last created agent on the platform*/
    private String LAST_CREATED;

    /** Platform cMsg server connection object. Messaging server runs within Hepatic normative agent*/
    cMsg cMsgSrv;

    /** Verbosity level defined in the rcConstants class. */
    private int verbose = rcConstants.debugInfo;

    /** Local host name */
    private static String LOCALHOST;

    /** Defines FIPA content language, used for inter-agent communications*/
    private Codec  codec = new SLCodec();


    /** Runtype of the current request for configuring the run.
     * This is the actual configuration file name coming from the rcGui.*/
//    private String UI_runtype_req;

    /** Session request from the rcGUI.*/
//    private String UI_session_req;

    /** CODA database jdbc url (for coda2 components).*/
    private String _URL;

    /** CODA database jdbc driver (for coda2 components).*/
    private String _DRIVER;

    /** CODA database user (for coda2 components).*/
    private String _USER;

    /** CODA database password (for coda2 components).*/
    private String _PASSWORD;

    /** Hepatic AID. N.B. Hard coded name for Hepatic.
     * Make sure Hepatic is started with this name during fornt-end startup. */
    private AID platformAdmin = new AID("Hepatic",AID.ISLOCALNAME);

    /** Map to keep correspondence between agent name and agent container name(both Strings) */
    private ConcurrentHashMap<String, String> CompContainer = new ConcurrentHashMap<String, String>();

    /** Map containing the configuration name (runtype) and CoolParser object.
     COOL parser object, used to hold cool agent model, described in a cool configuration file.*/
    private ConcurrentHashMap<String, CoolParser>ParsedCoolObjects = new ConcurrentHashMap<String, CoolParser>();

    /** Map containing the name of the rcClient physical component and its address (host and port)*/
    private ConcurrentHashMap<String, rcClientAddress>RcPhysicalClients = new ConcurrentHashMap<String, rcClientAddress>();

    /** Indicates the start of this agent. Prevents removing non existent runtype( supervisor) agents*/
    private boolean fresh_start;

    /** Container holds the names of the nodes where afecs containers are running.
     * N.B. containers have the same name as the hosting node. */
    private List Containers = new ArrayList();

    /**
     *  Setup method of this agent. Runs at the startup of the agent.
     */
    protected void setup(){
        String l = null;
        int index = 0;

        fresh_start = true;


// Get the name of the host where this agent is running. This is the same as main container host name
        try {
            l = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            e.printStackTrace();
        }
        try{
            index = l.indexOf(".");
        }catch(java.lang.NullPointerException e){e.printStackTrace();}
        if(index<=0)LOCALHOST = l;
        else LOCALHOST = l.substring(0,index);


// Create sessions directory
        boolean success = (new File(System.getenv("COOL_HOME")+"/sessions")).mkdir();
        if (success) {
            System.out.println("Warning: created sessions dir.");
        }

// Open platform topology file
        String cs;
        try {
            BufferedReader fin = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/containers"));
            while((cs=fin.readLine())!=null){
                Containers.add(cs);
            }
            fin.close();
        } catch (IOException e) {
//            e.printStackTrace();
        }
        catch (NumberFormatException ee) {
            ee.printStackTrace();
        }

// Registration with the DF
        DFAgentDescription dfd = new DFAgentDescription();
        ServiceDescription sd = new ServiceDescription();
        sd.setType("Cortex");

// Name is the given name when this agent was started from the command line. N.B. by design startup name is Cortex.
        sd.setName(getLocalName());
        sd.setOwnership("JLABCODA");
        dfd.setName(getAID());
        dfd.addServices(sd);
        try {
            DFService.register(this,dfd);
        } catch (FIPAException e) {
            String s = "Cortex-Error:\n registration with DF unsucceeded. Reason: "+e.getMessage();
            System.err.println(s);

// Registration failed delete the agent. This method is define in the Agent class.
            doDelete();
        }

//  Register the message content language. FIP_SL0 is used for inter-agent communications.
        getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);

// Register supported ontologies
        getContentManager().registerOntology(CodaRunControlOntology.getInstance());
        getContentManager().registerOntology(CodaPlatformAdminOntology.getInstance());

// Start the responder behaviour to respond the messages addressed to this agent.
        addBehaviour(new GenRespondBehaviour(this,verbose));

// Connect to cMsg server for getting messages from components outside of the afecs platform (e.g. RcGui).
        try {
            cMsgSrv = new cMsg(rcSubscription.PlatformcMsgServerUDL, "Cortex", rcSubscription.PlatformcMsgServerDescription);
            cMsgSrv.connect();

// Enable message reception
            cMsgSrv.start();

// Subscribe all the messages (type = *) addressed to its name ( subject = "Cortex");
            cMsgCallbackInterface cb = new GuiCallback();
            //cMsgSrv.subscribe(rcSubscription.CrtSubject, rcSubscription.CrtType, cb, null);
            cMsgSrv.subscribe(rcSubscription.CrtSubject, "*", cb, null);
        } catch (cMsgException e) {
            e.printStackTrace();
        }

// Sleep for a second. I do not remeber the reason...
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

// Open $COOL_HOME/.physcomponents file where previously registered clients name were saved and
// populate the RcPhysicalCilents hash table.
        rcClientAddress tmpad;
        String tmps, tmpcomp=null;
        StringTokenizer st;
        try {
            BufferedReader fin = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/.physcomponents"));
            while((tmps=fin.readLine())!=null){
                st = new StringTokenizer(tmps);
                if(st.hasMoreTokens())tmpcomp = st.nextToken();
                if(tmpcomp!=null){
                    tmpad = new rcClientAddress();
                    if(st.hasMoreTokens()) tmpad.setHost(st.nextToken());
                    if(st.hasMoreTokens()) tmpad.setPort(Integer.parseInt(st.nextToken()));
                    RcPhysicalClients.put(tmpcomp,tmpad);
                }
            }
            fin.close();
        } catch (FileNotFoundException e) {
            System.out.println("Cortes-Warning: Empty component registration database");
//            e.printStackTrace();
        }
        catch (NumberFormatException ee) {
            ee.printStackTrace();
        }
                 
        catch(IOException eee){
         eee.printStackTrace();
        }
// Start RcBroadcastServer to listen udp broadcast messages from the physical clients,
// willing to have a representative agent on the platform.
        cMsg cmsg;
        try {
            cmsg = new cMsg(rcSubscription.RcBServerUDL, rcSubscription.RcBServerName, rcSubscription.RcBServerDescription);
            cmsg.connect();

// Enable message reception
            cmsg.start();
            BroadcastCallback cb = new BroadcastCallback();
            cmsg.subscribe(rcSubscription.CrtRcBSubject, rcSubscription.CrtRcBType, cb, null);
        } catch (cMsgException e) {
            e.printStackTrace();
            System.exit(1);
        }

// Tell hepatic to start remote containers
        // if(!Containers.isEmpty()) InitiateCreateContainerConversation("some name");
    }


    /**
     * Ping the DF to see if an agent is NOT registered with the platform registration services.
     *
     * @param name                  agent name
     * @param timeout               sleep after checking the df
     * @param trials                trials
     * @return status               status of the search
     */
    private boolean searchDFabsence(String name, int timeout, int trials){
        int count = 0;
        try {
            DFAgentDescription dfd = new DFAgentDescription();
            ServiceDescription sd  = new ServiceDescription();
            sd.setName(name);
            dfd.addServices(sd);
            DFAgentDescription[] result = DFService.search(this, dfd);

            while((result.length > 0)){

                try {
                    Thread.sleep(timeout);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                count++;
                result = DFService.search(this, dfd);
                if(count>trials) return false;
            }

        } catch (FIPAException e) {
            e.printStackTrace();
        }
        return true;
    }


    /**
     * Ping the DF to see if an agent is registered with the platform registration services.
     *
     * @param name                  agent name
     * @param timeout               sleep after checking the df
     * @param trials                trials
     * @return status               status of the search
     */
    private boolean searchDFregister(String name, int timeout, int trials){
        int count = 0;
        try {
            DFAgentDescription dfd = new DFAgentDescription();
            ServiceDescription sd  = new ServiceDescription();
            sd.setName(name);
            dfd.addServices(sd);
            DFAgentDescription[] result = DFService.search(this, dfd);

            while((result.length <= 0)){

                try {
                    Thread.sleep(timeout);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                count++;
                result = DFService.search(this, dfd);
                if(count>trials) return false;
            }

        } catch (FIPAException e) {
            e.printStackTrace();
        }
        return true;
    }


    /**
     * This method sends type = RcGuiTypeTransition information to all
     * plexuses (external components, like rcGUI, subscribing to this type) through cMsg ipc
     *
     * @param info                   String message content sent to the plexuses
     * @param severity               severity of the message, i.e. info, warning, error
     * @param guicontrol             if it is true then info message will start with GUI:
     *                               indicating the GUI state change
     * @param session                session. necessary to address session specific GUIs.
     */
    private void informPlexuses(String info, String severity, boolean guicontrol, String session) {

        String subject;
        String type = rcSubscription.RcGuiTypeTransition;

        if(severity.equalsIgnoreCase(rcConstants.info)) {subject = rcSubscription.RcGuiSubjInfo+"_"+session;

// N.B. By design if guicontrol="true" the info must be one word, e.g. configured, and will have at the begining
// "GUI:" attached indicating that this is a state transition and GUI must be modified accordingly.
            if(guicontrol)info="GUI:"+info;
        }
        else if(severity.equalsIgnoreCase(rcConstants.warning)) subject = rcSubscription.RcGuiSubjWarning+"_"+session;
        else if(severity.equalsIgnoreCase(rcConstants.error)) subject = rcSubscription.RcGuiSubjError+"_"+session;
        else subject = rcSubscription.RcGuiSubjInfo+"_"+session;

// Create a message
        cMsgMessage msg = new cMsgMessage();
        msg.setSubject(subject);
        msg.setType(type);
        msg.setText(info);
        try {
            cMsgSrv.send(msg);
            cMsgSrv.flush(1);
        } catch (cMsgException e) {
            e.printStackTrace();
        }
    }

    /**
     * Overloaded method which will send an arbitrary message to the arbitrary subscriber outside of the platform.
     *
     * @param subject                 Subject of the message
     * @param type                    Type of the message
     * @param text                    Content of the message
     */
    private void informPlexuses(String subject, String type, String text){
        cMsgMessage msg = new cMsgMessage();
        msg.setSubject(subject);
        msg.setType(type);
        msg.setText(text);
        try {
            cMsgSrv.send(msg);
            cMsgSrv.flush(1);
        } catch (cMsgException e) {
            e.printStackTrace();
        }
    }

    /**
     * Method to parse COOL configuration files
     *
     * @param name                      The name of the cool configuration file
     * @param session                   The requested session where the requested configuration should run.
     * @return myParser                 CoolParser object
     */
    public CoolParser parseCoolConfig(String session, String name) {
        boolean streamline = true;

// Construct the cool file path
        String path = System.getenv("COOL_HOME")+"/"+name+".rdf";

// CoolParser object will contain configuration details (object model) described in the cool file.
        CoolParser myParser = new CoolParser(session,path,streamline);
        ParsedCoolObjects.put(name,myParser);

// Get the evironmental variables for coda DB ( coda2 components only).
        _URL = System.getenv("CODA_DB_URL");
        _DRIVER = System.getenv("CODA_DB_DRIVER");
        _USER = System.getenv("CODA_DB_USER");
        _PASSWORD = System.getenv("CODA_DB_PASSWORD");
        if(_USER==null)_USER=" ";
        if(_PASSWORD==null)_PASSWORD=" ";

        return myParser;
    }


    /**
     * Based on the parssed information (agent model of the configuration) Cortex sends
     * the message to the platform administrative normative agent (Hepatic) to start
     * requested agents on the platform.
     *
     * @param pcn      cool configuration name, the same as runtype
     * @return true   if succeeded and false otherwise
     */
    public boolean COOLcreateAgents(String session, String pcn) {

        String tmp;
        String contName;
        int c=0;

        CoolParser myParser = ParsedCoolObjects.get(pcn);
        try{
// Check the database in case we have coda2 components in the configuration
            if(myParser!=null && myParser.CodaOption!=null &&
                    myParser.CodaOption.getOption_Hascoda2component()!=null &&
                    myParser.CodaOption.getOption_Hascoda2component().equalsIgnoreCase("true")){

                openDB(session);

// Check if session exists in the sessions table.
                try {
                    if(session!=null && !DatabaseCommunication.checkSessionsTable(session)) {
                        informPlexuses("Cortex-error:\n  " +
                                "Requested session does not exist in the database. \n" +
                                "Configured coda2 components will not function properly.","error",false, session);
                        return false;
                    }
                    else if(pcn!=null && !DatabaseCommunication.checkRunTypesTable(pcn)) {
                        informPlexuses("Cortex-error:\n  " +
                                "Requested runtype does not exist in the database. \n" +
                                "Configured coda2 components will not function properly.","error",false, session);
                        return false;
                    }
                    else {
                    }
                } catch (RCException e) {
                    e.printStackTrace();
                } catch (NullPointerException ee){
                    ee.printStackTrace();}
            }
        }catch(NullPointerException eee){
            eee.printStackTrace();
            System.out.println("Error in the Cool file!");
        }


// If we have components described in the cool configuration file
        if(!myParser.CONTROL_COMPONENTS.isEmpty()){
            for(Iterator it=myParser.CONTROL_COMPONENTS.iterator(); it.hasNext();){
                CoolComponent agnt = (CoolComponent)it.next();

// Get the component name
                String cmpname = agnt.getComponent_Name();

// Get the component type
                String cmptype = agnt.getComponent_Type();

// Check the type of the component
                if (cmptype != null) {

// Get the node name where the agent should run (by default agents are running on the
// front-end, i.e. localhost). N.B. this agent, i.e. Cortex always runs in the fornt-end.
                    tmp = LOCALHOST;
                    if (agnt.getComponent_Node() != null) tmp = agnt.getComponent_Node().getNode_Name();
                    if ((tmp != null) && (!(tmp.equals(LOCALHOST)))) {

// Node name is not null and the node is not the same where main container (i.e. front-end) is running
// N.B. container name is the same as node name
                        contName = tmp;
                    }

// IF the node name is not defined in the cool file or node name is defined the same as platform node, then
// the name of the container will be Main-Container ( rca convention)
                    else contName = "Main-Container";

// Start sedning ACL messages to Hepatic to populate platform with the requested agents.
// Start the Supervisor agent for this run. The type of the supervisor agent described in the cool is RCS.
                    if ((cmptype.equalsIgnoreCase(rcConstants.rcSupervisorType))) {
                        System.out.println("Cortex-Info:\n Creating "+cmpname+" RC supervisor agent on the container = " + contName);
                        InitiateCreateAgentConversation(
                                cmpname,                                     // agent name
                                "org.jlab.coda.runcontrol.agents.Medulla",   // agent class
                                contName,                                    // container name
                                tmp);                                        // host/node name

// Start supervisor deputy agents. By design every container will have deputy
// (Thalamus) agent, assisting RCS agent of the specific runtype.
                        if(!Containers.isEmpty()){
                            for(int i=0;i<Containers.size();i++){
                                String cn = (String)Containers.get(i);
                                String dpn = cmpname+"_dp_"+cn;

// Create the deputy
                                InitiateCreateAgentConversation(dpn,
                                        "org.jlab.coda.runcontrol.agents.Thalamus",   // agent class
                                        cn,                                           // container name
                                        cn);                                          // host/node name
                            }
                        }
                    }

// State machine supervisor agent creation (agent type = SMS)
                    else if (cmptype.equalsIgnoreCase(rcConstants.statemachinesupervisorType)){
                        System.out.println("Cortex-Info:\n Creating "+cmpname+" SM supervisor agent on the container = " + contName);
                        InitiateCreateAgentConversation(
                                cmpname,                                     // agent name
                                "org.jlab.coda.runcontrol.agents.SMSAgent",   // agent class
                                contName,                                    // container name
                                tmp);                                        // host/node name
                    }

// State Winner crate specific agents (agent type = WNC)
                    else if (cmptype.equalsIgnoreCase(rcConstants.wienercrateType)){
                        System.out.println("Cortex-Info:\n Creating "+cmpname+" Wiener crate agent on the container = " + contName);
                        InitiateCreateAgentConversation(
                                cmpname,
                                "org.jlab.coda.runcontrol.agents.WienerCrateAgent",
                                contName,
                                tmp);
                        if(!searchDFregister(cmpname,rcConstants.sleepAfterCreate_Cortex,10)) {
                            informPlexuses("Cortex-error:\n  " +
                                    "Can not restart "+cmpname,"error",false, session);
                            return false;
                        }
                    }

// Start the StemCell agents for every component, except of cource for runtype supervisor component (type = RCS)
                    else {
                        c++;
                        if(c%rcConstants.CORTEX_CREATE_GUI_REPORT == 0)
                            informPlexuses("Cortex-Info: Creating agents. Please wait ..."+c,"info",false, session);

// See if there is a stemcell agent for this described component on the platform,i.e. configuration contains
// physical component that requested previously an agent representation to the rcBroadcastDomain Server.
                        if (RcPhysicalClients.containsKey(cmpname)) {

// Remove agent if it its current container is not the same as the requested one.
                            if((CompContainer.containsKey(cmpname)) && (!CompContainer.get(cmpname).equalsIgnoreCase(contName))){

// Remove agent from the platform
                                InitiateKillAgentConversation(cmpname);
                                if(!searchDFabsence(cmpname,rcConstants.sleepAfterKill_Cortex,10)){
                                    informPlexuses("Cortex-error:\n  " +
                                            "Can not remove "+cmpname,"error",false, session);
                                    return false;
                                }

// Start the stemcell agent on the correct container
                                InitiateCreateAgentConversation(
                                        cmpname,
                                        "org.jlab.coda.runcontrol.agents.StemCell",
                                        contName,
                                        tmp);
                                if(!searchDFregister(cmpname,rcConstants.sleepAfterCreate_Cortex,10)) {
                                    informPlexuses("Cortex-error:\n  " +
                                            "Can not restart "+cmpname,"error",false, session);
                                    return false;
                                }
                            }

// If this is the first request to start an agent, i.e. request coming from the configuration, then...
                            else if(!CompContainer.containsKey(cmpname)){

// Start the stemcell agent on the requested container
                                InitiateCreateAgentConversation(
                                        cmpname,
                                        "org.jlab.coda.runcontrol.agents.StemCell",
                                        contName,
                                        tmp);
                                if(!searchDFregister(cmpname,rcConstants.sleepAfterCreate_Cortex,10)) {
                                    informPlexuses("Cortex-error:\n  " +
                                            "Can not restart "+cmpname,"error",false, session);
                                    return false;
                                }
                            } else {

// Now all the agents are up and running in their containers.
// Sleep a while before sending init command to the agents, which will open rc
// domain connection ( private p2p connection, using cMsg api) to the physical client.
                                try {
                                    Thread.sleep(rcConstants.sleepBeforeInit_Cortex);
                                } catch (InterruptedException e) {
                                    e.printStackTrace();
                                }
                            }

// Record the name of the component and the container name (where the agent should run) in the map.
                            CompContainer.put(cmpname,contName);

// If agent represents coda2 component then read the process table, make sure
// that the component is registered in the coda DB.
                            if((agnt.getComponent_Representscoda2client()!=null) &&
                                    (agnt.getComponent_Representscoda2client().equalsIgnoreCase("true"))){
                                try {
                                    if(!DatabaseCommunication.checkComponent(cmpname)){
                                        informPlexuses("Cortex-error:\n  " +
                                                "Can not find coda2 component "+cmpname+
                                                " in the process table of the database.","error",false, session);
                                        return false;
                                    }
                                } catch (RCException e) {
                                    e.printStackTrace();
                                }
                            }

// Send codainit action message to the representative agent which will complete
// the connect request of the physical client.
// Attention: we assume that all the physical clients are started first, otherwise Cortex will complain
                            if(!cmptype.equalsIgnoreCase("file"))InitiateCodaInit(cmpname, RcPhysicalClients.get(cmpname));
                        } else {

// There is no any physical clinet with the same name described in the cool configuration file,
// that reqiested agent representation to the Cortex's RcBroadcastDomain server.
                            if(!cmptype.equalsIgnoreCase("file")){
                                informPlexuses("Cortex-Error:\n  physical client " + cmpname + " does not exist.","error", false, session);
                                return false;
                            }
                        }
                    }
                }
            }
            return true;

// There are no components described in the cool configuration file
        } else {
            informPlexuses("Cortex-Error:\n Missing component descriptions in the cool file.","error",false, session);
            return false;
        }
    }


    /**
     *  Method connects to the coda DB. For coda2 components only.
     * @param session      session of the request
     */
    public void openDB(String session){

// Connect to the database.
        try {
            DatabaseConnection.createConnection(_URL,_DRIVER,_USER,_PASSWORD);
        } catch (DBException e) {
            informPlexuses("Cortex-Error:\n  Can not create DB connection","error",false, session);
            e.printStackTrace();
        }
    }

    /**
     * Differentiates other then runcontrol system agents
     * @param myParser    Cool configuration file parsed object
     * @return status
     */
    public boolean differentiateControlSystemAgents(CoolParser myParser){
        String sprname=null;
        String cname, ctype;
        CoolComponent agnt;
        List agents = new ArrayList();
        Iterator it;
        // Object containing physical client host and port
        rcClientAddress ccladr;

// We are statrting to differentiate agents. Reset stopdiff
        STOPDIFF = false;
        if(!myParser.CONTROL_COMPONENTS.isEmpty()){

// For every described component in the cool file
            for(it=myParser.CONTROL_COMPONENTS.iterator(); it.hasNext();){
                agnt = (CoolComponent)it.next();
                cname = agnt.getComponent_Name();
                ctype = agnt.getComponent_Type();
                if (!ctype.equalsIgnoreCase(rcConstants.statemachinesupervisorType) &&
                        !ctype.equalsIgnoreCase(rcConstants.wienercrateType))  { //wiener crate uses java snmp directly
                    //no cmsg communications
//See if there is a registered rcPhysicalClient, i.e. physical component that requested
// connection to the rcBroadcastDomain for registration (i.e. RCDomain cMsg IPC between physical component and an agent)
                    if ((RcPhysicalClients.containsKey(cname))) {

// set physical client host and port to be sent to its representative agent.
                        ccladr = RcPhysicalClients.get(cname);
                        agnt.setComponent_ClientHost(ccladr.getHost());
                        agnt.setComponent_ClientPort(ccladr.getPort());
                    }
                }
                agents.add(agnt);
            }
// Check to see if someone wants the differentiation process to quit.
            if (!STOPDIFF) {

// Start agent differentiation
                it = agents.iterator();
                while(it.hasNext()){
                    agnt = (CoolComponent)it.next();
// If the described component has a type
                    if (agnt.getComponent_Type() != null) {

// Differentiate process ( teaching process) will set the name and priority for the agent, as well as
// other attributes described in the cool configuration file. After this agent will register with DF
                        DifferentiateStemCell(agnt, myParser.SlcOption);

// Sleep and give some time component to differentiate.
// Attention: seems the connection to the cMsg server is failing when there are many simultaneous connection requests.
                        try {
                            Thread.sleep(1000);
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        }

                        if(agnt.getComponent_Type().equalsIgnoreCase(rcConstants.statemachinesupervisorType)){
                            sprname = agnt.getComponent_Name();
                        }
                    }
                }
            }

// Sleep and give some time medulla to differentiate
            try {
                Thread.sleep(rcConstants.sleepForMedullaDiff);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }

// Supervisor (Medulla) registration process
            it = agents.iterator();
            while(it.hasNext()){
                agnt = (CoolComponent)it.next();

// Register every component in the control system with the SMS agent
                if (agnt.getComponent_Type() != null &&
                        !agnt.getComponent_Type().equalsIgnoreCase(rcConstants.statemachinesupervisorType)){

// Register agents with SMS.
                    RegisterAgentWithSMSAgent(sprname,agnt);
                }
            }
            return true;
        } else {
            return false;
        }
    }


    /**
     * Based on the cool description, Cortex will teach each StemCell agent
     * what to do during the CODA state transitions.
     *
     * @param myParser   Cool configuration parsed object
     * @param runtype    runcontrol runtype
     * @param session    runcotrnol session
     * @return status
     */
    public boolean differentiateRuncontrolAgents(CoolParser myParser, String runtype, String session){
        String cname, ctype;
        CoolComponent agnt;
        List agents = new ArrayList();
        StringBuffer complist = new StringBuffer();
/** Object containing physical client host and port*/
        rcClientAddress ccladr;

// We are statrting to differentiate agents. Reset stopdiff
        STOPDIFF = false;
        informPlexuses("Cortex-Info:\n Please wait. Agents are in the process of differentiation....","info",false, session);

        if(!myParser.CONTROL_COMPONENTS.isEmpty()){

// For every described coda component in the cool file
            for(Iterator it=myParser.CONTROL_COMPONENTS.iterator(); it.hasNext();){
                agnt = (CoolComponent)it.next();
                if(agnt.getComponent_Node()==null){

// If agent node name is not specified it means that StemCell agent is started on the main-container
// and is going to report to the supervisor agent directly ( not to the deputy).
//                    agnt.setComponent_Reportsto(session+"_"+runtype);//report to the Medulla: gran supervisor
                    agnt.setComponent_Reportsto(myParser.SUPERVISOR);//report to the Medulla: gran supervisor
                } else {

// If node name is specified  it will report to deputy. N.B. deputy name = session_runtype_dp_nodename
//                    String dpn = session+"_"+runtype+"_dp_"+agnt.getComponent_Node().getNode_Name();
                    String dpn = myParser.SUPERVISOR+"_dp_"+agnt.getComponent_Node().getNode_Name();
                    agnt.setComponent_Reportsto(dpn);// report to the deputy
                }
                cname = agnt.getComponent_Name();
                ctype = agnt.getComponent_Type();
                if(!ctype.equalsIgnoreCase("FILE")){
                    if (!ctype.equalsIgnoreCase(rcConstants.rcSupervisorType) &&
                            !ctype.equalsIgnoreCase(rcConstants.statemachinesupervisorType) &&
                            !ctype.equalsIgnoreCase(rcConstants.wienercrateType))  { //wiener crate uses java snmp directly
                        //no cmsg communications
//See if there is a registered rcPhysicalClient, i.e. physical component that requested
// connection to the rcBroadcastDomain for registration.
                        if ((RcPhysicalClients.containsKey(cname))) {

// set physical client host and port to be sent to its representative agent.
                            ccladr = RcPhysicalClients.get(cname);
                            agnt.setComponent_ClientHost(ccladr.getHost());
                            agnt.setComponent_ClientPort(ccladr.getPort());
                        } else {

// There is'nt any physical client with the same name described in the cool configuration file
// that requested agent representation to the Cortex's RcBroadcastDomain server
                            informPlexuses("Cortex-Warning:\n" +
                                    " cMsg physical client " + cname + " is not registered.",
                                    "warning", false, session);
                            return false;
                        }
                    }
                    agents.add(agnt);
                }
            }

// Check to see if someone wants the differentiation process to quit.
            if (!STOPDIFF) {

// Start agent differentiation
                Iterator it = agents.iterator();
                int c = 0;
                while(it.hasNext()){
                    agnt = (CoolComponent)it.next();

// If the described component has a type
                    if (agnt.getComponent_Type() != null) {
                        c++;
                        if(c%rcConstants.CORTEX_DIF_GUI_REPORT == 0){
                            informPlexuses("Cortex-Info: Differentiating agent  "+agnt.getComponent_Name()+" ...","info",false, session);
                        }


// Tell the stemcell agent to become the requested Coda component agent.
// Differentiate process ( teaching process) will set the name and priority for the agent, as well as
// other attributes described in the cool configuration file. After this agent will register with DF
                        DifferentiateStemCell(agnt, myParser.CodaOption);

// Sleep for a while before initiating CODA transition teaching processes
                        try {
                            Thread.sleep(rcConstants.sleepBeforeTeaching_Cortex);
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        }

// Teach what to do at every CODA state transition. Sending ontology actions Set... to agents
                        TeachTransition(agnt, new SetPreConfigure(), rcConstants.preconfigure, rcConstants.CMD_preconfigure);
                        TeachTransition(agnt, new SetConfigure(), rcConstants.configure, rcConstants.CMD_configure);
                        TeachTransition(agnt, new SetPostConfigure(), rcConstants.postconfigure, rcConstants.CMD_postconfigure);

                        TeachTransition(agnt, new SetPreDownload(), rcConstants.predownload, rcConstants.CMD_predownload);
                        TeachTransition(agnt, new SetDownload(), rcConstants.download, rcConstants.CMD_download);
                        TeachTransition(agnt, new SetPostDownload(), rcConstants.postdownload, rcConstants.CMD_postdownload);

                        TeachTransition(agnt, new SetPrePrestart(), rcConstants.preprestart, rcConstants.CMD_preprestart);
                        TeachTransition(agnt, new SetPrestart(), rcConstants.prestart, rcConstants.CMD_prestart);
                        TeachTransition(agnt, new SetPostPrestart(), rcConstants.postprestart, rcConstants.CMD_postprestart);

                        TeachTransition(agnt, new SetPreGo(), rcConstants.prego, rcConstants.CMD_prego);
                        TeachTransition(agnt, new SetGo(), rcConstants.go, rcConstants.CMD_go);
                        TeachTransition(agnt, new SetPostGo(), rcConstants.postgo, rcConstants.CMD_postgo);

                        TeachTransition(agnt, new SetPreEnd(), rcConstants.preend, rcConstants.CMD_preend);
                        TeachTransition(agnt, new SetEnd(), rcConstants.end, rcConstants.CMD_end);
                        TeachTransition(agnt, new SetPostEnd(), rcConstants.postend, rcConstants.CMD_postend);

                        TeachTransition(agnt, new SetInPause(), rcConstants.pause, rcConstants.CMD_pause);

                        TeachTransition(agnt, new SetInResume(), rcConstants.resume, rcConstants.CMD_resume);

                        TeachTransition(agnt, new SetReset(), rcConstants.reset, rcConstants.CMD_reset);

                    }
                }

// Sleep and give some time medulla to differentiate
                try {
                    Thread.sleep(rcConstants.sleepForMedullaDiff);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

// Medulla registration process
                it = agents.iterator();
                while(it.hasNext()){
                    agnt = (CoolComponent)it.next();

// Register every component other then RCS and FILE with Medulla
                    if (agnt.getComponent_Type() != null &&
                            !agnt.getComponent_Type().equalsIgnoreCase(rcConstants.rcSupervisorType) &&
                            !agnt.getComponent_Type().equalsIgnoreCase(rcConstants.statemachinesupervisorType)){

// Register agents with Medulla. Attention Medulla is
// started under the name of runType. Differentiation will only
// force Medulla to register with the DF under the name of runType.
                        RegisterAgentWithMedulla(myParser,agnt);

// Create component list for plexuses (e.g. rcGui)
                        complist.append(agnt.getComponent_Name()).append(" ");
                    }
                }
            }

// Broadcast the names for all registered DAQ components to GUIs
            informPlexuses(rcSubscription.RcGuiSubjInfo+"_"+session, rcSubscription.RcGuiTypeComponentNames,complist.toString());

// If we reach this point that means that differentiation succeeded
            return true;
        }

// There are no components described in the cool configuration file
        else {
            informPlexuses("Cortex-Error:\n Missing component descriptions in the cool file.","error",false, session);
            return false;
        }
    }

    /**
     * Method responsible for StemCell specialization.
     *
     * @param runtype                     Runtype of the dataproduction
     * @param session                     Seesion of the dataproduction
     * @return status
     */
    public boolean COOLdifferentiateAgents(String runtype, String session) {

        Boolean rcontrol = false, status = false;
        CoolComponent agnt;

        CoolParser myParser = ParsedCoolObjects.get(runtype);

        if(myParser!=null && !myParser.CONTROL_COMPONENTS.isEmpty()){
// See if this is a coda runcontrol configuration
            for(int i=0; i<myParser.CONTROL_COMPONENTS.size();i++){
                agnt = (CoolComponent)myParser.CONTROL_COMPONENTS.get(i);
                if(agnt.getComponent_Type().equalsIgnoreCase(rcConstants.rcSupervisorType)){
                    rcontrol = true;
                    break;
                }
            }

            if(rcontrol){
                status = differentiateRuncontrolAgents(myParser, runtype, session);
            }

            else {
                status = differentiateControlSystemAgents(myParser);
            }
        }
        return status;
    }


    /**
     * Sends differentiation coda runcontrol ontology message to the Component StemCell agent
     *
     * @param agnt                   CodaComponent object, i.e. representative agent on the platform
     * @param opt                    CoolOption object
     */
    public void DifferentiateStemCell(CoolComponent agnt, CoolOption opt){

//        System.out.println("DDD ... Cortex sends differentiate acl message to "+agnt.getComponent_Name());
// Define the AID of the receiver
        AID receiver = new AID(agnt.getComponent_Name(),AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Use the codaRunControl ontology
// Use the codaRunControl ontology
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill differentiation ontology action
            Differentiate diff = new Differentiate();
            diff.setDifferentiate_Component(agnt);
            diff.setDifferentiate_Option(opt);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(diff);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Sends CodaInit coda runcontrol ontology message to the Component StemCell agent
     *
     * @param agentname                   StemCell agent name, receiving init message
     * @param cladr                       rcClientAddress object
     */
    public void InitiateCodaInit(String agentname, rcClientAddress cladr){

// Define the AID of the receiver
        AID receiver = new AID(agentname,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Use the codaRunControl ontology
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill codainit ontology action
            CodaInit in = new CodaInit();
            in.setCodainit_Clienthost(cladr.getHost());
            in.setCodainit_Clientport(cladr.getPort());
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(in);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Cortex sends message to supervisor agent (Medulla) informing to register an
     * agent partisipating in the requested runtype.
     *
     * @param myParser                  Cool configuration file parsed object
     * @param agnt                      coda component (agent)
     */
    private void RegisterAgentWithMedulla(CoolParser myParser, CoolComponent agnt) {

// Create the AID of the supervisor agent as a receiver.
// N.B. Supervisor agent name is constructed  as session_runtype, so be very carefull
// describing runtypes for the components. They all should have the same runtype.
//        AID receiver = new AID(agnt.getComponent_Session()+"_"+agnt.getComponent_Runtype(),AID.ISLOCALNAME);
        AID receiver = new AID(myParser.SUPERVISOR,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// User parameter is used to indicate that the message is sent by the Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill ontology action
            AcceptComponent mAccept = new AcceptComponent();
            mAccept.setAcceptcomponent_Component(agnt);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(mAccept);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }

    /**
     * Cortex sends message to an SMS agent informing to register an
     * agent.
     *
     * @param spr                       supervisor agent name
     * @param agnt                      coda component (agent)
     */
    private void RegisterAgentWithSMSAgent(String spr,CoolComponent agnt) {

// Create the AID of the supervisor agent as a receiver.
// N.B. Supervisor agent name is constructed  as session_runtype, so be very carefull
// describing runtypes for the components. They all should have the same runtype.
        AID receiver = new AID(spr,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// User parameter is used to indicate that the message is sent by the Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill ontology action
            AcceptComponent mAccept = new AcceptComponent();
            mAccept.setAcceptcomponent_Component(agnt);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(mAccept);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }

    /**
     * Cortex sends message to the supervisor agent of the control (Medulla or SMSAgent) agent informing
     * to register a subcontrol system.
     * agent.
     *
     * @param spr                       supervisor agent name
     * @param subsys                    sub control system 
     */
    private void RegisterSystemWithSupervisor(String spr,CoolSystem subsys) {

// Create the AID of the supervisor agent as a receiver.
        AID receiver = new AID(spr,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// User parameter is used to indicate that the message is sent by the Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill ontology action
            AcceptSystem mAccept = new AcceptSystem();
            mAccept.setAcceptsystem_System(subsys);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(mAccept);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Cortex sends message to the supervisor agent of the control (Medulla or SMSAgent) agent informing
     * to register control system rules .
     * agent.
     *
     * @param spr                       supervisor agent name
     * @param rules                     List of all CoolRule objects
     */
    private void RegisterRulesWithSupervisor(String spr,List rules) {

// Create the AID of the supervisor agent as a receiver.
        AID receiver = new AID(spr,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// User parameter is used to indicate that the message is sent by the Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill ontology action
            AcceptRules mAccept = new AcceptRules();
            mAccept.setAcceptrules_List(rules);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(mAccept);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Cortex sends message to the supervisor agent of the control (Medulla or SMSAgent) agent informing
     * to register component described states of the control system.
     * agent.
     *
     * @param spr                       supervisor agent name
     * @param states                     List of all CoolState objects
     */
    private void RegisterStatesWithSupervisor(String spr,List states) {

// Create the AID of the supervisor agent as a receiver.
        AID receiver = new AID(spr,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// User parameter is used to indicate that the message is sent by the Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill ontology action
            AcceptStates mAccept = new AcceptStates();
            mAccept.setAcceptstates_List(states);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(mAccept);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * CODA state transition details are informed to the agent.
     * N.B. Transition names are hard coded, however basic transiton commands for cMsg based system can be defined
     * in the rcConstants file.Coda basic transition are not defined in the cool file.
     * The command type of those transitions are considered to be the same as CodaRuncontrol ontology components
     * <hasComponentIPC> predicate value.
     * N.B. transition name is also the state name
     *
     * @param agnt                    CodaComponent object
     * @param trdetails               CodaRunControl ontology action, which is a container of
     *                                the all processes attached to this transition.
     * @param transition              Enumerated coda state transition for which the teaching is organized
     * @param transitionStr           The name of the actual transition
     */
    private void TeachTransition(CoolComponent agnt, TransitionExtention trdetails, int transition, String transitionStr){

// Transition related process list, including basic coda transitions, as well as scripts.
        List cmList = new ArrayList();

// Create a receiver aid from the agent name
        AID receiver = new AID(agnt.getComponent_Name(),AID.ISLOCALNAME);

// ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// Cortex is sending this message. This is mentioned in the ACL message as a user defined parameter.
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Define state transition process, which will have command name the same as the transition name.
// N.B. CODA state transition processes are not described in the cool file.
            CoolProcess cmd = new CoolProcess();

// Basic coda transition state machine is generated by the supervisor agent (Medulla),
// so we dont have to teach him what to do.Also ROCS only have to know the established
// transition communication commands. These commands are hard codded here inorder to simplify
// the COOL configuration file and free the users from the ROC operational details.
            if((!agnt.getComponent_Type().equalsIgnoreCase(rcConstants.rcSupervisorType)))
            {

// Coda transition description for the ROC agents
                switch(transition){

// N.B. Here we define the basic coda transition commands, NOT the external attached processes.

// Download state transition
                    case rcConstants.download:

// ipc cMsg
                        if(agnt.getComponent_Ipc().equalsIgnoreCase("cMsg")){
                            cmd.setProcess_Commandname(rcConstants.CMD_download);
                            cmd.setProcess_Ipc("cMsg");
                            cmd.setProcess_Isstatetransition("true");

// Old dpsh command execution
                        } else if(agnt.getComponent_Ipc().equalsIgnoreCase("dpsh")){
                            cmd.setProcess_Commandname(rcConstants.CMD_download+" "+agnt.getComponent_Runtype());
                            cmd.setProcess_Ipc("dpsh");
                        }

// tcpServer command execution (depricated)
                        else if(agnt.getComponent_Ipc().equalsIgnoreCase("tcp")){
                            cmd.setProcess_Commandname(rcConstants.CMD_download+" "+agnt.getComponent_Runtype());
                            cmd.setProcess_Ipc("tcp");
                        }
                        break;

// Prestart state transition
                    case rcConstants.prestart:

// cMsg ipc.
                        if(agnt.getComponent_Ipc().equalsIgnoreCase("cMsg")){
                            cmd.setProcess_Commandname(rcConstants.CMD_prestart);
                            cmd.setProcess_Ipc("cMsg");
                            cmd.setProcess_Isstatetransition("true");

// Old dpsh command execution
                        } else if(agnt.getComponent_Ipc().equalsIgnoreCase("dpsh")){
                            cmd.setProcess_Commandname(rcConstants.CMD_prestart);
                            cmd.setProcess_Ipc("dpsh");
                        }

// tcpServer command execution  (depricated)
                        else if(agnt.getComponent_Ipc().equalsIgnoreCase("tcp")){
                            cmd.setProcess_Commandname(rcConstants.CMD_prestart);
                            cmd.setProcess_Ipc("tcp");
                        }
                        break;

// Go state transition
                    case rcConstants.go:

// cMsg ipc.
                        if(agnt.getComponent_Ipc().equalsIgnoreCase("cMsg")){
                            cmd.setProcess_Commandname(rcConstants.CMD_go);
                            cmd.setProcess_Ipc("cMsg");
                            cmd.setProcess_Isstatetransition("true");

// Old dpsh command execution
                        } else if(agnt.getComponent_Ipc().equalsIgnoreCase("dpsh")){
                            cmd.setProcess_Commandname(rcConstants.CMD_go);
                            cmd.setProcess_Ipc("dpsh");
                        }

// tcpServer command execution (depricated)
                        else if(agnt.getComponent_Ipc().equalsIgnoreCase("tcp")){
                            cmd.setProcess_Commandname(rcConstants.CMD_go);
                            cmd.setProcess_Ipc("tcp");
                        }
                        break;

// End state transition
                    case rcConstants.end:

// cMsg ipc
                        if(agnt.getComponent_Ipc().equalsIgnoreCase("cMsg")){
                            cmd.setProcess_Commandname(rcConstants.CMD_end);
                            cmd.setProcess_Ipc("cMsg");
                            cmd.setProcess_Isstatetransition("true");

// Old dpsh command execution
                        } else if(agnt.getComponent_Ipc().equalsIgnoreCase("dpsh")){
                            cmd.setProcess_Commandname(rcConstants.CMD_end);
                            cmd.setProcess_Ipc("dpsh");
                        }

// tcpServer command execution (depricated)
                        else if(agnt.getComponent_Ipc().equalsIgnoreCase("tcp")){
                            cmd.setProcess_Commandname(rcConstants.CMD_end);
                            cmd.setProcess_Ipc("tcp");
                        }
                        break;

// Reset state transition
                    case rcConstants.reset:

// cMsg ipc
                        if(agnt.getComponent_Ipc().equalsIgnoreCase("cMsg")){
                            cmd.setProcess_Commandname(rcConstants.CMD_reset);
                            cmd.setProcess_Ipc("cMsg");
                            cmd.setProcess_Isstatetransition("true");

// Old dpsh command execution
                        } else if(agnt.getComponent_Ipc().equalsIgnoreCase("dpsh")){
                            cmd.setProcess_Commandname(rcConstants.CMD_exit);
                            cmd.setProcess_Ipc("dpsh");
                        }

// tcpServer command execution (depricated)
                        else if(agnt.getComponent_Ipc().equalsIgnoreCase("tcp")){
                            cmd.setProcess_Commandname(rcConstants.CMD_exit);
                            cmd.setProcess_Ipc("tcp");
                        }
                        break;

// Pause state transition
                    case rcConstants.pause:

// cMsg ipc
                        if(agnt.getComponent_Ipc().equalsIgnoreCase("cMsg")){
                            cmd.setProcess_Commandname(rcConstants.CMD_pause);
                            cmd.setProcess_Ipc("cMsg");
                            cmd.setProcess_Isstatetransition("true");

// Old dpsh command execution. N.B. coda2 components are not supporting pause or resume
                        } else if(agnt.getComponent_Ipc().equalsIgnoreCase("dpsh")){
                        }

// tcpServer command execution (depricated)
                        else if(agnt.getComponent_Ipc().equalsIgnoreCase("tcp")){
                        }
                        break;

// Resume state transition
                    case rcConstants.resume:
// cMsg ipc
                        if(agnt.getComponent_Ipc().equalsIgnoreCase("cMsg")){
                            cmd.setProcess_Commandname(rcConstants.CMD_resume);
                            cmd.setProcess_Ipc("cMsg");
                            cmd.setProcess_Isstatetransition("true");

// Old dpsh command execution
                        } else if(agnt.getComponent_Ipc().equalsIgnoreCase("dpsh")){
                        }

// tcpServer command execution  (depricated)
                        else if(agnt.getComponent_Ipc().equalsIgnoreCase("tcp")){
                        }
                        break;
                }

// We define here an input data to the process.
// As for coda transition processes (sent through the cMsg) inputData is used to pass the parameters, like
// command = donwload and parameter = runtype.
                CoolData indata = new CoolData();
                indata.setData_Content(agnt.getComponent_Runtype());
                indata.setData_Type("String");
                cmd.setProcess_Inputdata(indata);

// Every Cool process has associated data. Since we are describing coda transition as a
// process here we must follow the cool convention and describe also data for transiton process.
// We define the data.content mandatory slot value as " "
                CoolData retData = new CoolData();
                retData.setData_Content("");
                retData.setData_Type("String");
                retData.setData_Cardinality("single");
                cmd.setProcess_Outputdata(retData);

// For Coda transition processes command execution is only one time execution and delay = 0
                cmd.setProcess_Periodicity(1);
                cmd.setProcess_Delay(0);

// Node of the actual physical component
                if(agnt.getComponent_Node()!=null){
                    CoolNode nd = new CoolNode();
                    nd.setNode_Name(agnt.getComponent_Node().getNode_Name());
                }

// Medulla agent name is the same as RunType.
// We create the coda component object for the supervisor agent (Medulla) here to let all agents report to him.
                CoolComponent rcomp = new CoolComponent();
                rcomp.setComponent_Name(agnt.getComponent_Session()+"_"+agnt.getComponent_Runtype());

// Add the basic coda transition process to the list.
                if(cmd.getProcess_Commandname() !=null)cmList.add(cmd);
            }

// Now we are adding other processes defined in the cool file, which are attached
// to the CODA basic transitions, described above.
            if(agnt.getComponent_Processes()!=null){
                for (Iterator i= agnt.getComponent_Processes().iterator(); i.hasNext();) {

// Get the transition related process, defined in the cool file
                    CoolProcess pr = (CoolProcess)i.next();
                    if(pr.getProcess_Ispartofstate() != null &&
                            pr.getProcess_Ispartofstate().equalsIgnoreCase(transitionStr)){

// Add the described process to the list
                        cmList.add(pr);
                    }
                }
            }
        } catch(Exception e) { e.printStackTrace(); }

// Create and fill ontology action object
        try{
            trdetails.setProcesses(cmList);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(trdetails);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Communication with the Hepatic asking to return the map of the platfoem,
     * i.e. the names of all registered agents on the afecs platform.
     */
    public void InitiateGetPlatformMapConversation () {

        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaPlatformAdminOntology.NAME);

// Preparing the message. REceiver is Hepatic agent
            msg.addReceiver(platformAdmin);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaPlatformAdminOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),platformAdmin.getLocalName());

// Ontology action definition
            GetPlatformMap mp = new GetPlatformMap();
            Action a = new Action();
            a.setActor(platformAdmin);
            a.setAction(mp);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Actual conversation with the Hepatic to create a new agent on the requested container of the platform.
     *
     * @param agent_name                The name of the agent. It is the same as the described component name
     * @param agent_class               Agent class, i.e. Medulla or StemCell
     * @param container_name            The name of the container, usually the same as the node name
     * @param container_node            Host name where the agent container, i.e. jvm will run.
     *                                  N.B. those two names should be the same ( by design)
     */
    public void InitiateCreateAgentConversation (String agent_name, String agent_class,
                                                 String container_name, String container_node) {

// If we already creaed agent on the platform return. This will prevent sending
// multiple request to the Hepatic to start the sanme agent.
        if(agent_name.equals(LAST_CREATED))return;
        LAST_CREATED = agent_name;
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaPlatformAdminOntology.NAME);

// Preparing the message. Receiver is Hepatic agent
            msg.addReceiver(platformAdmin);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaPlatformAdminOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),platformAdmin.getLocalName());

// Create CodaAgent ontology concept.
            CodaAgent mag = new CodaAgent();
            mag.setName(agent_name);
            mag.setClas(agent_class);
            CodaAgentType agtype = new CodaAgentType();
            agtype.setDaqRoc(true);
            mag.setType(agtype);

// Create coda container ontology concept.
            CodaContainer mct = new CodaContainer();
            mct.setName(container_name);
            mct.setNode(container_node);
            mct.setType(agtype);

// Start coda agent ontology action
            StartCodaAgent stag = new StartCodaAgent();
            stag.setContainer(mct);
            stag.setAgent(mag);
            Action a = new Action();
            a.setActor(platformAdmin);
            a.setAction(stag);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Actual conversation with the Hepatic to kill an agent
     *
     * @param agent_name               Agent name going to be removed from the platform
     */
    public void InitiateKillAgentConversation (String agent_name) {
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaPlatformAdminOntology.NAME);

// Preparing the message. Receiver is Hepatic agent
            msg.addReceiver(platformAdmin);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaPlatformAdminOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),platformAdmin.getLocalName());
            CodaAgent mag = new CodaAgent();
            mag.setName(agent_name);
            StopCodaAgent stag = new StopCodaAgent();
            stag.setAgent(mag);
            Action a = new Action();
            a.setActor(platformAdmin);
            a.setAction(stag);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Actual conversation with the superviosr agent (Medulla) to disconnect,
     * i.e. Medulla is going to release the session for all partisipating components under his control.
     *
     * @param agent_name               Medullas name which is composed from the runtype and the session.
     * @param session                  Session name
     */
    public void InitiateDisconnect (String agent_name, String session) {

// Define the AID of the receiver (supervisor agent)
        AID receiver = new AID(agent_name,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Use the codaRunControl ontology
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill disconnect action object
            Disconnect dis = new Disconnect();
            dis.setDisconnect_Session(session);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(dis);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Method informs suprvisor agent (Medulla) the last recorded run number for the particular session.
     *
     * @param session                  Session name
     * @param agent_name               Medullas name which is composed from the runtype and the session.
     * @param runnumber                Unpdated runnumber for the particular session
     */
    public void InitiateAcceptRunnumber (String session, String agent_name, int runnumber) {

// Define the AID of the receiver
        AID receiver = new AID(agent_name,AID.ISLOCALNAME);

// Create a ACL message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Use the codaRunControl ontology
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),receiver.getLocalName());

// Create and fill disconnect action object
            AcceptRunNumber arn = new AcceptRunNumber();
            arn.setAcceptrunnumber_Value(runnumber);
            arn.setAcceptrunnumber_Session(session);
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(arn);
            getContentManager().fillContent(msg, a);

// Fill the content of the message
        } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour to query
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }



    /**
     * Actual conversation with the Hepatic to start a new container
     *
     * @param filename                   filename where the container names
     *                                   (which are the same as the node names) are recorded.
     */

    public void InitiateCreateContainerConversation (String  filename) {
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaPlatformAdminOntology.NAME);

// Preparing the message. Receiver is Hepatic agent
            msg.addReceiver(platformAdmin);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaPlatformAdminOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),platformAdmin.getLocalName());

// StartRemoteContainer ontology action definition
            StartRemoteContainer stct = new StartRemoteContainer();
            stct.setFilename(filename);
            Action a = new Action();
            a.setActor(platformAdmin);
            a.setAction(stct);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Actual conversation with the Hepatic to kill a container
     *
     * @param container_name           The name of the container going to be removed from the platform
     */
    public void InitiateKillContainerConversation (String container_name) {
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaPlatformAdminOntology.NAME);

// Preparing the message. Receiver is Hepatic agent
            msg.addReceiver(platformAdmin);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaPlatformAdminOntology.NAME);

// Create a container concept
            CodaContainer mct = new CodaContainer();
            mct.setName(container_name);

// Create an action
            StopContainer stct = new StopContainer();
            stct.setContainer(mct);
            Action a = new Action();
            a.setActor(platformAdmin);
            a.setAction(stct);

// fill the content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Actual conversation with the Hepatic to clone an agent
     *
     * @param agent_name                The name of the agent. It is the same as the described component name
     * @param agent_new_name            The new name of the agent after it has been cloned.
     * @param agent_class               Agent class
     * @param container_name            The name of the container, where it should be cloned,
     *                                  (usually the same as the node name, where the container is running)
     */
    public void InitiateCloneAgentConversation (String agent_name,String agent_new_name, String agent_class,
                                                String container_name) {
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaPlatformAdminOntology.NAME);

// Preparing the message
            msg.addReceiver(platformAdmin);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaPlatformAdminOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),platformAdmin.getLocalName());

// Define a codaAgent concept
            CodaAgent mag = new CodaAgent();
            mag.setName(agent_name);
            mag.setNName(agent_new_name);

// coda container concept, where the agent is going to be cloned.
            CodaContainer mct = new CodaContainer();
            mct.setName(container_name);

// Clone coda agent action
            CloneCodaAgent clag = new CloneCodaAgent();
            clag.setContainer(mct);
            clag.setAgent(mag);
            Action a = new Action();
            a.setActor(platformAdmin);
            a.setAction(clag);

// Fill the content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Actual conversation with the Hepatic to move the agent
     *
     * @param agent_name                The name of the agent.
     * @param container_name            The name of the container, where an agent should be moved.
     */
    public void InitiateMoveAgentConversation (String agent_name, String container_name) {
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaPlatformAdminOntology.NAME);

// prepare the message object
            msg.addReceiver(platformAdmin);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaPlatformAdminOntology.NAME);

// Set the user defined parameter the name of this agent, i.e. Cortex
            msg.addUserDefinedParameter(this.getAID().getLocalName(),platformAdmin.getLocalName());

// Define the agent you want to move
            CodaAgent mag = new CodaAgent();
            mag.setName(agent_name);

// Define the container we want the agent be moved.
            CodaContainer mct = new CodaContainer();
            mct.setName(container_name);

// actual ontology action definitioin
            MoveCodaAgent mvag = new MoveCodaAgent();
            mvag.setContainer(mct);
            mvag.setAgent(mag);
            Action a = new Action();
            a.setActor(platformAdmin);
            a.setAction(mvag);

// Fill the content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * This is the implementatioin of the abstract method defined in the CAgent class.
     * This method is going to be called by the BaseicRequestBehaviour object to process
     * messages coming back as a result of this agent initiating the request
     * to some other agent on the platfrom (usually "done" message of the request).
     *
     * @param msg             ACLMessage object passed from the GenRequestBehaviour object.
     *                        Here we analyze that object.
     */
    protected void processInform(ACLMessage msg) {
        String st = msg.getContent();

        if(st!=null){
            String[] ss = st.split("\\s"); // split the string by spaces
            if (st.contains("fail")) informPlexuses(st,"error",false,ss[0]);
            if(st.contains("Differentiation failed")){

// Differentiation process failed. Stop further differentiation for other components.
                STOPDIFF = true;
                informPlexuses("Cortex Severe error:\n  Agents differentiation failed.","error",false,ss[0]);
            }
        }
    }


    /**
     * This is the implementatioin of the abstract method defined in the CAgent class.
     * This method is going to be called by the BaseicResponderBehaviour object to process
     * the requests to this agent.
     *
     * @param request         ACLMessage request object passed from the GenResponderBehaviour object.
     *                        Here we fill it and return to the GenResponderBehaviour object.
     * @return                ACLMessage object
     */
    protected ACLMessage processRequest(ACLMessage request, Action a) {

        String session = null;
        String runtype = null;
        String old_runtype;
        String supervName = null;
        int rnum = 0;

// Create the reply object
        ACLMessage reply = request.createReply();
        try {

            // DEPRICATED .....
// message sent by the Plexus agent (depricated), informing to start parsing cool file and creating agents
            if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.StartConfig")) {
                // DEPRICATED .....
// Start differentiation of the agents. Again request from the Plexus agent (depricated)
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Differentiate")) {

// Check the differentiation by checking the agent registration with DF. This might be the request from the Thalamus.
// Curently not used
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.CheckDifferentiate")) {

// Request from the Medulla. Cortex will read the runnumber from  the requested session file, increment the runnumber
// send it back to Medulla, and record the new runnumber in the file with the requested runtype.
// If there are coda2 components it will update also the sessions table of the database.
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.RecordNewRun")) {

                BufferedWriter mywriter;
                BufferedReader myreader;
                RecordNewRun nr =(RecordNewRun)a.getAction();
                session = nr.getRecordnewrun_Session();
                runtype = nr.getRecordnewrun_Runtype();
                old_runtype = null;
                rnum = 0;

                supervName = session+"_"+runtype;

                CoolParser myParser = ParsedCoolObjects.get(runtype);


                if(myParser !=null && myParser.CodaOption.getOption_Hascoda2component() != null &&
                        myParser.CodaOption.getOption_Hascoda2component().equalsIgnoreCase("true")){

// For coda2 components
// Check if session exists in the sessions table.
                    try {
                        if(!DatabaseCommunication.checkSessionsTable(session))
                            informPlexuses("Cortex-error:\n  " +
                                    "Requested session does not exist in the database. \n" +
                                    "Configured coda2 components will not function properly.","error",false, session);
                        else {

// There is session in the db. Read the db run number, increment it and update the db and file.
// N.B. db in case coda2 components has the priority, meaning that run number in the db
// will define the runnumber in the file.
                            rnum =  DatabaseCommunication.getRunNumber(session);
                            rnum = rnum+1;
                            DatabaseCommunication.setRunNumber(rnum,session);
                            try{

// Write an updated runnumber into the session file
                                mywriter = new BufferedWriter(new FileWriter(System.getenv("COOL_HOME")+"/sessions/"+session));
                                mywriter.write(rnum+" "+runtype);
                                mywriter.flush();
                                mywriter.close();
                                fresh_start = false;

// Send the new runnumber to Medulla
                                InitiateAcceptRunnumber(session,supervName,rnum);
                            } catch (FileNotFoundException e) {
                                e.printStackTrace();
                                informPlexuses("Cortex-Error:\n  Session file for "+session+" was not found ","error",false, session);
                            } catch (IOException e1) {
                                e1.printStackTrace();
                            }
                        }
                    } catch (RCException e) {
                        e.printStackTrace();
                    }

                } else {


// There are no configured coda2 components
// Read the session file
                    try{
                        myreader = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/sessions/"+session));

// Get the previously configured runtype for the required session
                        String l = myreader.readLine();
                        StringTokenizer stl = new StringTokenizer(l);

// N.B. file format is "runnumber runtype"
                        try{
                            if(stl.hasMoreTokens()) rnum = Integer.parseInt(stl.nextToken());
                        } catch(NumberFormatException ne){
                            ne.printStackTrace();
                        }
                        if(stl.hasMoreTokens()) old_runtype = stl.nextToken();

// Now we can remove the old runtype (supervisor) agent
                        if((!fresh_start) &&
                                (old_runtype!=null) &&
                                (!old_runtype.equalsIgnoreCase("unknown")) &&
                                (!old_runtype.equals(runtype))) {
                            InitiateKillAgentConversation(old_runtype);

// Kill supervisor all deputy agents
                            if(!Containers.isEmpty()){
                                for(int i=0;i<Containers.size();i++){
                                    String cn = (String)Containers.get(i);
                                    String dpn = old_runtype+"_dp_"+cn;
                                    InitiateKillAgentConversation(dpn);
                                }
                            }
                        }

// Close the file
                        myreader.close();

// Increment runnumber
                        rnum = rnum+1;

// Write an updated runnumber into the session file
                        mywriter = new BufferedWriter(new FileWriter(System.getenv("COOL_HOME")+"/sessions/"+session));
                        mywriter.write(rnum+" "+runtype);
                        mywriter.flush();
                        mywriter.close();
                        fresh_start = false;

// Send the new runnumber to new runtype supervisor agent (Medulla)
                        InitiateAcceptRunnumber(session,supervName,rnum);
                    } catch (FileNotFoundException e) {
                        e.printStackTrace();
                        informPlexuses("Cortex-Error:\n  Session file for "+session+" was not found ","error",false, session);
                    } catch (IOException e1) {
                        e1.printStackTrace();
                    }
                }
            } else {}

// Inform action done
            Done d = new Done();
            d.setAction(a);
            getContentManager().fillContent(reply, d);
            reply.setPerformative(ACLMessage.INFORM);

// Content of the message will be composed from the runtype and session
            reply.setContent(runtype +","+session);
        }
        catch (Codec.CodecException fe) {
            fe.printStackTrace();
            System.err.println(getLocalName()+" Fill/extract content unsucceeded. Reason:" + fe.getMessage());
        }
        catch (OntologyException oe){
            oe.printStackTrace();
            System.err.println(getLocalName()+" Ontology exception. Reason:" + oe.getMessage());
        }
        return reply;
    }


    /**
     * cMsg callback class for handle registration requests from the physical
     * clients outside of the platform.Those are request to start an agent
     * which will represent real world physical client on the afecs platform.
     *
     */
    class BroadcastCallback extends cMsgCallbackAdapter {

        /** Container name where the StemCell agents will start first.*/
        String contName;

        /** Holds the number of components asking a representation on the platform.*/
        int sz;

        /**
         * Actual callback method
         *
         * @param msg            Message that triggered this callback
         * @param userObject     User object
         */
        public void callback(cMsgMessage msg, Object userObject){

// Physical client name
            String rcClientName = msg.getSender();
            if((RcPhysicalClients!=null) && (cMsgSrv!=null)) {
                if(RcPhysicalClients.containsKey(rcClientName)) {
                    for(String rt:ParsedCoolObjects.keySet()){
                        CoolParser cp = ParsedCoolObjects.get(rt);
                        for(int i=0; i<cp.CONTROL_COMPONENTS.size();i++){
                            CoolComponent  comp = (CoolComponent)cp.CONTROL_COMPONENTS.get(i);
                            if(comp.getComponent_Name().equals(rcClientName)){
// Inform gui that we need to reset, because the component that has an agent representation
// asks for a new connection, meaning it had been crashed at some point.
// Send autoreset cMsg to Medulla
                                cMsgMessage msgo = new cMsgMessage();
                                msgo.setSubject(cp.SUPERVISOR);
                                msgo.setType(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_autoreset);
                                try {
                                    cMsgSrv.send(msgo);
                                    cMsgSrv.flush(1);
                                } catch (cMsgException e) {
                                    e.printStackTrace();
                                }
// Create a message
                                msgo = new cMsgMessage();
                                msgo.setSubject(rcSubscription.CrtSubject);
                                msgo.setType("CortexReset");
                                msgo.setText(rcClientName);
                                try {
                                    cMsgSrv.send(msgo);
                                    cMsgSrv.flush(1);
                                } catch (cMsgException e) {
                                    e.printStackTrace();
                                }
                                informPlexuses("CortexReset "+rcClientName,"error",false, cp.COOL_SESSION);
                                break;
                            }
                        }
                    }
                }
            }

// Register physical client with the platform
            rcClientAddress rcadr = new rcClientAddress();
            rcadr.setHost(msg.getSenderHost());
//            rcadr.setHost(msg.getText());
            rcadr.setPort(msg.getUserInt());
            System.out.println("<i> cMsg connect request from "+rcClientName+
                    " host = "+msg.getSenderHost()+
                    " port = "+msg.getUserInt());
//            System.out.println("<i> cMsg connect request from "+rcClientName+
//                    " host = "+msg.getText()+
//                    " port = "+msg.getUserInt());
            RcPhysicalClients.put(rcClientName, rcadr);

// Open a file in the $COOL_HOME and write RcPhysicalCilents hash table content
            rcClientAddress tmpad;
            String tmps;
            try {
                BufferedWriter fout = new BufferedWriter(new FileWriter(System.getenv("COOL_HOME")+"/.physcomponents"));
                Enumeration<String> en = RcPhysicalClients.keys();
                while(en.hasMoreElements()){
                    tmps = en.nextElement();
                    tmpad = RcPhysicalClients.get(tmps);
                    fout.write(tmps+" "+tmpad.getHost()+" "+tmpad.getPort()+"\n");
                }
                fout.flush();
                fout.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
            if(Containers.isEmpty())contName = "Main-Container";
            else{

// Equally distribute created StemCell agents over the described/started containers
                if(sz<Containers.size()){
                    contName = (String)Containers.get(sz);
                    sz++;
                } else {
                    contName = (String)Containers.get(0);
                    sz=1;
                }
            }

// Creating StemCell agent
            System.out.println("Cortex-Info:\n Creating agent = " + rcClientName + " on the "+ contName);
            InitiateCreateAgentConversation(
                    rcClientName,
                    "org.jlab.coda.runcontrol.agents.StemCell",
                    contName,
                    contName);
            if(!searchDFregister(rcClientName,rcConstants.sleepAfterCreate_Cortex,10)) {
//                informPlexuses("Cortex-error:\n  " +rcClientName+" agent can not be started.","error",false);
                System.out.println("Cortex-error:\n  " +rcClientName+" agent can not be started.");
                return ;
            }

//Now all the agents are up and running in their containers. Sleep a while before sending init,
// which will open rc domain connection to the physical client ( complete private p2p connection
// between representative agent and the real world physical client).
            try {
                Thread.sleep(rcConstants.sleepBeforeInit_Cortex);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }

// Add the component into the component/container hashmap
            if(!CompContainer.containsKey(rcClientName)) CompContainer.put(rcClientName,contName);

// send codainit action message to the agent which will complete connect request of the client.
            InitiateCodaInit(rcClientName, rcadr);
        }
    }


    /**
     * cMsg callback class for handling control messages from RcGui.
     * Those are outside platform messages through the platform cMsg server (Hepatic)
     * Message has the following structure:
     * "command  runtype:<String>  session:<String>  cool:<String>  remnode:<String>
     *
     */
    class GuiCallback extends cMsgCallbackAdapter {

        String message;
        String command;
        String type;
        BufferedReader myread;
        BufferedWriter mywriter;

        /**
         * Actual callback method.
         * @param msg             Message that triggered this callback.
         * @param userObject      usr object
         */
        public void callback(cMsgMessage msg, Object userObject){

            String prev_runtype = null;
            type = msg.getType();
            message = msg.getText();

// ... ReportSessions
            if (type.contains("ReportSessions")){

// Get the name of the gui who sent this request
                StringTokenizer st = new StringTokenizer(type,"_");
                st.nextToken();
                String gname = null;
                if (st.hasMoreTokens()) gname = st.nextToken();

// get list of the files in the sessions dir.
// N.B. Each file in the sessions dir contains the
// runnumber and the latest runtype for that session.
                StringBuffer stb = new StringBuffer();
                File dir = new File(System.getenv("COOL_HOME")+"/sessions");
                String[] children = dir.list();
                if(children == null){
                    System.out.println("UI-DEBUG: No sessions dir in the $COOL_HOME");
                } else {
                    for (String aChildren : children) {

// Append the file name to the string buffer
                        if (!aChildren.startsWith(".")) stb.append(aChildren).append(" ");
                    }

// Send the space separated names of the sessions to the gui.
                    cMsgMessage msgo = new cMsgMessage();
                    msgo.setSubject(rcSubscription.RcGuiSubjInfo+"_"+message);
                    msgo.setType(rcSubscription.RcGuiTypeRegisteredSessions+"_"+gname);
                    msgo.setText(stb.toString());
                    try {
                        cMsgSrv.send(msgo);
                        cMsgSrv.flush(1);
                    } catch (cMsgException e) {
                        e.printStackTrace();
                    }
                }

// ... ReportRunTypes
            }  else if(type.contains("ReportRunTypes")){

// Get the name of the gui who sent this request
                StringTokenizer st = new StringTokenizer(type,"_");
                st.nextToken();
                String gname = null;
                if (st.hasMoreTokens()) gname = st.nextToken();

// Get list of the files in the $COOL_HOME dir
                StringBuffer stb = new StringBuffer();
                File dir = new File(System.getenv("COOL_HOME"));
                String[] children = dir.list();
                if(children == null){
                    System.out.println("UI-DEBUG: $COOL_HOME is empty");
                } else {
                    for (String aChildren : children) {

// Append the file name to the string buffer
                        if (aChildren.endsWith(".rdf"))
                            stb.append(aChildren.substring(0, aChildren.indexOf(".rdf"))).append(" ");
                    }

// Send the space separated names of the runtypes to the gui.
                    cMsgMessage msgo = new cMsgMessage();
                    msgo.setSubject(rcSubscription.RcGuiSubjInfo+"_"+message);
                    msgo.setType(rcSubscription.RcGuiTypeConfigurations+"_"+gname);
                    msgo.setText(stb.toString());
                    try {
                        cMsgSrv.send(msgo);
                        cMsgSrv.flush(1);
                    } catch (cMsgException e) {
                        e.printStackTrace();
                    }
                }

// ...StartContainers
            } else if(type.contains("StartContainers")){

// Tell Hepatic to start topology of the platform containers, based on the $COOL_HOME/containers file
                // InitiateCreateContainerConversation("some name");

// ... rcGUI Control
            }else if(type.contains(rcSubscription.CrtType)){

// Parse the message content
                StringTokenizer st = new StringTokenizer(message," ");

// Get the command
                command = st.nextToken();

// -Configure request
                if(command.equalsIgnoreCase("Configure")){
                    String coolfile = null;
                    String  UI_runtype_req = null;
                    String  UI_session_req = null;
                    while(st.hasMoreTokens()){
                        String tmp = st.nextToken();

// Get the runtype and session from the message
                        if(tmp.startsWith("runtype")) UI_runtype_req = tmp.substring(tmp.lastIndexOf(":")+1);
                        else if(tmp.startsWith("session")) UI_session_req = tmp.substring(tmp.lastIndexOf(":")+1);

// Get the cool configuration file name
                        else if(tmp.startsWith("cool")) coolfile =tmp.substring(tmp.lastIndexOf(":")+1);
                    }

// Create a message and send requested runtype to all guis in that session
                    cMsgMessage msgRT = new cMsgMessage();
                    msgRT.setSubject(rcSubscription.RcGuiSubjInfo+"_"+UI_session_req);
                    msgRT.setType(rcSubscription.RcGuiTypeRunType);
                    msgRT.setText(UI_runtype_req);
                    try {
                        cMsgSrv.send(msgRT);
                        cMsgSrv.flush(1);
                    } catch (cMsgException e) {
                        e.printStackTrace();
                    }

// Read the session file in the COOL_HOME/session dir
                    try {
                        myread = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/sessions/"+UI_session_req));

// Get the previously configured runtype for the required session
                        String l = myread.readLine();
                        StringTokenizer stl = new StringTokenizer(l);

// File format is "runnumber runtype"
                        if(stl.hasMoreTokens())stl.nextToken();
                        if(stl.hasMoreTokens())prev_runtype = stl.nextToken();
                        myread.close();

// Send disconnect to the supervisor agent (Medulla) for that runtype
                        if((!fresh_start) &&
                                (prev_runtype!=null) &&
                                (!prev_runtype.equalsIgnoreCase("unknown")) &&
                                (!prev_runtype.equals(UI_runtype_req)))
                            InitiateDisconnect(prev_runtype,UI_session_req);
                    } catch (FileNotFoundException e) {
//                        e.printStackTrace();
                        System.out.println("Cortex-Info: Creating new session "+UI_session_req);
// Create session file. This is the first time request for this session
                        try {
                            mywriter = new BufferedWriter(new FileWriter(System.getenv("COOL_HOME")+"/sessions/"+UI_session_req));
                            mywriter.write("0 unknown");
                            mywriter.flush();
                            mywriter.close();
                        } catch (IOException e1) {
                            e1.printStackTrace();
                            return;
                        }
                    } catch (IOException e) {
                        e.printStackTrace();
                        return;
                    }

// Start to configure and populating the platform with the agents. Parse the cool file for this runtype.
                    CoolParser myParser = parseCoolConfig(UI_session_req, coolfile);


// Ask Hepatic to create agents
                    if(COOLcreateAgents(UI_session_req, UI_runtype_req)) {
                        informPlexuses("Cortex-Info:\n Agents were sucessfully created.","info",false, UI_session_req);

// Sleep for a wile before agents differentiation process.
// N.B. We must make sure that the  stemcell agents are up and running, i.e. Hepatic created them.
                        try {
                            Thread.sleep(rcConstants.sleepBeforeDifferentiation_Cortex);
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        }

// Start differentiation
                        if(COOLdifferentiateAgents(UI_runtype_req, UI_session_req)){
                        } else{
                            informPlexuses("Cortex-Error:\n  Agents differentiation failed.","error",false, UI_session_req);
                        }

                    } else {
                        informPlexuses("Cortex-Error:\n Creating agents failed.","error",false, UI_session_req);
                    }

// Sleep for a wile after differentiation. After the differentiation download button will be active.
// So, we do not wont to download when agents are not yet differentiated. N.B. asynch communications...
                    try {
                        Thread.sleep(rcConstants.sleepAfterDifferentiation_Cortex);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }

// If differentiation process was not quited, i.e. it whent through
                    if(!STOPDIFF)informPlexuses("configured","info",true, UI_session_req);

// See if there is a subcontrol systems configured for this control system (hierarchy)
                    if(myParser !=null && myParser.CONTROL_SUBSYSTEMS!=null && !myParser.CONTROL_SUBSYSTEMS.isEmpty()){
                        Iterator it = myParser.CONTROL_SUBSYSTEMS.iterator();
                        CoolSystem sys;
                        while(it.hasNext()){
                            sys = (CoolSystem)it.next();
                            RegisterSystemWithSupervisor(myParser.SUPERVISOR,sys);
                        }
                    }

// register Rules with the supervisor (control state machine additional to CODA state machine)
                    if(myParser.CONTROL_RULES!=null && !myParser.CONTROL_RULES.isEmpty()){
                        RegisterRulesWithSupervisor(myParser.SUPERVISOR,myParser.CONTROL_RULES);
                    }

// register component states ( other than CODA states) with the supervisor.
                    if(myParser.CONTROL_SUBSYSTEMS!=null && !myParser.CONTROL_STATES.isEmpty()){
                        RegisterRulesWithSupervisor(myParser.SUPERVISOR,myParser.CONTROL_STATES);
                    }
                }

/*
// -StartConfigure request (depricated)
                else if(command.equalsIgnoreCase("StartConfig")){
                    String f = null;
                    while(st.hasMoreTokens()){
                        String tmp = st.nextToken();
                        if(tmp.startsWith("cool")) f = tmp.substring(tmp.lastIndexOf(":")+1);
                        else if(tmp.startsWith("session")) UI_session_req = tmp.substring(tmp.lastIndexOf(":")+1);
                        parseCoolConfig(UI_session_req, f);
                        COOLcreateAgents();
                        if(myParser.CodaOption.getOption_Hascoda2component().equalsIgnoreCase("true")){
                            openDB();

// Check if session exists in the sessions table. If not create a record
                            try {
                                if(!DatabaseCommunication.checkSessionsTable(UI_session_req))
                                    informPlexuses("Cortex-error:\n  " +
                                            "Requested session does not exist in the database. \n" +
                                            "Configured coda2 components can not function properly.","error",false);

                            } catch (RCException e) {
                                e.printStackTrace();
                            }
                        }
                    }

// -Differentiate request (not used)               
                }

                else if(command.equalsIgnoreCase("Differentiate")){
                    while(st.hasMoreTokens()){
                        String tmp = st.nextToken();
                        if(tmp.startsWith("runtype")) UI_runtype_req = tmp.substring(tmp.lastIndexOf(":")+1);
                        else if(tmp.startsWith("session")) UI_session_req = tmp.substring(tmp.lastIndexOf(":")+1);
                    }

// Actual agents differentiation method
                    COOLdifferentiateAgents(UI_runtype_req,UI_session_req);

                } else if(command.equalsIgnoreCase("CheckDifferentiated")){

// Report the active runtype to GUI for the required session
                }

*/
                else if(command.equalsIgnoreCase(rcConstants.CMD_ReportActiveRunType)){
                    String ar = null;
                    String l;
                    while(st.hasMoreTokens()){

// Get the session
                        String sess =  st.nextToken();

// Read the session file
                        BufferedReader myreader;
                        try {
                            myreader = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/sessions/"+sess));

// Get the previously configured runtype for the required session
                            l = myreader.readLine();
                        } catch (FileNotFoundException e) {
//                            e.printStackTrace();
                            System.out.println("Cortex-Warning: session "+sess+" is not registered with the platform.");
                            return;
                        } catch (IOException e) {
                            e.printStackTrace();
                            return;
                        }
                        StringTokenizer stl = new StringTokenizer(l);

// File format is "runnumber runtype"
                        stl.nextToken();

// Get active runtype for that session
                        if(stl.hasMoreTokens()) ar = stl.nextToken();

// N.B. if you remeber the runtype is constructed like "session_runtype". So, the regular expression
// here removes the leading session_.
                        assert ar != null;
                        ar = ar.replaceFirst("(?:"+sess+"_)+", "");

// Create a message and send runtype to all guis in that session
                        cMsgMessage msgRT = new cMsgMessage();
                        msgRT.setSubject(rcSubscription.RcGuiSubjInfo+"_"+sess);
                        msgRT.setType(rcSubscription.RcGuiTypeRunType);
                        msgRT.setText(ar);
                        try {
                            cMsgSrv.send(msgRT);
                            cMsgSrv.flush(1);
                        } catch (cMsgException e) {
                            e.printStackTrace();
                        }
                    }

// -DisconnectRuntype request
                } else if(command.equalsIgnoreCase(rcConstants.CMD_DisconnectRunType)){
                    String superv = null;
                    String sess = null;
                    while(st.hasMoreTokens()){

// Get the supervisor agent name and the session
                        superv =  st.nextToken();
                        sess = st.nextToken();

// Update the sessions file with runtype = undefined
// First read the runnumber
                        int rnum=0;
                        try{
                            myread = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/sessions/"+sess));

// Get the previously configured runtype for the required session
                            String l = myread.readLine();
                            StringTokenizer stl = new StringTokenizer(l);

// File format is "runnumber runtype"
                            try{
                                if(stl.hasMoreTokens()) rnum = Integer.parseInt(stl.nextToken());
                            } catch(NumberFormatException ne){
                                ne.printStackTrace();
                            }

// Write undefined as a runtype
                            mywriter = new BufferedWriter(new FileWriter(System.getenv("COOL_HOME")+"/sessions/"+sess));
                            mywriter.write(rnum+" undefined");
                            mywriter.flush();
                            mywriter.close();
                        } catch (FileNotFoundException e) {
                            e.printStackTrace();
                            informPlexuses("Cortex-Error:\n  Session file for "+sess+" was not found ","error",false, sess);
                        } catch (IOException e1) {
                            e1.printStackTrace();
                        }
                    }

// Initiate disconnect
                    if(superv!=null) InitiateDisconnect(superv, sess);
                }
            }
        }
    }


    /**
     * Utility class extends thread and executes the differentiateStemCell method
     * (agent (CoolComponent object) passed in the constructor) in the separate thread.
     */
/*
    public class SCDifferentiate extends Thread{
        CoolComponent comp;


        public SCDifferentiate(CoolComponent c){
            comp = c;
        }

        public void run(){
            DifferentiateStemCell(comp, myParser.CodaOption);
        }
    }
*/
}
