/*---------------------------------------------------------------------------*
*  Copyright (c) 2005        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.    *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.agents;

import rca.content.lang.Codec;
import rca.content.onto.OntologyException;
import rca.content.onto.basic.Action;
import rca.domain.FIPANames;
import rca.lang.acl.ACLMessage;
import rca.lang.acl.MessageTemplate;
import rca.proto.SimpleAchieveREResponder;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.CodaRunControlOntology;

/**
 * Agent responder behaviour.
 * This class incapsulates agent communication FIPA_REQUEST protocol.
 *
 * @author Vardan Gyurjyan
 *         Date: Jan 13, 2007
 *         Time: 3:45:51 PM
 */
public class GenRespondBehaviour extends SimpleAchieveREResponder {

    // Object copy of the CAgent class, passed to the constructor
    private CAgent myCAgent;

    /**
     * Constructor
     *
     * @param ag the refference to the Agent object instansiating this class.
     * @param vrb int indicating the verbosity level.
     */
    public GenRespondBehaviour(CAgent ag, int vrb){

// Call the parent constructor, and specify the message templet sellecting only the messages having
// FIPA_REQUEST interaction protocol.
        super(ag, MessageTemplate.and(MessageTemplate.and( MessageTemplate.MatchProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST),
                MessageTemplate.MatchOntology(CodaRunControlOntology.NAME)),
                MessageTemplate.MatchPerformative(ACLMessage.QUERY_IF)));
        myCAgent = ag;
    }


    /**
     * This method gets called after request has bin reached to the agent. Agent will first examine the user
     * defined parameter of the message ( ussually afecs sender puts the name of the
     * receiver agent as a user defined parameter).In case this message is for this particular
     * agent (passed in the constructor), corresponding reply message will be generated.
     *
     * @param request ACLMessage message
     * @return   ACLMessage
     */
    public ACLMessage prepareResponse(ACLMessage request) {

//        System.out.println(myCAgent.getLocalName()+" got the acl message");
// Prepare a dummy ACLMessage used to create the content of all reply messages for particular request message
        ACLMessage msg = request.createReply();
        if(request.getUserDefinedParameter(request.getSender().getLocalName()) != null) {
            if((request.getUserDefinedParameter(request.getSender().getLocalName()).equalsIgnoreCase(myCAgent.getLocalName()) ||
                    (request.getUserDefinedParameter(request.getSender().getLocalName()).equalsIgnoreCase("broadcast")))) {
                msg.setPerformative(ACLMessage.AGREE);
            }   else {
                msg.setPerformative(ACLMessage.NOT_UNDERSTOOD);
                request.reset();
            }
        }
        return msg;
    }


    /**
     * This method gets the ontology action object from the request object and passes it with the
     * created reply object to the agent for processing.
     *
     * @param request  ACLMessage object
     * @param response ACLMessage object
     * @return   ACLMessage object. If the processRequest method of the CAgent fails it will return the null object.
     */
    public ACLMessage prepareResultNotification(ACLMessage request, ACLMessage response){
        try {
//            System.out.println(myCAgent.getLocalName()+" preparing the ACL response");

// Get the ontology action object from the request
            Action a = (Action)myCAgent.getContentManager().extractContent(request);

// Process this action in the agent
            return myCAgent.processRequest(request,a);
        }
        catch (Codec.CodecException fe) {
            System.err.println(myCAgent.getLocalName()+
                    " Fill/extract content unsucceeded. Reason:" + fe.getMessage());
        }
        catch (OntologyException oe){
            System.err.println(myCAgent.getLocalName()+
                    " getRoleName() unsucceeded. Reason:" + oe.getMessage());
        }
        return null;
    }
}

