/*----------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-12H        *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5800             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package org.jlab.coda.runcontrol.agents;

import rca.content.lang.Codec;
import rca.content.lang.sl.SLCodec;
import rca.content.onto.OntologyException;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.core.AID;
import rca.core.Agent;
import rca.core.ContainerID;
import rca.core.behaviours.CyclicBehaviour;
import rca.core.behaviours.SenderBehaviour;
import rca.core.behaviours.SequentialBehaviour;
import rca.domain.AMSService;
import rca.domain.DFService;
import rca.domain.FIPAAgentManagement.*;
import rca.domain.FIPAException;
import rca.domain.FIPANames;
import rca.domain.JADEAgentManagement.CreateAgent;
import rca.domain.JADEAgentManagement.JADEManagementOntology;
import rca.domain.JADEAgentManagement.KillAgent;
import rca.domain.JADEAgentManagement.KillContainer;
import rca.domain.introspection.*;
import rca.domain.mobility.CloneAction;
import rca.domain.mobility.MobileAgentDescription;
import rca.domain.mobility.MobilityOntology;
import rca.domain.mobility.MoveAction;
import rca.lang.acl.ACLMessage;
import rca.lang.acl.MessageTemplate;
import rca.proto.SimpleAchieveREInitiator;
import rca.proto.SimpleAchieveREResponder;
import rca.util.leap.List;
import org.jlab.coda.runcontrol.ontologies.CodaPlatform.*;
import org.jlab.coda.runcontrol.util.rcObjectTransport;
import org.jlab.coda.runcontrol.util.rcConstants;
import org.jlab.coda.runcontrol.util.rcSubscription;
import org.jlab.coda.runcontrol.util.sh.SysExe;
import org.jlab.coda.cMsg.cMsgDomain.server.cMsgNameServer;
import org.jlab.coda.cMsg.cMsgMessage;
import org.jlab.coda.cMsg.cMsgException;
import org.jlab.coda.cMsg.cMsg;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;


/**
 *
 * <em>Hepatic</em> class is the administrative agent on the
 * <b>AFECS</b> Agent Platform.
 * It provides suitable behaviours to interact with the AMS,
 * registering for interesting events and requesting actions
 * when needed. Starting and stoping agents and/or agent containers.
 *
 * author Vardan Gyurjyan - JLAB DAQ group
 * version $Date: 2007/11/15 13:55:43 $ $Revision: 0.6 $
 */
public class Hepatic extends Agent  {

    /** Verbose level of std prfintouts*/
    private int verbose = rcConstants.debugInfo;

    /** Define the ACL content language, used for iter-agent communications */
    private Codec codec = new SLCodec();

    /** Defining the subscribe ACL meassages object for AMS subscription*/
    private ACLMessage AMSSubscription       = new ACLMessage(ACLMessage.SUBSCRIBE);

    /** Defining the cancel ACL meassages object, used to cancel AMS communications*/
    private ACLMessage AMSCancellation       = new ACLMessage(ACLMessage.CANCEL);

    /** Defining the request ACL meassages object, used to send requests to AMS*/
    private ACLMessage AMSRequest            = new ACLMessage(ACLMessage.REQUEST);

    /** Defines the sequential behaviour for AMS events subscription*/
    private SequentialBehaviour AMSSubscribe = new SequentialBehaviour();

    /** Map to keep correspondence between agent and container  (AID and Strings) */
    private ConcurrentHashMap<AID, String> AgentAIDContainer = new ConcurrentHashMap<AID, String>();

    /** Map to keep correspondence between agent-name and container  (String and  Strings) */
    private ConcurrentHashMap<String, String> AgentNameContainer = new ConcurrentHashMap<String, String>();

    /** Map to keep correspondence between agent name and AID  (String and AID)*/
    private ConcurrentHashMap<String, AID> AidAgen = new ConcurrentHashMap<String, AID>();

    /** Map to keep correspondence between agent and agent_class names (both Strings) */
    private ConcurrentHashMap<String, String> AgentClass = new ConcurrentHashMap<String, String>();

    /** Map to keep correspondence between agent and IP of the host where agent is running (both Strings) */
    private ConcurrentHashMap<String, String> AgentIp = new ConcurrentHashMap<String, String>();

    /** cMsg domain server. This is the platform messaging server,
     * used for external platform publish-subscribe communications. */
    cMsg cMsgSrv;


    /**
     * Interface introduces a handle method for handling the AMS events (agent dead, agent born, etc.)
     */
    public static interface EventHandler {
        void handle(Event ev);
    }


    /**
     * Setup methid of the <b>Hepatic</b>
     * Here we start cMsg server for outside platform messaging.
     * This method also starts the <em>Hepatic</em> behaviours, allowing this agent
     * to carry on its duties within <em><b>AFECS</b></em> agent platform.
     */
    public final void setup() {
        cMsgNameServer server;

// Get agent arguments.
// N.B> arguments are passed to the agent in this way name-of-agent(par1 par2...)
//        Object argus[] = getArguments();

// Start platform cMsg server
        server = new cMsgNameServer(Integer.parseInt(rcSubscription.cMsgTcpPort),
                Integer.parseInt(rcSubscription.cMsgUdpPort),
                false, true, null, null, 0);
// Start server
        server.startServer(null);

// Connect to the cMsg server just started.
        try {
            cMsgSrv = new cMsg(rcSubscription.PlatformcMsgServerUDL, "Hepatic", rcSubscription.PlatformcMsgServerDescription);
            cMsgSrv.connect();
        } catch (cMsgException e) {
            e.printStackTrace();
        }

// Start receiving messages
        cMsgSrv.start();

// Open the file and read the usr-name and passwd. This is used for running remote containers. (depricated)
/*
        try{

            BufferedReader f = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/.rccoda/.uid/.remoteid.txt"));
            f.close();
        } catch(IOException e){
            System.out.println("Hepatic_Warning: No permission to run remote containers.");
        }
*/
        
// Register wit the DF
        DFAgentDescription dfd = new DFAgentDescription();
        ServiceDescription sd = new ServiceDescription();

// Add to the service descriptor the type, name and owner.
        sd.setType("Hepatic");
        sd.setName(getLocalName());
        sd.setOwnership("JLABCODA");
        dfd.setName(getAID());
        dfd.addServices(sd);
        try {
            DFService.register(this,dfd);
        } catch (FIPAException e) {

            if(verbose>=rcConstants.debugError)
                System.err.println(getLocalName()+" registration with DF unsucceeded. Reason: "+e.getMessage());

// Registration failed delet agent
            doDelete();
        }

// Register the content language sytax for SL, used for iter-agent communications
        getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);

// Register the supported ontologies, i.e. semantice of the content language symbols.
        getContentManager().registerOntology(CodaPlatformAdminOntology.getInstance());
        getContentManager().registerOntology(FIPAManagementOntology.getInstance());
        getContentManager().registerOntology(JADEManagementOntology.getInstance());
        getContentManager().registerOntology(IntrospectionOntology.getInstance());
        getContentManager().registerOntology(MobilityOntology.getInstance());

// Fill AMS subscription ACL messages fields going to be send to the AMS.
// Sender is this agent, t.e. RCHepaptic
        AMSSubscription.setSender(getAID());

// We want to send this message only to AMS.
        AMSSubscription.clearAllReceiver();

// Set the receiver to be AMS
        AMSSubscription.addReceiver(getAMS());

// Content language syntax is SL0
        AMSSubscription.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);

// Ontology for this specific conversation is introspection
        AMSSubscription.setOntology(IntrospectionOntology.NAME);

// Reply to Hepatic with
        AMSSubscription.setReplyWith("tool-subscription");

// As a conversation ID we pass the name of this agent
        AMSSubscription.setConversationId(getLocalName());

// We want to subscribe the platform events.
        AMSSubscription.setContent("platform-events");

// The same way fill the ANS cancel message object
        AMSCancellation.setSender(getAID());
        AMSCancellation.clearAllReceiver();
        AMSCancellation.addReceiver(getAMS());
        AMSCancellation.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
        AMSCancellation.setOntology(IntrospectionOntology.NAME);
        AMSCancellation.setReplyWith("tool-cancellation");
        AMSCancellation.setConversationId(getLocalName());

// Fill the request message object to AMS. In this case according to the
// FIPA protocole we do not need to set the content.
        AMSRequest.setSender(getAID());
        AMSRequest.clearAllReceiver();
        AMSRequest.addReceiver(getAMS());
        AMSRequest.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
        AMSRequest.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
        AMSRequest.setOntology(JADEManagementOntology.NAME);

// Schedule platform events subscription ( first subbehaviour)
        AMSSubscribe.addSubBehaviour(new SenderBehaviour(this, AMSSubscription));

// Schedule handling incoming 'inform' messages from AMS (second subbehaviour)
        AMSSubscribe.addSubBehaviour(new RCApDiagnos());

// Now schedule this complex Behaviour for execution
        addBehaviour(AMSSubscribe);

// Schedule a new behaviour for handling inner platform agents requests (for e.g. Cortex)
        addBehaviour(new RCHepaticResponder(this));
    }


    /**
     * This method creates and sends cMsg message.
     * This is used to send information to all plexuses (rcGUI).
     *
     * @param subject         Subject of the message. In some casses indicates the ceverity
     *                        of the message. for e.g. ERROR, WARNING, INFO, etc.
     * @param type :          Type of the message. In some casses it indicates the name of the transition
     *                        during which the message was generated.
     * @param info            Message content (text)
     */
    private void informPlexuses(String subject, String type, String info) {

        cMsgMessage msg = new cMsgMessage();
        msg.setSubject(subject);
        msg.setType(type);
        msg.setText(info);
        try {
            cMsgSrv.send(msg);
            cMsgSrv.flush(1);
        } catch (cMsgException e) {
            e.printStackTrace();
        }
    }


    /**
     * Hepatic calls this method before shutdown.
     */
    protected final void takeDown() {

// Informs AMS that we will be gone shortly.
        send(AMSCancellation);
    }


    /**
     * Utility method returns AMS request message object.
     *
     * @return ACLMesage message containing AMS request details filled at setup.
     */
    protected ACLMessage getRequest() {
        return AMSRequest;
    }


    /**
     * Method starts the remote container.
     * under construction.....
     *
     * @param agname          The list of agent names
     * @param agclass         The list of agent classes
     * @param cntnode         Agent container node name
     */
    public void startRemCo(List agname, List agclass, String cntnode) {

// Agent name
        String tmpName;

// Agent class
        String tmpClass;

// Main container host name
        String hostMC=null;

// Iterators to iterate the agent and class name lists
        Iterator iagname;
        Iterator iagclass;

// Get the host name of the main container
        try {
            hostMC = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {e.printStackTrace();}

// Create a remote command string
        StringBuffer sshcommand = new StringBuffer("  ");

// Start creating the ssh command
        if(agname!=null && agclass!=null) {

// Get the iterators
            iagname = agname.iterator();
            iagclass = agclass.iterator();
            if(iagclass!=null && iagname!=null){

// Iterate over the agent names and classes.
// N.B if agent class or name is not defined this agent will not be started.
                while(iagclass.hasNext() && iagname.hasNext()){

// Get the name of the agent
                    tmpName = (String)iagname.next();

// Get the class of the agent
                    tmpClass = (String)iagclass.next();

// Build the remote command
                    sshcommand.append(" ").append(tmpName).append(":").append(tmpClass);

// Fill the map keeping correspondence between agent_name and class_name.
                    if(!AgentClass.containsKey(tmpName)) AgentClass.put(tmpName, tmpClass);

// Fill the map, keeping correspondence between an agent_name and a
// node name of the host, where agent is running.
                    if(!AgentIp.containsKey(tmpName))   AgentIp.put(tmpName, cntnode);
                }
            }
        }

// Compose the final command line
        String cmdLine = sshcommand.toString()+" >& /dev/null &";
        cmdLine = cmdLine.trim();
        if(verbose>=rcConstants.debugInfo) System.out.println("Hepatic:command -> "+cmdLine);
    }


    /**
     * This creates a new agent on the container, by sending a message to the AMS.
     *
     * @param ct              Container where the agent will start
     * @param ag              CodaAgent object
     */
    public void newAgent(CodaContainer ct, CodaAgent ag) {

// Get the agent name*
        String agentName = ag.getName();

// Get the class name
        String className = ag.getClas();

// Get the container name
        String containerName = ct.getName();

// Get the node name where the agent will run.
// N.B. by design node and container have the same name.
        String address = ct.getNode();

// Check if an agent is already registered with the Hepatic.
        if (AgentClass.containsValue(agentName)) {
            System.out.println("Hepatic : Warning " + agentName + " is registered on the platform");
            return;
        }

// Otherwise register: fill the map keeping correspondence between agent_name and class_name.
        if(!AgentClass.containsKey(agentName)) AgentClass.put(agentName, className);

// Fill the map keeping correspondence between agent_name and
// node name of the host where agent is running.
        if(!AgentIp.containsKey(agentName))   AgentIp.put(agentName, address);

// Fill the map containing the correspondence between agent name and container name.
        AgentNameContainer.put(agentName,containerName);

// Ontological class representing create agent action.
        CreateAgent ca = new CreateAgent();
        ca.setAgentName(agentName);
        ca.setClassName(className);

// Set container id
        ca.setContainer(new ContainerID(containerName, null));
        try {
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(ca);

// Create request message
            ACLMessage requestMsg = getRequest();

// Set ontology
            requestMsg.setOntology(JADEManagementOntology.NAME);

// Fill the content
            getContentManager().fillContent(requestMsg, a);

// Schedule create agent behaviour
            addBehaviour(new RCAmsRequest("CreateAgent", requestMsg));
        } catch (Exception fe) {
            fe.printStackTrace();
        }
    }


    /**
     * Overloaded method to start a new agent on the container.
     *
     * @param agentName       Agent name
     * @param className       Agent class
     * @param containerName   Container name
     * @param address         Node name of the container where the agent will run
     */
    public void newAgent(String agentName, String className, String containerName, String address) {

// Check if the agent is already registered with the Hepatic.
        if (AgentClass.containsValue(agentName)) {
            if(verbose>=rcConstants.debugWarning)
                System.out.println("Hepatic : Warning " + agentName + " is registered on the platform");
            return;
        }

// Otherwise register: fill the map keeping correspondence between agent_name and agent class_name.
        if(!AgentClass.containsKey(agentName)) AgentClass.put(agentName, className);

// Fill the map keeping correspondence between agent_name and node name of the host where agent is running.
        if(!AgentIp.containsKey(agentName))   AgentIp.put(agentName, address);

// Ontological class representing the createagent action.
        CreateAgent ca = new CreateAgent();

// Building the create-agent ontological action object
        ca.setAgentName(agentName);
        ca.setClassName(className);
        ca.setContainer(new ContainerID(containerName, null));
        try {
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(ca);

// Get request message object
            ACLMessage requestMsg = getRequest();

// Set ontology
            requestMsg.setOntology(JADEManagementOntology.NAME);

// Fill ACL message content*/
            getContentManager().fillContent(requestMsg, a);

// Schedule create agent behaviour
            addBehaviour(new RCAmsRequest("CreateAgent", requestMsg));
        } catch (Exception fe) {
            fe.printStackTrace();
        }
    }


    /**
     * Suspends agent on the AFECS platform
     *
     * @param name AID of the agent
     */
    public void suspendAgent(AID name) {

// Create AMS description object
        AMSAgentDescription amsd = new AMSAgentDescription();

// Set the agent name going to be suspended
        amsd.setName(name);

// Set the required state
        amsd.setState(AMSAgentDescription.SUSPENDED);

// Create 'modify' ontology action
        Modify m = new Modify();

// Build 'modify' ontology action object
        m.setDescription(amsd);
        try {
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(m);

// Fill ACL mesage specific details
            AMSRequest.setOntology(JADEManagementOntology.NAME);
            getContentManager().fillContent(AMSRequest, a);

// Schedule behaviour for execution.
            addBehaviour(new RCAmsRequest("SuspendAgent", AMSRequest));
        } catch (Exception fe) {
            fe.printStackTrace();
        }
    }


    /**
     * Resumes an agent on the AFECS paltform
     *
     * @param name: AID of the agent
     */
    public void resumeAgent(AID name) {

        AMSAgentDescription amsd = new AMSAgentDescription();
        amsd.setName(name);
        amsd.setState(AMSAgentDescription.ACTIVE);
        Modify m = new Modify();
        m.setDescription(amsd);
        try {
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(m);
            AMSRequest.setOntology(JADEManagementOntology.NAME);
            getContentManager().fillContent(AMSRequest, a);
            addBehaviour(new RCAmsRequest("ResumeAgent", AMSRequest));
        } catch (Exception fe) {
            fe.printStackTrace();
        }
    }

    /**
     * Killes an agent on the AFECS platform
     *
     * @param name: AID of the agent
     */
    public void killAgent(AID name) {
        KillAgent ka = new KillAgent();
        ka.setAgent(name);
        try {
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(ka);
            AMSRequest.setOntology(JADEManagementOntology.NAME);
            getContentManager().fillContent(AMSRequest, a);
            addBehaviour(new RCAmsRequest("KillAgent", AMSRequest));
        } catch (Exception fe) {
            fe.printStackTrace();
        }

// Deregister agent
        DfDeregister(name);
        AmsDeregister(name);

// Get actual name of the agent
        String tmpName = name.getLocalName();

// Remove from Hepatic registery maps this agent.
        AgentAIDContainer.remove(name);
        AgentNameContainer.remove(name.getLocalName());
        AidAgen.remove(tmpName);
        AgentClass.remove(tmpName);
        AgentIp.remove(tmpName);
    }


    /**
     * This method kills entire container with all its agents.
     *
     * @param name: Name of the container.
     */
    public void killContainer(String name) {

// List of the agents to be removed
        java.util.List<Object> tmpSetRemoveAgentAID  = new java.util.ArrayList<Object>();

// Container name
        String tmpContainer;

// Agent name
        String tmpName;

// Agents AID
        AID tmpAIDAgentName;

// Temporary AID variable
        AID tmpAid;

// Ontology action defined in the jade management ontology
        KillContainer kc = new KillContainer();

// Pass the name of the container
        kc.setContainer(new ContainerID(name, null));
        try {
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(kc);

// AMS request message ontology
            AMSRequest.setOntology(JADEManagementOntology.NAME);
            getContentManager().fillContent(AMSRequest, a);

// Schedule this behaviour for execution.
// N.B. the agents will be automatically killed if the JRE is killed, DHAAA ...?
            addBehaviour(new RCAmsRequest("KillContainer", AMSRequest));
        } catch (Exception fe) {
            fe.printStackTrace();
        }

// Deregister all the killed agents from tha AMS and DF registery systems.
        if (!AgentAIDContainer.isEmpty()) {

// Iterate over AgentAIDContainer to get recorded AID's
            for (AID aid : AgentAIDContainer.keySet()) {

// Get agent AUD
                tmpAIDAgentName = aid;

// Get container name where that particular agent was running
                tmpContainer = AgentAIDContainer.get(tmpAIDAgentName);

// Deregister all agents on that container
                if (tmpContainer.equals(name)) {

// Put AID agent name in to the tmp set
                    tmpSetRemoveAgentAID.add(tmpAIDAgentName);

// Deregister agent
                    DfDeregister(tmpAIDAgentName);
                    AmsDeregister(tmpAIDAgentName);
                }
            }

// Deregister all the killed agents from the Hepatic registery system.
// Iterate over just created removed agents list
            for (Object aTmpSetRemoveAgentAID : tmpSetRemoveAgentAID) {

// Get AID of the agent
                tmpAid = (AID) aTmpSetRemoveAgentAID;

// Get the name of the agent
                tmpName = tmpAid.getLocalName();

// Remove the agent from the Hepatic registery system
                if (AgentAIDContainer.containsKey(tmpAid)) AgentAIDContainer.remove(tmpAid);
                if (AgentNameContainer.containsKey(tmpName)) AgentNameContainer.remove(tmpName);
                if (AidAgen.containsKey(tmpName)) AidAgen.remove(tmpName);
                if (AgentClass.containsKey(tmpName)) AgentClass.remove(tmpName);
                if (AgentIp.containsKey(tmpName)) AgentIp.remove(tmpName);
            }
        }
    }


    /**
     * Clones an agent on the AFECS platform
     *
     * @param name:           AID of the agent
     * @param newName:        new name of the cloned agent
     * @param container:      containter name where the agent will be cloned
     */
    public void cloneAgent(AID name, String newName, String container) {
        AID tmpAid;
        String tmpName;

// Ontology action for cloning
        CloneAction cloneAct = new CloneAction();

// Define ontology concept
        MobileAgentDescription desc = new MobileAgentDescription();

// Define an agent name going to be cloned
        desc.setName(name);

// Create the destination container id
        ContainerID dest = new ContainerID(container, null);
        desc.setDestination(dest);

// Fll the actual cloning action
        cloneAct.setMobileAgentDescription(desc);
        cloneAct.setNewName(newName);
        try{
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(cloneAct);
            ACLMessage requestMsg = getRequest();
            requestMsg.setOntology(MobilityOntology.NAME);
            getContentManager().fillContent(requestMsg,a);

// Schedule this behaviour for execution
            addBehaviour(new RCAmsRequest("CloneAgent",requestMsg));
        }catch(Exception fe){fe.printStackTrace();}

// Get the name of the initial cloning agent
        tmpName = name.getLocalName();

// Register the cloned agent
// First record cloned agents class name, which is the same as the clonee class
        AgentClass.put( newName, AgentClass.get(tmpName));

// Get the containers IP address and put it into the map.
// To do that we look already recorded container ip names,
// since we can clone agent to the existing/running container only.
        for (AID aid : AgentAIDContainer.keySet()) {
            tmpAid = aid;
            if ((AgentAIDContainer.get(tmpAid)).equals(container)) {
                tmpName = tmpAid.getName();

// Get the IP address of the container from the map.*/
                AgentIp.put(newName, AgentIp.get(tmpName));
            }
        }
    }


    /**
     * Move an agent on the AFECS platform
     *
     * @param name:           AID of the agent
     * @param container:      name of the container
     */
    public void moveAgent(AID name, String container) {

        MoveAction moveAct = new MoveAction();
        MobileAgentDescription desc = new MobileAgentDescription();
        desc.setName(name);
        ContainerID dest = new ContainerID(container, null);
        desc.setDestination(dest);
        moveAct.setMobileAgentDescription(desc);
        try{
            Action a = new Action();
            a.setActor(getAMS());
            a.setAction(moveAct);
            ACLMessage requestMsg = getRequest();
            requestMsg.setOntology(MobilityOntology.NAME);
            getContentManager().fillContent(requestMsg,a);
            addBehaviour(new RCAmsRequest("MoveAgent",requestMsg));
        }catch(Exception fe){fe.printStackTrace();}
        AgentAIDContainer.remove(name);
        AgentAIDContainer.put(name,container);
        AgentNameContainer.remove(name.getLocalName());
        AgentNameContainer.put(name.getLocalName(),container);
    }


    /**
     * Deregisters an agent from the DF
     * @param theaid:         AID of the agent.
     */
    public void DfDeregister(AID theaid) {
        try {
            DFAgentDescription amsd = new DFAgentDescription();
            amsd.setName(theaid);
            DFService.deregister(this, amsd);
        } catch (FIPAException fe) {
            fe.printStackTrace();
        }
    }


    /**
     * Deregister an agent from the AMS
     *
     * @param theaid:        AID of the agent
     */
    public void AmsDeregister(AID theaid) {
        try {
            AMSAgentDescription amsd = new AMSAgentDescription();
            amsd.setName(theaid);
            AMSService.deregister(this, amsd);
        } catch (FIPAException fe) {
            fe.printStackTrace();
        }
    }


    /**
     * Shut donw entire AFECS platform
     */
    public void shutDownPlatform() {
        killContainer("Main-Container");
    }


    /**
     * <em>AMSListenerBehaviour</em> inner class is resposible for
     * receiving notifications by AMS.
     *
     * @author Vardan Gyurjyan - JLAB DAQ group
     * @version $Date: 2006/11/15 13:55:43 $ $Revision: 1.6 $
     */

    protected abstract class AMSListenerBehaviour extends CyclicBehaviour {

        // Define a message temple
        private MessageTemplate listenTemplate;

        // MAp of the AMS notification (event) handlers Ignore case for event names
        private Map handlers = new TreeMap(String.CASE_INSENSITIVE_ORDER);


        protected AMSListenerBehaviour() {

// I'm interested in messages having SL0 content language and introspection ontology
            MessageTemplate mt1 = MessageTemplate.MatchLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            MessageTemplate mt2 = MessageTemplate.MatchOntology(IntrospectionOntology.NAME);
            MessageTemplate mt12 = MessageTemplate.and(mt1, mt2);

// Language templet also will match with inreplyto string
            mt1 = MessageTemplate.MatchInReplyTo("tool-subscription");

// Introspection templet will match with performative
            mt2 = MessageTemplate.MatchPerformative(ACLMessage.INFORM);

// Add all  this requirements together.
            listenTemplate = MessageTemplate.and(mt1, mt2);
            listenTemplate = MessageTemplate.and(listenTemplate, mt12);

// Fill the event handler table, using a deferred operation.
            installHandlers(handlers);
        }


        /**
         * This method has to be implemented by concrete subclasses,
         * filling the <code>Map</code> passed as parameter with
         * implementations of the <code>EventHandler</code> interface,
         * using the name of the event as key (see the <code>Event</code>
         * interface.
         *
         * @param handlersTable     The table that associates each event name
         *                          with a proper handler.
         */
        protected abstract void installHandlers(Map handlersTable);


        /**
         * Handles inform message from the AMS
         */
        public void action() {
            ACLMessage current = receive(listenTemplate);
            if (current != null) {
                try {
                    Occurred o = (Occurred)getContentManager().extractContent(current);
                    EventRecord er = o.getWhat();
                    Event ev = er.getWhat();
                    String eventName = ev.getName();
                    EventHandler h = (EventHandler) handlers.get(eventName);
                    if (h != null)
                        h.handle(ev);
                } catch (ClassCastException fe) {
                    fe.printStackTrace();
                } catch (Exception cce) {
                    cce.printStackTrace();
                }
            } else
                block();
        }
    } // End of AMSListenerBehaviour class


    /**
     * <em>RCApDiagnos</em> inner class is resposible for getting
     * AMS messages
     *
     * @author Vardan Gyurjyan - JLAB DAQ group
     * @version $Date: 2006/11/15 13:55:43 $ $Revision: 1.6 $
     */
    protected class RCApDiagnos extends AMSListenerBehaviour {

        // Implement the abstract method of the AMSListenerBehaviour
        protected void installHandlers(Map handlersTable) {

// Fill the event handler table, i.e. what we do in case say addcontainer message is coming
// to the Hepatic from AMS, informing thet someone started a new contaioner.
// ... addcontainer case
            handlersTable.put(IntrospectionVocabulary.ADDEDCONTAINER, new EventHandler() {
                public void handle(Event ev) {
                    AddedContainer ac = (AddedContainer) ev;
                    ContainerID cid = ac.getContainer();
                    String name = cid.getName();
                    String address = cid.getAddress();
                    Calendar mycal = new GregorianCalendar();
                    String d =mycal.getTime().toString();
                    if(verbose>=rcConstants.debugInfo){
                        System.out.println(d.substring(0,d.indexOf("E"))+ " : container "+name + " is started on the " + address);
                    }
                }
            });

// ... Handle the removed container case
            handlersTable.put(IntrospectionVocabulary.REMOVEDCONTAINER, new EventHandler() {
                public void handle(Event ev) {
                    AID tmpAIDAgentName;
                    AID tmpAid;
                    java.util.List<Object> tmpSetAgentName       = new java.util.ArrayList<Object>();
                    java.util.List<Object> tmpSetClassName       = new java.util.ArrayList<Object>();
                    java.util.List<Object> tmpSetRemoveAgentAID  = new java.util.ArrayList<Object>();
                    String tmpContainer;
                    String tmpAgentName;
                    String tmpClassName;
                    String tmpName;

// Get ontology concept
                    RemovedContainer rc = (RemovedContainer) ev;
                    ContainerID cid = rc.getContainer();

// Get removed container name
                    String name = cid.getName();
                    Calendar mycal = new GregorianCalendar();
                    String d =mycal.getTime().toString();
                    if(verbose>=rcConstants.debugInfo) {
                        System.out.println(d.substring(0,d.indexOf("E"))+ " :Removed container - " + name);
                    }

// Update accordingly Hepatic registration
                    if (!AgentAIDContainer.isEmpty()) {
                        for (AID aid : AgentAIDContainer.keySet()) {

// Get an agent name
                            tmpAIDAgentName = aid;
                            tmpAgentName = tmpAIDAgentName.getLocalName();

// Get corresponding container name from the Map
                            tmpContainer = AgentAIDContainer.get(tmpAIDAgentName);
                            if (tmpContainer.equals(name)) {

// Deregister all agents on that container
// Put agent name in to the tmp set of agent names (String)
                                tmpSetAgentName.add(tmpAgentName);

//  Put AID agent name in to the tmp set
                                tmpSetRemoveAgentAID.add(tmpAIDAgentName);

// Get an agents class name and put it in to the tmp set of agent class names
                                tmpClassName = AgentClass.get(tmpAgentName);
                                tmpSetClassName.add(tmpClassName);

// Deregister agent at the platform level
                                DfDeregister(tmpAIDAgentName);
                                AmsDeregister(tmpAIDAgentName);
                            }
                        }

// remove from the Hepatic registration
                        if(tmpSetRemoveAgentAID != null) {
                            for (Object aTmpSetRemoveAgentAID : tmpSetRemoveAgentAID) {
                                tmpAid = (AID) aTmpSetRemoveAgentAID;
                                tmpName = tmpAid.getLocalName();
                                AgentAIDContainer.remove(tmpAid);
                                AgentNameContainer.remove(tmpName);
                                AidAgen.remove(tmpName);
                                AgentClass.remove(tmpName);
                                AgentIp.remove(tmpName);
                            }
                        }
                    }
                }
            });

// ... Handle agent is born message
            handlersTable.put(IntrospectionVocabulary.BORNAGENT, new EventHandler() {
                public void handle(Event ev) {
                    BornAgent ba = (BornAgent) ev;
                    ContainerID cid = ba.getWhere();
                    String container = cid.getName();
                    AID agent = ba.getAgent();
                    String name = agent.getLocalName();
                    Calendar mycal = new GregorianCalendar();
                    String d =mycal.getTime().toString();
                    if(verbose>=rcConstants.debugInfo)
                        System.out.println(d.substring(0,d.indexOf("E"))+" : Agent " + name + " is born on the " + container);

// Fill the Map. Check if it was requested by Hepatic
                    if(AgentClass.containsKey(name)){
                        if (!AgentAIDContainer.containsKey(agent)) AgentAIDContainer.put(agent, container);
                        if (!AgentNameContainer.containsKey(agent.getLocalName())) AgentNameContainer.put(agent.getLocalName(), container);
                        if (!AidAgen.containsKey(name)) AidAgen.put(name, agent);
                    }
                }
            });

// ... Handle agent is dead message
            handlersTable.put(IntrospectionVocabulary.DEADAGENT, new EventHandler() {
                public void handle(Event ev) {
                    DeadAgent da = (DeadAgent) ev;
                    ContainerID cid = da.getWhere();
                    String container = cid.getName();
                    AID agent = da.getAgent();
                    String name = agent.getLocalName();
                    Calendar mycal = new GregorianCalendar();
                    String d =mycal.getTime().toString();
                    if(verbose>=rcConstants.debugInfo)
                        System.out.println(d.substring(0,d.indexOf("E"))+" : Agent " + name + " died on the " + container);
                    if (AgentAIDContainer.containsKey(agent)) {

// Deregister agent on that container
                        DfDeregister(agent);
                        AmsDeregister(agent);

// Restart the dead agent, if agent is registered with Hepatic
// Here we restart the dead agent. For now comment it out
                        // if(AgentClass.containsKey(name)) {
                        //newAgent(name, (String) (AgentClass.get(name)), container, (String) (AgentIp.get(name)));
                        //                        }
                    }
                }
            });
        }
    } // End of RCApDiagnos class


    /**
     * <em>RCAmsRequest</em> inner class is resposible for sending requests to AMS.
     *
     * @author Vardan Gyurjyan - JLAB DAQ group
     * @version $Date: 2006/11/15 13:55:43 $ $Revision: 1.6 $
     */
    private class RCAmsRequest extends SimpleAchieveREInitiator {

        private String actionName;

        /**
         * Constructor
         *
         * @param an          String representing an action name
         * @param request     Request message
         */
        public RCAmsRequest(String an, ACLMessage request) {
            super(Hepatic.this, request);
            actionName = an;
        }

        /**
         * FIPA communication protocol specific method. This will be called in case
         * AMS sends back "not understood" message.
         * @param reply              ACL message from AMS
         */
        protected void handleNotUnderstood(ACLMessage reply) {
            System.out.println("Hepatic : NOT-UNDERSTOOD received from AMS during " + actionName + " " +reply.getContent());
        }

        /**
         * FIPA communication protocol specific method. This will be called in case
         * AMS sends back "refuse" message.
         * @param reply              ACL message from AMS
         */

        protected void handleRefuse(ACLMessage reply) {
            System.out.println("Hepatic : REFUSE received from AMS during " + actionName + " " +reply.getContent());
        }

        /**
         * FIPA communication protocol specific method. This will be called in case
         * AMS sends back "agree" message.
         * @param reply              ACL message from AMS
         */
        protected void handleAgree(ACLMessage reply) {
        }


        /**
         * FIPA communication protocol specific method. This will be called in case
         * AMS sends back "failure" message.
         * @param reply              ACL message from AMS
         */
        protected void handleFailure(ACLMessage reply) {
            System.out.println("Hepatic : FAILURE received  from AMS during " + actionName + " " +reply.getContent());
            if(reply.getContent().contains("already-registered ")) {
                StringTokenizer st = new StringTokenizer(reply.getContent());
                while (st.hasMoreTokens()){

                    if(st.nextToken().equalsIgnoreCase(":agent-name"))
                        informPlexuses(rcSubscription.Hepatic, rcSubscription.RcGuiTypeStemCell,
                                "Name Conflict Detected for the request to start "+st.nextToken());
                }
            }
        }


        /**
         * FIPA communication protocol specific method. This will be called in case
         * AMS sends back "inform" message.
         * @param reply              ACL message from AMS
         */

        protected void handleInform(ACLMessage reply) {
        }
    } // End of RCAmsRequest class


    /**
     * <em>RCHepaticResponder</em> inner class is resposible for
     * responding upcoming requests to the Hepatic agent
     * (inter-agent communications via ACL).
     *
     * @author Vardan Gyurjyan - JLAB DAQ group
     * @version $Date: 2006/11/15 13:55:43 $ $Revision: 1.6 $
     */
    private class RCHepaticResponder extends SimpleAchieveREResponder {

        // StartRemoteContainer ontology action
        private StartRemoteContainer stct;

        /**
         * Constructor
         *
         * @param myAgent            an Agent object passed to the constructor
         */
        public RCHepaticResponder(Agent myAgent){
            super(myAgent, MessageTemplate.and(
                    MessageTemplate.MatchProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST),
                    MessageTemplate.MatchOntology(CodaPlatformAdminOntology.NAME)));
        }

        /**
         * Overiding a method from the SimpleAchieveREResponder
         *
         * @param request           ACL message
         * @return                  ACL message
         */
        public ACLMessage prepareResponse(ACLMessage request) {

// Prepare a dummy ACLMessage used to create the content of all reply messages
            ACLMessage msg = request.createReply();

// The QUERY message could be a QUERY-REF. In this case reply
// with NOT_UNDERSTOOD
            if (msg.getPerformative() != ACLMessage.QUERY_IF){
                msg.setPerformative(ACLMessage.REFUSE);
                String content = "Unknown performative";
                msg.setContent(content);
                block(100);
            }   else {
                msg.setPerformative(ACLMessage.AGREE);
                String content = "I'll do it";
                msg.setContent(content);
            }
            return msg;
        }

        /**
         * Overiding a method from  SimpleAchieveREResponder
         *
         * @param request          ACL message
         * @param response         ACL message
         * @return                 reply ACL message
         */
        public ACLMessage prepareResultNotification(ACLMessage request, ACLMessage response){
            ACLMessage reply = request.createReply();
            ACLMessage objmsg = request.createReply();
            try {
                Action a = (Action)getContentManager().extractContent(request);

// ... Start a new agent
                if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.StartCodaAgent")) {
                    StartCodaAgent stag = (StartCodaAgent) a.getAction();
                    CodaAgent ag       = stag.getAgent();
                    CodaContainer ct   = stag.getContainer();

// If agent already exists on the platform there is no need to start it again.
                    if(!AidAgen.containsKey(ag.getName()))
                        newAgent(ct, ag);
                }

// ... Stop agent
                else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.StopCodaAgent")) {
                    StopCodaAgent stpag = (StopCodaAgent) a.getAction();
                    CodaAgent ag       = stpag.getAgent();
                    AID agid = new AID(ag.getName(),AID.ISLOCALNAME);
                    killAgent(agid);
                }

// ... Clone the agent
                else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.CloneCodaAgent")) {
                    CloneCodaAgent clag = (CloneCodaAgent) a.getAction();
                    CodaAgent ag       = clag.getAgent();
                    CodaContainer ct   = clag.getContainer();
                    AID agid = new AID(ag.getName(),AID.ISLOCALNAME);
                    cloneAgent(agid, ag.getNName(), ct.getName());
                }

// ... Move the agent
                else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.MoveCodaAgent")) {
                    MoveCodaAgent mvag = (MoveCodaAgent) a.getAction();
                    CodaAgent ag       = mvag.getAgent();
                    CodaContainer ct   = mvag.getContainer();
                    AID agid = new AID(ag.getName(),AID.ISLOCALNAME);
                    moveAgent(agid, ct.getName());
                }

// ... Start new container
                else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.StartRemoteContainer")) {
                    Calendar mycal = new GregorianCalendar();
                    String d =mycal.getTime().toString();
                    if(verbose>=rcConstants.debugInfo)
                        System.out.println(d.substring(0,d.indexOf("E"))+
                                " : Hepatic is starting platform containers. Be patient... ");

// Here we read the $COOL_HOME/containers file and start remote containers accordingly.
                    SysExe myExe = new SysExe();
//                    myExe.run_o_unix(System.getenv("AFECS_HOME")+"/bin/system/startcontainers & > /dev/null/&");
//@todo coment out to stop container auto-creation
//                    myExe.run_o_unix(System.getenv("AFECS_HOME")+"/bin/system/startcontainers");
                    System.out.println("Wait until platform containers will report...");
                }

// ... Start remote container and populate with agents
                else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.StartRemoteContainerAgents")) {
                    StartRemoteContainerAgents strct = (StartRemoteContainerAgents) a.getAction();
                    List names    = strct.getStrca_Agentnamelist();
                    List classes   = strct.getStrca_Agentclasslist();
                    String remcontnode      = strct.getStrca_Containernode();
                    startRemCo(names,classes,remcontnode);
                }

// ... Stop container
                else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.StopContainer")) {
                    StopContainer stpct = (StopContainer) a.getAction();
                    CodaContainer ct   = stpct.getContainer();
                    killContainer(ct.getName());
                }

// ... Platform map
                else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                        "CodaPlatform.GetPlatformMap")) {
                    try{

// Inform action done for getplatformmap. In this case we send a serialized object
                        Done d = new Done();
                        d.setAction(a);
                        myAgent.getContentManager().fillContent(objmsg, d);
                        objmsg.setPerformative(ACLMessage.INFORM);
                        objmsg.setContentObject(new rcObjectTransport(AgentAIDContainer));
                        objmsg.setLanguage("JavaSerialization");
                        objmsg.setOntology("AnyButFIPA_SL0");
                        objmsg.addUserDefinedParameter("Hepatic","PlatformMap");
                    } catch (IOException e ) {
                        e.printStackTrace();
                    }
                    return objmsg;
                }
                else {
                }

//Inform action done
                Done d = new Done();
                d.setAction(a);
                myAgent.getContentManager().fillContent(reply, d);
                reply.setPerformative(ACLMessage.INFORM);
            }
            catch (Codec.CodecException fe) {
                System.err.println(myAgent.getLocalName()+" Fill/extract content unsucceeded. Reason:" + fe.getMessage());
            }
            catch (OntologyException oe){
                System.err.println(myAgent.getLocalName()+" getRoleName() unsucceeded. Reason:" + oe.getMessage());
            }
            return reply;
        }
    }// End of RCHepaticResponder 
} // END of container class Hepatic
