/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.agents;

import org.jlab.coda.runcontrol.util.*;
import org.jlab.coda.runcontrol.util.sh.SysExe;
import org.jlab.coda.runcontrol.util.db.DatabaseCommunication;
import org.jlab.coda.runcontrol.util.db.RCException;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.*;
import org.jlab.coda.cMsg.*;
import rca.domain.FIPAAgentManagement.DFAgentDescription;
import rca.domain.FIPAAgentManagement.ServiceDescription;
import rca.domain.DFService;
import rca.domain.FIPAException;
import rca.domain.FIPANames;
import rca.content.lang.Codec;
import rca.content.lang.sl.SLCodec;
import rca.content.AgentAction;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.content.onto.OntologyException;
import rca.core.AID;
import rca.lang.acl.ACLMessage;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArraySet;
import java.io.*;

/**
 * Control system supervisor agent. Incapsulates coda state machine.
 * Object is configured using the ACL messages sent by <code>Cortex<code>.
 *
 * Author: gurjyan
 * Date: Jun 9, 2006
 * Time: 2:00:52 PM
 *
 */
public class Medulla extends CAgent{

    /** The name of this agent */
    private String myName;

    /** RCS CodaComponent object (supervisor) obtained during the differentiate command */
    private CoolComponent itsMe;

    /** Rules of this CODA Run Control system*/
    private rca.util.leap.List _myRules;

    /** All the registered states for the components of this CODA run control system*/
    private rca.util.leap.List _myStates;

    /** Physical client default reporting interval*/
    private static String PCRI = "1.0";

    /** Verbosity level defined in the rcConstants class.*/
    private int verbose = rcConstants.debugError;

    /** AFECS platform cMsg domain server (running as part of the <em>Hepaic</em> */
    cMsg cMsgSrv;

    /** cMsg subject for logging messages */
    private String LogSubject = rcSubscription.logSubject;

    /** cMsg type for logging messages*/
    private String LogType = rcSubscription.logType;

    /** This variable indicats a reset request */
    private volatile boolean _ResetRequest = false;

    /** Indicates if component is registered with the df */
    private boolean _RegisteredFlag = false;

    /** Linked lists of the CoolScheduler for processes and stop/pause events */
    private rca.util.leap.List _schedulerProcessList;
    private rca.util.leap.List _schedulerStopList;
    private rca.util.leap.List _schedulerPauseList;

    /** This variable indicates the index of the scheduled process in the list*/
    private int _schedulerIndex = -1000;

    /** Gets assighned to true if there is scheduler defined in the cool configuration file */
    private boolean _schedulerIsDefined = false;

    /** Private field keeps the SetPreConfigure object, containing the set of the processes
     * necessary to executeexecute at preconfigure transition.*/
    private SetPreConfigure  setpreconfigure;

    /** Private field keeps the SetConfigure object, containing the set of the processes
     * necessary to execute at configure transition.*/
    private SetConfigure     setconfigure;

    /** Private field keeps the SetPostConfigure object, containing the set of the processes
     * necessary to execute at postconfigure transition.*/
    private SetPostConfigure setpostconfigure;

    /** Private field keeps the SetPreDownload object, containing the set of the processes
     * necessary to execute at predownload transition.*/
    private SetPreDownload   setpredownload;

    /** Private field keeps the SetDownload object, containing the set of the processes
     * necessary to execute at download transition.*/
    private SetDownload      setdownload;

    /** Private field keeps the SetPostDownload object, containing the set of the processes
     * necessary to execute at postdownload transition.*/
    private SetPostDownload  setpostdownload;

    /** Private field keeps the SetPrePrestart object, containing the set of the processes
     * necessary to execute at preprestart transition.*/
    private SetPrePrestart   setpreprestart;

    /** Private field keeps the SetPrestart object, containing the set of the processes
     * necessary to execute at prestart transition.*/
    private SetPrestart      setprestart;

    /** Private field keeps the SetPostPrestart object, containing the set of the processes
     * necessary to execute at postprestart transition.*/
    private SetPostPrestart  setpostprestart;

    /** Private field keeps the SetPreGo object, containing the set of the processes
     * necessary to execute at prego transition.*/
    private SetPreGo         setprego;

    /** Private field keeps the SetGo object, containing the set of the processes
     * necessary to execute at go transition.*/
    private SetGo            setgo;

    /** Private field keeps the SetPostGo object, containing the set of the processes
     * necessary to execute at postgo transition.*/
    private SetPostGo        setpostgo;

    /** Private field keeps the SetPreEnd object, containing the set of the processes
     * necessary to execute at preend transition.*/
    private SetPreEnd        setpreend;

    /** Private field keeps the SetEnd object, containing the set of the processes
     * necessary to execute at end transition.*/
    private SetEnd           setend;

    /** Private field keeps the SetpostEnd object, containing the set of the processes
     * necessary to execute at postend transition.*/
    private SetPostEnd       setpostend;

    /** Private field keeps the SetInPause object, containing the set of the processes
     * necessary to execute at pause transition.*/
    private SetInPause       setinpause;

    /** Private field keeps the SetInResume object, containing the set of the processes
     * necessary to execute at resume transition.*/
    private SetInResume      setinresume;

    /** Private field keeps the SetReset object, containing the set of the processes
     * necessary to execute at reset transition.*/
    private SetReset         setreset;

    /** State of the run. By default set to be "reseted" */
    private volatile String _RunState = rcConstants.CMD_reset_ok;

    /** This flag is used to prevent update from the physical clients for pre and post transitions.*/
    private volatile boolean updateFromClient = true;

    /** Current runcontrol run number from the CODA database. */
//    public String  _RunNumber = "0";

    /** Current runcontrol run number from the .runnumber file in the $COOL_HOME dir. */
    public String  _RunNumberF = "0";

    /** Object that transforms a message's content to and from a specific language
     * representation. The language in this case is SL0. */
    private Codec  codec = new SLCodec();

    /** The name of the runcontrol session. */
    private String  _session;

    /** The name of the runcontrol run type. */
    private String  _runtype;
    private String  _origRunType;

    /** Private variable controled by the UI.
     * It controls autostart of the run, described in the cool file. */
    private boolean  _guiAutoStart   = true;

    /** Control system Option ontology object obtained during the differentiate process */
    private CoolOption _myOption;

    /** Subcontrol systems for this system*/
    private CopyOnWriteArraySet<CoolSystem> _mySubSystems = new CopyOnWriteArraySet<CoolSystem>();

    /** Hashtable containing the read-out-controller name as the key and its status (String) as the value. */
    private ConcurrentHashMap<String, String>  Rocs = new ConcurrentHashMap<String, String>();

    /** Hashtable containing the trigger supervisor name as the key and its status (String) as the value. */
    private ConcurrentHashMap<String, String>  Tss = new ConcurrentHashMap<String, String>();

    /** ConcurrentHashMap containing the event builder name as the key and its status (String) as the value. */
    private ConcurrentHashMap<String, String>  Ebs = new ConcurrentHashMap<String, String>();

    /** ConcurrentHashMap containing the event recorder name as the key and its status (String) as the value. */
    private ConcurrentHashMap<String, String>  Ers = new ConcurrentHashMap<String, String>();

    /** ConcurrentHashMap containing the slow control name as the key and its status (String) as the value. */
    private ConcurrentHashMap<String, String>  Slcs = new ConcurrentHashMap<String, String>();

    /** ConcurrentHashMap containing the analyzer name as the key and its status (String) as the value. */
    private ConcurrentHashMap<String, String>  Anas = new ConcurrentHashMap<String, String>();

    /** ConcurrentHashMap containing (String) the coda component name as the key and its status (String)
     * as the value. This ConcurrentHashMap contains all the coda critical components.*/
    private ConcurrentHashMap<String, String>  DAQ = new ConcurrentHashMap<String, String>();

    /** ConcurrentHashMap containing (String) the registered component name as the key and its
     * state (String) as the value.
     * This ConcurrentHashMap contains all the components, (including coda, slow control, analyses, SMS, etc.)
     * stored in the other ConcurrentHashMaps.*/
    private ConcurrentHashMap<String, String>  ALL = new ConcurrentHashMap<String, String>();

    /** ConcurrentHashMap containing (String), registered component name as the key and the
     * (Integer) number of fails, i.e. how many times component was ot responding during the
     * specified timeout period.*/
    private ConcurrentHashMap<String, Integer>  Fails = new ConcurrentHashMap<String, Integer>();

    /** ConcurrentHashMap containing String key-the name of the process (agent:statistics),
     * and Vector as a value, containing Producer, CodaTimeStamp and CoolData objects.*/
    private ConcurrentHashMap<String, Vector>  MonitorProcesses = new ConcurrentHashMap<String, Vector>();

    /** This is a container of ConcurrentHashMap objects each of which contain the name
     * (key) and status (value) of all components for different priorities.
     * The key of this ConcurrentHashMap is the priority.*/
    private ConcurrentHashMap<Integer, ConcurrentHashMap<String,String>> Components = new ConcurrentHashMap<Integer, ConcurrentHashMap<String,String>>();

    /** Hash map of component names and their types */
    private ConcurrentHashMap<String, String> ComponentTypes = new ConcurrentHashMap<String, String>();

    /** HashMap contains the names and corresponding reporting intervals for configured components*/
    private ConcurrentHashMap<String,String> CompReportInterval = new ConcurrentHashMap<String,String>();

    /** This sorted set will keep the sorted priorities which will later be used to access the Components*/
    private SortedSet<Integer> Priorities = new TreeSet<Integer>();

    /** This map keeps the refference information (command-name as a String, thread as an object)
     * about the cyclic running, monitoring threads, attached to this RCS component */
    private ConcurrentHashMap<String, Thread> cyclicThreads = new ConcurrentHashMap<String, Thread>();

    /** Instance of the class, incapsulating the Unix shell IPC commands*/
    private SysExe mySysExe;

    /** Object of the run statistics cMsg broadcasting thread (inner class). The usual subscribers are GUIs*/
    private MonitorOn myMonitorOn;

    /** List containing the names of the SMS agents*/
    private List<String> _mySMS = new ArrayList<String>();

    /** Indicats that state machine rules are registered*/
    private boolean SM_RULES_ACCEPTED = false;

    /**
     * This method is for application specific startup code. When this method is
     * called the agent has been already registered with the Agent Platform AMS
     * and is able to send and receive messages. However, no behaviour scheduling is
     * active yet. This method is essential to at least add a behaviour to the
     * agent, in order for it to be able to do anything.
     */
    protected void setup() {

// Register the supported content language
        getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);

// Register the supported ontologi
        getContentManager().registerOntology(CodaRunControlOntology.getInstance());

// This is the given name when the agent was started by the Hepatic.
        myName = getLocalName();
        _runtype = myName;

// Limit agent internal queue size
        this.setQueueSize(2000);

// Adds the responder behaviour, which will be scheduled for the execution in the agent thread.
// This is the behaviour where we process request information from the other agents. The actual
// processing method is defined in this class but called from the GenResponderBehaviour class.
        addBehaviour(new GenRespondBehaviour(this, verbose));
    }


    /**
     * Sets the _RunState
     *
     * @param st            State of the run.
     */
    public synchronized void setRunState(String st){
        _RunState = st;
    }


    /**
     * Returns the _RunState
     *
     * @return _RunState     The state of the run.
     */
    public synchronized String getRunState() {
        return _RunState;
    }


    /**
     * Reads the file ".runnumber" in the $COOL_HOME dir
     *
     * @return runnumber        String value of the runnumber or "error"
     */
    public String getRunNumber(){

        String rn;

// Open .runnumber file in the $COOL_HOME dir and read the runnumber
        try {
            BufferedReader br = new BufferedReader(new FileReader(System.getenv("COOL_HOME")+"/.runnumber"));
            if((rn = br.readLine())!=null) {
                br.close();
            }
        } catch (IOException e) {
            e.printStackTrace();
            rn = "error";
        }
        return rn;
    }


    /**
     * Reads and increments the runnumber in the ".runnumber" file located in the $COOL_HOME dir.
     *
     * @return runnumber
     */
    public String incrementRunNumber(){

        String rnew = null;
        Integer r;

// Get the runnumber from the file
        String rold = getRunNumber();
        if(!rold.equalsIgnoreCase("error")){

// Increment the runnumber
            r= Integer.valueOf(rold);
            r=r+1;
            rnew = r.toString();

// Write the string value of the incremented runnumber back into the file
            try {
                BufferedWriter bw = new BufferedWriter(new FileWriter(System.getenv("COOL_HOME")+"/.runnumber"));
                bw.write(rnew);
                bw.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        return rnew;
    }


    /**
     * Informs the state of the component to the receiver: SMS agent.
     * This method sends to the receiver a acceptState coda ontology ACL message, describing the producer
     * i.e. the components that produces this information (state) and a timestamp.
     *
     * @param stateName               The state name of the component
     * @param severity of the message
     */
    public void informStateSMS(String stateName, String severity) {
        String tmps;
        AID tmpaid;
        if(_mySMS!=null && !_mySMS.isEmpty()){
            for (String a_mySMS : _mySMS) {
                tmps = a_mySMS;

// Create the acl message object
                ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
                try {
                    getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                    tmpaid = new AID(tmps, AID.ISLOCALNAME);

// Preparing the message
                    msg.addReceiver(tmpaid);
                    msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                    msg.setOntology(CodaRunControlOntology.NAME);
                    msg.addUserDefinedParameter(myName, tmpaid.getLocalName());

// Create a codacomponent ontology concept object
                    CoolComponent cm = new CoolComponent();
                    cm.setComponent_Name(myName);
                    cm.setComponent_Type(rcConstants.rcSupervisorType);
                    cm.setComponent_Runtype(_runtype);
                    cm.setComponent_Session(_session);

// Create a Producer ontology concept object
                    Producer pr = new Producer();
                    pr.setProducer_Component(cm);

// Create a coda timestamp object
                    CodaTimeStamp tm = new CodaTimeStamp();
                    tm.setTimestamp_Value(new Date());

// Create coolState object
                    CoolState myState = new CoolState();
                    myState.setState_Name(stateName);
                    myState.setState_Isfor(cm);

// Create a acceptStatus ontology action object
                    AcceptState myaccept = new AcceptState();
                    myaccept.setAcceptstate_Producer(pr);
                    myaccept.setAcceptstate_Timestamp(tm);
                    myaccept.setAcceptstate_State(myState);
                    myaccept.setAcceptstate_Severity(severity);

// Create an action transport object
                    Action a = new Action();
                    a.setActor(tmpaid);
                    a.setAction(myaccept);

// Fill the message content
                    getContentManager().fillContent(msg, a);
                } catch (Exception e) {
                    e.printStackTrace();
                }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
                addBehaviour(new GenRequestBehaviour(this, msg, verbose));
            }
        }
    }


    /**
     * This method send an ACL ontology action message to an agent on the AFECS platform.
     *
     * @param transition               AgentAction object
     * @param component                Name of an agent on the platform
     */
    public void sendACLOntologyAction(AgentAction transition, String component) {

// Create an AID of the component going to receive the transition
        AID receiver = new AID(component,AID.ISLOCALNAME);

// Create an ACL message
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Use the coda runcontrol ontology
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Add receiver to the message
            msg.addReceiver(receiver);

// We are using fipa_sl0 content language
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);

// Set the ontology
            msg.setOntology(CodaRunControlOntology.NAME);

// Add user defined parameter. The user parameter is the name of this agent, i.e.
// control system supervisor name.
            msg.addUserDefinedParameter(getAID().getLocalName(),receiver.getLocalName());

// Create an action object
            Action a = new Action();

// Add actor to the action object
            a.setActor(receiver);

// Set an actual action, passed as a parameter to the action object
            a.setAction(transition);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        }
        catch(Exception e) {
            e.printStackTrace();
        }

// Create and add a behaviour to the message queue
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * This method is used to send to the Cortex a message,
     * asking to record the new run conditions for this session,
     * and return the new (incremented) runnumber.
     *
     * @param session               session name
     * @param runtype               runtype name, i.e. Medulla object name
     */
    public void recordNewRun(String session, String runtype) {

// Create an AID of the component going to receive the transition
        AID receiver = new AID("Cortex",AID.ISLOCALNAME);

// Create an ACL message
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Use the coda runcontrol ontology
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Add receiver to the message
            msg.addReceiver(receiver);

// We are using fipa_sl0 content language
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);

// Set the ontology
            msg.setOntology(CodaRunControlOntology.NAME);

// Add user defined parameter. The user parameter is the name of this agent, i.e. control system supervisor name
            msg.addUserDefinedParameter(getAID().getLocalName(),receiver.getLocalName());
            RecordNewRun rn = new RecordNewRun();
            rn.setRecordnewrun_Session(session);
            rn.setRecordnewrun_Runtype(runtype);

// Create an action object
            Action a = new Action();

// Add actor to the action object
            a.setActor(receiver);

// Set actual action, passed as a parameter to the action object
            a.setAction(rn);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        }
        catch(Exception e) {
            e.printStackTrace();
        }

// Create and add a behaviour to the message queue
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * This method uses cMsg connection directly to the physical clients
     * (bypassing clients representative agent on the platform) to set their reporting intervals.
     */
    public void setReportingIntervals(){

// Go over all registered component agents
        for(String tmpCompName : ALL.keySet()) {
            if ((tmpCompName != null) && CompReportInterval.containsKey(tmpCompName) )
                informCmsgClient(tmpCompName,
                        rcSubscription.SessionControlType+"/"+rcConstants.CMD_SetInterval,
                        true,CompReportInterval.get(tmpCompName),false, null);
        }
    }


    /**
     * Thi method will send to all agents in the same session the "exit" acl message,
     * which will be then forwarded through private P2P channel (cMsg rcdomain) to
     * the physical components. This will force physical components to die grasefully.
     */
    public void broadcastComponentExit(){

// AgentAction object
        AgentAction transConcept;

// Create exit object
        Exit  transition = new Exit();
        transition.setExit_Session(_session);
        transition.setExit_Runtype(_runtype);
        transConcept = transition;

// Go over all registered component agents
        for(String tmpCompName : ALL.keySet()) {

// Send reset transition to the agent.
            if (tmpCompName != null)
                sendACLOntologyAction(transConcept, tmpCompName);
        }
    }


    /**
     * Sends ACL message to all subcontrol system supervisors, asking to start reporting their statuses.
     */
    public void SubSystemsStartReporting(){

// AgentAction object
        AgentAction transConcept = null;

// Create a reset object
        ResultNotify  transition = new ResultNotify();
        transition.setResultNotify_Component(itsMe);
        transConcept = transition;

        // Go over all subsystems registered with this Medulla
        for(CoolSystem sys: _mySubSystems){
            // send acl message to the subcontrol system supervisor
            sendACLOntologyAction(transConcept,sys.getSystem_Supervisor());
        }
    }

    /**
     * Sends ACL message to all subcontrol system supervisors, asking to stop reporting their statuses.
     */
    public void SubSystemsStopReporting(){

// AgentAction object
        AgentAction transConcept = null;

// Create a reset object
        StopReporting  transition = new StopReporting();
        transConcept = transition;

        // Go over all subsystems registered with this Medulla
        for(CoolSystem sys: _mySubSystems){
            // send acl message to the subcontrol system supervisor
            sendACLOntologyAction(transConcept,sys.getSystem_Supervisor());
        }
    }

    /**
     * Iner platform communication.
     * This method is used to send coda transition to all CODA component agents. After sending transition
     * it will wait completion statuses from all components and will return the overal status of the coda
     * transition execution. We are sending coda transitons in a particular order, based on the priorities
     * of the components.
     * N.B. pre , post and reset transitions do not follow the component priorities
     *
     * @param trans_name                       The transition name (String)
     * @return                                 The status of the execution (boolean)
     */
    public boolean broadcastCodaTransition(String trans_name) {

// AgentAction object
        AgentAction transConcept = null;

// List of the agents names that are marked as being representatives of the failed components,
// and we are not going to wait for the complition of the desired transition from those agents.
        List<String> nowait = new ArrayList<String>();

// Status of the execution is set to false by default
        boolean succeeded = false;

// Component name
        String CompName;

// Tmp string
        String logingString;

// Hash table holding the components of the same priority
        ConcurrentHashMap samePriorityHash;


// Reset components ConcurrentHashMap by assigning status of the components to undefined
        resetComponentsStatus(null);

// ... RESET transition.
        if( trans_name.equals(rcConstants.CMD_reset)){

// Stop broadcasting monitored information.
// N.B. "end" related transitions will not stop the broadcasting thread
            if(myMonitorOn!=null){
                if(!trans_name.contains("end")) myMonitorOn.shouldrun = false;
            }

// Create a reset object
            Reset  transition = new Reset();
            transition.setReset_Session(_session);
            transition.setReset_Runtype(_runtype);
            transConcept = transition;

// Go over all registered component agents
            for(String tmpCompName : ALL.keySet()) {

// See if the component had failed
                if((Fails.containsKey(tmpCompName)) &&
                        (Fails.get(tmpCompName) >= rcConstants.FAILS_NUMBER)) {

// Add this component agent name into the list of nowait agents
                    nowait.add(tmpCompName);
                } else {

// Send reset transition to the agent.
                    if ((tmpCompName != null) && (transConcept !=null))
                        sendACLOntologyAction(transConcept, tmpCompName);
                }
            }

            succeeded = true;
/*
            if (!succeeded) {
                logingString = "Medulla-Error:\n "+trans_name + " failed";

// Log the message and inform to all plexuses, i.e. all user components
                LogIt(logingString);

// Inform plexuses
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,logingString, false,null);
                dalogMsg("error",logingString);
                return false;
            }
*/
        }


// ... PRE AND POST transitions will be getting all agents, indipendent of their priorities.
        else if( (!trans_name.equals(rcConstants.CMD_prestart)) &&
                (trans_name.contains("pre") || trans_name.contains("post"))){

// Check the transition name
            if(trans_name.equals(rcConstants.CMD_preconfigure)) {
                PreConfigure transition = new PreConfigure();
                transition.setPreconfigure_Session(_session);
                transition.setPreconfigure_Runtype(_runtype);
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_postconfigure)) {
                PostConfigure transition = new PostConfigure();
                transition.setPostconfigure_Session(_session);
                transition.setPostconfigure_Runtype(_runtype);
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_predownload)) {
                PreDownload transition = new PreDownload();
                transition.setPredownload_Session(_session);
                transition.setPredownload_Runtype(_runtype);
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_postdownload)) {
                PostDownload transition = new PostDownload();
                transition.setPostdownload_Session(_session);
                transition.setPostdownload_Runtype(_runtype);
                transConcept = transition;

// At the post-download we send reporting intervals
                setReportingIntervals();
            }

// ... "preprestart" transition request. N.B. At first we increment the runnumber
            else if(trans_name.equals(rcConstants.CMD_preprestart)) {

// Get runnumber from the database. Backward compatibility
//                try {
                    if(_myOption.getOption_Hascoda2component().equalsIgnoreCase("true")){

// Increment the runnumber in the database
//                        int r = DatabaseCommunication.getRunNumber(_session);
//                        r=r+1;
//                        _RunNumber = Integer.toString(r);
//                        DatabaseCommunication.setRunNumber(r, _session);
                    }

// Request to Cortex to record the new run in this session, and ask for runnumber
                String rtp = _runtype.substring(_session.length()+1);
                    recordNewRun(_session, rtp);
//                } catch (RCException e) {
//                    informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,
//                            "Medulla-Error:\n DB communication Error",false,null);
//                    dalogMsg("error","DB communication error");
//
//                    e.printStackTrace();
//                    return false;
//                }
                PrePrestart transition = new PrePrestart();
                transition.setPreprestart_Session(_session);
                transition.setPreprestart_Runtype(_runtype);
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_postprestart)) {
                PostPrestart transition = new PostPrestart();
                transition.setPostprestart_Session(_session);
                transition.setPostprestart_Runtype(_runtype);
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_prego)) {
                PreGo transition = new PreGo();
                transition.setPrego_Session(_session);
                transition.setPrego_Runtype(_runtype);
                transition.setPrego_Runnumber(Integer.parseInt(_RunNumberF));
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_postgo)) {
                PostGo transition = new PostGo();
                transition.setPostgo_Session(_session);
                transition.setPostgo_Runtype(_runtype);
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_preend)) {
                PreEnd transition = new PreEnd();
                transition.setPreend_Session(_session);
                transition.setPreend_Runtype(_runtype);
                transConcept = transition;
            }
            else if(trans_name.equals(rcConstants.CMD_postend)) {
                PostEnd transition = new PostEnd();
                transition.setPostend_Session(_session);
                transition.setPostend_Runtype(_runtype);
                transConcept = transition;

// ... Unknown transition name
            } else {
                logingString = "Medulla-Error: Unknown transition \n "+trans_name;

// Log the message and inform to all plexuses, i.e. all user interface components
                LogIt(logingString);

// Inform plexuses
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,logingString, false,null);
                dalogMsg("error",logingString);
                return false;
            }

// Go over all registered component agents
            for(String tmpCompName : ALL.keySet()) {

// See if the component had failed
                if((Fails.containsKey(tmpCompName)) &&
                        (Fails.get(tmpCompName) >= rcConstants.FAILS_NUMBER)) {

// Add this component agent name into the list of nowait agents
                    nowait.add(tmpCompName);
                } else {

// Send the pre/post transition to the agent.
                    if ((tmpCompName != null) && (transConcept !=null))
                        sendACLOntologyAction(transConcept, tmpCompName);
                }
            }

// Wait the completion messages back.
            //@todo changed ALL to DAQ at 08/04/2008            
            succeeded = wait4Completion(trans_name, rcConstants.TIMEOUT, DAQ, nowait);
            if (!succeeded) {
                logingString = "Medulla-Error:\n "+trans_name + " failed";

// Log the message and inform to all plexuses, i.e. all user components
                LogIt(logingString);

// Inform plexuses
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,logingString, false,null);
                dalogMsg("error",logingString);
                return false;
            }
        }

// ... basic CODA transitions
        else {
// end, pause and reset CODA transitions will follow opposit priority schema
            if(trans_name.equalsIgnoreCase(rcConstants.CMD_end) ||
                    trans_name.equals(rcConstants.CMD_pause) ||
                    trans_name.equals(rcConstants.CMD_resume)) {
                if(trans_name.equalsIgnoreCase(rcConstants.CMD_end)){
                    End  transition = new End();
                    transition.setEnd_Session(_session);
                    transition.setEnd_Runtype(_runtype);
                    transConcept = transition;
                }
                else if(trans_name.equals(rcConstants.CMD_pause)) {
                    Pause transition = new Pause();
                    transition.setPause_Session(_session);
                    transition.setPause_Runtype(_runtype);
                    transConcept = transition;
                }
                else if(trans_name.equals(rcConstants.CMD_resume)) {
                    Resume transition = new Resume();
                    transition.setResume_Session(_session);
                    transition.setResume_Runtype(_runtype);
                    transConcept = transition;
                }

// Stores the priority
                int prioritygroup;

// From the Priority table get an array of integers representing reverse order of the priorities
// Get the size of the Priorities sorted set
                int s = Priorities.size();
                int j = s;

// Create reverse priorities array
                Integer[] reversePriority = new Integer[j];

// Iterate over the Priorities sorted set
                for (Integer Priority : Priorities) {

// Fill the reverse order array
                    reversePriority[--j] = Priority;
                }

// March through the reverse priorities
                for(int i=0;i<s;i++) {

// Get the priority
                    prioritygroup = reversePriority[i];

// From the Components table using the priority as a key get the ConcurrentHashMap
// of component names and statusses.
                    samePriorityHash = Components.get(prioritygroup);
                    if(!samePriorityHash.isEmpty()){

// Iterate through the all processses for specific priority
                        for (Object o : samePriorityHash.keySet()) {

// Get the component name
                            CompName = (String) o;

// See if the component failed
                            if((Fails.containsKey(CompName)) &&
                                    (Fails.get(CompName) >= rcConstants.FAILS_NUMBER)) {

// Add this component agent name into the list of nowait agents
                                nowait.add(CompName);
                            } else {

// Send reset command to the agent in case it is not failed.
                                if ((CompName != null) && (transConcept !=null))
                                    sendACLOntologyAction(transConcept, CompName);
                            }
                        }

// For the pause and resume transitions we need to send message to the highest priority group only.
// So if it is the case we need to return, becasuse there is no need to pause the lower
// priority components. For e.g. TS being the highest priority component will execute the pause by simply
// disabling "go" bit of the TS. For now we do not need to wait successful completion.
                        if((trans_name.equalsIgnoreCase(rcConstants.CMD_pause)) ||
                                (trans_name.equalsIgnoreCase(rcConstants.CMD_resume))) return true;

// We send end transitions to all components within the same priority group.
// Now we need to wait for the completion.
// We pass as a parameter the hsatable containing the name of the component
// and their statusses for the same priority.
                        succeeded = wait4Completion(trans_name, rcConstants.TIMEOUT, samePriorityHash, nowait);

// If any of the priority group component fails to end return false
                        if(!succeeded) {
                            logingString = "Medulla-Error: \n"+prioritygroup+" components "+ trans_name+" failed";

// Log the message and inform to all plexuses, i.e. all user components
                            LogIt(logingString);

// Send message to all user components
                            informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,logingString,false,null);
                            dalogMsg("error",logingString);
                            return false;
                        }
                    }
                }
            }

// ... download, prestart and go will follow the priority schema described in the cool
            else {

// Check the transition name
                if(trans_name.equals(rcConstants.CMD_configure)) {
                    Configure transition = new Configure();
                    transition.setConfigure_Session(_session);
                    transition.setConfigure_Runtype(_runtype);
                    transConcept = transition;
                }
                else if(trans_name.equals(rcConstants.CMD_download)) {
                    Download transition = new Download();
                    transition.setDownload_Session(_session);
                    transition.setDownload_Runtype(_runtype);
                    transConcept = transition;
                }
                else if(trans_name.equals(rcConstants.CMD_prestart)) {
                    Prestart transition = new Prestart();
                    transition.setPrestart_Session(_session);
                    transition.setPrestart_Runtype(_runtype);
                    transConcept = transition;
                }
                else if(trans_name.equals(rcConstants.CMD_go)) {
                    Go transition = new Go();
                    transition.setGo_Session(_session);
                    transition.setGo_Runtype(_runtype);
                    transConcept = transition;
                }
                else {
                    logingString = "Medulla-Error: Unknown transition \n "+trans_name;

// Log the message and inform to all plexuses, i.e. all user components
                    LogIt(logingString);

// Inform plexuses
                    informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,logingString, false,null);
                    dalogMsg("error",logingString);
                    return false;
                }

// Get the iterator on the priorities sorted set
                for (Integer prioritygroup : Priorities) {

// Get the hash table of the components in the same priority group
                    samePriorityHash = Components.get(prioritygroup);

// Get iterator on that ConcurrentHashMap
                    for (Object o : samePriorityHash.keySet()) {

// Get the component agent name
                        CompName = (String) o;

// See if the component failed
                        if( (Fails.containsKey(CompName)) &&
                                (Fails.get(CompName) >= rcConstants.FAILS_NUMBER)) {

// Add this component agent name into the list of nowait agents
                            nowait.add(CompName);

                        } else {

// Send the transition to the agent in case it is not failed.
                            if ((CompName != null) && (transConcept !=null))
                                sendACLOntologyAction(transConcept, CompName);
                        }

// Log the action, by sending cMsgMessage, subject: loging_subject, type: loging_type
                        String tmp = myName + ": INFO => sent " + trans_name + " TO " + CompName;
                        LogIt(tmp);
                    }

// Now that we sent to all the components in the same priority group requested transition,
// wait the completion messages back.
                    succeeded = wait4Completion(trans_name, rcConstants.TIMEOUT, samePriorityHash, nowait);
                    if (!succeeded) {
                        logingString = "Medulla-Error:\n priority group = " + prioritygroup + " components " + trans_name + " failed";

// Log the message and inform to all plexuses, i.e. all user components
                        LogIt(logingString);
                        informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,logingString, false,null);
                        dalogMsg("error",logingString);
                        return false;
                    }

                } // loop for the normal priorities
            }//download, prestart, go
        } // end basic coda transitions
        return succeeded;
    }


    /**
     * This method is used to wait for the completion of a transition
     * after the command to start the transition has been sent to a group of the CODA
     * component agents with the same priority.
     *
     * @param transaction    name of the particular transition to monitor
     * @param timeout        amount of time (int) to wait before returning a transition failure
     * @param hash           ConcurrentHashMap containing ComponntName and Status
     * @param exclude        ArrayList of component names that we do not have to  wait
     * @return execution     status (bollean).
     */
    public boolean wait4Completion(String transaction, int timeout, ConcurrentHashMap hash, List exclude){

// Temporary variables for componanent name and some string
        String tmpStatus, tmpCompName, tmpString;

// Returne value
        boolean success = true;

// Loop for i=timeout time
        for (int i=1; i<=timeout;i++) {

// For every new iteration set success to true
            success = true;

// Get an iterator on the hash table of the component statuses passed as a parameter.
// This is a sinchronized table which will be updated when the component will report his
// status to Medulla (this agent).
            for (Object o : hash.keySet()) {

// Get the name of the component
                tmpCompName = (String) o;

// Get the status
                tmpStatus = (String) hash.get(tmpCompName);

// See if we do not have to wait for this component status. Excluded components are
// the components that fail to report within given timeout.
                if ((exclude != null) && (exclude.contains(tmpCompName))) {

// In this case we return success wihtout actually waiting for the component
                } else {

// N.B. Successful execution of the transition by the component agent means sending back
// status = transition
                    if (!tmpStatus.equalsIgnoreCase(transaction)) {

// The recorded status in the ConcurrentHashMap is not what we are looking for
                        success = false;

// Every 5sec report to the user that we still are waiting for the completion of the transition
                        if ((i % rcConstants.PREPROST_NOTREPORTING) == 0) {

// As usual at pre or post transitions we are executing external processes
                            if (((transaction.startsWith("pre")) &&
                                    (!transaction.equalsIgnoreCase("prestart"))) ||
                                    (transaction.startsWith("post"))) {
                                tmpString = "Medulla_Warning: \n"+ i / 10 +
                                        " sec. waiting for external process, executed by the "+tmpCompName+
                                        " to complete. "+tmpStatus;
                            } else {
                                tmpString = "Medulla_Warning: \n" + i / 10 + " sec. waiting " +
                                        tmpCompName + " to " + transaction;
                            }

// Log the message and inform to all plexuses, i.e. all user interface components
                            LogIt(tmpString);
// Send message to all user components
                            informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,tmpString,false,null);
                            dalogMsg("warning",tmpString);
                        }
                    }
                }
            }

// For all components in the priority group if we have satisfing response return true
            if(success) return true;

// Otherwise wait for a sec.
            else{
                try {
                    Thread.sleep(100);
                } catch (InterruptedException e) {e.printStackTrace();}
            }

// If we got "reset" request from the gui then abort waiting the transition and return false.
            if(_ResetRequest) return false;
            else if(getRunState().equalsIgnoreCase(rcConstants.CMD_end_ok))return true;

        } // end of the loop  waiting for timeout
        return success;
    }


    /**
     * This method is used to reset the component statuses ( by setting them rcConstants.undefined)
     * in the <em>Components</em> hash table OR will reset the passed ConcurrentHashMap with the
     * String,String structure.
     * When we are starting a new transition we will get the component status information directly
     * from the component agent, and will update the <em>Components</em> ConcurrentHashMap accordingly.
     *
     * @param htable          ConcurrentHashMap of the components (name-String, status-String)
     *                        need to be reset.
     *                        if ConcurrentHashMap is passed as null, then this method will make all the
     *                        component statuses rcConstants.Udefined.
     */
    public void resetComponentsStatus(ConcurrentHashMap<String,String> htable) {

// Tmp string for component name (the same as agent name)
        String tmpCompName;

// Tmp ConcurrentHashMap
        ConcurrentHashMap<String,String> th;

// In Case no ConcurrentHashMap is passed as a parameter
        if(htable==null){

// Iterate over the priorities
            for (Integer prioritygroup : Priorities) {

// Get the hastable of all components within the same priority group.
                th = Components.get(prioritygroup);

// Iterate over that ConcurrentHashMap
                for (Object o : th.keySet()) {

// Get component agent name
                    tmpCompName = (String) o;

// Assign undefined the status for that component
                    th.put(tmpCompName, rcConstants.undefined);
                }
            }

// Deal with the passed ConcurrentHashMap
        } else {

// Iterate over that ConcurrentHashMap
            for (Object o : htable.keySet()) {

// Get component agent name
                tmpCompName = (String) o;

// Assign undefined the status for that component
                htable.put(tmpCompName, rcConstants.undefined);
            }
        }
    }

    /**
     * Sending dalog message
     * @param severity      severity of the message
     * @param message       content of the message
     */
    private void dalogMsg(String severity, String message){
        String subject=null;
        String type = null;
        cMsgPayloadItem sever = null, state = null;

        if(rcSubscription.expid!=null && _session!=null && _origRunType!=null ){
            cMsgMessage msg = new cMsgMessage();
            subject = rcSubscription.expid+"/"+_session+"/"+_origRunType+"/"+rcConstants.rcSupervisorType +"/"+myName;
            try {
                // payload item state
                state = new cMsgPayloadItem("state",getRunState());
                if (severity.equalsIgnoreCase("info")) {
                    type = "dalog/1";
                    sever = new cMsgPayloadItem("severity","INFO.1");

                }
                else if(severity.equalsIgnoreCase("warning")) {
                    type = "dalog/5";
                    sever = new cMsgPayloadItem("severity","WARN.5");
                }
                else if(severity.equalsIgnoreCase("error")) {
                    type = "dalog/9";
                    sever = new cMsgPayloadItem("severity","ERROR.9");
                }
                else if(severity.equalsIgnoreCase("severeerror")){
                    type = "dalog/13";
                    sever = new cMsgPayloadItem("severity","ERRORS.13");
                }
                if(subject!=null && type !=null){
                    cMsgPayloadItem p_expid = new cMsgPayloadItem("EXPID",rcSubscription.expid);
                    cMsgPayloadItem p_session = new cMsgPayloadItem("session",_session);
                    cMsgPayloadItem p_config = new cMsgPayloadItem("config",_origRunType);
                    cMsgPayloadItem p_type = new cMsgPayloadItem("codaClass",rcConstants.rcSupervisorType);
                    cMsgPayloadItem p_codaName = new cMsgPayloadItem("codaName",myName);
                    msg.addPayloadItem(p_expid);
                    msg.addPayloadItem(p_session);
                    msg.addPayloadItem(p_config);
                    msg.addPayloadItem(p_codaName);
                    msg.addPayloadItem(p_type);

                    msg.setSubject(subject);
                    msg.setType(type);
                    msg.setText(message);
                    msg.addPayloadItem(sever);
                    msg.addPayloadItem(state);
                    cMsgSrv.send(msg);
                    cMsgSrv.flush(1);
                }
            } catch (cMsgException e) {
                e.printStackTrace();
            }

//        System.out.println(subject+" "+type);
        }

    }

    /**
     * This method sends the status type information to cMsg clients
     * (plexuses or user interfaces, physical components. etc.) through cMsg (cMsg domain)ipc
     *
     * @param subject         Subject of the message.  For plexus casses indicates the ceverity
     *                        of the message. for e.g. ERROR, WARNING, INFO, etc.
     *                        N.B. subject is constructed having preceeding "_session"   
     * @param type :          Type of the message. In some casses (namely for plexuses)
     *                        this indicates the name of the transition during which the
     *                        message was generated.
     * @param sendText        Boolean indicates that the we are sending the text message
     * @param info            Message content (text) 
     * @param sendBinary      Boolean indicates that the we are sending the binary message
     * @param binArray        Binary content of the message
     */
    private void informCmsgClient(String subject, String type,
                                  boolean sendText, String info, boolean sendBinary, byte[] binArray) {

// Create the message
        cMsgMessage msg = new cMsgMessage();
        msg.setSubject(subject);
        msg.setType(type);

// Text
        if (sendText) {
            msg.setText(info);
        }

// Binary
        else if (sendBinary) {
            System.out.println("Sending byte array\n");
            if(binArray!=null) msg.setByteArrayNoCopy(binArray);
        }
        try {
            cMsgSrv.send(msg);
            cMsgSrv.flush(1);
        } catch (cMsgException e) {
            e.printStackTrace();
        }
    }

    /**
     * Sends cmsg message to all registered components
     * @param type
     * @param txt
     */
    private void broadcastCmsgClients(String type, String txt){

// Go over all registered component agents
        for(String tmpCompName : ALL.keySet()) {
// Create the message
            cMsgMessage msg = new cMsgMessage();
            msg.setSubject(tmpCompName);
            msg.setType(type);

// Text
            if (txt!=null) {
                msg.setText(txt);
            }
            try {
                cMsgSrv.send(msg);
                cMsgSrv.flush(1);
            } catch (cMsgException e) {
                e.printStackTrace();
            }

        }
    }

    /**
     * This method sends a cMsg message to the loging services using cMsg domain server
     *
     * @param txt             loging string
     */
    private void LogIt(String txt){
        cMsgMessage msg = new cMsgMessage();
        msg.setSubject(LogSubject);
        msg.setType(LogType);
        msg.setText(txt);
        try {
            cMsgSrv.send(msg);
            cMsgSrv.flush(1);
        } catch (cMsgException e) {
            e.printStackTrace();
        }
    }


    /**
     * Informs user interface components about the failed client.
     * It also starts run end sequence.
     *
     * @param componentName           The name of the failed component
     */

    public void endrun(String componentName) {

// Create a string message content, which will be sent to all plexuses
        String endinfo = "Medulla-SevereError:\n Component "+ componentName+
                " is not reporting. Click End button to end the run.\n "+
                "Attention: Before starting the new run do not forget to reset "+componentName+
                "\n Start the new run by clicking the Reset button first.";

        informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,endinfo,false,null);

        dalogMsg("severeerror",endinfo);
// Calling this method will just inform UI. If you need really to end the run uncoment the following line.
//      startEnd();
    }


    /**
     * This method will start broadcasting the coda transitions to the component agents in the separate threads.
     * External processes attached to this agent ( Medulla ) will be executed in the separate threads as well.
     * Based on the proceses defined in the cool configuration file this method will start working threads to
     * satisfy the configuration requirements.
     *
     * @param trext                 TransitionExtention object created during the Cortex teaching process
     *                              (setPreconfigure, etc..). This object contains the <i>process</i> ontology
     *                              concept, which is attached to the CODA transition.
     * @param transition            CODA state transition
     */
    private void organizeCodaTransition(TransitionExtention trext, String transition){

        String commandName;

// Status of the trasnition execution
        boolean trstatus;
        int loop=0;
        Iterator it;


// Check if Medulla is programmed  to do something during this transition, i.e. Cortex told this agent
//  to execute the process described in the cool configuration file, attached to this particular transition.
// N.B. associated processes will use shell commands (command line interface) only.
        if((trext!=null) && (trext.getProcesses())!=null) {

// Iterate over all registered processes with this transition
            it = trext.getProcesses().iterator();
            while(it.hasNext()) {

// Get the process
                CoolProcess pr = (CoolProcess)it.next();

// Get the command name accociated with this process
                commandName = pr.getProcess_Commandname();

// Get the periodisity of the process
                if(pr.getProcess_Periodicity()!=null)loop = pr.getProcess_Periodicity();

// Run the process n = loop time
                if(loop>=0){
                    RunProcess myRun = new RunProcess(transition, pr);
                    myRun.start();
                }

// Negative loop means loop for ever. We need to get handle on this threads to stop it gracefully.
                else {
                    if(!cyclicThreads.containsKey(myName +":"+commandName)){
                        RunProcess myRun = new RunProcess(transition, pr);
                        cyclicThreads.put(myName +":"+commandName, myRun);
                        myRun.start();
                    }
                }
            }
        }

// Check if this is reset transition and set _ResetRequest
        _ResetRequest = (transition.equalsIgnoreCase(rcConstants.CMD_reset)) ||
                (transition.equalsIgnoreCase(rcConstants.CMD_autoreset));

// Now start broadcasting coda transition requests to the all component agents in the separate thread.
        if(transition.equalsIgnoreCase(rcConstants.CMD_preconfigure)){
            broadcastCodaTransition(rcConstants.CMD_preconfigure);
        } else if (transition.equalsIgnoreCase(rcConstants.CMD_configure)) {
            broadcastCodaTransition(rcConstants.CMD_configure);
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_postconfigure)){
            broadcastCodaTransition(rcConstants.CMD_postconfigure);
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_predownload)){

// ... Start predownload
            trstatus = broadcastCodaTransition(rcConstants.CMD_predownload);
            if(trstatus){
                setRunState(rcConstants.CMD_predownload_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_predownload_ok,false,null);
                dalogMsg("info",rcConstants.CMD_predownload_ok);
                informStateSMS(rcConstants.CMD_predownload_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Predownload_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:predownload_failed",false,null);
                dalogMsg("error","predownload_failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_download)){

// ... Start download
            trstatus =  broadcastCodaTransition(rcConstants.CMD_download);
            if(trstatus){
                setRunState(rcConstants.CMD_download_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_download_ok,false,null);
                dalogMsg("info",rcConstants.CMD_download_ok);
                informStateSMS(rcConstants.CMD_download_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Download_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:download_failed",false,null);
                dalogMsg("error","download failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_postdownload)){

// ... Start postdownload
            trstatus =  broadcastCodaTransition(rcConstants.CMD_postdownload);
            if(trstatus){
                setRunState(rcConstants.CMD_postdownload_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_postdownload_ok,false,null);
                dalogMsg("info",rcConstants.CMD_postdownload_ok);
                informStateSMS(rcConstants.CMD_postdownload_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Postdownload_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:postdownload_failed",false,null);
                dalogMsg("error","postdownload failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_preprestart)){

// Reset the Fails ConcurrentHashMap
            for (String s : Fails.keySet()) {
                Fails.put(s,0);
            }

// ... Start preprestart
            trstatus =  broadcastCodaTransition(rcConstants.CMD_preprestart);
            if(trstatus){
                setRunState(rcConstants.CMD_preprestart_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_preprestart_ok,false,null);
                dalogMsg("info",rcConstants.CMD_preprestart_ok);
                informStateSMS(rcConstants.CMD_preprestart_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Preprestart_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:preprestart_failed",false,null);
                dalogMsg("error","preprestart failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_prestart)) {

// ... Start prestart
            trstatus =  broadcastCodaTransition(rcConstants.CMD_prestart);
            if(trstatus){
                setRunState(rcConstants.CMD_prestart_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_prestart_ok,false,null);
                dalogMsg("info",rcConstants.CMD_prestart_ok);
                informStateSMS(rcConstants.CMD_prestart_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Prestart_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:prestart_failed",false,null);
                dalogMsg("error","prestart failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_postprestart)) {

// ...Start postprestart
            trstatus =  broadcastCodaTransition(rcConstants.CMD_postprestart);
            if(trstatus){
                setRunState(rcConstants.CMD_postprestart_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_postprestart_ok,false,null);
                dalogMsg("info",rcConstants.CMD_postprestart_ok);
                informStateSMS(rcConstants.CMD_postprestart_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Postprestart_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:postprestart_failed",false,null);
                dalogMsg("error","postprestart failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_prego)){

// ...Start prego
            trstatus =  broadcastCodaTransition(rcConstants.CMD_prego);
            if(trstatus){
                setRunState(rcConstants.CMD_prego_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_prego_ok,false,null);
                dalogMsg("info",rcConstants.CMD_prego_ok);
                informStateSMS(rcConstants.CMD_prego_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Prego_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:prego_failed",false,null);
                dalogMsg("error","prego failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_go)) {

// ... Start go
            trstatus =  broadcastCodaTransition(rcConstants.CMD_go);
            if(trstatus){
                setRunState(rcConstants.CMD_go_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_go_ok,false,null);
                dalogMsg("info",rcConstants.CMD_go_ok);
                informStateSMS(rcConstants.CMD_go_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Go_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:go_failed",false,null);
                dalogMsg("error","go failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_postgo)) {

// ... Start postgo
            trstatus =  broadcastCodaTransition(rcConstants.CMD_postgo);
            if(trstatus){
                setRunState(rcConstants.CMD_postgo_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_postgo_ok,false,null);
                dalogMsg("info",rcConstants.CMD_postgo_ok);
                informStateSMS(rcConstants.CMD_postgo_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Postgo_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:postgo_failed",false,null);
                dalogMsg("error","postgo failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_preend)){

// ...Start preend
            trstatus =  broadcastCodaTransition(rcConstants.CMD_preend);
            if(trstatus){
                setRunState(rcConstants.CMD_preend_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_preend_ok,false,null);
                dalogMsg("info",rcConstants.CMD_preend_ok);
                informStateSMS(rcConstants.CMD_preend_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Preend_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:preend_failed",false,null);
                dalogMsg("error","preend failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_end)){

// ... Start end
            trstatus =  broadcastCodaTransition(rcConstants.CMD_end);
            if(trstatus){
                setRunState(rcConstants.CMD_end_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_end_ok,false,null);
                dalogMsg("info",rcConstants.CMD_end_ok);
                informStateSMS(rcConstants.CMD_end_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("End_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:end_failed",false,null);
                dalogMsg("error","end failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if (transition.equalsIgnoreCase(rcConstants.CMD_postend)) {

// ... Start postend
            trstatus =  broadcastCodaTransition(rcConstants.CMD_postend);
            if(trstatus){
                setRunState(rcConstants.CMD_postend_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_postend_ok,false,null);
                dalogMsg("info",rcConstants.CMD_postend_ok);
                informStateSMS(rcConstants.CMD_postend_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Postend_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:postend_failed",false,null);
                dalogMsg("error","postend failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
        else if(transition.equalsIgnoreCase(rcConstants.CMD_reset)){

// Reset the Fails ConcurrentHashMap
            for (String s : Fails.keySet()) {
                Fails.put(s,0);
            }

// ... Start reset
            trstatus =  broadcastCodaTransition(rcConstants.CMD_reset);
            if(trstatus){
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Reset_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:reset_failed",false,null);
                dalogMsg("error","reset failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
// ... Auto reset
        else if(transition.equalsIgnoreCase(rcConstants.CMD_autoreset)){

// reset the Fails ConcurrentHashMap
            for (String s : Fails.keySet()) {
                Fails.put(s,0);
            }

// Start actual reset
            trstatus =  broadcastCodaTransition(rcConstants.CMD_reset);
            if(trstatus){
                setRunState(rcConstants.CMD_reset_ok);
            } else {
                setRunState("Reset_failed");
            }
        }
// ... Start pause
        else if (transition.equalsIgnoreCase(rcConstants.CMD_pause)) {

            trstatus =  broadcastCodaTransition(rcConstants.CMD_pause);
            if(trstatus){
                setRunState(rcConstants.CMD_pause_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_pause_ok,false,null);
                dalogMsg("info",rcConstants.CMD_pause_ok);
                informStateSMS(rcConstants.CMD_pause_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Paused_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:pause_failed",false,null);
                dalogMsg("error","pause failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }

// ... Start resume
        else if (transition.equalsIgnoreCase(rcConstants.CMD_resume)){
            trstatus =  broadcastCodaTransition(rcConstants.CMD_resume);
            if(trstatus){
                setRunState(rcConstants.CMD_postgo_ok);
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_resume_ok,false,null);
                dalogMsg("info",rcConstants.CMD_resume_ok);
                informStateSMS(rcConstants.CMD_resume_ok,rcConstants.info);
            } else if(_ResetRequest) {
                setRunState(rcConstants.CMD_reset_ok);
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:"+rcConstants.CMD_reset_ok,false,null);
                dalogMsg("info",rcConstants.CMD_reset_ok);
                informStateSMS(rcConstants.CMD_reset_ok,rcConstants.info);
            } else {
                setRunState("Resume_failed");
                informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeTransition,true,"GUI:resume_failed",false,null);
                dalogMsg("error","resume failed");
                informStateSMS(getRunState(),rcConstants.error);
            }
        }
    }


    /**
     * This method whill check the requested state and move entire run control to that state.
     * Mehod is used during inter-agent communication requests ( depricated)
     * N.B. another way to transition to the coda state machine states is through the cMsg
     * callback mechanism (initiated from the messages coming from the rcGui).
     *
     * @param state                 required state of the runcontrol supervisor agent.
     *                              The states supported are listed in the
     *                              <em>rcConstants.CMD_" "_ok</em> constants.
     */
    public void moveToRCState(String state){
        String st;

        System.out.println(" Poop 1 ... moving to .... "+state +" from "+getRunState());

        if (state.equalsIgnoreCase(rcConstants.CMD_download)){
            st = getRunState();
            if(st.equalsIgnoreCase(rcConstants.CMD_reset_ok))
                startDownload();
        }
        else if (state.equalsIgnoreCase(rcConstants.CMD_prestart)){
            st = getRunState();
            if(st.equalsIgnoreCase(rcConstants.CMD_postdownload_ok) ||
                    st.equalsIgnoreCase(rcConstants.CMD_postend_ok))
                startPrestart();
        }
        else if (state.equalsIgnoreCase(rcConstants.CMD_go)){
            st = getRunState();
            if(st.equalsIgnoreCase(rcConstants.CMD_postprestart_ok))
                startGo();
        }
        else if (state.equalsIgnoreCase(rcConstants.CMD_end)){
            st = getRunState();
            if((st.equalsIgnoreCase(rcConstants.CMD_postgo_ok)) ||
                    (st.equalsIgnoreCase(rcConstants.CMD_postprestart_ok)) )
                startEnd();
        }
        else if (state.equalsIgnoreCase(rcConstants.CMD_pause)){
            st = getRunState();
            if(st.equalsIgnoreCase(rcConstants.CMD_postgo_ok))
                startPause();
        }
        else if (state.equalsIgnoreCase(rcConstants.CMD_resume)){
            st = getRunState();
            if(st.equalsIgnoreCase(rcConstants.CMD_pause_ok))
                startResume();
        }
    }

    /**
     * Informs receiver to move to the state
     * This method sends to the receiver a <clas>CoolActionMoveTo</clas> coda ontology ACL message,
     * describing the desired state. This method is used by the SMS agent.
     *
     * @param state               The CoolState object of the component
     */
    public void moveToState(CoolState state) {

        String receiver = null;
        AID tmpaid;

        if(state.getState_Isfor() == null && state.getState_Isfor().getComponent_Name().equals(myName)){
            if(!getRunState().equals(state.getState_Name())){
                moveToRCState(state.getState_Name());
//                getRunState() = state.getState_Name();
//                myStateObject = state;
                dalogMsg("Info",getRunState());
            }
        }
        else{
            receiver = state.getState_Isfor().getComponent_Name();
            if(receiver !=null){
// Create the acl message object
                ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
                try{
                    getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                    tmpaid = new AID(receiver,AID.ISLOCALNAME);

// Preparing the message
                    msg.addReceiver(tmpaid);
                    msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                    msg.setOntology(CodaRunControlOntology.NAME);
                    msg.addUserDefinedParameter(myName,tmpaid.getLocalName());

// Create CoolActionMoveTo ontology action object
                    CoolActionMoveTo mt = new CoolActionMoveTo();
                    mt.setActionMoveto_State(state);

// Create an action transport object
                    Action a = new Action();
                    a.setActor(tmpaid);
                    a.setAction(mt);

// Fill the message content
                    getContentManager().fillContent(msg, a);
                } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
                addBehaviour(new GenRequestBehaviour(this, msg, 0));
            }
        }
    }



    /**
     * This is the implementation of the abstract method, defined in the CAgent class.
     * This method is going to be called byt the BaseicRequestBehaviour object to process
     * the iform messages coming back to this agent as a result of the initiating the request
     * to some agent on the platfrom.
     *
     * @param msg             ACLMessage object passed from the GenRequestBehaviour object.
     */
    protected void processInform(ACLMessage msg) {
    }


    /**
     * This method is mainly for handling requests from the Cortex and component representative agents.
     * This is the implementatioin of the abstract method defined in the CAgent class.
     * This method is going to be called by the BaseicResponderBehaviour object to process
     * the request to this agent.
     * This method actualy will start a separate thread to execute external unix or
     * vxworks executable functions.
     * N.B. This is the case when requests are coming through the agent communication channel.
     *
     * @param request                   ACLMessage request object passed from the GenResponderBehaviour object.
     *                                  Here we fill it and return it to the GenResponderBehaviour object.
     * @param a                         Action Object
     * @return                          ACLMessage object
     */
    protected ACLMessage processRequest(ACLMessage request, Action a) {

// Start the thread which does the actual work
        ProcessRequestThread mt = new ProcessRequestThread(request, a);
        mt.start();

// Create an acl message objects
        ACLMessage reply = request.createReply();

// Temporary message content string
        String tmpStr = _session;
        try {

// Construct done ontology action and send it to the initiator
            Done d = new Done();
            d.setAction(a);
            getContentManager().fillContent(reply, d);
            reply.setPerformative(ACLMessage.INFORM);
            reply.setContent(tmpStr);
        }
        catch (Codec.CodecException e) {
            System.err.println(this.getLocalName()+" Fill/extract content unsucceeded. Reason:" + e.getMessage());
        }
        catch (OntologyException e){
            System.err.println(this.getLocalName()+" Ontology exception. Reason:" + e.getMessage());
        }
        return reply;
    }


    /**
     * Method takes acceptedata codaruncontrol ontology action and does the simple check
     * ( message contains the word <em>rcConstants.failed</em> or not) of the data sent
     * by the component agent. This information is used to see if we need to control the
     * run (e.g. autoend, etc.).
     *
     * @param myAcceptData              Ontology <b>acion</b> object, containing data from
     *                                  the component agent
     */
    private void updateData(AcceptData myAcceptData){

        Vector<Object> tmpVector;

// This vector will hold the Data, Producer and Timestamp of the Monitoring processes.
        tmpVector = new Vector<Object>();

// Clear the vector (not necessary , but...
        tmpVector.clear();

// Populate the vector with this information
        tmpVector.add(0,myAcceptData.getAcceptdata_Producer());
        tmpVector.add(1,myAcceptData.getAcceptdata_Timestamp());
        tmpVector.add(2,myAcceptData.getAcceptdata_Data());

// Fill MonitorProcesses hashmap
        MonitorProcesses.put(myAcceptData.getAcceptdata_Process(),tmpVector);

// If we are in the postgo (last stage of the go) state and the data content is not "failed"
        if((getRunState().equalsIgnoreCase(rcConstants.CMD_postgo_ok)) &&
                (!myAcceptData.getAcceptdata_Processstate().equalsIgnoreCase(rcConstants.failed))){

// Update  ALL component ConcurrentHashMap the state
            if(updateFromClient) {
                ALL.put(myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Name(),myAcceptData.getAcceptdata_Processstate());

// Update DAQ component ConcurrentHashMap the state
                if(myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.rocType)
                        || myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.ebType)
                        || myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.erType)
                        || myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.tsType)
                        ) {
//                if(updateFromClient)
                    DAQ.put(myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Name(),myAcceptData.getAcceptdata_Processstate());
                }
            }
// Reset the fails number
            Fails.put(myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Name(), 0);
        }
    }


    /**
     * Method takes DataQuant codaruncontrol ontology action and updates the local hash tables.
     *
     * @param myAcceptData             Ontology <b>concept</b> object containing data from
     *                                 the component agent
     */
    private void updateData(DataQuant myAcceptData){

        Vector<Object> tmpVector;

// This vector will hold the Data, Producer and Timestamp of the Monitoring processes.*/
        tmpVector = new Vector<Object>();

// Clear the vector (not necessary , but...
        tmpVector.clear();

// Populate the vector with this information
        tmpVector.add(0,myAcceptData.getDataquant_Producer());
        tmpVector.add(1,myAcceptData.getDataquant_Timestamp());
        tmpVector.add(2,myAcceptData.getDataquant_Data());

// Fill MonitorProcesses hashmap
        MonitorProcesses.put(myAcceptData.getDataquant_Process(),tmpVector);

// If we are in the postgo (last stage of the go) state and the data content is not "failed"
        if((getRunState().equalsIgnoreCase(rcConstants.CMD_postgo_ok)) &&
                (!myAcceptData.getDataquant_Processstate().equalsIgnoreCase(rcConstants.failed))){

// Update  ALL component ConcurrentHashMap the state
            if(updateFromClient) {
                ALL.put(myAcceptData.getDataquant_Producer().getProducer_Component().getComponent_Name(),myAcceptData.getDataquant_Processstate());

// Update DAQ component ConcurrentHashMap the state
                if(myAcceptData.getDataquant_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.rocType)
                        || myAcceptData.getDataquant_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.ebType)
                        || myAcceptData.getDataquant_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.erType)
                        || myAcceptData.getDataquant_Producer().getProducer_Component().getComponent_Type().equalsIgnoreCase(rcConstants.tsType)
                        ) {
//                if(updateFromClient)
                    DAQ.put(myAcceptData.getDataquant_Producer().getProducer_Component().getComponent_Name(),myAcceptData.getDataquant_Processstate());
                }
            }

// Reset the fails number
            Fails.put(myAcceptData.getDataquant_Producer().getProducer_Component().getComponent_Name(), 0);
        }
    }


    /**
     * Method resets the <em>MonitorProcess</em> ConcurrentHashMap, i.e. data for all registered
     * processes will be 0'd.
     * Methods goes over the MonitorProcess hashtabel and replaces the existing data vectors for all
     * components with the vector containing an empty CoolData object
     */
    private void zeroData(){

// This will hold the Data, Producer and Timestamp of the Monitoring processes.
        Vector<Object> tmpVector = new Vector<Object>();

// Clear the vector
        tmpVector.clear();

// Fill MonitorProcesses hashmap
// Go over the MonitorProcess hashtabel and replace existing data vector for the
// component with the vector containing an empty CoolData object
        for (String o : MonitorProcesses.keySet()) {
            Vector tmpvec = MonitorProcesses.get(o);
            Producer tmpProducer = (Producer)tmpvec.elementAt(0);

// Get and print the time when the measurement was taken
            CodaTimeStamp mTimeStamp = (CodaTimeStamp)tmpvec.elementAt(1);

// Populate the vector with this information
            tmpVector.add(0,tmpProducer);
            tmpVector.add(1,mTimeStamp);

// Set the data object
            CoolData myData = new CoolData();
            myData.setData_Type("Integer Float Long Double");
            myData.setData_Cardinality("multiple");
            myData.setData_Semantics("Event_Number Event_Rate Number_of_longs Data_Rate");
            myData.setData_Content("0 0.0 0.0 0.0");
            tmpVector.add(2,myData);
            MonitorProcesses.put(o,tmpVector);
        }
    }


    /**
     * Method handles the differentiate request to the Medulla agent, sent from the Cortex agent.
     * Here we connect to the cMsg server (run within Hepatic) and subscribe specific messages.
     * Medulla uses Differentiate action, described in the CodaRunControl ontology to turn
     * to runcontrol agent.
     *
     * @param myDifferentiate               Differentiate codaruncontrol ontology action object
     */
    private void handleDifferentiate(Differentiate myDifferentiate) {


// Clear local hashtables
        if(_mySubSystems !=null) _mySubSystems.clear();
        if(Components!=null)Components.clear();
        if(ComponentTypes!=null)ComponentTypes.clear();
        if(CompReportInterval!=null)CompReportInterval.clear();
        if(Priorities!=null)Priorities.clear();
        if(cyclicThreads!=null)cyclicThreads.clear();
        if(Fails!=null)Fails.clear();
        if(ALL!=null)ALL.clear();
        if(DAQ!=null)DAQ.clear();
        if(Rocs!=null)Rocs.clear();
        if(Tss!=null)Tss.clear();
        if(Ebs!=null)Ebs.clear();
        if(Ers!=null)Ers.clear();
        if(Slcs!=null)Slcs.clear();
        if(Anas!=null)Anas.clear();
        if(_myRules !=null) _myRules.clear();
        if(_myStates !=null) _myStates.clear();
        SM_RULES_ACCEPTED = false;

        try{

// Disconnect first from the cMsg server and connect again.
            if(cMsgSrv!=null)cMsgSrv.disconnect();
            cMsgSrv = new cMsg(rcSubscription.PlatformcMsgServerUDL, myName, rcSubscription.PlatformcMsgServerDescription);
            cMsgSrv.connect();

// Enable message reception
            cMsgSrv.start();

// Subscribe to subject/type  for transition requests from RcGui
            cMsgCallbackInterface cb = new TransitionRequestCallback();
            cMsgSrv.subscribe(myName, rcSubscription.RunTransitionType +"/*", cb, null);

// Subscribe to subject/type  for reset requests from RcGui
            cMsgCallbackInterface cbreset = new SupervisorControlCallback();
            cMsgSrv.subscribe(myName, rcSubscription.SupervisorControlType +"/*", cbreset, null);
        } catch (cMsgException ee) {
            ee.printStackTrace();
        }

// Get atributes of this agent described in the cool
        itsMe = myDifferentiate.getDifferentiate_Component();
        _myOption = myDifferentiate.getDifferentiate_Option();
        String _agentType = itsMe.getComponent_Type();

// Attach to the session
        _session = itsMe.getComponent_Session();
        _origRunType = itsMe.getComponent_Runtype();
// Broadcast the options datafile to plexuses
        String dtf = _myOption.getOption_Datafile();
        if(dtf!=null)
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeDataFile,true,dtf,false,null);

// Check if the RC component name in the cool configuration file and the name of this created agent
// are the same
        if(!itsMe.getComponent_Name().equals(myName)) {
            informCmsgClient(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeRunType,true,
                    "Medulla-Error:\n Agent name is not consistent with the cool",false, null);
            dalogMsg("severeerror","Agent name is not consistent with the cool");

        }

// See if we have scheduler defined in the options concept and there are processes defined which needs to be
// executed befor starting the new run (the process list, get the stop/pause list and start the run, etc.)
        if((_myOption.getOption_Scheduler()!=null) &&
                (_myOption.getOption_Scheduler().getScheduler_ProcessList()!=null) &&
                (!_myOption.getOption_Scheduler().getScheduler_ProcessList().isEmpty())){
            _schedulerIsDefined = true;
            _schedulerProcessList = _myOption.getOption_Scheduler().getScheduler_ProcessList();
            _schedulerStopList = _myOption.getOption_Scheduler().getScheduler_StopafterList();
            _schedulerPauseList = _myOption.getOption_Scheduler().getScheduler_PauseafterList();
            _schedulerIndex = 0;
            _guiAutoStart = true;
// Tell gui to enable autostart button
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeAutoStart,true,"true",false,null);
        }


 // See if autostart is enabled from cool
        else if(_myOption.getOption_Hasautostart()!=null && _myOption.getOption_Hasautostart().equalsIgnoreCase("true")){
            _guiAutoStart = true;
// Tell gui to enable autostart button
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeAutoStart,true,"true",false,null);
        }

 // See if startrun is enabled from cool
        if(_myOption.getOption_Hasstartrun()!=null && _myOption.getOption_Hasstartrun().equalsIgnoreCase("true")){
// Tell gui to enable startrun button
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeStartRun,true,"true",false,null);
        }else {
// Tell gui to disable startrun button
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeStartRun,true,"false",false,null);
        }

// See if eventlimit is set from cool
        if(_myOption.getOption_Eventlimit()>0){
// Tell gui to set event limit
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeEventLimit,true,
                    Integer.toString(_myOption.getOption_Eventlimit()),false,null);
        }else {
// Tell gui to disable event limit
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeEventLimit,true,
                    "0",false,null);
        }

// create object for shell command execution
        mySysExe = new SysExe();

// Register with the DF
        if(!_RegisteredFlag){
            _RegisteredFlag = true;
            DFAgentDescription dfd = new DFAgentDescription();
            ServiceDescription sd = new ServiceDescription();
            sd.setType(_agentType);
            sd.setName(myName);
            sd.setOwnership("JLABCODA");
            dfd.setName(getAID());
            dfd.addServices(sd);
            try {
                DFService.register(this,dfd);
            }
            catch (FIPAException e) {
                System.err.println(getLocalName()+" registration with DF unsucceeded. Reason: "+e.getMessage());
                doDelete();
                _RegisteredFlag = false;
            }
        }

    }


    /**
     * Method to organize Configure transition.
     * N.B. for pre  and post transitions we execute attached processes for this agent only, and
     * the main transition will be organized as a triplet for the supervised agents. For the main
     * transition the attached processes are executed for this agent as well.
     *
     * @see #organizeCodaTransition
     */
    public void startConfigure() {
        organizeCodaTransition(setpreconfigure,rcConstants.CMD_preconfigure);
        organizeCodaTransition(setconfigure,rcConstants.CMD_configure);
        organizeCodaTransition(setpostconfigure,rcConstants.CMD_postconfigure);
    }


    /**
     * Method to organize Download transition.
     * N.B. We will wait in the loop untill the transition succeeded and then execute the next transiton
     * from the triplet. However if transition failes or reset request is coming from RcGui we exit
     * the waiting loop.
     * N.B. reset request from the RcGui is handled in the separeate callback.
     * At the beginig we send the names and number of the components in the current configuration to all
     * subscribed plexuses (rcGUI)
     */
    public void startDownload() {

// Hash table holding the components of the same priority
        ConcurrentHashMap tmphash;
        String tmpCompName;
        StringBuffer tsb = new StringBuffer();

        updateFromClient = true;

// Some times we do not need to configure and will use previously configured run ( no differentiate) and will
// start from download. So, we need again to check to see if there is scheduler defined in the cool.        
// Start from download.
        if((_myOption.getOption_Scheduler()!=null) &&
                (_myOption.getOption_Scheduler().getScheduler_ProcessList()!=null) &&
                (!_myOption.getOption_Scheduler().getScheduler_ProcessList().isEmpty())){
            _schedulerIsDefined = true;
            _schedulerProcessList = _myOption.getOption_Scheduler().getScheduler_ProcessList();
            _schedulerStopList = _myOption.getOption_Scheduler().getScheduler_StopafterList();
            _schedulerPauseList = _myOption.getOption_Scheduler().getScheduler_PauseafterList();
            _schedulerIndex = 0;
            _guiAutoStart = true;
// Tell gui to enable autostart button
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeAutoStart,true,"true",false,null);

        }

// See if eventlimit is set from cool
        if(_myOption.getOption_Eventlimit()>0){
// Tell gui to set event limit
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeEventLimit,true,
                    Integer.toString(_myOption.getOption_Eventlimit()),false,null);
        }else {
// Tell gui to disable event limit
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeEventLimit,true,
                    "0",false,null);
        }
        

// Inform the component names and component count
// Broadcast the number of all registered components
        informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeComponentCount,true,Integer.toString(DAQ.size()),false,null);

// Get the iterator on the priorities (sorted set)
        for (Integer Priority : Priorities) {

// Get the hash table of the components in the same priority group
            tmphash = Components.get(Priority);

// Get iterator on that ConcurrentHashMap
            for (Object o : tmphash.keySet()) {

// Get the component agent name
                tmpCompName = (String) o;

// See if the component is DAQ component
                if (DAQ.containsKey(tmpCompName) ) {

// Apend the component name
                    tsb.append(tmpCompName).append(":").append(ComponentTypes.get(tmpCompName));
                    tsb.append(" ");
                }
            }
        }

// Broadcast the names for all registered DAQ components
        informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeComponentNames,true,tsb.toString(),false,null);

// Start organizing the transition
        organizeCodaTransition(setpredownload,rcConstants.CMD_predownload);
        if(_ResetRequest || getRunState().contains("failed")) return;

        organizeCodaTransition(setdownload,rcConstants.CMD_download);
        if(_ResetRequest || getRunState().contains("failed")) return;

        organizeCodaTransition(setpostdownload,rcConstants.CMD_postdownload);
        if(_ResetRequest || getRunState().contains("failed"))return;
    }


    /**
     * Method to organize Prestart transition. Here too, at the beginig we send the names and
     * the number of the components in the current configuration to user interface components.
     */
    public void startPrestart() {

// Hash table holding the components of the same priority
        ConcurrentHashMap tmphash;
        String tmpCompName;
        StringBuffer tsb = new StringBuffer();

        updateFromClient = true;

// Inform the component names and component count
// Broadcast the number of all registered components
        informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeComponentCount,true,Integer.toString(DAQ.size()),false,null);

// Get the iterator on the priorities sorted set
        for (Integer Priority : Priorities) {

// Get the hash table of the components in the same priority group
            tmphash = Components.get(Priority);

// Get iterator on that ConcurrentHashMap
            for (Object o : tmphash.keySet()) {

// Get the component agent name
                tmpCompName = (String) o;

// See if the component is DAQ component and it failed
                if (DAQ.containsKey(tmpCompName) ) {

// Apend the component name
                    tsb.append(tmpCompName).append(":").append(ComponentTypes.get(tmpCompName));
                    tsb.append(" ");
                }
            }
        }

// Broadcast DAQ component names
        informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeComponentNames,true,tsb.toString(),false,null);

// Prestart transition triplet
        organizeCodaTransition(setpreprestart,rcConstants.CMD_preprestart);
        if(_ResetRequest || getRunState().contains("failed")) return;

        organizeCodaTransition(setprestart,rcConstants.CMD_prestart);
        if(_ResetRequest || getRunState().contains("failed")) return;

        organizeCodaTransition(setpostprestart,rcConstants.CMD_postprestart);
        if(_ResetRequest || getRunState().contains("failed"))return;
    }


    /**
     * Method to organize Go transition triplet.
     * Checks the scheduler of the options consept of the control and executes
     * the process before starting the new run.
     *
     */
    public void startGo() {

// See if we are scheduled to run scheduled process at the start of the new run
// N.B. _schedulerIndex will be incremented at the end transition
        if((_schedulerIndex>=0) &&
                (_schedulerProcessList!=null) &&
                (!_schedulerProcessList.isEmpty()) &&
                (_schedulerIndex<_schedulerProcessList.size()))
        {
            CoolProcess pr = (CoolProcess)_schedulerProcessList.get(_schedulerIndex);

// Execute the first external process in the scheduled processes list before starting the new run
            mySysExe.run_o_unix(pr.getProcess_Commandname());

// Nothing is scheduled
        } else _schedulerIndex = -1000;

// Go triplet
        organizeCodaTransition(setprego,rcConstants.CMD_prego);
        if(_ResetRequest || getRunState().contains("failed")) return;
        organizeCodaTransition(setgo,rcConstants.CMD_go);
        if(_ResetRequest || getRunState().contains("failed")) return;

// Activate the run
        organizeCodaTransition(setpostgo,rcConstants.CMD_postgo);
        if(_ResetRequest || getRunState().contains("failed")) return;

// We are in the active state, zero the data in the MonitorProcesses ConcurrentHashMap
        zeroData();

// Start broadcasting monitored statistics info  thread which will periodically send data to the plexuses
        if(myMonitorOn!=null)myMonitorOn.shouldrun = false;
        myMonitorOn = new MonitorOn();
        myMonitorOn.shouldrun = true;
        myMonitorOn.start();

// Tell subcontrol systems if any with this configuration to start reporting status information
        SubSystemsStartReporting();
    }


    /**
     * starts the pause transition. After the run is paused  it will check if there is scheduler described in the cool,
     * and that it has pause schema (i.e. scheduler states change for every pause transition). Then it checks to see
     * if there are scheduled processes in the scheduler, and will execute the next process in the process list.
     * If the scheduler process list is empty it will dissable the scheduler (_schedulerIndex = -1000).
     */
    public void startPause(){
        System.out.println(" Poop 2 ... start pause ....");
        organizeCodaTransition(setinpause,rcConstants.CMD_pause);

// N.B. scheduler index indicates that there is scheduler described in the option consept of the control
        if((_schedulerIndex>=0) &&
                (_schedulerPauseList!=null) &&
                (!_schedulerPauseList.isEmpty()) &&
                (_schedulerIndex<_schedulerPauseList.size())){

// Scheduler described has pause schema. get the next process from the process list.
            _schedulerIndex++;
            if( (_schedulerProcessList!=null) &&
                    (!_schedulerProcessList.isEmpty()) &&
                    _schedulerIndex<_schedulerProcessList.size()){
                CoolProcess pr = (CoolProcess)_schedulerProcessList.get(_schedulerIndex);

// Execute the process
                mySysExe.run_o_unix(pr.getProcess_Commandname());

// Auto resume the run
                startResume();

// End of the process list or empty process list: dissable the scheduler
            } else if(_schedulerIsDefined) {
                _schedulerIndex = -1000;
                _guiAutoStart = false;
// Tell gui to desable autostart button
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeAutoStart,true,"false",false,null);
            } else _schedulerIndex = -1000;
        }
    }


    /**
     * Starts the resume transition
     */
    public void startResume(){
        organizeCodaTransition(setinresume,rcConstants.CMD_resume);
    }


    /**
     * Method to organize End transition.
     * It will auto start the new run if it is described so in the cool option concept.
     * Before auto starting it will check if there are scheduled processes in the scheduler and will increment
     * the scheduler index preparing for startGo method to execute the scheduled process.
     */
    public void startEnd() {

        updateFromClient = false;

// Local boolean which will be set to false in case there were DAQ problems.
        boolean permitautostart = true;

        organizeCodaTransition(setpreend,rcConstants.CMD_preend);
        if(_ResetRequest || getRunState().contains("failed")) return;
        updateFromClient = true;
        organizeCodaTransition(setend,rcConstants.CMD_end);
        if(_ResetRequest || getRunState().contains("failed")) return;
        updateFromClient = false;
        organizeCodaTransition(setpostend,rcConstants.CMD_postend);
        if(_ResetRequest || getRunState().contains("failed")) return;

// Wait this many sec before stoping the monitor broadcasting thread. This will keep gui updated after the end.
        try {
            Thread.sleep(rcConstants.update_after_end);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        if(myMonitorOn!=null)myMonitorOn.shouldrun = false;

// Increment scheduler index if it was scheduled the series of run executions
        _schedulerIndex++;
        if((_schedulerIndex>=0) && (_schedulerIndex<_schedulerProcessList.size())) { }//do nothing...
        else if(_schedulerIsDefined) {
            _schedulerIndex = -1000;
            _guiAutoStart = false;
// Tell gui to desable autostart button
            informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeAutoStart,true,"false",false,null);
        }  else _schedulerIndex = -1000;

// Autostart the new run in case it is defined in the cool config file.
// N.B. make sure do not auto start in case there is a DAQ problem, so check the fails table first.
// Get iterator on the DAQ components ConcurrentHashMap
        for (String s : DAQ.keySet()) {

// See if the component is DAQ component and it failed
            if ((Fails.containsKey(s)) &&
                    (Fails.get(s) >= rcConstants.FAILS_NUMBER)) {
                permitautostart = false;
            }
        }

// Start the new run if options hasAutoStart has set to "true", and UI did not ovewrite it.
        if(_guiAutoStart &&
                permitautostart &&
                (_myOption.getOption_Hasautostart()!=null) &&
                (_myOption.getOption_Hasautostart().equalsIgnoreCase("true"))) {

// Start prestart triplet
            startPrestart();

// Start go triplet
            startGo();
        } else {
// Tell subcontrol systems if any with this configuration to stop reporting status information
            SubSystemsStopReporting();
        }
    }


    /**
     * Reports run statistics to the Plexuses.
     *
     * @return evt_number               String representation of the event number
     */
    public String reportRunStatistics(){

        String evt_number="0",evt_rate="0", data_rate="0",average_evt_rate="0", average_data_rate="0";
        String processName, componentName, tmptype, tmpcontent, tmpsemantics, tmps;



// Create a stringbuffer for message content
        StringBuffer sb = new StringBuffer();

// Sending string messages of the components statistics information.
// N.B. following is the internal format used:
//  session runtype runstate component_name:evt_number:evt_rate:data_rate:number_of_long component_name....

// Start creating the message string going to be sent to the RcGui
        sb.append(_session).append(" ").append(_runtype).append(" ").append(getRunState()).append(" ");

// Go over the MonitorProcess hashtabel and constract the strings according to the internal format
        for (Object o : MonitorProcesses.keySet()) {
            try {

// Get the process name
                processName = (String) o;

// Get the process
                Vector tmpvec = MonitorProcesses.get(processName);

// Get the producer
                Producer tmpProducer = (Producer)tmpvec.elementAt(0);

// Get the data
                CoolData data = (CoolData)tmpvec.elementAt(2);

// Get the name of the component
                componentName = tmpProducer.getProducer_Component().getComponent_Name();

// Analyse the data
                if (data.getData_Cardinality().equalsIgnoreCase("multiple")) {

// Get the data content
                    String dcontent = data.getData_Content();

// Data has to be interpreted as a space separated values of the type defined. Here we create
// StringTokenized on the data_content data_type and data_semanitcs.
                    StringTokenizer stdc = new StringTokenizer(dcontent);
                    StringTokenizer stdt = new StringTokenizer(data.getData_Type());
                    StringTokenizer stds = new StringTokenizer(data.getData_Semantics());
                    while (stdc.hasMoreTokens() && stdt.hasMoreTokens() && stds.hasMoreTokens()) {
                        tmptype = stdt.nextToken();
                        tmpcontent = stdc.nextToken();
                        tmpsemantics = stds.nextToken();
                        if (tmpcontent.equalsIgnoreCase(rcConstants.failed)) {
                            tmps ="Medulla_Warning: "+componentName+" has no data";
                            informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,tmps,false,null);
                        } else {
                            if ((tmptype.equalsIgnoreCase("integer")) &&
                                    (tmpsemantics.equalsIgnoreCase("Event_Number"))) {
                                try {

// Event number
                                    evt_number=tmpcontent;
                                } catch (NumberFormatException e) {
                                    e.printStackTrace();
                                }
                            } else if ((tmptype.equalsIgnoreCase("long")) &&
                                    (tmpsemantics.equalsIgnoreCase("Number_of_longs"))) {

// Nothing for the number of longs. May be some time in the future....
                            } else if ((tmptype.equalsIgnoreCase("float")) &&
                                    (tmpsemantics.equalsIgnoreCase("Event_Rate"))) {
                                try {

// Event rate
                                    evt_rate = tmpcontent;
                                } catch (NumberFormatException e) {
                                    e.printStackTrace();
                                }
                            } else if ((tmptype.equalsIgnoreCase("double")) &&
                                    (tmpsemantics.equalsIgnoreCase("Data_Rate"))) {
                                try {

// Data rate
                                    data_rate = tmpcontent;
                                } catch (NumberFormatException e) {
                                    e.printStackTrace();
                                }
                            } else if ((tmptype.equalsIgnoreCase("float")) &&
                                    (tmpsemantics.equalsIgnoreCase("Average_Event_Rate"))) {
                                try {

// Average event rate
                                    average_evt_rate = tmpcontent;
                                } catch (NumberFormatException e) {
                                    e.printStackTrace();
                                }
                            } else if ((tmptype.equalsIgnoreCase("double")) &&
                                    (tmpsemantics.equalsIgnoreCase("Average_Data_Rate"))) {
                                try {

// Average data rate
                                    average_data_rate = tmpcontent;
                                } catch (NumberFormatException e) {
                                    e.printStackTrace();
                                }
                            } else if (tmptype.equalsIgnoreCase("string")) {
                            }
                        }
                    }
                }
            } catch (NullPointerException n) {
                n.printStackTrace();
                return "-1";
            }

// Construct a string for the current component
            sb.append(componentName).append(":")
                    .append(evt_number).append(":")
                    .append(evt_rate).append(":")
                    .append(data_rate).append(":")
                    .append(average_evt_rate).append(":")
                    .append(average_data_rate).append(" ");
        }

// Broadcast statistics data to plexuses
        informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeRunStatistics,true,sb.toString(),false,null);
// Broadcast to entire world the status of this particular control system under this supervisor control.
// Create dalog message
        if(getRunState().contains("failed")) {
            try{
                cMsgMessage dlg = new cMsgMessage();
                dlg.setSubject(rcSubscription.expid+"/"+_session+"/"+_origRunType+"/");
                dlg.setType("dalog/11");
//        else
//            dlg.setType("dalog/3");
                dlg.setText(getRunState());
// Send the message
                cMsgSrv.send(dlg);
                cMsgSrv.flush(1);
            } catch (cMsgException e) {
                e.printStackTrace();
            }
        }

        return evt_number;
    }

    /**
     * Comparing two states
     *
     * @param st1 state1
     * @param st2 state2
     * @return the result of the comparison
     */

    public boolean eQuals(CoolState st1, CoolState st2){
        boolean result = false;
        // see if the components are the same for which the states are being compared
//        if(st1.getState_Isfor().getComponent_Name().equals(st2.getState_Isfor().getComponent_Name())){

            // Compare the state names
            if(st1.getState_Name().equals(st2.getState_Name())){
                result = true;
            }
//        }
        return result;
    }


    /**
     * Informs receiver to execute a process
     * This method sends to the receiver a <clas>CoolActionExecute</clas> coda ontology ACL message,
     * describing the desired process. This method is used also by SMS agents.
     *
     * @param receiver            The name of the receiver
     * @param process             The CoolProcess object
     */
    public void executeProcess(String receiver, CoolProcess process) {

        AID tmpaid;

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);
            tmpaid = new AID(receiver,AID.ISLOCALNAME);

// Preparing the message
            msg.addReceiver(tmpaid);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(myName,tmpaid.getLocalName());

// Create CoolActionMoveTo ontology action object
            CoolActionExecute ex = new CoolActionExecute();
            ex.setActionExecute_Process(process);

// Create an action transport object
            Action a = new Action();
            a.setActor(tmpaid);
            a.setAction(ex);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, 0));
    }

    /**
     * AFECS Rule syntaxparser
     *
     * @param r  CoolRule object
     */
    public void SMAlgorithm(CoolRule r){

        boolean flag =true;
        Iterator it;

// ANDIFSTATE
        if(r.getRule_Andif()!=null){
            it = r.getRule_Andif().iterator();
            // flag indicates that recorded state-names for the required
            // components are the same as required ANDIFSTATE name;
            while(it.hasNext()){
                CoolState andifState = (CoolState)it.next();

                if(!ALL.containsKey(andifState.getState_Isfor().getComponent_Name())){
                    flag =false;
                } else {
                    String recStName = ALL.get(andifState.getState_Isfor().getComponent_Name());
                    if(!recStName.equals(andifState.getState_Name())){
                        flag =false;
                    }
                }
            }

            // all ANDIFSTATE rules are satisfied
            if(flag){
                if(r.getRule_Thenmoveto()!=null){
                    it = r.getRule_Thenmoveto().iterator();
                    while(it.hasNext()){
                        CoolState tms = (CoolState)it.next();
                        moveToState(tms);
                    }
                }
                if(r.getRule_Thenexecute()!=null){
                    it = r.getRule_Thenexecute().iterator();
                    while(it.hasNext()){
                        CoolProcess tmp = (CoolProcess)it.next();
                        executeProcess(myName,tmp);
                    }
                }
            }
        }


// ORIF
        else if(r.getRule_Orif()!=null){
            it = r.getRule_Orif().iterator();
            flag = false;
            while(it.hasNext()){
                CoolState orIfState = (CoolState)it.next();
                if(ALL.containsKey(orIfState.getState_Isfor().getComponent_Name())){
                    String recStName = ALL.get(orIfState.getState_Isfor().getComponent_Name());
                    if(recStName.equals(orIfState.getState_Name())){
                        flag =true;
                    }
                }
            }
            // at least one ORIFSTATE  rule is satisfied
            if(flag){
                if(r.getRule_Thenmoveto()!=null){
                    it = r.getRule_Thenmoveto().iterator();
                    while(it.hasNext()){
                        CoolState tms = (CoolState)it.next();
                        moveToState(tms);
                    }
                }
                if(r.getRule_Thenexecute()!=null){
                    it = r.getRule_Thenexecute().iterator();
                    while(it.hasNext()){
                        CoolProcess tmp = (CoolProcess)it.next();
                        executeProcess(myName,tmp);
                    }
                }
            }
        }

// THENMOVETO action
        else if(r.getRule_Thenmoveto()!=null){
            it = r.getRule_Thenmoveto().iterator();
            while(it.hasNext()){
                CoolState tms = (CoolState)it.next();
                moveToState(tms);
            }
        }
// THENEXECUTE action
        else if(r.getRule_Thenexecute()!=null){
            it = r.getRule_Thenexecute().iterator();
            while(it.hasNext()){
                CoolProcess tmp = (CoolProcess)it.next();
                executeProcess(myName,tmp);
            }
        }
        else{
            System.out.println("SM Rule Error: Syntax error in the rule.");
        }

    }

    /**
     *
     * @param stateIn       iunput state to be considered inside of the configured rule.
     */
    public void execrule(CoolState stateIn){
        Iterator rulesIt, it;
        CoolRule r;

        rulesIt = _myRules.iterator();
        while(rulesIt.hasNext()){

            r = (CoolRule)rulesIt.next();

// -> IF rule
            if(r.getRule_If()!=null && eQuals(stateIn,r.getRule_If())){
                SMAlgorithm(r);

// -> ELSEIF rule
                if(r.getRule_Elseif()!=null && eQuals(stateIn, r.getRule_Elseif())){
                    SMAlgorithm(r);
                }

// -> ELSE rule
                if(r.getRule_Else()!=null){
                    if(r.getRule_Elsemoveto()!=null){
                        it = r.getRule_Elsemoveto().iterator();
                        while(it.hasNext()){
                            CoolState tms = (CoolState)it.next();
                            moveToState(tms);
                        }
                    }
                    if(r.getRule_Elseexecute()!=null){
                        it = r.getRule_Elseexecute().iterator();
                        while(it.hasNext()){
                            CoolProcess tmp = (CoolProcess)it.next();
                            executeProcess(myName,tmp);
                        }
                    }
                }

// -> WHILE rule
            } else if(r.getRule_While()!=null && eQuals(stateIn,r.getRule_While())){

                it = ALL.keySet().iterator();
                while(it.hasNext()){
                    String compName = (String)it.next();
                    String compState = ALL.get(compName);
                    if(compName.equals(r.getRule_While().getState_Isfor().getComponent_Name()) &&
                            compState.equals(r.getRule_While().getState_Name())){
                        if(r.getRule_Whilemoveto()!=null){
                            it = r.getRule_Whilemoveto().iterator();
                            while(it.hasNext()){
                                CoolState tms = (CoolState)it.next();
                                moveToState(tms);
                            }
                        }
                        if(r.getRule_Whileexecute()!=null){
                            it = r.getRule_Whileexecute().iterator();
                            while(it.hasNext()){
                                CoolProcess tmp = (CoolProcess)it.next();
                                executeProcess(myName,tmp);
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Utility method to dump the rules
     *
     * @param l    List of CoolRule objects
     */
    public void dumprule(rca.util.leap.List l){
        rca.util.leap.Iterator it, itt;
        CoolRule r;
        CoolProcess p;

        rca.util.leap.List t;
        itt = l.iterator(); while(itt.hasNext()){
        r = (CoolRule)itt.next();

        System.out.println("======== Rule ========");

//................. IF
        if(r.getRule_If()!=null){
            System.out.println("if state = "+r.getRule_If().getState_Name()+" of "+r.getRule_If().getState_Isfor().getComponent_Name() );
        }

        //orif
        if(r.getRule_Orif()!=null && !r.getRule_Orif().isEmpty()){
            t=r.getRule_Orif();


            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" orif state = "+s.getState_Name());
        }
        }

        // andif
        if(r.getRule_Andif()!=null && !r.getRule_Andif().isEmpty()){
            t=r.getRule_Andif();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" andif state = "+s.getState_Name());
        }
        }

        // thenmoveto
        if(r.getRule_Thenmoveto()!=null && !r.getRule_Thenmoveto().isEmpty()){
            t=r.getRule_Thenmoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" then moveto state = "+s.getState_Name());
        }
        }

        // thenexecute
        if(r.getRule_Thenexecute()!=null && !r.getRule_Thenexecute().isEmpty()){
            t=r.getRule_Thenexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" then execute process = "+p.getProcess_Commandname());
        }
        }

// .............. ELSEIF
        if(r.getRule_Elseif()!=null){
            System.out.println("elseif state = "+r.getRule_Elseif().getState_Name());
        }

        // elseifmoveto
        if(r.getRule_Elseifmoveto()!=null && !r.getRule_Elseifmoveto().isEmpty()){
            t=r.getRule_Elseifmoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" elseif moveto state = "+s.getState_Name());
        }
        }

        // elseifexecute
        if(r.getRule_Elseifexecute()!=null && !r.getRule_Elseifexecute().isEmpty()){
            t=r.getRule_Elseifexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" elseif execute process = "+p.getProcess_Commandname());
        }
        }


// .............. ELSE
        if(r.getRule_Else()!=null){
            System.out.println("else");
        }

        // elsemoveto
        if(r.getRule_Elsemoveto()!=null && !r.getRule_Elsemoveto().isEmpty()){
            t=r.getRule_Elsemoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" else moveto state = "+s.getState_Name());
        }
        }

        // elseexecute
        if(r.getRule_Elseexecute()!=null && !r.getRule_Elseexecute().isEmpty()){
            t=r.getRule_Elseexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" else execute process = "+p.getProcess_Commandname());
        }
        }

// .............. WHILE
        if(r.getRule_While()!=null){
            System.out.println("while state = "+r.getRule_While().getState_Name());
        }

        // whilemoveto
        if(r.getRule_Whilemoveto()!=null && !r.getRule_Whilemoveto().isEmpty()){
            t=r.getRule_Whilemoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" while moveto state = "+s.getState_Name());
        }
        }

        // whileexecute
        if(r.getRule_Whileexecute()!=null && !r.getRule_Whileexecute().isEmpty()){
            t=r.getRule_Whileexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" while execute process = "+p.getProcess_Commandname());
        }
        }
    }
    }


// ... Inner Class ...

    /**
     * This class is responsible for communication with the actual process or the hardware (component) using
     * IPC (inner process communication) protocols: DP, tcpServer, shell and cMsg (cMsg domain).
     * This will also organize the monitoring/executing of the external processes ( for e.g. attached scripts)
     * Those processes will be attached to the particular transition for this agent ( RCS).
     * N.B. supervisor agent will not execute contiuous processes (design choice).
     *
     * @author Vardan Gyurjyan
     *         Date: Jan 26, 2006
     *         Time: 10:00:30 PM
     */
    private class RunProcess extends Thread {

        /** Keeps transition name ( download, prestart,etc.)*/
        private String _transition;

        /** Command line string of the process.*/
        private String _commandName;

        /** Command type, defins the IPC protocole. For e.g. dp, cmsg, etc. */
        private String _commandType;

        /** Process execution result string in case process succeded.*/
        private String _expectedOk;

        /** How many times the process will be repeted. Negative value means loop for ever.*/
        private int _loop = 1;

        /** How long we need to wait before the next execution of the process in the loop.*/
        private int _delay = 0;


        /**
         * Constructor. Initializes the global fields of this class.
         *
         * @param cm                  Process command name.
         *                            For the CODA component this will be a a name of the CODA transition.
         * @param proc                <em>CoolProcess</em> object from the <em>CodaRunControl ontology</em>.
         */
        public RunProcess( String cm, CoolProcess proc){
            _transition = cm;

// Getting command line string from the passed "CoolProcess" ontology concept.
            _commandName = proc.getProcess_Commandname();

// Get the command type.
            if(proc.getProcess_Ipc()!=null)
                _commandType = proc.getProcess_Ipc();
            if(proc.getProcess_Outputdata()!=null){

// Get expected_ok string, defined in the cool configuration file
                if(proc.getProcess_Outputdata().getData_Content()!=null)
                    _expectedOk = proc.getProcess_Outputdata().getData_Content();
            }

// Command loop parameters
            if(proc.getProcess_Periodicity()!=null)
                _loop = proc.getProcess_Periodicity();
            if(proc.getProcess_Delay()!=null)
                _delay = proc.getProcess_Delay();
        }


        /**
         * Sends the message to cMsg server (Hepatic domain) for communicating with the external components.
         * N.B. this is still not implemented. It is assumed that cool will contain information about
         * subject and type of the subscribtion of the external component to be contacted.
         *
         * @return status
         */
        public boolean runcMsg(){
            cMsgMessage msg = new cMsgMessage();
            msg.setSubject("subject defined in the cool");
            msg.setType("type defined in the cool");
            msg.setText(_commandName);
            try {
                cMsgSrv.send(msg); // send the mesage
            } catch (cMsgException e) {
                e.printStackTrace();
                return false;
            }
            return true;
        }


        /**
         * This method executes the Shell command using the global fields of this class,
         * assigned at the constructor.
         *
         * @param command                 Shell command name beeing executed
         * @param expectedOk              This string will be compared with the
         *                                captured IO of the shell execution
         * @return                        <em>rcConstants.failed</em> or the name of the
         *                                current request transition
         */
        public String runShell(String command, String expectedOk){

// Local variables
            String res = null;
            String result;
            StringTokenizer st;
            List<String> l;
            BufferedReader myReader;

// Unix shell command execution IO stream, returned by the SysExe object.
            InputStream iostream;

// Actual execution of the command. Create a list;
            l = new ArrayList<String>();

// Tokenize the command.
            st = new StringTokenizer(command);

// Add the tokens into the list
            while(st.hasMoreTokens())l.add(st.nextToken());

// Create a string array
            String[] ta = new String[l.size()];

// Convert the list into the array of strings
            l.toArray(ta);
            String s =myName+"_Info: executing "+command+" ...";
            System.out.println(s);
            informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeStemCell,true,s,false,null);

// Execute the unix command and get an IO inputstream of the stdio
            iostream = mySysExe.run_unix_stream(ta);
            if(iostream!=null){
                try {

// Create buffereader using the iostream.
                    myReader = new BufferedReader(new InputStreamReader(iostream));

// Atention: Check command will return only one line summary.
                    res =myReader.readLine();
                    myReader.close();
                } catch (IOException e) {
                    e.printStackTrace();
                    return rcConstants.failed;
                }

// N.B. returned null iostream might be ok for some processes
// We do the following if expectedok is defined for the command
                if(expectedOk !=null){

// See if the execution of the main command (not the check command) is ok.
                    if(!res.equalsIgnoreCase(expectedOk)) result = rcConstants.failed;
                    else result = _transition;
                }
                else result = _transition;
            } else result = rcConstants.failed;
            s =myName+"_Info: completed "+command+": stdio => \n"+res;
            System.out.println(s);
            informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeStemCell,true,s,false,null);
            return result;
        }


        /**
         *   Overriding the run method of the Thread class
         */
        public void run() {

// Temporary String holds the process execution result.
            String tmpstr;

// If process is not continuous
            if(_loop >= 0) {
                for(int i = 0;i < _loop; i++) {
                    tmpstr = null;

// ... Command type is cMsg. Send a message, and if it failes exit this thread.
                    if(_commandType.equalsIgnoreCase("cMsg")){
                        if(!runcMsg())return;
                    }

// ... Command type is Unix, i.e. the external process is unix process
                    else if (_commandType.equalsIgnoreCase("shell")) {
                        tmpstr = runShell(_commandName, _expectedOk);
                    }

// The Medulla agent does not understand other than cMsg and shell ipc protocols
                    else {
                        System.err.println("Unknown command type: "+ _commandType);
                    }

// Inform the execution results to the Plexuses.
                    if(tmpstr!=null)  informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeRunStatus,true,
                            myName+" "+ _commandName +" execution_status = "+tmpstr,false,null);

// We sleep before the next external process execution
                    try {
                        sleep(_delay *1000);
                    } catch (Exception e) {e.printStackTrace();}
                }

// Cyclic process execution: loop<0
            } else {
// Supervisor agent do not support continuous monitoring processes.
            }
        }
    }


// ... Inner Class ...

    /**
     * This class defines the callback to be run when Coda transition request message arrives from RcGui, or
     * some other user component ouside of the AFECS platform
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 26, 2006
     *         Time: 10:00:30 PM
     */
    private class TransitionRequestCallback extends cMsgCallbackAdapter {

        /**
         * Callback method definition.
         *
         * @param msg                       Message received from rcGui (cMsg domain).
         *                                  It has the following format:
         *                                  <em>command runtype:"String" session:"String"</em>
         * @param userObject                Object passed as an argument which was set when the
         *                                  client orginally subscribed to a subject and type of
         *                                  the message.
         */
        public void callback(cMsgMessage msg, Object userObject) {

// Get the message type
            String mtype = msg.getType();

// Get the current run state
            String state = getRunState();

// Now basic coda transition requests from plexus (User interface)
// Configure triplet is depricated, since Medulla is not getting configure requests.
// N.B. Configure request is sent by rcGui to Cortex.
// The first cmsg message that gets Medulla from rcGui is download
// N.B. the _RunState by default is set to "reseted"
            if(mtype.equalsIgnoreCase(rcSubscription.RunTransitionType +"/"+rcConstants.CMD_configure)) {
                if(state.equalsIgnoreCase(rcConstants.CMD_preconfigure_ok))
                    startConfigure();
            }
            else if(mtype.equalsIgnoreCase(rcSubscription.RunTransitionType +"/"+rcConstants.CMD_download)) {
                if(state.equalsIgnoreCase(rcConstants.CMD_reset_ok)||
                        (state.equalsIgnoreCase("postdownload_failed")) ||
                        (state.equalsIgnoreCase("predownload_failed")))
                    startDownload();
            }
            else if(mtype.equalsIgnoreCase(rcSubscription.RunTransitionType +"/"+rcConstants.CMD_prestart)) {
                if(state.equalsIgnoreCase(rcConstants.CMD_postdownload_ok) ||
                        state.equalsIgnoreCase(rcConstants.CMD_postend_ok)||
                        (state.equalsIgnoreCase("postprestart_failed")) ||
                        (state.equalsIgnoreCase("preprestart_failed")))
                    startPrestart();
            }
            else if(mtype.equalsIgnoreCase(rcSubscription.RunTransitionType +"/"+rcConstants.CMD_go)) {
                if(state.equalsIgnoreCase(rcConstants.CMD_postprestart_ok) ||
                        (state.equalsIgnoreCase("postgo_failed")) ||
                        (state.equalsIgnoreCase("prego_failed")))
                    startGo();
            }
            else if(mtype.equalsIgnoreCase(rcSubscription.RunTransitionType +"/"+rcConstants.CMD_end)) {
                if((state.equalsIgnoreCase(rcConstants.CMD_postgo_ok)) ||
                        (state.equalsIgnoreCase(rcConstants.CMD_postprestart_ok)) ||
                        (state.equalsIgnoreCase("postend_failed")) ||
                        (state.equalsIgnoreCase("preend_failed")))

// Clear internal queue of this agent. N.B. The internal queue of this agent will be field with the messages
// sent by the representative agents.
                    clearInternalQueue();
                startEnd();
            }
            else if(mtype.equalsIgnoreCase(rcSubscription.RunTransitionType +"/"+rcConstants.CMD_pause)) {
                if(state.equalsIgnoreCase(rcConstants.CMD_postgo_ok))
                    startPause();
            }
            else if(mtype.equalsIgnoreCase(rcSubscription.RunTransitionType +"/"+rcConstants.CMD_resume)) {
                if(state.equalsIgnoreCase(rcConstants.CMD_pause_ok))
                    startResume();
            }
        }

        public boolean maySkipMessages() {return false;}

        public boolean mustSerializeMessages() {return true;}

        public int  getMaximumThreads() {return 200;}
    }


// ... Inner Class...

    /**
     * This class defines the callback to be run to serve other then coda transition requests.
     * An exception is <em>reset</em> transition which is served in this callback.
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 26, 2006
     *         Time: 10:00:30 PM
     */
    private class SupervisorControlCallback extends cMsgCallbackAdapter {

        /**
         * Callback method definition.
         *
         * @param msg                         Message received from the cMsg domain server.
         *                                    It has the following format: for e.g.
         *                                    <em>reset runtype:"String" session:"String"</em>
         * @param userObject                  Object passed as an argument, which was set when the
         *                                    client orginally subscribed to a subject and type of
         *                                    message.
         */

        public void callback(cMsgMessage msg, Object userObject) {

// Get the message type and content
            String mtype = msg.getType();
            String mtext = msg.getText();

// Get the current run state
            String sta = getRunState();

// ... Report run status request
            if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_ReportRunStatus)){
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session,
                        rcSubscription.RcGuiTypeRunStatus,true,sta,false,null);
                informStateSMS(sta,rcConstants.info);
            }

// ... Report runumber request
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_ReportRunNumber)) {
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session,
                        rcSubscription.RcGuiTypeRunNumber,true,_RunNumberF,false,null);
            }

// ... Report runtype request
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_ReportRunType)) {
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session,
                        rcSubscription.RcGuiTypeRunType,true,_runtype,false,null);
            }

// ... Report run session request
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_ReportSession)) {
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session,
                        rcSubscription.RcGuiTypeSession,true,_session,false,null);
            }

// ... Autostart request
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_AutoStartTrue)) {
                _guiAutoStart = true;
            }
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_AutoStartFalse)) {
                _guiAutoStart = false;
            }

// ... Reporting run statistics information
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_ReporRunStatistics )) {
                reportRunStatistics();
            }

// ... Report components under medullas control
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_ReportComponents)) {

// Hash table holding the components of the same priority
                ConcurrentHashMap tmphash;
                String tmpCompName;
                StringBuffer tsb = new StringBuffer();

// Inform the component names and component count
// Broadcast the number of all registered components
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session,
                        rcSubscription.RcGuiTypeComponentCount,true,Integer.toString(DAQ.size()),false,null);

// Get the iterator on the priorities sorted set
                for (Integer Priority : Priorities) {

// Get the hash table of the components in the same priority group
                    tmphash = Components.get(Priority);

// Get iterator on that ConcurrentHashMap
                    for (Object o : tmphash.keySet()) {

// Get the component agent name
                        tmpCompName = (String) o;

// See if the component is DAQ component
                        if (DAQ.containsKey(tmpCompName) ) {

// Apend the component name
                            tsb.append(tmpCompName).append(":").append(ComponentTypes.get(tmpCompName));
                            tsb.append(" ");
                        }
                    }
                }

// Broadcast the names for all registered DAQ components
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session,
                        rcSubscription.RcGuiTypeComponentNames,true,tsb.toString(),false,null);
            }

// ReportInterval. GUI asks to report the names and reporting intervals for all registered agents for this control system.
// The message content will have the following format
// name1:reportinginterval1 name2:reportinginterval2 ....
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_reportInterval)) {
                StringBuffer rib = new StringBuffer();
                String n, guiname;
                guiname = mtext;
                Enumeration en = CompReportInterval.keys();
                while(en.hasMoreElements()){
                    n = (String)en.nextElement();
                    rib.append(n).append(":").append(CompReportInterval.get(n));
                    rib.append(" ");
                }

// Send this to the gui, to the one that intiated the request.
// For that reason we add the name of the intiator gui in the type.
                informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session,
                        rcSubscription.RcGuiTypeCompRepInterval+"_"+guiname,true,rib.toString(),false,null);
            }

// ... SetInterval request from the Gui...
// The format of the message is [name interval]            
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_SetInterval)) {
                String name,interval;
                String txt = msg.getText();
                StringTokenizer st = new StringTokenizer(txt);
                name = st.nextToken();
                interval = st.nextToken();
                if(name!=null && interval!=null){

// Send cMsg domain message to the physical component.
// Attention: this is the only case when RCS agent talks directly to the physical component.
                    informCmsgClient(name,rcSubscription.SessionControlType+"/"+rcConstants.CMD_SetInterval,true,interval,false, null);

// Update the hashtable for the intervals
                    CompReportInterval.remove(name);
                    CompReportInterval.put(name,interval);
                }
            }


// ... Component kill request from the GUI...
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_componentKill)) {
//@todo kill all agents under its supervision                
            }

// ... Reset request from the GUI...
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_reset)) {

// ... Stop broadcasting monitored information
                if(myMonitorOn!=null)myMonitorOn.shouldrun = false;
                organizeCodaTransition(setreset,rcConstants.CMD_reset);

// Executed befor starting the new run get the process list, get the stop/pause list and start the run.
                if((_myOption.getOption_Scheduler()!=null) &&
                        (_myOption.getOption_Scheduler().getScheduler_ProcessList()!=null) &&
                        (!_myOption.getOption_Scheduler().getScheduler_ProcessList().isEmpty())){
                    _schedulerIsDefined = true;
                    _schedulerProcessList = _myOption.getOption_Scheduler().getScheduler_ProcessList();
                    _schedulerStopList = _myOption.getOption_Scheduler().getScheduler_StopafterList();
                    _schedulerPauseList = _myOption.getOption_Scheduler().getScheduler_PauseafterList();
                    _schedulerIndex = 0;
                    _guiAutoStart = false;
// Tell gui to desable autostart button
                    informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeAutoStart,true,"false",false,null);
                }
            }

// ... Autoreset request from the GUI...
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_autoreset)) {

// Stop broadcasting monitored information
                if(myMonitorOn!=null)myMonitorOn.shouldrun = false;
                organizeCodaTransition(setreset,rcConstants.CMD_autoreset);

// Executed befor starting the new run get the process list, get the stop/pause list and start the run.
                if((_myOption.getOption_Scheduler()!=null) &&
                        (_myOption.getOption_Scheduler().getScheduler_ProcessList()!=null) &&
                        (!_myOption.getOption_Scheduler().getScheduler_ProcessList().isEmpty())){
                    _schedulerIsDefined = true;
                    _schedulerProcessList = _myOption.getOption_Scheduler().getScheduler_ProcessList();
                    _schedulerStopList = _myOption.getOption_Scheduler().getScheduler_StopafterList();
                    _schedulerPauseList = _myOption.getOption_Scheduler().getScheduler_PauseafterList();
                    _schedulerIndex = 0;
                    _guiAutoStart = false;
// Tell gui to desable autostart button
                    informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeAutoStart,true,"false",false,null);
                }
            }

// ... Component Exit request from the Gui...
            else if(mtype.equalsIgnoreCase(rcSubscription.SupervisorControlType +"/"+rcConstants.CMD_Exit)) {
// ... Stop broadcasting monitored information
                if(myMonitorOn!=null)myMonitorOn.shouldrun = false;
                broadcastCmsgClients(rcSubscription.SessionControlType+"/"+rcConstants.CMD_Exit,null);

            }

        }

        public boolean maySkipMessages() {return false;}

        public boolean mustSerializeMessages() {return true;}

        public int  getMaximumThreads() {return 200;}
    }


// ... Inner Class ...

    /**
     * This class starts periodinc broadcasting of the run statistics information to the plexuses.
     *
     * @author Vardan Gyurjyan
     *         Date: Jul 13, 2006
     *         Time: 10:00:30 PM
     */
    public class MonitorOn extends Thread {

        /** Control of this thread*/
        public volatile boolean shouldrun;


        public void run(){

            String evt_number;

// Execute starEnd only once.
            boolean autoendstarted = false;
            while(shouldrun){

// In case of the reset request from the rcGui stop this thread
                if(_ResetRequest) return;

// Report plexuses the run statistics information
                evt_number = reportRunStatistics();

// Autstop the run if it is scheduled in the scheduler (defined in the cool configuration file).
// Here we check the event limits
                if((_schedulerIndex>=0) &&
                        (_schedulerStopList!=null) &&
                        (!_schedulerStopList.isEmpty()) &&
                        (_schedulerIndex<_schedulerStopList.size())){

// Autostop the run after n events
                    if((!(autoendstarted)) &&
                            (Integer.parseInt(evt_number)>(Integer.parseInt((String)_schedulerStopList.get(_schedulerIndex))))){

// Execute end run in the separate thread
                        Thread myT = new Thread(){
                            public void run(){
                                startEnd();
                            }
                        };
                        myT.start();
                        autoendstarted = true;
                    }

// Autopause the run if the scheduler is defined to do so
                } else if((_schedulerIndex>=0) &&
                        (_schedulerPauseList!=null) &&
                        (!_schedulerPauseList.isEmpty()) &&
                        (_schedulerIndex<_schedulerPauseList.size())
                        ){
                    if(Integer.parseInt(evt_number)>(Integer.parseInt((String)_schedulerPauseList.get(_schedulerIndex)))) startPause();
                } else {

// Now this is the case when scheduler is not defined but we have defined
// auto-end after defined event limit in the "Options" concept of the control.
// Check if we reached the event limit request described in the options object
                    if((!(autoendstarted)) && (_myOption !=null) &&
                            (_myOption.getOption_Eventlimit()>0) &&
                            (Integer.parseInt(evt_number)> _myOption.getOption_Eventlimit())) {

// Execute end run in the separate thread
                        Thread myT = new Thread(){
                            public void run(){
                                startEnd();
                            }
                        };
                        myT.start();
                        autoendstarted = true;
                    }
                }
                try {
                    Thread.sleep(rcConstants.mbroadcast_period);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }


// ... Inner Class ...

    /**
     * Thread that runs to satisfy responder behaviour of this agent
     *
     */
    private class ProcessRequestThread extends Thread {
        private ACLMessage request;
        private Action a;

        public ProcessRequestThread(ACLMessage r, Action ap){
            request = r;
            a=ap;
        }

        public void run(){

            CoolComponent tmpComp;

//            System.out.println("PPP ... "+myName+" got the ACL message "+a.getAction().getClass().getName());

// Temporary ConcurrentHashMap containing name and status of the components with the same priority code.
            ConcurrentHashMap<String, String> tmpcmp;

// ... Differentiation
            if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Differentiate")) {
                Differentiate mydif = (Differentiate)a.getAction();
                handleDifferentiate(mydif);
            }

// ... AcceptStates from Cortex.
// This request informs this SMS aget what are the all registered states within the control system,
// for which this agent is registered to be the supervisor.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptStates")) {
                AcceptStates ast = (AcceptStates)a.getAction();
                if(ast.getAcceptstates_List()!=null &&  !(ast.getAcceptstates_List().isEmpty())){
                    _myStates = ast.getAcceptstates_List();
                }
            }

// ... AcceptRules from Cortex.
// This request informs this SMS aget what are the all registered rules within the control system (state machine),
// for which this agent is registered to be the supervisor.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptRules")) {
                AcceptRules rl = (AcceptRules)a.getAction();
                if(rl.getAcceptrules_List()!=null &&  !(rl.getAcceptrules_List().isEmpty())){
                    _myRules = rl.getAcceptrules_List();
                    dumprule(_myRules);
                    SM_RULES_ACCEPTED = true;
                }
            }

// ... ResultNotify.
// This request tells this agent to report its state to the SMS agent.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.ResultNotify")) {
                ResultNotify rnt = (ResultNotify)a.getAction();
                if(rnt.getResultNotify_Component()!=null && rnt.getResultNotify_Component().getComponent_Name()!=null)
                    _mySMS.add(rnt.getResultNotify_Component().getComponent_Name());
            }

// ... ActionMoveTo.
// This request tells this agent to move to the state.
// Remember this is the RCS agent which has coda state machine hard coded, so it will support only
// states having the names = rcConstants.CMD_""_ok...
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.CoolActionMoveTo")) {
                CoolActionMoveTo mt = (CoolActionMoveTo)a.getAction();

// get the state
                CoolState st = mt.getActionMoveto_State();

// get the name of the state
                String stname = st.getState_Name();
                moveToRCState(stname);
            }

// Following are commands sent by cortex, telling what to do at every transition
// ... Configure triplet
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPreConfigure")) {
                setpreconfigure = (SetPreConfigure)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetConfigure")) {
                setconfigure = (SetConfigure)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostConfigure")) {
                setpostconfigure = (SetPostConfigure)a.getAction();

// ... Download triplet
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPredownload")) {
                setpredownload = (SetPreDownload)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetDownload")) {
                setdownload = (SetDownload)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostDownload")) {
                setpostdownload = (SetPostDownload)a.getAction();

// ... Prestart triplet
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPreprestart")) {
                setpreprestart = (SetPrePrestart)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPrestart")) {
                setprestart = (SetPrestart)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostPrestart")) {
                setpostprestart = (SetPostPrestart)a.getAction();

// ... Go triplet
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPrego")) {
                setprego = (SetPreGo)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetGo")) {
                setgo = (SetGo)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostGo")) {
                setpostgo = (SetPostGo)a.getAction();

// ... Reset
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetReset")) {
                setreset = (SetReset)a.getAction();

// ... End triplet
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPreend")) {
                setpreend = (SetPreEnd)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetEnd")) {
                setend = (SetEnd)a.getAction();
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostEnd")) {
                setpostend = (SetPostEnd)a.getAction();

// ... Pause
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetInPause")) {
                setinpause = (SetInPause)a.getAction();

// ... Resume
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetInResume")) {
                setinresume = (SetInResume)a.getAction();
            }

// ... AcceptRunNumber message from Cortex
            else if(a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptRunNumber")) {
                AcceptRunNumber arn = (AcceptRunNumber)a.getAction();
                if(arn.getAcceptrunnumber_Session().equals(_session)){
                    _RunNumberF = Integer.toString(arn.getAcceptrunnumber_Value());

// Broadcast the runnumber to plexuses
                    informCmsgClient(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeRunNumber,true,_RunNumberF,false,null);
                }
            }

// ... Disconnect message from Cortex
            else if(a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Disconnect")) {
                Disconnect mydisc = (Disconnect)a.getAction();
                if(mydisc.getDisconnect_Session().equals(_session)){

// Send disconnect to all components under this medulla object control
// Go over all registered component agents
                    for(String tmpCompName : ALL.keySet()) {

// Send disconnect transition to the agent.
                        if ((tmpCompName != null) && (mydisc !=null))
                            sendACLOntologyAction(mydisc, tmpCompName);
                    }

// @Todo
// Broadcast a dalog cMsg message to all listeners that this control system (i.e. supervisor and its agents ) are
// going to be disconnected from the platform.
                    cMsgMessage dlg = new cMsgMessage();
                    dlg.setSubject(rcSubscription.DalogSupervisor);
                    dlg.setType(rcSubscription.Retire);
                    try {
                        cMsgPayloadItem sessionItem = new cMsgPayloadItem("session", _session);
                        cMsgPayloadItem supervisorItem = new cMsgPayloadItem("supervisor", myName);
                        String[] cmps = new String[ALL.size()];
                        ALL.keySet().toArray(cmps);
                        cMsgPayloadItem componentsItem = new cMsgPayloadItem("components", myName);
                        dlg.addPayloadItem(sessionItem);
                        dlg.addPayloadItem(supervisorItem);
                        dlg.addPayloadItem(componentsItem);

// Send the message
                        cMsgSrv.send(dlg);
                        cMsgSrv.flush(1);

                    } catch (cMsgException e) {
                        e.printStackTrace();
                    }


                }

// Release session
                _session = "";
                try{

// Disconnect from cMsg server
                    if(cMsgSrv!=null) cMsgSrv.disconnect();
                } catch (cMsgException e) {
                    e.printStackTrace();
                }
            }

// ...  AcceptSystem
// Get the system configured to be a subcontrol system of the hierarchy
// ... AcceptSystem
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptSystem")) {
                AcceptSystem myAccept = (AcceptSystem)a.getAction();
                _mySubSystems.add(myAccept.getAcceptsystem_System());
            }
// ... AcceptComponent
// Get the component which is configured for this particular run. This is request from the Cortex.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptComponent")) {
                AcceptComponent myAccept = (AcceptComponent)a.getAction();

// Get component
                tmpComp = myAccept.getAcceptcomponent_Component();

// Fill the ComponentTypes hashmap
                ComponentTypes.put(tmpComp.getComponent_Name(), tmpComp.getComponent_Type());

// Fill the CompReportInterval hashmap with the cool values, otherwise default values = 1.0 sec.
                if(tmpComp.getComponent_Reportinginterval()!=null){
                    CompReportInterval.put(tmpComp.getComponent_Name(), tmpComp.getComponent_Reportinginterval());
                }
                else {
                    CompReportInterval.put(tmpComp.getComponent_Name(),PCRI);
                }

// If this is not a FILE type component
                if(!tmpComp.getComponent_Type().equalsIgnoreCase("FILE")){

// Get the priority of the component
                    Integer tmpPriority = tmpComp.getComponent_Priority();

// Add to the set of the priorities
                    if(tmpPriority!=null)Priorities.add(tmpPriority);

// If Components hash table is not empty and there is a ConcurrentHashMap of the components having the
// same priority, then the status of the new component will be "justborn"
                    if((!Components.isEmpty()) && (tmpcmp = Components.get(tmpPriority))!=null) {
                        tmpcmp.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                    }
                    else {

// Otherwise create a new ConcurrentHashMap for this priority
                        tmpcmp = new ConcurrentHashMap<String,String>();

// Add this component to the ConcurrentHashMap
                        tmpcmp.put(tmpComp.getComponent_Name(),rcConstants.justborn);

// Add the ConcurrentHashMap to the Components tabe
                        Components.put(tmpPriority, tmpcmp);
                    }

// Now if the type of the component is ROC
                    if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.rocType)){
                        Rocs.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        DAQ.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        ALL.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                    }

// TS component, etc. you got it
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.tsType)) {
                        Tss.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        DAQ.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        ALL.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.ebType)) {
                        Ebs.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        DAQ.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        ALL.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.erType)) {
                        Ers.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        DAQ.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        ALL.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.slowcontrolType)) {
                        Slcs.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        ALL.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.analysisType)) {
                        Anas.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                        ALL.put(tmpComp.getComponent_Name(),rcConstants.justborn);
                    }
                }

// @Todo
// Broadcast a dalog cMsg message to all listeners that this control system (i.e. supervisor and its agents ) are
// statrting a new subcontrol system
//                cMsgMessage dlg = new cMsgMessage();
//                dlg.setSubject(rcSubscription.DalogSupervisor);
//                dlg.setType(rcSubscription.Employ);
//                try {
//                    cMsgPayloadItem sessionItem = new cMsgPayloadItem("session", _session);
//                    cMsgPayloadItem supervisorItem = new cMsgPayloadItem("supervisor", myName);
//                    String[] cmps = new String[ALL.size()];
//                    ALL.keySet().toArray(cmps);
//                    cMsgPayloadItem componentsItem = new cMsgPayloadItem("components", cmps);
//                    dlg.addPayloadItem(sessionItem);
//                    dlg.addPayloadItem(supervisorItem);
//                    dlg.addPayloadItem(componentsItem);
//
//// Send the message
//                    cMsgSrv.send(dlg);
//                    cMsgSrv.flush(1);
//
//                    System.out.println(" Sending dalog ... "+_session+" "+myName);
//                    for(int i = 0; i<cmps.length;i++) System.out.println(cmps[i]);
//
//                } catch (cMsgException e) {
//                    e.printStackTrace();
//                }

            }

// ... AccpetGroupStatus
// Deputy suoervisor sends the group of the status informations
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptGroupStatus")) {
                int nfail;

// Get the ontology action object
                AcceptGroupStatus myGroupStatus = (AcceptGroupStatus)a.getAction();

// List of the statusquant objects sent by the deputy supervisor
                rca.util.leap.List groupstates = myGroupStatus.getAcceptgroupstatus_Status();
                rca.util.leap.Iterator it = groupstates.iterator();
                while(it.hasNext()){
                    StatusQuant mq = (StatusQuant)it.next();

// Get the component
                    tmpComp = mq.getStatusquant_Producer().getProducer_Component();
                    if(tmpComp !=null){

// Get the status string
                        String tmpstatus = mq.getStatusquant_Status();

// Record the status in the concurent hashmaps
// Get the priority of the component sending the status info
                        Integer tmpPriority = tmpComp.getComponent_Priority();

// Update all relevant ConcurrentHashMaps
                        if((!Components.isEmpty()) && (tmpcmp = Components.get(tmpPriority))!=null) {
                            tmpcmp.put(tmpComp.getComponent_Name(),tmpstatus);
                        }
                        else {
                            tmpcmp = new ConcurrentHashMap<String,String>();
                            tmpcmp.put(tmpComp.getComponent_Name(),tmpstatus);
                        }
                        Components.put(tmpPriority, tmpcmp);
                        if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.rocType)){
                            Rocs.put(tmpComp.getComponent_Name(),tmpstatus);
                            DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                            ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                        }
                        else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.tsType)) {
                            Tss.put(tmpComp.getComponent_Name(),tmpstatus);
                            DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                            ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                        }
                        else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.ebType)) {
                            Ebs.put(tmpComp.getComponent_Name(),tmpstatus);
                            DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                            ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                        }
                        else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.erType)) {
                            Ers.put(tmpComp.getComponent_Name(),tmpstatus);
                            DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                            ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                        }
                        else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.slowcontrolType)) {
                            Slcs.put(tmpComp.getComponent_Name(),tmpstatus);
                            ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                        }
                        else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.analysisType)) {
                            Anas.put(tmpComp.getComponent_Name(),tmpstatus);
                            ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                        }

// Get the status severity
                        String tmpseverity = mq.getStatusquant_Severity();
                        if((tmpseverity!=null) && (tmpseverity.equalsIgnoreCase("Warning"))){

// Warning means that component is not reporting. set Fails ConcurrentHashMap
// Set or increment fails number in the Fails table
                            if(Fails.containsKey(tmpComp.getComponent_Name())){
                                nfail = Fails.get(tmpComp.getComponent_Name());
                                nfail++;
                                Fails.put(tmpComp.getComponent_Name(),nfail);
                            } else Fails.put(tmpComp.getComponent_Name(), 1);
                            String s = "Medulla-Warning:\n "+tmpstatus ;
                            informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,s,false,null);
                            dalogMsg("warning",s);
                        } else if ((tmpseverity!=null) && (tmpseverity.equalsIgnoreCase("Error"))){

// Make sure that durnig the end we skip this componenet because in any cases he will not able
// to respond to the end transition.
                            Fails.put(tmpComp.getComponent_Name(), rcConstants.FAILS_NUMBER);
                            String s = "Medulla-Error:\n"+tmpstatus;
                            informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,s,false,null);
                            dalogMsg("error",s);

// End the run
//                        endrun(tmpComp.getComponent_Name());

                        } else {}
                    } else {
// oops the name of the component is null

                        String s = "Medulla-Error:\n Component name is null.";
                        System.out.println(s);
                        informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,s,false,null);
                        dalogMsg("error",s);
                    }
                }
            }

// Component agents are sending back state info to this RC supervisor agent. Sending components are non CODA components,
// most probably subcontrol system SMS agents
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptState")) {

                if(SM_RULES_ACCEPTED){
// Get the state from the ontology action object
                    AcceptState astate = (AcceptState)a.getAction();
                    CoolState state = astate.getAcceptstate_State();

// Get the component sending the status information
                    tmpComp = astate.getAcceptstate_Producer().getProducer_Component();
                    if(tmpComp !=null){
                        System.out.println(" Medulla .... accepted state ... "+tmpComp.getComponent_Name()+" "+state.getState_Name());

// Add to the ALL component hash
                        ALL.put(tmpComp.getComponent_Name(),state.getState_Name());

// If control has rools then
                        if(itsMe!=null){
                            if(_myRules!=null && !_myRules.isEmpty()){
                                StateMachine st = new StateMachine(state);
                                st.start();
                            }
                        }
                    }
                }
            }

// ... AcceptStatus
// CODA Component agents are sending back status info to Medulla
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptStatus")) {
                int nfail;

// Get the ontology action object
                AcceptStatus myAcceptStatus = (AcceptStatus)a.getAction();

// Get the component sending the status information
                tmpComp = myAcceptStatus.getAcceptstatus_Producer().getProducer_Component();
                if(tmpComp !=null){

// Get the status string
                    String tmpstatus = myAcceptStatus.getAcceptstatus_Status();

// Record the status in the concurent hashmaps
// Get the priority of the component sending the status info
                    Integer tmpPriority = tmpComp.getComponent_Priority();

// Update all relevant ConcurrentHashMaps
                    if((!Components.isEmpty()) && (tmpcmp = Components.get(tmpPriority))!=null) {
                        tmpcmp.put(tmpComp.getComponent_Name(),tmpstatus);
                    }
                    else {
                        tmpcmp = new ConcurrentHashMap<String,String>();
                        tmpcmp.put(tmpComp.getComponent_Name(),tmpstatus);
                    }
                    Components.put(tmpPriority, tmpcmp);
                    if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.rocType)){
                        Rocs.put(tmpComp.getComponent_Name(),tmpstatus);
                        DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                        ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.tsType)) {
                        Tss.put(tmpComp.getComponent_Name(),tmpstatus);
                        DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                        ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.ebType)) {
                        Ebs.put(tmpComp.getComponent_Name(),tmpstatus);
                        DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                        ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.erType)) {
                        Ers.put(tmpComp.getComponent_Name(),tmpstatus);
                        DAQ.put(tmpComp.getComponent_Name(),tmpstatus);
                        ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.slowcontrolType)) {
                        Slcs.put(tmpComp.getComponent_Name(),tmpstatus);
                        ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                    }
                    else if(tmpComp.getComponent_Type().equalsIgnoreCase(rcConstants.analysisType)) {
                        Anas.put(tmpComp.getComponent_Name(),tmpstatus);
                        ALL.put(tmpComp.getComponent_Name(),tmpstatus);
                    }

// Get the status severity
                    String tmpseverity = myAcceptStatus.getAcceptstatus_Severity();
                    if((tmpseverity!=null) && (tmpseverity.equalsIgnoreCase("Warning"))){

// Warning means that component is not reporting.
// Set or increment fails number in the Fails table
                        if(Fails.containsKey(tmpComp.getComponent_Name())){
                            nfail = Fails.get(tmpComp.getComponent_Name());
                            nfail++;
                            Fails.put(tmpComp.getComponent_Name(),nfail);
                        } else Fails.put(tmpComp.getComponent_Name(), 1);
                        String s = "Medulla-Warning:\n "+tmpstatus ;
                        informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,s,false,null);
                        dalogMsg("warning",s);
                    } else if ((tmpseverity!=null) && (tmpseverity.equalsIgnoreCase("Error"))){

// Make sure that durnig the end we skip this componenet because
// in any cases he will not able to respond to the end transition.
                        Fails.put(tmpComp.getComponent_Name(), rcConstants.FAILS_NUMBER);
                        String s = "Medulla-Error:\n"+tmpstatus;
                        informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,s,false,null);
                        dalogMsg("error",s);

// end the run
//                        endrun(tmpComp.getComponent_Name());

                    } else {}
                } else {
// oops the name of the component is null
                    String s = "Medulla-Error:\n Component name is null.";
                    System.out.println(s);
                    informCmsgClient(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeTransition,true,s,false,null);
                    dalogMsg("error",s);
                }
            }

// ... AccpetData
// Component agents are sending back data (string) to Medulla
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptData")) {
                AcceptData myAcceptData = (AcceptData)a.getAction();

// Analyze data and tell plexuses to update their graphs.
                updateData(myAcceptData);
            }

// ... AcceptGroupDtata
// Message from the deputy supervisor
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptGroupData")) {
                AcceptGroupData myGroupData = (AcceptGroupData)a.getAction();

// List of the DataQuant objects sent by the deputy supervisor (Thalaus) agent
                rca.util.leap.List dl = myGroupData.getAcceptgroupdata_Data();
                rca.util.leap.Iterator it = dl.iterator();
                while(it.hasNext()){
                    DataQuant dq = (DataQuant)it.next();

// Analyze data and tell plexuses to update their graphs.
                    updateData(dq);
                }
            }

// ... AcceptTsData
// TS Component sends data (string) to Thalamus
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptTsData")) {

// Data structure from the TS component agent
                String LiveTime=null, Bit1=null, Bit2=null, Bit3=null, Bit4=null, Bit5=null, Bit6=null,
                        Bit7=null, Bit8=null, Bit9=null, Bit10=null, Bit11=null, Bit12=null ;

// If we are in the active state and we are not requested to end the run
                if(getRunState().equalsIgnoreCase(rcConstants.CMD_postgo_ok)){
                    AcceptTsData myAcceptTsData = (AcceptTsData)a.getAction();

// Get the string from the TS agent and separate the tokens by " "
                    StringTokenizer stk = new StringTokenizer(myAcceptTsData.getAccepttsdata_Data().getData_Content());
                    if(stk.hasMoreTokens())LiveTime = stk.nextToken();
                    if(stk.hasMoreTokens())Bit1 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit2 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit3 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit4 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit5 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit6 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit7 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit8 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit9 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit10 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit11 = stk.nextToken();
                    if(stk.hasMoreTokens())Bit12 = stk.nextToken();
                } else {
                    LiveTime = "0";
                    Bit1 = "0";
                    Bit2 = "0";
                    Bit3 = "0";
                    Bit4 = "0";
                    Bit5 = "0";
                    Bit6 = "0";
                    Bit7 = "0";
                    Bit8 = "0";
                    Bit9 = "0";
                    Bit10 = "0";
                    Bit11 = "0";
                    Bit12 = "0";
                }
                informCmsgClient(rcSubscription.RcGuiSubjTsData+"_"+_session, rcSubscription.RcGuiTypeTransition,true,LiveTime+
                        " "+Bit1+" "+Bit2+" "+Bit3+" "+Bit4+" "+Bit5+" "+
                        Bit6+" "+Bit7+" "+Bit8+" "+Bit9+" "+Bit10+" "+Bit11+" "+Bit12, false, null);
            }
            else {
            }
        }
    }

    /**
     * Inner class responsible for organizing control system state machine.
     */
    private class StateMachine extends Thread{

        public CoolState mys;

        StateMachine(CoolState st){
            mys = st;
        }

        public void run(){
            execrule(mys);
        }
    }


}
