/*---------------------------------------------------------------------------*
*  Copyright (c) 2008        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.agents;

import rca.content.lang.Codec;
import rca.content.lang.sl.SLCodec;
import rca.content.onto.OntologyException;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.content.AgentAction;
import rca.core.AID;
import rca.domain.DFService;
import rca.domain.FIPAAgentManagement.DFAgentDescription;
import rca.domain.FIPAAgentManagement.ServiceDescription;
import rca.domain.FIPAException;
import rca.domain.FIPANames;
import rca.lang.acl.ACLMessage;
import org.jlab.coda.runcontrol.ontologies.CodaPlatform.CodaPlatformAdminOntology;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.*;
import org.jlab.coda.runcontrol.util.*;
import org.jlab.coda.runcontrol.util.sh.SysExe;
import org.jlab.coda.cMsg.*;
import java.io.*;
import java.util.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArraySet;


/**
 * This is the basic SMSAgent agent class, representative of the physical component in the platform. It incapsulates
 * inter porocess communication protocols (dp, shell, cMsg) used to  communicate with the component.
 * This class is used also to build and deploy State Machine Supervisor (SMS) agents
 *
 * @author Vardan Gyurjyan
 *         Date: June 7, 2006
 *         Time: 1:00:30 PM
 */
public class SMSAgent extends CAgent {

    /** Reference to the CoolComponent object.*/
    private CoolComponent itsMe;

    /** All the registered states for the components of the control system*/
    private rca.util.leap.List _myStates;

    /** Rules of the control system*/
    private rca.util.leap.List _myRules;

    /** AID of the supervisor agent.  */
    private AID mySupervisor;

    /** cMsg domain server (Hepatic) connection object */
    cMsg cMsgSrv;

    /** The name of this agent which is the same as the representing physical client name*/
    private String myName;

    /** Instance of the class incapsulating Unix shell IPC commands*/
    public SysExe mySysExe;

    /** The current state name of the agent*/
    private String myPrivateState = rcConstants.booted;

    /** The current state object of the agent*/
    private CoolState myStateObject;

    /** Internal field showing the agents registration status with the DF of the AFECS platform.*/
    private boolean  _RegisteredFlag    = false;

    /** The type of the agent, i.e. ROC EB, ER, etc.*/
    private String           _agentType;

    /** The name of the agent, i.e. R1 EB1, ER1, etc.*/
    private String           _agentName;

    /** Priority of thie agent.*/
    private Integer          _agentPriority = null;

    /** Reporting interval for this agent */
    private int _reportingInterval;

    /** State reporting thread*/
    private StateReport _stateReportingThread;

    /** Session of the agent. The default value is set to the env deffinition*/
    private String _agentSession = System.getenv("SESSION");

    /** Runtype of the agent, i.e. configuration name.*/
    private String _agentRuntype ="";

    /** Indicats that state machine rules are registered*/
    private boolean SM_RULES_ACCEPTED = false;

    /** Indicates that syaye machine algorithm is in progress*/
    private boolean SM_ACTIVE = false;

    /** Defines the content language, used for inter-agent communications*/
    private Codec  codec = new SLCodec();

    /** Hashtable containing the state machine supervisor agent controled components names and their states */
    private ConcurrentHashMap<String, String>  ComponentStates = new ConcurrentHashMap<String, String>();

    /** Subcontrol systems for this system*/
    private CopyOnWriteArraySet<CoolSystem> _mySubSystems = new CopyOnWriteArraySet<CoolSystem>();

    /** List containing the names of the SMS agents*/
    private List<String> _mySMS = new ArrayList<String>();


    /**
     * Setup method of the agent. It organizes the
     * ontology registration, and agent behaviour scheduling.
     */
    protected void setup(){

// This is the given name when the agent was started by the Hepatic.
        myName = getLocalName();

// Register supported languages
        getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);

// Register the AFECS platform administration ontology for Hepatic conversations
        getContentManager().registerOntology(CodaPlatformAdminOntology.getInstance());

// Register the coda run control ontology for Cortex and Medulla conversations
        getContentManager().registerOntology(CodaRunControlOntology.getInstance());

// Add the responder behaviour.
// This is the behaviour where we process request information from the other agents in the platform.
// The actual processing method is defined in this class but called from the GenResponderBehaviour object
        addBehaviour(new GenRespondBehaviour(this, 0));

// DF registration
        if(!_RegisteredFlag){
            DFAgentDescription dfd = new DFAgentDescription();
            dfd.setName(getAID());
            ServiceDescription sd = new ServiceDescription();
            sd.setType("SMSAgent");
            sd.setName(myName);
            sd.setOwnership("JLABCODA");
            dfd.addServices(sd);
            try {
                DFService.register(this,dfd);
                _RegisteredFlag = true;
            } catch (FIPAException e) {
                System.err.println(getLocalName()+" registration with DF not succeeded. Reason: "+e.getMessage());
                _RegisteredFlag = false;
            }
        }

// Connect cMsg domain server (runs within Hepatic). N.B. we have noticed performance decline for the case when
// we have >500 agents conencting to the cMsg domain server.
        connectTocMsg();
    }


    /**
     * This method is called at the agent shutdown
     */
    protected void takeDown()
    {
        try { DFService.deregister(this); }
        catch (Exception e) {e.printStackTrace();}
    }

    /**
     * Sends ACL message to all subcontrol system supervisors, asking to start reporting their statuses.
     */
    public void SubSystemsStartReporting(){

// AgentAction object
        AgentAction transConcept = null;

// Create a reset object
        ResultNotify  transition = new ResultNotify();
        transition.setResultNotify_Component(itsMe);
        transConcept = transition;

        // Go over all subsystems registered with this Medulla
        for(CoolSystem sys: _mySubSystems){
            // send acl message to the subcontrol system supervisor
            sendTransition(transConcept,sys.getSystem_Supervisor());
        }
    }

    /**
     * Sends ACL message to all subcontrol system supervisors, asking to stop reporting their statuses.
     */
    public void SubSystemsStopReporting(){

// AgentAction object
        AgentAction transConcept = null;

// Create a reset object
        StopReporting  transition = new StopReporting();
        transConcept = transition;

        // Go over all subsystems registered with this Medulla
        for(CoolSystem sys: _mySubSystems){
            // send acl message to the subcontrol system supervisor
            sendTransition(transConcept,sys.getSystem_Supervisor());
        }
    }

    /**
     * This method send an ACL message to an agent on the AFECS platform.
     *
     * @param transition               AgentAction object
     * @param component                Name of an agent on the platform
     */
    public void sendTransition (AgentAction transition, String component) {

// Create an AID of the component going to receive the transition
        AID receiver = new AID(component,AID.ISLOCALNAME);

// Create an ACL message
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{

// Use the coda runcontrol ontology
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Add receiver to the message
            msg.addReceiver(receiver);

// We are using fipa_sl0 content language
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);

// Set the ontology
            msg.setOntology(CodaRunControlOntology.NAME);

// Add user defined parameter. The user parameter is the name of this agent, i.e.
// control system supervisor name.
            msg.addUserDefinedParameter(getAID().getLocalName(),receiver.getLocalName());

// Create an action object
            Action a = new Action();

// Add actor to the action object
            a.setActor(receiver);

// Set an actual action, passed as a parameter to the action object
            a.setAction(transition);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        }
        catch(Exception e) {
            e.printStackTrace();
        }

// Create and add a behaviour to the message queue
        addBehaviour(new GenRequestBehaviour(this, msg, rcConstants.debugError));
    }


    /**
     * Sending dalog message
     * @param severity      severity of the message
     * @param message       content of the message
     */
    private void dalogMsg(int severity, String message){

        String subject;
        String type = null;
        cMsgPayloadItem sever = null, state;

//        System.out.println("SMS .... dalogmsg ........ "+myName +" - "+_agentType);
        if(rcSubscription.expid!=null && _agentSession !=null && _agentRuntype !=null && _agentType!=null && myName!=null){
            cMsgMessage msg = new cMsgMessage();
            subject = rcSubscription.expid+"/"+ _agentSession +"/"+ _agentRuntype +"/"+_agentType+"/"+myName;
            try {
                // payload item state
                state = new cMsgPayloadItem("state", myPrivateState);
                if(severity<5){
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","INFO."+severity);

                }
                else if(severity>=5 && severity<9) {
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","WARN."+severity);
                }
                else if(severity>=9 && severity<13) {
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","ERROR."+severity);
                }
                else if(severity>=13 && severity<17){
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","ERRORS."+severity);
                }
                if(subject!=null && type !=null){
                    cMsgPayloadItem p_expid = new cMsgPayloadItem("EXPID",rcSubscription.expid);
                    cMsgPayloadItem p_session = new cMsgPayloadItem("session", _agentSession);
                    cMsgPayloadItem p_config = new cMsgPayloadItem("config", _agentRuntype);
                    cMsgPayloadItem p_type = new cMsgPayloadItem("codaClass",_agentType);
                    cMsgPayloadItem p_codaName = new cMsgPayloadItem("codaName",myName);
                    msg.addPayloadItem(p_expid);
                    msg.addPayloadItem(p_session);
                    msg.addPayloadItem(p_config);
                    msg.addPayloadItem(p_codaName);
                    msg.addPayloadItem(p_type);

                    msg.setSubject(subject);
                    msg.setType(type);
                    msg.setText(message);
                    msg.addPayloadItem(sever);
                    msg.addPayloadItem(state);
                    cMsgSrv.send(msg);
                    cMsgSrv.flush(1);
                }
            } catch (cMsgException e) {
                e.printStackTrace();
            }
            //        System.out.println(subject+" "+type);
        }
    }


    /**
     * This is the implementatioin of the abstract method defined in the CAgent class.
     * This method is going to be called by the BaseicRequestBehaviour object to process
     * the iform messages coming back to this agent as a result of initiating the request
     * to some agent on the platfrom.
     *
     * @param msg      ACLMessage object passed from the GenRequestBehaviour object.
     */
    protected void processInform(ACLMessage msg) {
    }


    /**
     * Method connects to the cMsg Server started by the Hepatic. This will allow this agent
     * to respond directly to the cMsg requests from outside of the platform.
     */
    protected void connectTocMsg(){

        try {
            if((cMsgSrv!=null) && (cMsgSrv.isConnected())) {
// We are connected do nothing
            }

// Start a new connection
            else {

// Disconnect first if there is cMsgSrv object
                if(cMsgSrv!=null) cMsgSrv.disconnect();
                cMsgSrv = new cMsg(rcSubscription.PlatformcMsgServerUDL, myName, rcSubscription.PlatformcMsgServerDescription);
                cMsgSrv.connect();

// Enable message reception
                cMsgSrv.start();

            }
        } catch (cMsgException ee) {
            try {
                cMsgSrv.disconnect();
                cMsgSrv = null;
            } catch (cMsgException e) {
                cMsgSrv = null;
                e.printStackTrace();
            }
            ee.printStackTrace();
        }
    }



    /**
     * This method gets called whenever the ACL message (inter-agent communication) is reached to this agent.
     * This is the implementatioin of the abstract method defined in the <clas>CAgent</clas>, which is
     * going to be called by the <clas>BaseicResponderBehaviour</clas> object to process the request to
     * this agent. Most of the ACL message request are coming from the supervisor agent
     * <clas>Medulla</clas> and/or <clas>Cortex</clas>.
     *
     * @param request      ACLMessage request object, passed from the GenResponderBehaviour object.
     * @return             ACLMessage object
     */
    protected ACLMessage processRequest(ACLMessage request,Action a) {


        CoolOption option;
        String tmpStr = _agentSession;


// Create an agent communication reply object
        ACLMessage reply = request.createReply();
        try {

// This set of the messages are usually coming from the Cortex.
// Cortex uses Differentiate action to turn this "SMSAgent" agent in to the specific type of an agent, which will
// represent the physical client in the AFECS platform. Differentiated (specialized) agents will register
// their _agentTypes with the DF.
//==>
            if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Differentiate")) {

                String reportsto;

// Clear local hashtables
                if(ComponentStates != null) ComponentStates.clear();
                if(_mySMS!=null)_mySMS.clear();
                if(_mySubSystems !=null) _mySubSystems.clear();
                if(_myStates!=null)_myStates.clear();
                if(_myRules !=null) _myRules.clear();
                SM_RULES_ACCEPTED = false;

// Get the differentiate ontology object
                Differentiate myDifferentiate = (Differentiate)a.getAction();

// Intialization of the variables
                option          = myDifferentiate.getDifferentiate_Option();
                itsMe           = myDifferentiate.getDifferentiate_Component();
                _agentType      = itsMe.getComponent_Type();
                _agentName      = itsMe.getComponent_Name();


// Check the state of the agent. If it is booted re-asign the session, based on the cortex configure request.
                if(myPrivateState !=null){
                    if ( ((myPrivateState.equals(rcConstants.booted)) || (_agentSession.equals("")) ) ){
                        _agentSession = itsMe.getComponent_Session();


// If it is not booted then we can not change already registered session for the agent.
                    }
                    else if(!_agentSession.equals(itsMe.getComponent_Session())){
                        String s ="Request to configure "+myName+" for the session = "+itsMe.getComponent_Session()+
                                "\n Request is denied: Agent "+myName+" is already configured in the session = *"+ _agentSession +"*";
                        System.out.println(s);
                        informPlexuses(rcSubscription.Stemcell, rcSubscription.RcGuiTypeStemCell,s);
                        dalogMsg(14,s);
                        s = "Severe error";
                        informPlexuses(rcSubscription.RcGuiSubjError+itsMe.getComponent_Session(), rcSubscription.RcGuiTypeStemCell,s);

                        try{

// Create and send done action message (ACL message)  to the cortex, informing that this agent is
// already registered in the different session
                            Done d = new Done();
                            d.setAction(a);
                            getContentManager().fillContent(reply, d);
                        } catch (Codec.CodecException e) {
                            e.printStackTrace();
                        } catch (OntologyException e) {
                            e.printStackTrace();
                        }
                        reply.setPerformative(ACLMessage.INFORM);
                        reply.setContent(_agentSession+" "+myName+" Differentiation failed");
                        return reply;
                    }

// If myPrivateState is null then set the session
                } else {
                    _agentSession = itsMe.getComponent_Session();

                }
                _agentRuntype = itsMe.getComponent_Runtype();
                _agentPriority  = itsMe.getComponent_Priority();

                // get reporting interval (reporting periodicity)
                try{
                    if(itsMe.getComponent_Reportinginterval()!=null){
                        _reportingInterval = Integer.parseInt(itsMe.getComponent_Reportinginterval());
                        // start monitoring thread
                        if(_stateReportingThread!=null){
                            _stateReportingThread.shouldrun=false;
                        }
                        // Start the monitoring thread
                        _stateReportingThread = new StateReport(_reportingInterval);
                        _stateReportingThread.start();

                    }
                } catch(NumberFormatException e){ e.printStackTrace();}

// Dalog message cool file control
//                DALOG_FILTER = component.getComponent_Filtersdalog();

// Tell subcontrol systems if any registered with this configuration to start reporting status information
                SubSystemsStartReporting();

            }

// ... AcceptComponent from Cortex
// Get the component which is configured for this particular run. This is request from the Cortex.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptComponent")) {
                AcceptComponent myAccept = (AcceptComponent)a.getAction();

// Get component
                CoolComponent tmpComp = myAccept.getAcceptcomponent_Component();

// Fill the ComponentStates hashmap
                ComponentStates.put(tmpComp.getComponent_Name(),rcConstants.unknown);

// Tell that component to report periodically its state to this agent.
// Here we construct AID from the name of the component,
                reportMe(new AID(tmpComp.getComponent_Name(),AID.ISLOCALNAME));
            }

// ... AcceptStates from Cortex.
// This request informs this SMS aget what are the all registered states within the control system,
// for which this agent is registered to be the supervisor.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptStates")) {
                AcceptStates ast = (AcceptStates)a.getAction();
                if(ast.getAcceptstates_List()!=null &&  !(ast.getAcceptstates_List().isEmpty())){
                    _myStates = ast.getAcceptstates_List();
                }
            }

// ... AcceptRules from Cortex.
// This request informs this SMS aget what are the all registered rules within the control system (state machine),
// for which this agent is registered to be the supervisor.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptRules")) {
                AcceptRules rl = (AcceptRules)a.getAction();
                if(rl.getAcceptrules_List()!=null &&  !(rl.getAcceptrules_List().isEmpty())){
                    _myRules = rl.getAcceptrules_List();
                    dumprule(_myRules);
                    SM_RULES_ACCEPTED = true;
                }
            }

// ... ResultNotify.
// This request tells this agent to report its state not only to Medulla but also to other SMS agents.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.ResultNotify")) {
                ResultNotify rnt = (ResultNotify)a.getAction();
                if(rnt.getResultNotify_Component()!=null && rnt.getResultNotify_Component().getComponent_Name()!=null)
                    _mySMS.add(rnt.getResultNotify_Component().getComponent_Name());
            }

// ... ActionMoveTo. This request tells this agent to move to the state.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.CoolActionMoveTo")) {

                CoolActionMoveTo mt = (CoolActionMoveTo)a.getAction();

// Get the state
                CoolState st = mt.getActionMoveto_State();

// Get the name of the state
                String stname = st.getState_Name();
                // actually move to the requested state if we are not already in that state
                if(!stname.equals(myPrivateState)){
                    // get the states I depend on for this state
                    rca.util.leap.List dst = st.getState_Dependson();
                    if(dst!=null){
                        Iterator lit = dst.iterator();
                        while(lit.hasNext()){
                            CoolState tmpst = (CoolState)lit.next();
                            String desiredStateName = tmpst.getState_Name();
                            String compNameUnderControl = tmpst.getState_Isfor().getComponent_Name();
                            if(ComponentStates.containsKey(compNameUnderControl)){
                                String recordedStateName  = ComponentStates.get(compNameUnderControl);
                                if(!recordedStateName.equals(desiredStateName)){
                                    // send a moveTo state message to the agent under control of this SMS
                                    moveToState(compNameUnderControl,tmpst);
                                }
                            }
                        }
                    }
                }
            }

// Component agents are sending back status info to this agent. N.B. in this case this agent is SMS.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.AcceptState")) {

                if(SM_RULES_ACCEPTED){

// Get the state from the ontology action object
                    AcceptState astate = (AcceptState)a.getAction();
                    CoolState state = astate.getAcceptstate_State();

// Get the component sending the status information
                    CoolComponent tmpComp = astate.getAcceptstate_Producer().getProducer_Component();
                    if(tmpComp !=null){
//                        System.out.println(" SMS .... accepted state ... "+tmpComp.getComponent_Name()+" "+state.getState_Name());

// Add to the componentStates hash
                        ComponentStates.put(tmpComp.getComponent_Name(),state.getState_Name());

// If control has rools then
                        if(itsMe!=null){
                            if(_myRules!=null && !_myRules.isEmpty() && !SM_ACTIVE){
                                StateMachine st = new StateMachine(state);
                                st.start();
                            }
                        }
                    }
                }
            }

// Create and send done action message (ACL message)  to the client agent.
            Done d = new Done();
            d.setAction(a);
            getContentManager().fillContent(reply, d);
        } catch (Codec.CodecException e) {
            e.printStackTrace();
        } catch (OntologyException e) {
            e.printStackTrace();
        }
        reply.setPerformative(ACLMessage.INFORM);

// tmpStr contains information for e.g. runtype , etc.
        reply.setContent(tmpStr);
        return reply;
    }


    /**
     * Informs status to receiver (Medulla most of the time).
     * This method sends to the receiver an <clas>AcceptStatus</clas> coda ontology ACL message,
     * describing the producer, i.e. the process that produces information and timestamp.
     *
     * @param status               The state of the component
     * @param transition           The requested transition
     * @param severity             The severity of the message
     * @param receiver             The receiver of the message
     */
    public void informStatus(String status, String transition,String severity, AID receiver) {

// Current state of this agent
        if( (transition !=null) && (!transition.equals("")))
            myPrivateState = transition ;
        else {
            myPrivateState =rcConstants.booted;
            return;
        }

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Create a codacomponent ontology concept object
            CoolComponent cm = new CoolComponent();
            cm.setComponent_Name(_agentName);
            cm.setComponent_Type(_agentType);
            cm.setComponent_Priority(_agentPriority);
            cm.setComponent_Runtype(_agentRuntype);
            cm.setComponent_Session(_agentSession);

// Create a Producer ontology concept object
            Producer pr = new Producer();
            pr.setProducer_Component(cm);

// Create a coda timestamp object
            CodaTimeStamp tm = new CodaTimeStamp();
            tm.setTimestamp_Value(new Date());

// Create a acceptStatus ontology action object
            AcceptStatus myaccept = new AcceptStatus();
            myaccept.setAcceptstatus_Producer(pr);
            myaccept.setAcceptstatus_Timestamp(tm);
            myaccept.setAcceptstatus_Status(status);
            myaccept.setAcceptstatus_Transition(transition);
            if(severity!=null)myaccept.setAcceptstatus_Severity(severity);

// Create an action transport object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(myaccept);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, 0));
    }


    /**
     * Informs state to receiver (In this case this is the SMS agent).
     * This method sends to the receiver a <clas>AcceptState</clas> coda ontology ACL message, describing the producer
     * i.e. the components that produces information and timestamp.
     *
     * @param stateName               The state of the component
     * @param severity                of the component
     */
    public void informStateSMS(String stateName, String severity) {
        String tmps;
        AID tmpaid;
        if(_mySMS!=null && !_mySMS.isEmpty()){
            for (String a_mySMS : _mySMS) {
                tmps = a_mySMS;

// Create the acl message object
                ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
                try {
                    getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                    tmpaid = new AID(tmps, AID.ISLOCALNAME);

// Preparing the message
                    msg.addReceiver(tmpaid);
                    msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                    msg.setOntology(CodaRunControlOntology.NAME);
                    msg.addUserDefinedParameter(_agentName, tmpaid.getLocalName());

// Create a codacomponent ontology concept object
                    CoolComponent cm = new CoolComponent();
                    cm.setComponent_Name(_agentName);
                    cm.setComponent_Type(_agentType);
                    cm.setComponent_Priority(_agentPriority);
                    cm.setComponent_Runtype(_agentRuntype);
                    cm.setComponent_Session(_agentSession);

// Create a Producer ontology concept object
                    Producer pr = new Producer();
                    pr.setProducer_Component(cm);

// Create a coda timestamp object
                    CodaTimeStamp tm = new CodaTimeStamp();
                    tm.setTimestamp_Value(new Date());

// Create coolState object
                    CoolState myState = new CoolState();
                    myState.setState_Name(stateName);
                    myState.setState_Isfor(cm);

// Create a acceptStatus ontology action object
                    AcceptState myaccept = new AcceptState();
                    myaccept.setAcceptstate_Producer(pr);
                    myaccept.setAcceptstate_Timestamp(tm);
                    myaccept.setAcceptstate_State(myState);
                    myaccept.setAcceptstate_Severity(severity);

// Create an action transport object
                    Action a = new Action();
                    a.setActor(tmpaid);
                    a.setAction(myaccept);

// Fill the message content
                    getContentManager().fillContent(msg, a);
                } catch (Exception e) {
                    e.printStackTrace();
                }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
                addBehaviour(new GenRequestBehaviour(this, msg,0));
            }
        }
    }


    /**
     * Informs receiver to move to the state
     * This method sends to the receiver a <clas>CoolActionMoveTo</clas> coda ontology ACL message,
     * describing the desired state. This method is used by the SMS agent.
     *
     * @param receiver            The name of the receiver
     * @param state               The CoolState object of the component
     */
    public void moveToState(String receiver, CoolState state) {

//        System.out.println("... moveTO ..... "+receiver+" "+state.getState_Name());
        AID tmpaid;

        if(receiver.equals(myName)){
            if(!myPrivateState.equals(state.getState_Name())){
                myPrivateState = state.getState_Name();
                myStateObject = state;
                dalogMsg(state.getState_Severity(), myPrivateState);
                System.out.println(myName+" ............. STAET CHANGE TO .......... "+myPrivateState);
            }
        }
        else{
// Create the acl message object
            ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
            try{
                getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                tmpaid = new AID(receiver,AID.ISLOCALNAME);

// Preparing the message
                msg.addReceiver(tmpaid);
                msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                msg.setOntology(CodaRunControlOntology.NAME);
                msg.addUserDefinedParameter(_agentName,tmpaid.getLocalName());

// Create CoolActionMoveTo ontology action object
                CoolActionMoveTo mt = new CoolActionMoveTo();
                mt.setActionMoveto_State(state);

// Create an action transport object
                Action a = new Action();
                a.setActor(tmpaid);
                a.setAction(mt);

// Fill the message content
                getContentManager().fillContent(msg, a);
            } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
            addBehaviour(new GenRequestBehaviour(this, msg, 0));
        }
    }


    /**
     * Informs receiver to execute a process
     * This method sends to the receiver a <clas>CoolActionExecute</clas> coda ontology ACL message,
     * describing the desired process. This method is used by the SMS agent.
     *
     * @param receiver            The name of the receiver
     * @param process             The CoolProcess object
     */
    public void executeProcess(String receiver, CoolProcess process) {

        AID tmpaid;

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);
            tmpaid = new AID(receiver,AID.ISLOCALNAME);

// Preparing the message
            msg.addReceiver(tmpaid);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,tmpaid.getLocalName());

// Create CoolActionMoveTo ontology action object
            CoolActionExecute ex = new CoolActionExecute();
            ex.setActionExecute_Process(process);

// Create an action transport object
            Action a = new Action();
            a.setActor(tmpaid);
            a.setAction(ex);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, 0));
    }


    /**
     * Method sends the data to the receiver agent, with the complete description of the
     * command name, used to retrive the datat from the physical component, state of the execution, etc.
     *
     * @param receiver           Agent who will receive the message
     * @param data               Process execution resulting data
     * @param state              Status or state of the component
     * @param commandname        Actual command name of the process, being executed to retrive the data
     */
    public void informResult(CoolData data, String state, String commandname, AID receiver) {

        if(this.countBlockedMessages()<=10){

// Create and fill the acl message object
            ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF) ;
            try{

// Use codaruncontrol ontology
                getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                msg.addReceiver(receiver);
                msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                msg.setOntology(CodaRunControlOntology.NAME);
                msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Define the component concept sending the result
                CoolComponent cm = new CoolComponent();
                cm.setComponent_Name(_agentName);
                cm.setComponent_Type(_agentType);
                cm.setComponent_Priority(_agentPriority);
                cm.setComponent_Runtype(_agentRuntype);
                cm.setComponent_Session(_agentSession);

// Define the producer concept, necessary for the acceptData action object
                Producer pr = new Producer();
                pr.setProducer_Component(cm);

// Get the time of this response
                CodaTimeStamp tm = new CodaTimeStamp();
                tm.setTimestamp_Value(new Date());

// Create and fill AcceptData codaruncontrol ontology action.
                AcceptData myaccept = new AcceptData();
                myaccept.setAcceptdata_Data(data);
                myaccept.setAcceptdata_Producer(pr);
                myaccept.setAcceptdata_Process(_agentName+":"+commandname);
                myaccept.setAcceptdata_Timestamp(tm);
                myaccept.setAcceptdata_Processstate(state);

// Create an action object
                Action a = new Action();
                a.setActor(receiver);
                a.setAction(myaccept);

// Fill the content of the message
                getContentManager().fillContent(msg, a);
            } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour
            addBehaviour(new GenRequestBehaviour(this, msg, 0));
        }
    }



    /**
     * Method sends <clas>ResultNotify</clas> ontology action to the agents under SMS control.
     * This is used to organize State Machine Supervisor algorithm.
     *
     * @param receiver                Agent who will receive the message
     */
    public void reportMe(AID receiver) {

// Create and fill the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF) ;
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Define the component concept sending the result
            CoolComponent cm = new CoolComponent();
            cm.setComponent_Name(_agentName);
            cm.setComponent_Type(_agentType);
            cm.setComponent_Priority(_agentPriority);
            cm.setComponent_Runtype(_agentRuntype);
            cm.setComponent_Session(_agentSession);

            ResultNotify rn = new ResultNotify();
            rn.setResultNotify_Component(cm);

// Action object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(rn);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

//Create and add a behaviour
        addBehaviour(new GenRequestBehaviour(this, msg, 0));
    }






    /**
     * This method executes the Shell command using the global fields of this class,
     * assigned at the constructor.
     *
     * @param command                 Shell command name beeing executed
     * @param expectedOk              This string will be compared with the
     *                                captured IO of the shell execution
     * @return                        rcConstants.failed or the name of the requested command name
     */
    public String runShellN(String command, String expectedOk){

        String res = null;
        String result;
        StringTokenizer st;
        List<String> l;
        BufferedReader myReader;

// Unix shell command execution IO stream, returned by the SysExe object.
        InputStream iostream;

// Actual execution of the command
        l = new ArrayList<String>();

// Tokenize the command.
        st = new StringTokenizer(command);

// Add the tokens into the list
        while(st.hasMoreTokens())l.add(st.nextToken());

// Create a string array
        String[] ta = new String[l.size()];

// Convert the list into the array of strings
        l.toArray(ta);
        String s =myName+"_Info: executing "+command+" ...";
        System.out.println(s);
        informPlexuses(rcSubscription.RcGuiSubjWarning+"_"+ _agentSession, rcSubscription.RcGuiTypeStemCell,s);
        dalogMsg(1,s);

// Execute the unix command and get an IO inputstream of the stdio
        iostream = mySysExe.run_unix_stream(ta);
        if(iostream!=null){
            try {

// Create buffereader using the iostream.
                myReader = new BufferedReader(new InputStreamReader(iostream));

// Atention: Check command will return only one line summary.
                res =myReader.readLine();
                myReader.close();
            } catch (IOException e) {
                e.printStackTrace();
                return rcConstants.failed;
            }

// N.B. returned null iostream might be ok for some processes
// We do the following if expectedok is defined for the command
            if(expectedOk !=null){

// See if the execution of the main command (not the check command) is ok.
                if(!res.equalsIgnoreCase(expectedOk)) result = rcConstants.failed;
                else result = command;
            } else result = command;
        } else result = rcConstants.failed;
        s =myName+"_Info: completed "+command+": stdio => \n"+res;
        System.out.println(s);
        informPlexuses(rcSubscription.RcGuiSubjWarning+"_"+ _agentSession, rcSubscription.RcGuiTypeStemCell,s);
        dalogMsg(1,s);
        return result;
    }


    /**
     * This method sends the status type information to all plexuses or user interfaces through cMsg (cMsg domain)ipc
     *
     * @param subject         Subject of the message. In some casses indicates the ceverity
     *                        of the message. for e.g. ERROR, WARNING, INFO, etc.
     * @param type :          Type of the message. In some casses it indicates the name of the transition
     *                        during which the message was generated.
     * @param info            Message content (String)
     */
    private void informPlexuses(String subject, String type, String info) {

        if((cMsgSrv!=null) && (cMsgSrv.isConnected())) {
            cMsgMessage msg = new cMsgMessage() ;
            msg.setSubject(subject);
            msg.setType(type);

            msg.setText(info);
            try {
                cMsgSrv.send(msg);
                cMsgSrv.flush(1);
            } catch (cMsgException ee) {
                try {
                    cMsgSrv.disconnect();
                    cMsgSrv = null;
                } catch (cMsgException e) {
                    cMsgSrv = null;
                    e.printStackTrace();
                }
                ee.printStackTrace();
            }
        }
    }

    /**
     * Comparing two states
     *
     * @param st1 state1
     * @param st2 state2
     * @return the result of the comparison
     */

    public boolean eQuals(CoolState st1, CoolState st2){
        boolean result = false;

        // see if the components are the same for which the states are being compared
        if(st1.getState_Isfor().getComponent_Name().equals(st2.getState_Isfor().getComponent_Name())){

            // Compare the state names
            if(st1.getState_Name().equals(st2.getState_Name())){
                result = true;
            }
        }
        return result;
    }

    /**
     * Utility method to dump the rules
     *
     * @param l    List of CoolRule objects
     */
    public void dumprule(rca.util.leap.List l){
        rca.util.leap.Iterator it, itt;
        CoolRule r;
        CoolProcess p;

        rca.util.leap.List t;
        itt = l.iterator(); while(itt.hasNext()){
        r = (CoolRule)itt.next();

        System.out.println("======== Rule ========");

//................. IF
        if(r.getRule_If()!=null){
            System.out.println("if state = "+r.getRule_If().getState_Name());
        }

        //orif
        if(r.getRule_Orif()!=null && !r.getRule_Orif().isEmpty()){
            t=r.getRule_Orif();


            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" orif state = "+s.getState_Name());
        }
        }

        // andif
        if(r.getRule_Andif()!=null && !r.getRule_Andif().isEmpty()){
            t=r.getRule_Andif();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" andif state = "+s.getState_Name());
        }
        }

        // thenmoveto
        if(r.getRule_Thenmoveto()!=null && !r.getRule_Thenmoveto().isEmpty()){
            t=r.getRule_Thenmoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" then moveto state = "+s.getState_Name());
        }
        }

        // thenexecute
        if(r.getRule_Thenexecute()!=null && !r.getRule_Thenexecute().isEmpty()){
            t=r.getRule_Thenexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" then execute process = "+p.getProcess_Commandname());
        }
        }

// .............. ELSEIF
        if(r.getRule_Elseif()!=null){
            System.out.println("elseif state = "+r.getRule_Elseif().getState_Name());
        }

        // elseifmoveto
        if(r.getRule_Elseifmoveto()!=null && !r.getRule_Elseifmoveto().isEmpty()){
            t=r.getRule_Elseifmoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" elseif moveto state = "+s.getState_Name());
        }
        }

        // elseifexecute
        if(r.getRule_Elseifexecute()!=null && !r.getRule_Elseifexecute().isEmpty()){
            t=r.getRule_Elseifexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" elseif execute process = "+p.getProcess_Commandname());
        }
        }


// .............. ELSE
        if(r.getRule_Else()!=null){
            System.out.println("else");
        }

        // elsemoveto
        if(r.getRule_Elsemoveto()!=null && !r.getRule_Elsemoveto().isEmpty()){
            t=r.getRule_Elsemoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" else moveto state = "+s.getState_Name());
        }
        }

        // elseexecute
        if(r.getRule_Elseexecute()!=null && !r.getRule_Elseexecute().isEmpty()){
            t=r.getRule_Elseexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" else execute process = "+p.getProcess_Commandname());
        }
        }

// .............. WHILE
        if(r.getRule_While()!=null){
            System.out.println("while state = "+r.getRule_While().getState_Name());
        }

        // whilemoveto
        if(r.getRule_Whilemoveto()!=null && !r.getRule_Whilemoveto().isEmpty()){
            t=r.getRule_Whilemoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" while moveto state = "+s.getState_Name());
        }
        }

        // whileexecute
        if(r.getRule_Whileexecute()!=null && !r.getRule_Whileexecute().isEmpty()){
            t=r.getRule_Whileexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" while execute process = "+p.getProcess_Commandname());
        }
        }
    }
    }

    /**
     * AFECS Rule syntaxparser
     *
     * @param r  CoolRule object
     */
    public void SMAlgorithm(CoolRule r){

        boolean flag =true;
        Iterator it;

// ANDIFSTATE
        if(r.getRule_Andif()!=null){
            it = r.getRule_Andif().iterator();
            // flag indicates that recorded state-names for the required
            // components are the same as required ANDIFSTATE name;
            while(it.hasNext()){
                CoolState andifState = (CoolState)it.next();

                if(!ComponentStates.containsKey(andifState.getState_Isfor().getComponent_Name())){
                    flag =false;
                } else {
                    String recStName = ComponentStates.get(andifState.getState_Isfor().getComponent_Name());
                    if(!recStName.equals(andifState.getState_Name())){
                        flag =false;
                    }
                }
            }

            // all ANDIFSTATE rules are satisfied
            if(flag){
                if(r.getRule_Thenmoveto()!=null){
                    it = r.getRule_Thenmoveto().iterator();
                    while(it.hasNext()){
                        CoolState tms = (CoolState)it.next();
                        moveToState(tms.getState_Isfor().getComponent_Name(),tms);
                    }
                }
                if(r.getRule_Thenexecute()!=null){
                    it = r.getRule_Thenexecute().iterator();
                    while(it.hasNext()){
                        CoolProcess tmp = (CoolProcess)it.next();
                        executeProcess(myName,tmp);
                    }
                }
            }
        }


// ORIF
        else if(r.getRule_Orif()!=null){
            it = r.getRule_Orif().iterator();
            flag = false;
            while(it.hasNext()){
                CoolState orIfState = (CoolState)it.next();
                if(ComponentStates.containsKey(orIfState.getState_Isfor().getComponent_Name())){
                    String recStName = ComponentStates.get(orIfState.getState_Isfor().getComponent_Name());
                    if(recStName.equals(orIfState.getState_Name())){
                        flag =true;
                    }
                }
            }
            // at least one ORIFSTATE  rule is satisfied
            if(flag){
                if(r.getRule_Thenmoveto()!=null){
                    it = r.getRule_Thenmoveto().iterator();
                    while(it.hasNext()){
                        CoolState tms = (CoolState)it.next();
                        moveToState(tms.getState_Isfor().getComponent_Name(),tms);
                    }
                }
                if(r.getRule_Thenexecute()!=null){
                    it = r.getRule_Thenexecute().iterator();
                    while(it.hasNext()){
                        CoolProcess tmp = (CoolProcess)it.next();
                        executeProcess(myName,tmp);
                    }
                }
            }
        }

// THENMOVETO action
        else if(r.getRule_Thenmoveto()!=null){
            it = r.getRule_Thenmoveto().iterator();
            while(it.hasNext()){
                CoolState tms = (CoolState)it.next();
                moveToState(tms.getState_Isfor().getComponent_Name(),tms);
            }
        }
// THENEXECUTE action
        else if(r.getRule_Thenexecute()!=null){
            it = r.getRule_Thenexecute().iterator();
            while(it.hasNext()){
                CoolProcess tmp = (CoolProcess)it.next();
                executeProcess(myName,tmp);
            }
        }
        else{
            System.out.println("SM Rule Error: Syntax error in the rule.");
        }

    }

    /**
     *
     * @param stateIn       iunput state to be considered inside of the configured rule.
     */
    public void execrule(CoolState stateIn){
        Iterator rulesIt, it;
        CoolRule r;

        rulesIt = _myRules.iterator();
        while(rulesIt.hasNext()){

            r = (CoolRule)rulesIt.next();

// -> IF rule
            if(r.getRule_If()!=null && eQuals(stateIn,r.getRule_If())){
                SMAlgorithm(r);

// -> ELSEIF rule
                if(r.getRule_Elseif()!=null && eQuals(stateIn, r.getRule_Elseif())){
                    SMAlgorithm(r);
                }

// -> ELSE rule
                if(r.getRule_Else()!=null){
                    if(r.getRule_Elsemoveto()!=null){
                        it = r.getRule_Elsemoveto().iterator();
                        while(it.hasNext()){
                            CoolState tms = (CoolState)it.next();
                            moveToState(tms.getState_Isfor().getComponent_Name(),tms);
                        }
                    }
                    if(r.getRule_Elseexecute()!=null){
                        it = r.getRule_Elseexecute().iterator();
                        while(it.hasNext()){
                            CoolProcess tmp = (CoolProcess)it.next();
                            executeProcess(myName,tmp);
                        }
                    }
                }

// -> WHILE rule
            } else if(r.getRule_While()!=null && eQuals(stateIn,r.getRule_While())){

                it = ComponentStates.keySet().iterator();
                while(it.hasNext()){
                    String compName = (String)it.next();
                    String compState = ComponentStates.get(compName);
                    if(compName.equals(r.getRule_While().getState_Isfor().getComponent_Name()) &&
                            compState.equals(r.getRule_While().getState_Name())){
                        if(r.getRule_Whilemoveto()!=null){
                            it = r.getRule_Whilemoveto().iterator();
                            while(it.hasNext()){
                                CoolState tms = (CoolState)it.next();
                                moveToState(myName,tms);
                            }
                        }
                        if(r.getRule_Whileexecute()!=null){
                            it = r.getRule_Whileexecute().iterator();
                            while(it.hasNext()){
                                CoolProcess tmp = (CoolProcess)it.next();
                                executeProcess(myName,tmp);
                            }
                        }
                    }
                }
            } 
        }
    }



/// ... Inner Class ...


    /**
     * Inner class responsible for organizing control system state machine.
     */
    private class StateMachine extends Thread{

        public CoolState mys;

        StateMachine(CoolState st){
            mys = st;
        }

        public void run(){
            SM_ACTIVE = true;
            execrule(mys);
            SM_ACTIVE = false;
        }
    }

    /**
     * Inner class responsible periodic reporting of the state of this agent.
     */
    private class StateReport extends Thread{
        /** Thread control variable */
        public volatile boolean shouldrun = true;

        /** Period of the monitoring in seconds*/
        private int period;


        StateReport(int p){
            period = p;
            System.out.println(_agentName+"  running ... "+shouldrun);
        }

        public void run(){
            while(shouldrun){
                try {
                    Thread.sleep(period*1000);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                if(myStateObject!=null){
                    // report user interfaces
                    dalogMsg(myStateObject.getState_Severity(), myStateObject.getState_Name());
                    // report to the control hierarchy up
                    informStateSMS(myStateObject.getState_Name(),"info");
                }
            }
        }
    }


}

