/*---------------------------------------------------------------------------*
*  Copyright (c) 2008        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.agents;

import rca.content.lang.Codec;
import rca.content.lang.sl.SLCodec;
import rca.content.onto.OntologyException;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.core.AID;
import rca.domain.DFService;
import rca.domain.FIPAAgentManagement.DFAgentDescription;
import rca.domain.FIPAAgentManagement.ServiceDescription;
import rca.domain.FIPAException;
import rca.domain.FIPANames;
import rca.lang.acl.ACLMessage;
import org.jlab.coda.runcontrol.ontologies.CodaPlatform.CodaPlatformAdminOntology;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.*;
import org.jlab.coda.runcontrol.util.*;
import org.jlab.coda.runcontrol.util.db.DatabaseConnection;
import org.jlab.coda.runcontrol.util.sh.SysExe;
import org.jlab.coda.runcontrol.util.tcp.JtcpClient;
import org.jlab.coda.runcontrol.util.dp.JdpClient;
import org.jlab.coda.runcontrol.util.dp.DpliteException;
import org.jlab.coda.cMsg.*;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeoutException;


/**
 * This is the basic StemCell agent class, representative of the physical component in the platform. It incapsulates
 * inter porocess communication protocols (dp, shell, cMsg) used to  communicate with the component.
 * This class is used also to build and deploy State Machine Supervisor (SMS) agents
 *
 * @author Vardan Gyurjyan
 *         Date: June 7, 2006
 *         Time: 1:00:30 PM
 */
public class StemCell extends CAgent {

    /** Verbosity level defined in the rcConstants class. */
    private static int verbose = rcConstants.debugError;

    /** keeps download, prestart, go and end states. Defined to prevent continuous state reportings to Medulla,
     * trigered by the continuous statistics messages coming from the physical client.
     * Default value is set to "unknown"*/
    private String _myState_ = rcConstants.unknown;

    /** cMsg domain server (Hepatic) connection object */
    cMsg cMsgSrv;

    /** The name of this agent which is the same as the representing physical client name*/
    private String myName;

    /** The current state of the agent*/
    private String myState = rcConstants.booted;

    /** cMsg server (rc domain) object, used to organize P2P/private communication
     * between this agent and the representing real world component. */
    private cMsg rcDomainSrv;

    /** Thread to watch the coda component statistics message arrivals*/
    private messageWatcher myWatch;

    /** Private variable used to count the number of seconds before the next rc domain cmessage arrives to this agent.*/
    private volatile int COUNT;

    /** Private variable to calculate the average data rate. */
    private double data_rate_a;

    private int num_of_reportings;

    /** Instance of the class incapsulating Unix shell IPC commands*/
    public SysExe mySysExe;

    /** Object of the JtcpClient class used to communicate through tcp to the component*/
    private JtcpClient mytcp;

    /** Object of the JdpClient class used to communicate through dp to the component*/
    private JdpClient mydp;

    /** Flag used during the reset transition for thread synchronization*/
    public boolean RESETTIME = false;

    /** AID of the runcontrol supervisor agent.
     * N.B. The name of the Medulla_aid type agent is alwais equal to CODA runtype*/
    private AID Medulla_aid;

    /** Requested transition from the Medulla_aid (supervisor)*/
    private volatile String _requestedTransition;

    /** Hashmap stores all working thread objects, executing continuosly monitoring
     * processes, attached to the certain CODA state transitions.
     * N.B. the processes described in the cool are started by this agent, i.e. if there is
     * a periodic process described in the cool, then agent should periodically execute the
     * process specific commands.
     * The key of the hashmap is constructed using the agent name and the process command name
     * <em>_agentName:_commandName</em> */
    private ConcurrentHashMap<String,RunProcess> cyclicThreads = new ConcurrentHashMap<String,RunProcess>();

    /** Internal field showing the agents registration status with the DF of the AFECS platform.*/
    private boolean  _RegisteredFlag    = false;

    /** Indicates that this agent is the representative for the coda2 component.
     *  N.B. those are components working with DB*/
    private String           _coda2component;

    /** The type of the agent, i.e. ROC EB, ER, etc.*/
    private String           _agentType;

    /** The name of the agent, i.e. R1 EB1, ER1, etc.*/
    private String           _agentName;

    /** Priority of thie agent.*/
    private Integer          _agentPriority = null;

    /** Session of the agent. The default value is set to the env deffinition*/
    private String           _session = System.getenv("SESSION");

    /** Runtype of the agent, i.e. configuration name.*/
    private String           _runtype ="";

    /** This integer defines how many times foreign process will be executed. -1 means forever.*/
    private int              _loop  = 1;

    /** Private field keeps the SetPreConfigure object, containing the set of the processes
     * necessary to execute at preconfigure transition.*/
    private SetPreConfigure  setpreconfigure;

    /** Private field keeps the SetConfigure object, containing the set of the processes
     * necessary to execute at configure transition.*/
    private SetConfigure     setconfigure;

    /** Private field keeps the SetPostConfigure object, containing the set of the processes
     * necessary to execute at postconfigure transition.*/
    private SetPostConfigure setpostconfigure;

    /** Private field keeps the SetPredownload object, containing the set of the processes
     * necessary to execute at predownload transition.*/
    private SetPreDownload   setpredownload;

    /** Private field keeps the SetDownload object, containing the set of the processes
     * necessary to execute at download transition.*/
    private SetDownload      setdownload;

    /** Private field keeps the SetPostDownload object, containing the set of the processes
     * necessary to execute at postdownload transition.*/
    private SetPostDownload  setpostdownload;

    /** Private field keeps the SetPrePrestart object, containing the set of the processes
     * necessary to execute at preprestart transition.*/
    private SetPrePrestart   setpreprestart;

    /** Private field keeps the SetPrestart object, containing the set of the processes
     * necessary to execute at prestart transition.*/
    private SetPrestart      setprestart;

    /** Private field keeps the SetPostPrestart object, containing the set of the processes
     * necessary to execute at postprestart transition.*/
    private SetPostPrestart  setpostprestart;

    /** Private field keeps the SetPreGo object, containing the set of the processes
     * necessary to execute at prego transition.*/
    private SetPreGo         setprego;

    /** Private field keeps the SetGo object, containing the set of the processes
     * necessary to execute at go transition.*/
    private SetGo            setgo;

    /** Private field keeps the SetPostGo object, containing the set of the processes
     * necessary to execute at postgo transition.*/
    private SetPostGo        setpostgo;

    /** Private field keeps the SetPreEnd object, containing the set of the processes
     * necessary to execute at preend transition.*/
    private SetPreEnd        setpreend;

    /** Private field keeps the SetEnd object, containing the set of the processes
     * necessary to execute at end transition.*/
    private SetEnd           setend;

    /** Private field keeps the SetpostEnd object, containing the set of the processes
     * necessary to execute at postend transition.*/
    private SetPostEnd       setpostend;

    /** Private field keeps the SetInPause object, containing the set of the processes
     * necessary to execute at pause transition.*/
    private SetInPause       setinpause;

    /** Private field keeps the SetInResume object, containing the set of the processes
     * necessary to execute at resume transition.*/
    private SetInResume      setinresume;

    /** Private field keeps the SetReset object, containing the set of the processes
     * necessary to execute at reset transition.*/
    private SetReset         setreset;

    /** Defines the content language, used for inter-agent communications*/
    private Codec  codec = new SLCodec();

    /** Coda component specific statistics: event_number*/
    private int    event_number;

    /** Coda component specific statistics: event_rate*/
    private float  event_rate;

    /** Coda component specific statistics: number_of_longs*/
    private long   number_of_long;

    /** Coda component specific statistics: data_rate*/
    private double data_rate;

    /** Variables used to calculate the time after which message needs to be sent to the Medulla*/
    private long prevTime;

    /** Variables used to calculate the final averages*/
    private long runStartTime;

    /** Event number after the end is clicked*/
    int end_evt_num ;


    /** List containing the names of the SMS agents*/
    private List<String> _mySMS = new ArrayList<String>();

    /** XML SAX parser default handler*/
    private final DefaultHandler handler = new RcStatusHandler();

    /** Dalog message control variables*/
    private String DALOG_FILTER;

    /** cMsg subscription tmp objects*/
    private Object ob1, ob2, ob3, ob4, ob5, ob6, ob7;

    /**
     * Setup method of the agent. It organizes the
     * ontology registration, and agent behaviour scheduling.
     */
    protected void setup(){

// This is the given name when the agent was started by the Hepatic.
        myName = getLocalName();

// Register supported languages
        getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);

// Register the AFECS platform administration ontology for Hepatic conversations
        getContentManager().registerOntology(CodaPlatformAdminOntology.getInstance());

// Register the coda run control ontology for Cortex and Medulla conversations
        getContentManager().registerOntology(CodaRunControlOntology.getInstance());

// Add the responder behaviour.
// This is the behaviour where we process request information from the other agents in the platform.
// The actual processing method is defined in this class but called from the GenResponderBehaviour object
        addBehaviour(new GenRespondBehaviour(this, verbose));

// DF registration
        if(!_RegisteredFlag){
            DFAgentDescription dfd = new DFAgentDescription();
            dfd.setName(getAID());
            ServiceDescription sd = new ServiceDescription();
            sd.setType("StemCell");
            sd.setName(myName);
            sd.setOwnership("JLABCODA");
            dfd.addServices(sd);
            try {
                DFService.register(this,dfd);
                _RegisteredFlag = true;
            } catch (FIPAException e) {
                System.err.println(getLocalName()+" registration with DF not succeeded. Reason: "+e.getMessage());
                _RegisteredFlag = false;
            }
        }

// Connect cMsg domain server (runs within Hepatic). N.B. we have noticed performance decline for the case when
// we have >500 agents conencting to the cMsg domain server.        
        connectTocMsg();
    }


    /**
     * This method is called at the agent shutdown
     */
    protected void takeDown()
    {
        try { DFService.deregister(this); }
        catch (Exception e) {e.printStackTrace();}
    }


    /**
     * Sets the _requestedTransition global variable
     *
     * @param st requestedTransition.
     */
    public synchronized void setRequestedTransition(String st){
        _requestedTransition = st;
    }

    /**
     * Returns the transition that this agent is in.
     *
     * @return _requestedTransition
     */
    public synchronized String getRequestedTransition() {
        return _requestedTransition;
    }

    /**
     * Sending dalog message
     * @param severity      severity of the message
     * @param message       content of the message
     */
    private void dalogMsg(String severity, String message){
        /*
                String subject;
                String type = null;
                cMsgPayloadItem sever = null, state;

                System.out.println(".................... "+myName +" - "+_agentType);
                if(rcSubscription.expid!=null && _session!=null && _runtype!=null && _agentType!=null && myName!=null){
                    cMsgMessage msg = new cMsgMessage();
                    subject = rcSubscription.expid+"/"+_session+"/"+_runtype+"/"+_agentType+"/"+myName;
                    try {
                        // payload item state
                        state = new cMsgPayloadItem("state",_myState_);
                        if (severity.equalsIgnoreCase("info")) {
                            type = "dalog/1";
                            sever = new cMsgPayloadItem("severity","INFO|1");

                        }
                        else if(severity.equalsIgnoreCase("warning")) {
                            type = "dalog/5";
                            sever = new cMsgPayloadItem("severity","WARN|5");
                        }
                        else if(severity.equalsIgnoreCase("error")) {
                            type = "dalog/9";
                            sever = new cMsgPayloadItem("severity","ERROR|9");
                        }
                        else if(severity.equalsIgnoreCase("severeerror")){
                            type = "dalog/13";
                            sever = new cMsgPayloadItem("severity","ERRORS|13");
                        }
                        if(subject!=null && type !=null){

                            msg.setSubject(subject);
                            msg.setType(type);
                            msg.setText(message);
                            msg.addPayloadItem(sever);
                            msg.addPayloadItem(state);
                            cMsgSrv.send(msg);
                            cMsgSrv.flush(1);
                        }
                    } catch (cMsgException e) {
                        e.printStackTrace();
                    }
        //        System.out.println(subject+" "+type);
                }
        */
    }

    /**
     * This method will organize communication with the physical component.
     * Based on the proceses defined in the cool configuration file this method will start working
     * threads for each described process for particular transition to satisfy the configuration requirements.
     * CODA transitions as well as attached external processes are described in the cool file, and are stored in the
     * objects of the transition specific classes, called set[transition]. Those classes implement
     * <clas>TransitionExtension</clas>
     * interface having two methods for setting and getting the list of the processes.
     * N.B. if nothing is described in the cool for this transition it will inform back to supervisor agent
     * the name of the requested transition.

     * @param trext        TransitionExtention object created during the Cortex teaching process (setPreconfigure, etc..).
     *                     This object contains the <clas>CoolProcess</clas> ontology concept, which is attached to the
     *                     CODA transition.
     * @param transition   CODA state transition
     */
    public void organizeTransition (TransitionExtention trext, String transition){

// String discribing the command beeing executed usinb the specific IPC, described in the cool.
        String  commandName;

// Setting the global var indicating the requested transition from Medulla
        setRequestedTransition(transition);

// Check if the agent is programmed to do something during this transition.
        if((trext!=null) && (trext.getProcesses())!=null) {

// Iterate over all registered processes with this transition
            Iterator it = trext.getProcesses().iterator();
            while(it.hasNext()) {

// Get the process
                CoolProcess pr = (CoolProcess)it.next();

// Check if the command is null
                if(pr.getProcess_Commandname()!=null){

// Get the command accociated with this process
                    commandName = pr.getProcess_Commandname();

// Get the periodicity of this command
                    if(pr.getProcess_Periodicity()!=null)
                        _loop = pr.getProcess_Periodicity();
                }

// Inform plexuses that there is process defined in the cool without defining the process command
                else {
                    String tt =myName+ ": command name for the process is not defined.";
                    System.out.println(tt);
                    informStatus(tt,getRequestedTransition(),"Warning", Medulla_aid);
                    return;
                }

// Start the thread to convey the request to the physical component
                if(_loop>=0){
                    RunProcess myRun = new RunProcess(transition, pr);
                    myRun.start();
                }

// Negative _loop means loop for ever. We need to get handle on this threads to stop them gracefully, so register
// them in the hashmap.
                else {

// cyclicThreads hashmap contains all periodic processes for this particular component.
                    if(!cyclicThreads.containsKey(_agentName+":"+commandName)){

// There is no active working thread object registered in the cyclicThread map.
// Create the object and add it to the map
                        RunProcess myRun = new RunProcess(transition, pr);
                        cyclicThreads.put(_agentName+":"+commandName, myRun);

// N.B. cyclick process name is defined as:  agentname:commandname;
// Start the tread
                        myRun.shouldRun=true;
                        myRun.start();
                    }

// Thread object exists and registered
                    else {
                        RunProcess rp = cyclicThreads.get(_agentName+":"+commandName);

// Activate the thread
                        if(rp != null) rp.shouldRun = true;
                    }
                }
            }

// Reset transition is hard programmed in the Cortex, and for this particular transition we do not wait.
            if(transition.equalsIgnoreCase(rcConstants.CMD_reset)) {
                informStatus(transition, transition,null, Medulla_aid);
                informStateSMS(transition);
            }
        } else{

// Send back to Medulla the name of the transition. This is done because Medulla is waiting for the completion
// of the transiton and transition related processes. If it was nothing scheduled for this ( nothing in the COOL file)
// transition just send back the name of the transition to satisfy Medulla.
            informStatus(transition, transition,null, Medulla_aid);
            informStateSMS(transition);
        }
    }


    /**
     * This is the implementatioin of the abstract method defined in the CAgent class.
     * This method is going to be called by the BaseicRequestBehaviour object to process
     * the iform messages coming back to this agent as a result of initiating the request
     * to some agent on the platfrom.
     *
     * @param msg      ACLMessage object passed from the GenRequestBehaviour object.
     */
    protected void processInform(ACLMessage msg) {
    }


    /**
     * Method connects to the cMsg Server started by the Hepatic. This will allow this agent
     * to respond directly to the cMsg requests from outside of the platform.
     */
    protected void connectTocMsg(){

        try {
            if((cMsgSrv!=null) && (cMsgSrv.isConnected())) {
// We are connected do nothing
            }

// Start a new connection
            else {

// Disconnect first if there is cMsgSrv object
                if(cMsgSrv!=null) cMsgSrv.disconnect();
                cMsgSrv = new cMsg(rcSubscription.PlatformcMsgServerUDL, myName, rcSubscription.PlatformcMsgServerDescription);
                cMsgSrv.connect();

// Enable message reception
                cMsgSrv.start();

/* //@todo comented out cMsg sunscription  for performance reasons

// Unsubscribe
                if(cMsgSrv!=null) {
                    if(ob1!=null)cMsgSrv.unsubscribe(ob1);
                    if(ob2!=null)cMsgSrv.unsubscribe(ob2);
                    if(ob3!=null)cMsgSrv.unsubscribe(ob3);
                    if(ob4!=null)cMsgSrv.unsubscribe(ob4);
                    if(ob5!=null)cMsgSrv.unsubscribe(ob5);
                }

// N.B. at first we will subscribe all the subscriptions: coda/info, session/control, run/control and run/transition.
// Subjects for this subscriptions are the name of this agent
// ... conda/info
                cMsgCallbackInterface codainfo = new CodaInfoCallback();
                ob1 = cMsgSrv.subscribe(myName, rcSubscription.CodaInfoType +"/*", codainfo, null);

// Here we subscribe also to the subject=CodaObject and type=coda/info/* to receive messages addressed to all CodaObjects
                ob2 = cMsgSrv.subscribe("CodaObject", rcSubscription.CodaInfoType +"/*", codainfo, null);

// ... session/control
// N.B. we need also to subscribe to the subject=_session and type=session/contrl/* to receive the messages
// addressed to the agents in the same session. We are goin to do that at differentiation statge when an
// actual session will be available.
                cMsgCallbackInterface sessioncontrol = new SessionControlCallback();
                ob3 = cMsgSrv.subscribe(myName, rcSubscription.SessionControlType +"/*", sessioncontrol, null);

// ... run/control
                cMsgCallbackInterface runcontrol = new RunControlCallback();
                ob4 = cMsgSrv.subscribe(myName, rcSubscription.RunControlType +"/*", runcontrol, null);

// ... run/transition
                cMsgCallbackInterface runtransition = new RunTransitionCallback();
                ob5 = cMsgSrv.subscribe(myName, rcSubscription.RunTransitionType +"/*", runtransition, null);

*/
            }

        } catch (cMsgException ee) {
            try {
                cMsgSrv.disconnect();
                cMsgSrv = null;
            } catch (cMsgException e) {
                cMsgSrv = null;
                e.printStackTrace();
            }
            ee.printStackTrace();
        }
    }


    /**
     * Method starts rcDomain cMsg server.
     * Remember: every real world component representative agent has cMsg server
     *  running locally, which is used for private P2P communications.
     *
     * @param host              Physical client host name
     * @param port              Physical client port number
     * @return                  status of the connection
     */
    protected boolean  startRcCmsgServer(String host, int port){

// Status of the connection
        boolean status = true;
        try{

// Start the cMsg (rc domain) server
// N.B. physical client needs to know the udp host and port of this agent to be
// able to send information back to the agent (for e.g. statistics messages or done messages of the transitions).
// rcDomain is taking care of that by defining and sending to the physical client udp host and port of this rcDomainSrv.



// Define the rc universal domain locator string
            String rcsUDL = rcSubscription.RcServerUDL+host+":"+port;


// Disconnect first if the rcDomainSrv object exists
            if(rcDomainSrv!=null) rcDomainSrv.disconnect();
//            System.out.println(myName+"-DEBUG: Agent of "+myName+" starts rcDomain server .......");
//            dalogMsg("info","Agent of "+myName+" starts rcDomain server." );

// Start the server
            rcDomainSrv = new cMsg(rcsUDL, rcSubscription.RcServerName, rcSubscription.RcServerDescription);

            System.out.println("agent of "+myName +" initiates connect to "+rcSubscription.RcServerUDL+host+":"+port);

// Complete the P2P connection
            rcDomainSrv.connect();
            System.out.println("<i> Agent of "+myName+" connected to "+rcDomainSrv.getString()+" at "+host+" "+port+" .......");
            dalogMsg("info","<i> Agent of "+myName+" connected to "+rcDomainSrv.getString()+" at "+host+" "+port);

// Check the name of the clinet that we connected.
// N.B. The name of the cilent is the same as the name of its agent.
// The name of the cilent is passed through the connection object.
            if(!(rcDomainSrv.getString().equals(myName))) {

// This is the different client on the same host and using the known port. Disconnect!
// N.B. Client port is passed from Cortex, where it was recorded during the
// client broadcasting its request to have a representative agent on the control platform.
                rcDomainSrv.disconnect();
                status = false;
                String s ="cMsgRcDomain-ERROR: Agent "+myName+" conflicts with "+rcDomainSrv.getString();
                System.out.println(s);
                informPlexuses(rcSubscription.Stemcell, rcSubscription.RcGuiTypeStemCell,s);
                dalogMsg("error",s);
            } else {

// Start receiving messages
                if(rcDomainSrv!=null) rcDomainSrv.start();

// These messages are coming from the client. Create callbacks for them.
// Create monitored messages (sent from the physical component) callback object. This is where we translate
// arrived message into ACL for Medulla 
                RunStatisticsXMLCallback mcb = new RunStatisticsXMLCallback();

// Subscribe physical component statistics messages. This will start a
// separate thread, where every time the statistcs message arrives it will
// call callback method of the mcb object.
// N.B. subject of this subscription is always the name of the component
                if(rcDomainSrv!=null) rcDomainSrv.subscribe(myName, rcSubscription.RCReport,mcb,null);

// Callback for the dalog messages coming from the physical client.
                DalogCallback dalogcb = new DalogCallback();

// Actual subscription to the components dalog messages
//                if(rcDomainSrv!=null) rcDomainSrv.subscribe(myName, rcSubscription.Dalog,dalogcb,null);
                if(rcDomainSrv!=null) rcDomainSrv.subscribe("*", rcSubscription.Dalog,dalogcb,null);

// Below are subscriptions that agent will handle them just simply passing it to the
// cMsg server for outside platform users.
                ResponseCallback rtr = new ResponseCallback();
                if(rcDomainSrv!=null) rcDomainSrv.subscribe(myName, rcSubscription.Response+"/*", rtr,null);

// Callback for "don't care" messages, coming from the client. In this case we pass it to the cMsg server.
                DontCareCallback dc = new DontCareCallback();
                if(rcDomainSrv!=null) rcDomainSrv.subscribe(null, null, dc,null);
            }
        } catch (cMsgException ee) {
            ee.printStackTrace();
            status = false;
            String s ="Physical component "+myName+" is not responding";
            informPlexuses(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
            dalogMsg("error",s);
            try {
                rcDomainSrv.disconnect();
                rcDomainSrv = null;
            } catch (cMsgException e) {
                rcDomainSrv = null;
                e.printStackTrace();
                status = false;
            }
        }
        return status;
    }


    /**
     * This method gets called whenever the ACL message (inter-agent communication) is reached to this agent.
     * This is the implementatioin of the abstract method defined in the <clas>CAgent</clas>, which is
     * going to be called by the <clas>BaseicResponderBehaviour</clas> object to process the request to
     * this agent. Most of the ACL message request are coming from the supervisor agent
     * <clas>Medulla</clas> and/or <clas>Cortex</clas>.
     *
     * @param request      ACLMessage request object, passed from the GenResponderBehaviour object.
     * @return             ACLMessage object
     */
    protected ACLMessage processRequest(ACLMessage request,Action a) {

// Reference to the CoolComponent object.
        CoolComponent component;

// Reference to the CoolOption object.
        CoolOption option;
        String tmpStr = _session;

// Host name of the physical component (represented by this agent).
        String  physClientHost;

// Port number of the physical component.
        int  physClientPort;

// Create an agent communication reply object
        ACLMessage reply = request.createReply();
        try {
//            System.out.println("PPP ... "+myName+" got the ACL message "+a.getAction().getClass().getName());
                              
// This set of the messages are usually coming from the Cortex.
// Cortex uses Differentiate action to turn this "stemcell" agent in to the specific type of an agent, which will
// represent the physical client in the AFECS platform. Differentiated (specialized) agents will register
// their _agentTypes with the DF.
            if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Differentiate")) {

//                System.out.println("DDD... "+myName+" got differentiate");
// Get the differentiate ontology object
                Differentiate myDifferentiate = (Differentiate)a.getAction();

// Intialization of the variables
                option          = myDifferentiate.getDifferentiate_Option();
                _coda2component = option.getOption_Hascoda2component();
                component       = myDifferentiate.getDifferentiate_Component();
                _agentType      = component.getComponent_Type();
                _agentName      = component.getComponent_Name();
                String reportsto = component.getComponent_Reportsto();


// Check the state of the agent. If it is booted re-asign the session, based on the cortex configure request.
                if(myState!=null){
                    if ( ((myState.equals(rcConstants.booted)) || (_session.equals("")) ) ){
                        _session  = component.getComponent_Session();

/* //@todo comented out cMsg sunscription  for performance reasons
// Here we unsubscribe and subscribe to the subject=_session and type=session/control/*
                        cMsgCallbackInterface sessioncontrol = new SessionControlCallback();
                        try {
                            if(cMsgSrv!=null){
                                if(ob6 != null) cMsgSrv.unsubscribe(ob6);
                                ob6 = cMsgSrv.subscribe(_session, rcSubscription.SessionControlType +"/*", sessioncontrol, null);
                            }
                        } catch (cMsgException e) {
                            e.printStackTrace();
                        }
*/

// If it is not booted then we can not change already registered session for the agent.
                    }
                    else if(!_session.equals(component.getComponent_Session())){
                        String s ="Request to configure "+myName+" for the session = "+component.getComponent_Session()+
                                "\n Request is denied: Agent "+myName+" is already configured in the session = *"+_session+"*";
                        System.out.println(s);
                        informPlexuses(rcSubscription.Stemcell, rcSubscription.RcGuiTypeStemCell,s);
                        dalogMsg("severeerror",s);
                        s = "Severe error";
                        informPlexuses(rcSubscription.RcGuiSubjError+component.getComponent_Session(), rcSubscription.RcGuiTypeStemCell,s);
                        try{

// Create and send done action message (ACL message)  to the cortex, informing that this agent is
// already registered in the different session
                            Done d = new Done();
                            d.setAction(a);
                            getContentManager().fillContent(reply, d);
                        } catch (Codec.CodecException e) {
                            e.printStackTrace();
                        } catch (OntologyException e) {
                            e.printStackTrace();
                        }
                        reply.setPerformative(ACLMessage.INFORM);
                        reply.setContent(_session+" "+myName+" Differentiation failed");
                        return reply;
                    }

// If myState is null then set the session
                } else {
                    _session  = component.getComponent_Session();

/* //@todo comented out cMsg subscription for performance reasons
                    cMsgCallbackInterface sessioncontrol = new SessionControlCallback();
// Here we unsubscribe and subscribe to the subject=_session and type=session/control/*. Probably is not needed...
                    try {
                        if(cMsgSrv!=null){
                            if(ob7 != null)  cMsgSrv.unsubscribe(ob7);
                            ob7 = cMsgSrv.subscribe(_session, rcSubscription.SessionControlType +"/*", sessioncontrol, null);
                        }
                    } catch (cMsgException e) {
                        e.printStackTrace();
                    }
*/
                }
                _runtype        = component.getComponent_Runtype();
                _agentPriority  = component.getComponent_Priority();

// Coda component "statistics" message watcher. We stop this thread in order to start it at "go" transition.
                if(myWatch!=null){
                    myWatch.shouldrun=false;
                }

// Report to the supervisor defined by the Cortex through setComponent_Reoprtsto ontology action
                Medulla_aid = new AID(component.getComponent_Reportsto(),AID.ISLOCALNAME);

// Create object for tcp communication if the component is described in the cool as coda2 component
// N.B. will be depricated
                try {
                    if((_coda2component.equalsIgnoreCase("true")) &&
                            (component.getComponent_Ipc().equalsIgnoreCase("tcp")))mytcp = new JtcpClient(_agentName);
                } catch (rcException e) {
                    e.printStackTrace();
                }

// Create object for dp communications if the component is described in the cool as coda2 component
                    if((_coda2component.equalsIgnoreCase("true")) &&
                            (component.getComponent_Ipc().equalsIgnoreCase("dpsh"))) {
                        try {
                        mydp = new JdpClient(_agentName);
                        } catch (rcException e) {
                            e.printStackTrace();
                        }
                    }

// Dalog message cool file control
                DALOG_FILTER = component.getComponent_Filtersdalog();

// ... CodaInit request from Cortex
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.CodaInit")) {

// cMsg (rc domain) connection status
                boolean connected;

// Get codainit ontology object
                CodaInit myinit = (CodaInit)a.getAction();

// Create an object for shell command execution
                mySysExe = new SysExe();
                physClientHost  = myinit.getCodainit_Clienthost();
                physClientPort  = myinit.getCodainit_Clientport();

// Connect cMsg domain server and subscribe control messages coming from outside of the platform
// commented out for performace reasons
                connectTocMsg();

// start the cMsg internal server (rc domain)
                connected = startRcCmsgServer(physClientHost, physClientPort);

// If we successfully connected the the client through rcDomain
                if(connected){
                } else {

// If the connection is not succeeded than create the following string which will be returned to the
// Medulla through agent communication channels that this agent failed to represent the desired client.
                    tmpStr = myName+" Differentiation failed";
                    System.out.println(myName+" cannot connect to the component");
                }

// The following set of messages are coming from the Cortex
// Message telling what to do for Preconfigure CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPreConfigure")) {
                setpreconfigure = (SetPreConfigure)a.getAction();

// Message telling what to do for Configure CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetConfigure")) {
                setconfigure = (SetConfigure)a.getAction();

// Message telling what to do for Postconfigure CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostConfigure")) {
                setpostconfigure = (SetPostConfigure)a.getAction();

// Message telling what to do for Predownload CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPredownload")) {
                setpredownload = (SetPreDownload)a.getAction();

// Message telling what to do for Download CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetDownload")) {
                setdownload = (SetDownload)a.getAction();

// Message telling what to do for Postdownload CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostDownload")) {
                setpostdownload = (SetPostDownload)a.getAction();

// Message telling what to do for Preprestart CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPreprestart")) {
                setpreprestart = (SetPrePrestart)a.getAction();

// Message telling what to do for Prestart CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPrestart")) {
                setprestart = (SetPrestart)a.getAction();

// Message telling what to do for Postprestart CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostPrestart")) {
                setpostprestart = (SetPostPrestart)a.getAction();

// Message telling what to do for Prego CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPrego")) {
                setprego = (SetPreGo)a.getAction();

// Message telling what to do for Go CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetGo")) {
                setgo = (SetGo)a.getAction();

// Message telling what to do for Postgo CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostGo")) {
                setpostgo = (SetPostGo)a.getAction();

// Message telling what to do for Reset CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetReset")) {
                setreset = (SetReset)a.getAction();

// Message telling what to do for Preend CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPreend")) {
                setpreend = (SetPreEnd)a.getAction();

// Message telling what to do for End CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetEnd")) {
                setend = (SetEnd)a.getAction();

// Message telling what to do for Postend CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetPostEnd")) {
                setpostend = (SetPostEnd)a.getAction();

// Message telling what to do for Pause CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetInPause")) {
                setinpause = (SetInPause)a.getAction();

// Message telling what to do for Resume CODA transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetInResume")) {
                setinresume = (SetInResume)a.getAction();

// Message telling what to in case of the active state of the component
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.SetInActive")) {
            }

// The following set of the messages are usually coming from the Medulla (basicly those are
// coda state machine transitions). Here we always check if the request from Medulla is for
// the specified session, i.e. supervisor agent session is consistent with the agents session.

// ... Disconnect message. This indicates that agent and component will release their session and will be available to
// be configured in other sessions.            
            else if(a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Disconnect")) {
                Disconnect mydisc = (Disconnect)a.getAction();
                if(mydisc.getDisconnect_Session().equals(_session)){
                    myState = rcConstants.booted;
                    _session = "";
                }
            }

// Preconfigure state transition
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PreConfigure")) {

// Set the global var indicating that reset is not allowed
                RESETTIME = false;

// Get the preconfigure coda transition object from the acl action object, implementing the
// TransitionExtention interface, filled during the teaching messages sent by Cortex (set...)
                PreConfigure myPreConfigure = (PreConfigure)a.getAction();
                if (myPreConfigure.getPreconfigure_Session().equalsIgnoreCase(_session)){

// Organize the transition
                    organizeTransition(setpreconfigure,"preconfigure");
                }

// ... Configure state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Configure")) {
                Configure myConfigure = (Configure)a.getAction();
                if (myConfigure.getConfigure_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setconfigure,"configure");
                }

// ... Postconfigure state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PostConfigure")) {
                PostConfigure myPostConfigure = (PostConfigure)a.getAction();
                if (myPostConfigure.getPostconfigure_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpostconfigure,"postconfigure");
                }

// ... Predownload state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PreDownload")) {

// No reset allowed
                RESETTIME = false;
                PreDownload myPreDownload = (PreDownload)a.getAction();
                if (myPreDownload.getPredownload_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpredownload,rcConstants.CMD_predownload);
                }

// Set the session of the client
                controlClient(rcSubscription.SessionControlType, rcConstants.CMD_SetSession, _session);

//  Here we tell component to set reporting interval = 1.0
                controlClient(rcSubscription.SessionControlType, rcConstants.CMD_SetInterval,
                        rcConstants.DefaultReportingInterval);

//  Here we tell client to start sending periodic updates on its running statistics information
                controlClient(rcSubscription.SessionControlType, rcConstants.CMD_StartReporting," ");

// ... Download state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Download")) {
                Download myDownload = (Download)a.getAction();
                if (myDownload.getDownload_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setdownload,rcConstants.CMD_download);
//                    System.out.println("DDD... "+myName+" organizes download");
                }

// ... Postdownload state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PostDownload")) {
                PostDownload myPostDownload = (PostDownload)a.getAction();
                if (myPostDownload.getPostdownload_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpostdownload,rcConstants.CMD_postdownload);
                }

// ... Preprestart state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PrePrestart")) {
                PrePrestart myPrePrestart = (PrePrestart)a.getAction();
                if (myPrePrestart.getPreprestart_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpreprestart,rcConstants.CMD_preprestart);
                }

// ... Prestart state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Prestart")) {
                Prestart myPrestart = (Prestart)a.getAction();
                if (myPrestart.getPrestart_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setprestart,rcConstants.CMD_prestart);
                }

// ... Postprestart state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PostPrestart")) {
                PostPrestart myPostPrestart = (PostPrestart)a.getAction();
                if (myPostPrestart.getPostprestart_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpostprestart,rcConstants.CMD_postprestart);
                }

// ... Prego state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PreGo")) {
                PreGo myPreGo = (PreGo)a.getAction();
                if (myPreGo.getPrego_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setprego,rcConstants.CMD_prego);

// Send runnumber to the physical client
                    controlClient(rcSubscription.RunControlType, rcConstants.CMD_SetRunNumber,
                            String.valueOf(myPreGo.getPrego_Runnumber()));
                }

// ... Go state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Go")) {
                Go myGo = (Go)a.getAction();
                if (myGo.getGo_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setgo,rcConstants.CMD_go);

// Start a thread watching the mesage arrivals from its clinet
                    myWatch = new messageWatcher();
                    myWatch.start();
                }

// ... Postgo state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PostGo")) {
                PostGo myPostGo = (PostGo)a.getAction();
                if (myPostGo.getPostgo_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpostgo,rcConstants.CMD_postgo);
                }

// ... Preend state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PreEnd")) {
                PreEnd myPreEnd = (PreEnd)a.getAction();
                if (myPreEnd.getPreend_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpreend,"preend");
                }

// ... End state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.End")) {
                End myEnd = (End)a.getAction();
                if (myEnd.getEnd_Session().equalsIgnoreCase(_session)){

// Stop watching messages
                    if(myWatch!=null) myWatch.shouldrun = false;
                    organizeTransition(setend,"end");
                }

// ... Postend state transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.PostEnd")) {
                PostEnd myPostEnd = (PostEnd)a.getAction();
                if (myPostEnd.getPostend_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setpostend,"postend");
                }

// ... Reset transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Reset")) {
                Reset myReset = (Reset)a.getAction();
                if (myReset.getReset_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setreset,rcConstants.CMD_reset);

// Stop watching messages from the physical client
                    if(myWatch!=null) myWatch.shouldrun = false;

// Stop waiting components to respond.
                    RESETTIME = true;
                }

// ... Pause transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Pause")) {
                Pause myPause = (Pause)a.getAction();
                if (myPause.getPause_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setinpause,rcConstants.CMD_pause);
                }

// ... Resume transition
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Resume")) {
                RESETTIME = false;
                Resume myResume = (Resume)a.getAction();
                if (myResume.getResume_Session().equalsIgnoreCase(_session)){
                    organizeTransition(setinresume,"resume");
                }

// Following 3 commands are synchronous commands, meaning initiator will handle inform message, without
// lunching a handling thread. Information will be send back with the done message.
// ... Report the runtype.
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.ReportRunType")) {
                tmpStr =  _runtype;

// ... Report the session.
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.ReportSession")) {
                tmpStr = _session;

// ... Pause continuously running monitoring threads
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.StopMonitoring")) {
                StopMonitoring myStop = (StopMonitoring)a.getAction();
                String processname = myStop.getStopmonitoring_Processname();

// Get the working thread for this perocess
                RunProcess myRunStop = cyclicThreads.get(processname);

// Pause the thread
                if(myRunStop != null){
                    myRunStop.shouldPause = true;
                }

// ... Resume the monitoring process
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.ResumeMonitoring")) {
                ResumeMonitoring myresume = (ResumeMonitoring)a.getAction();
                String processname = myresume.getResumemonitoring_Processname();

// Get the working thread for this perocess
                RunProcess myRunResume = cyclicThreads.get(processname);

// Resume the thread
                if(myRunResume !=null)myRunResume.shouldPause = false;

// ... Stop the monitoring process.
// N.B. monitoring processes are described in the cool and started at the differentiation
// process (executed at the configure coda transition).
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.KillMonitoring")) {
                KillMonitoring mykill = (KillMonitoring)a.getAction();
                String processname = mykill.getKillmonitoring_Processname();

// Get the monitoring thread object for this process
                RunProcess myRunKill = cyclicThreads.get(processname);

// Remove this thread object from the map
                cyclicThreads.remove(processname);

// Stop the thread, which will be garbage collected by the system.
                if(myRunKill!=null) {
                    myRunKill.shouldRun = false;
                }

// ... Exit. send cMsg (P2P) "exit" to the physical component.
// For coda3 components this will mean close all the threads and die.
            } else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Exit")) {
                runcMsg(myName, rcSubscription.SessionControlType +"/"+rcConstants.CMD_Exit,"exit");
            }

// ... ResultNotify.
// This request tells this agent to report its state not only to Medulla but also tho the SMS agent.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.ResultNotify")) {
                ResultNotify rnt = (ResultNotify)a.getAction();
                if(rnt.getResultNotify_Component()!=null && rnt.getResultNotify_Component().getComponent_Name()!=null)
                    _mySMS.add(rnt.getResultNotify_Component().getComponent_Name());
            }

// ... ActionMoveTo. This request tells this agent to move to the state.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.CoolActionMoveTo")) {

                String subject=null, type=null, comandname, expectedok=null;

                CoolActionMoveTo mt = (CoolActionMoveTo)a.getAction();

// Get the state
                CoolState st = mt.getActionMoveto_State();

// Get the name of the state
                String stname = st.getState_Name();

// Get the process that should be executed to acieve the state
                rca.util.leap.List prlist = st.getState_Achievedthrough();

// Execute all the listed processes in the list sequentially
                if(prlist!=null && !prlist.isEmpty()){
                    for(Iterator it = prlist.iterator(); it.hasNext();){
                        CoolProcess pr = (CoolProcess)it.next();
                        comandname = pr.getProcess_Commandname();
                        if (pr.getProcess_Outputdata().getData_Content()!=null)
                            expectedok = pr.getProcess_Outputdata().getData_Content();

// Get the type of the process, i.e. process IPC ( inter process communication)
                        if(pr.getProcess_Ipc().equalsIgnoreCase("cMsg")){
                            if(pr.getProcess_Subscribesto()!=null)
                                subject = pr.getProcess_Subscribesto().getPublisher_Sendsubject();
                            if(pr.getProcess_Sendsto()!=null)
                                type = pr.getProcess_Sendsto().getPublisher_Sendtype();
                            if(subject!=null && type !=null && comandname!=null)

// If cMsg succeeded report the state to the SMS agent.
                                if(runcMsg(subject,type,comandname))
                                    informStateSMS(stname);
                        } else if(pr.getProcess_Ipc().equalsIgnoreCase("dpsh")) {
                            if(comandname!=null) {

// If dp succeeded report the state to the SMS agent.
                                if(runDpN(comandname,expectedok).equalsIgnoreCase(comandname))
                                    informStateSMS(stname);
                            }
                        } else if(pr.getProcess_Ipc().equalsIgnoreCase("shell")) {
                            if(comandname!=null) {

// If shell succeeded report the state to the SMS agent.
                                if(runShellN(comandname,expectedok).equalsIgnoreCase(comandname))
                                    informStateSMS(stname);
                            }
                        }
                    }
                }

            }

// Create and send done action message (ACL message)  to the client agent.
            Done d = new Done();
            d.setAction(a);
            getContentManager().fillContent(reply, d);
        } catch (Codec.CodecException e) {
            e.printStackTrace();
        } catch (OntologyException e) {
            e.printStackTrace();
        }
        reply.setPerformative(ACLMessage.INFORM);

// tmpStr contains information for e.g. runtype , etc.
        reply.setContent(tmpStr);
        return reply;
    }


    /**
     * Informs status to receiver (Medulla most of the time).
     * This method sends to the receiver an <clas>AcceptStatus</clas> coda ontology ACL message,
     * describing the producer, i.e. the process that produces information and timestamp.
     *
     * @param status               The state of the component
     * @param transition           The requested transition
     * @param severity             The severity of the message
     * @param receiver             The receiver of the message
     */
    public void informStatus(String status, String transition,String severity, AID receiver) {

// Current state of this agent
        if( (transition !=null) && (!transition.equals("")))
            myState = transition ;
        else {
            myState =rcConstants.booted;
            return;
        }

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Create a codacomponent ontology concept object
            CoolComponent cm = new CoolComponent();
            cm.setComponent_Name(_agentName);
            cm.setComponent_Type(_agentType);
            cm.setComponent_Priority(_agentPriority);
            cm.setComponent_Runtype(_runtype);
            cm.setComponent_Session(_session);

// Create a Producer ontology concept object
            Producer pr = new Producer();
            pr.setProducer_Component(cm);

// Create a coda timestamp object
            CodaTimeStamp tm = new CodaTimeStamp();
            tm.setTimestamp_Value(new Date());

// Create a acceptStatus ontology action object
            AcceptStatus myaccept = new AcceptStatus();
            myaccept.setAcceptstatus_Producer(pr);
            myaccept.setAcceptstatus_Timestamp(tm);
            myaccept.setAcceptstatus_Status(status);
            myaccept.setAcceptstatus_Transition(transition);
            if(severity!=null)myaccept.setAcceptstatus_Severity(severity);

// Create an action transport object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(myaccept);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Informs status to receiver (In this case this is the SMS agent).
     * This method sends to the receiver a <clas>AcceptState</clas> coda ontology ACL message, describing the producer
     * i.e. the components that produces information and timestamp.
     *
     * @param stateName               The state of the component
     */
    public void informStateSMS(String stateName) {
        String tmps;
        AID tmpaid;
        if(_mySMS!=null && !_mySMS.isEmpty()){
            for (String a_mySMS : _mySMS) {
                tmps = a_mySMS;

// Create the acl message object
                ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
                try {
                    getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                    tmpaid = new AID(tmps, AID.ISLOCALNAME);

// Preparing the message
                    msg.addReceiver(tmpaid);
                    msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                    msg.setOntology(CodaRunControlOntology.NAME);
                    msg.addUserDefinedParameter(_agentName, tmpaid.getLocalName());

// Create a codacomponent ontology concept object
                    CoolComponent cm = new CoolComponent();
                    cm.setComponent_Name(_agentName);
                    cm.setComponent_Type(_agentType);
                    cm.setComponent_Priority(_agentPriority);
                    cm.setComponent_Runtype(_runtype);
                    cm.setComponent_Session(_session);

// Create a Producer ontology concept object
                    Producer pr = new Producer();
                    pr.setProducer_Component(cm);

// Create a coda timestamp object
                    CodaTimeStamp tm = new CodaTimeStamp();
                    tm.setTimestamp_Value(new Date());

// Create coolState object
                    CoolState myState = new CoolState();
                    myState.setState_Name(stateName);
                    myState.setState_Isfor(cm);
                    
// Create a acceptStatus ontology action object
                    AcceptState myaccept = new AcceptState();
                    myaccept.setAcceptstate_Producer(pr);
                    myaccept.setAcceptstate_Timestamp(tm);
                    myaccept.setAcceptstate_State(myState);

// Create an action transport object
                    Action a = new Action();
                    a.setActor(tmpaid);
                    a.setAction(myaccept);

// Fill the message content
                    getContentManager().fillContent(msg, a);
                } catch (Exception e) {
                    e.printStackTrace();
                }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
                addBehaviour(new GenRequestBehaviour(this, msg, verbose));
            }
        }
    }


    /**
     * Informs receiver to move to the state
     * This method sends to the receiver a <clas>CoolActionMoveTo</clas> coda ontology ACL message,
     * describing the desired state. This method is used by the SMS agent.
     *
     * @param receiver            The name of the receiver
     * @param state               The CoolState object of the component
     */
    public void moveToState(String receiver, CoolState state) {

        AID tmpaid;

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);
            tmpaid = new AID(receiver,AID.ISLOCALNAME);

// Preparing the message
            msg.addReceiver(tmpaid);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,tmpaid.getLocalName());

// Create CoolActionMoveTo ontology action object
            CoolActionMoveTo mt = new CoolActionMoveTo();
            mt.setActionMoveto_State(state);

// Create an action transport object
            Action a = new Action();
            a.setActor(tmpaid);
            a.setAction(mt);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Informs receiver to execute a process
     * This method sends to the receiver a <clas>CoolActionExecute</clas> coda ontology ACL message,
     * describing the desired process. This method is used by the SMS agent.
     *
     * @param receiver            The name of the receiver
     * @param process             The CoolProcess object
     */
    public void executeProcess(String receiver, CoolProcess process) {

        AID tmpaid;

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);
            tmpaid = new AID(receiver,AID.ISLOCALNAME);

// Preparing the message
            msg.addReceiver(tmpaid);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,tmpaid.getLocalName());

// Create CoolActionMoveTo ontology action object
            CoolActionExecute ex = new CoolActionExecute();
            ex.setActionExecute_Process(process);

// Create an action transport object
            Action a = new Action();
            a.setActor(tmpaid);
            a.setAction(ex);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Method sends the data to the receiver agent, with the complete description of the
     * command name, used to retrive the datat from the physical component, state of the execution, etc.
     *
     * @param receiver           Agent who will receive the message
     * @param data               Process execution resulting data
     * @param state              Status or state of the component
     * @param commandname        Actual command name of the process, being executed to retrive the data
     */
    public void informResult(CoolData data, String state, String commandname, AID receiver) {

        if(this.countBlockedMessages()<=10){

// Create and fill the acl message object
            ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF) ;
            try{

// Use codaruncontrol ontology
                getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                msg.addReceiver(receiver);
                msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                msg.setOntology(CodaRunControlOntology.NAME);
                msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Define the component concept sending the result
                CoolComponent cm = new CoolComponent();
                cm.setComponent_Name(_agentName);
                cm.setComponent_Type(_agentType);
                cm.setComponent_Priority(_agentPriority);
                cm.setComponent_Runtype(_runtype);
                cm.setComponent_Session(_session);

// Define the producer concept, necessary for the acceptData action object
                Producer pr = new Producer();
                pr.setProducer_Component(cm);

// Get the time of this response
                CodaTimeStamp tm = new CodaTimeStamp();
                tm.setTimestamp_Value(new Date());

// Create and fill AcceptData codaruncontrol ontology action.
                AcceptData myaccept = new AcceptData();
                myaccept.setAcceptdata_Data(data);
                myaccept.setAcceptdata_Producer(pr);
                myaccept.setAcceptdata_Process(_agentName+":"+commandname);
                myaccept.setAcceptdata_Timestamp(tm);
                myaccept.setAcceptdata_Processstate(state);

// Create an action object
                Action a = new Action();
                a.setActor(receiver);
                a.setAction(myaccept);

// Fill the content of the message
                getContentManager().fillContent(msg, a);
            } catch(Exception e) { e.printStackTrace(); }

// Create and add a behaviour
            addBehaviour(new GenRequestBehaviour(this, msg, verbose));
        }
    }


    /**
     * Method sends data coming from the trigger supervisor component to the receiver agent.
     * Uses special ontology object <clas>AcceptTsData</clas> from the codaruncontrol ontology
     *
     * @param receiver                Agent who will receive the message
     * @param data                    The process execution result
     */
    public void informTsResult(CoolData data, AID receiver) {

// Create and fill the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF) ;
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Define the component concept sending the result
            CoolComponent cm = new CoolComponent();
            cm.setComponent_Name(_agentName);
            cm.setComponent_Type(_agentType);
            cm.setComponent_Priority(_agentPriority);
            cm.setComponent_Runtype(_runtype);
            cm.setComponent_Session(_session);

// Define the producer concept, necessary for the acceptData action object
            Producer pr = new Producer();
            pr.setProducer_Component(cm);

// Get the time of this response
            CodaTimeStamp tm = new CodaTimeStamp();
            tm.setTimestamp_Value(new Date());

// Create and fill AcceptData codaruncontrol ontology action.
            AcceptTsData myaccept = new AcceptTsData();
            myaccept.setAccepttsdata_Data(data);
            myaccept.setAccepttsdata_Producer(pr);
            myaccept.setAccepttsdata_Timestamp(tm);

// Action object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(myaccept);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

//Create and add a behaviour
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Method sends <clas>ResultNotify</clas> ontology action to the agents under SMS control.
     * This is used to organize State Machine Supervisor algorithm.
     *
     * @param receiver                Agent who will receive the message
     */
    public void reportMe(AID receiver) {

// Create and fill the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF) ;
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Define the component concept sending the result
            CoolComponent cm = new CoolComponent();
            cm.setComponent_Name(_agentName);
            cm.setComponent_Type(_agentType);
            cm.setComponent_Priority(_agentPriority);
            cm.setComponent_Runtype(_runtype);
            cm.setComponent_Session(_session);

            ResultNotify rn = new ResultNotify();
            rn.setResultNotify_Component(cm);

// Action object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(rn);

// Fill the content of the message
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

//Create and add a behaviour
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Controls the coda object( real world physical client), by sending cMsg message to the
     * subject = myName and type = rcSubscription....
     * P2P communication (rcDomain).
     * N.B. type of the message is composed from the first 2 parameters to this method.
     *
     * @param control                    session or runcontrol
     * @param ctrl                       clontrol subject (setSession, etc...)
     * @param text                       text of the message
     * @return status
     */
    public boolean controlClient(String control, String ctrl, String text){

// Create the cMsgMessage object
        cMsgMessage msg = new cMsgMessage();

// cMsg subject is the name of the agent ( which is the same as physical component name)
        msg.setSubject(myName);
        msg.setType(control+"/"+ctrl);
        msg.setText(text);
        try {
            if( (rcDomainSrv!=null) && (rcDomainSrv.isConnected())) rcDomainSrv.send(msg); // send the mesage
        } catch (cMsgException ee) {
            try {
                rcDomainSrv.disconnect();
                rcDomainSrv = null;
            } catch (cMsgException e) {
                rcDomainSrv = null;
                e.printStackTrace();
            }
            ee.printStackTrace();
            return false;
        }
        return true;
    }


    /**
     * Sends the cMsg (P2P) to the physical client.
     * N.B. we are using cMsg as a asynchronous communication, i.e. the result will be available
     * in the callback of associated subscribtion. By design, for this release, checkcommand is not
     * suported for cMsg type communications.
     *
     * @param subject                         subject of the message
     * @param type                            type of the message
     * @param txt                             text of the message
     * @return status
     */
    public boolean runcMsg(String subject, String type, String txt){

// Create the cMsgMessage object
        cMsgMessage msg = new cMsgMessage();

// cMsg subject is the name of the agent ( which is the same as physical component name)
        msg.setSubject(subject);

// cmsg type = run/transition/* requires response from the cilent,
// those are download, prestart..... Another type is rc/control/startReporting which means that after getting the
// message component will start internal thread of continuous monitoring, e.g. statistics
        msg.setType(type);
        msg.setText(txt);
        try {
            if( (rcDomainSrv!=null) && (rcDomainSrv.isConnected())) {
                rcDomainSrv.send(msg);
//                System.out.println(myName+ " ... -> subject = "+subject+" type = "+type+" text = "+txt);
            }
        } catch (cMsgException ee) {
            String s = myName+": cMsg communication error.";
            informPlexuses(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
            try {
                rcDomainSrv.disconnect();
                rcDomainSrv = null;
            } catch (cMsgException e) {
                rcDomainSrv = null;
                e.printStackTrace();
            }
            ee.printStackTrace();
            return false;
        }
        return true;
    }


    /**
     * This method executes the tcpServer command using the global fields of this class,
     * assigned at the constructor. Going to be depricated. DEPRICATED
     *
     * @param command                 Shell command name beeing executed
     * @param expectedOk              This string will be compared with the
     *                                captured IO of the shell execution.
     * @return                        rcConstants.failed or the name of the current request transition
     */
    public String runTcp(String command, String expectedOk){

// Result of the execution
        String res;

// Transition name or rcConstants.failed
        String result;

// Check to see if tcp communication object is created and if this agent is coda2 component agent
        if((mytcp==null) && (!_coda2component.equalsIgnoreCase("true"))){
            System.out.println(myName+"-ERROR:  inconsistent cool configuration.IPC=tcp and component is not ddefined as coda2 component.");
            return rcConstants.failed;
        }

// Actual execution of the command
        try{
            res = mytcp.vxcmd(1, command);
        } catch(Exception e){
            e.printStackTrace();
            return rcConstants.failed;
        }

// See if the execution of the main command (not the check command) is ok.
        if(expectedOk !=null){
            if((res!=null) && (!res.equalsIgnoreCase(expectedOk))) result = rcConstants.failed;
            else result = getRequestedTransition();
        } else result = getRequestedTransition();

// N.B. we return _requestedTransition, because the we will be getting the true status in the callback, where the client
// is sending continuosly the statistics information.
        return result;
    }


    /**
     * This method executes the DP command using the global fields of this class,
     * assigned at the constructor.
     *
     * @param command                 Shell command name beeing executed
     * @param expectedOk              This string will be compared with the
     *                                captured IO of the shell execution
     * @return                        rcConstants.failed or the name of the current request transition
     */
    public String runDp(String command, String expectedOk){

// Transition name or rcConstants.failed
        String result;

// Check to see if dp communication object is created and if this agent is coda2 component agent
        if((mydp==null) && (!_coda2component.equalsIgnoreCase("true"))){
            System.out.println(myName+"-ERROR:  inconsistent cool configuration." +
                    "IPC=dpsh and component is not defined as coda2 component.");
            return rcConstants.failed;
        }

// Actual execution of the command
        try{
            result = mydp.sendCmd(command);
        } catch(rcException e){
            e.printStackTrace();
            String s = myName+": DP communication error.";
            informPlexuses(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
            return rcConstants.failed;
        }

// See if the execution of the main command (not the check command) is ok.
        if(expectedOk !=null){
            if((result!=null) && (!result.equalsIgnoreCase(expectedOk))) result = rcConstants.failed;
            else result = getRequestedTransition();
        } else result = getRequestedTransition();

// N.B. we return _requestedTransition, because the we will be getting the true status in the callback, where the client
// is sending continuosly the statistics information.
        return result;
    }


    /**
     * This method executes the DP command using the global fields of this class,
     * assigned at the constructor.
     *
     * @param command                 Shell command name beeing executed
     * @param expectedOk              This string will be compared with the
     *                                captured IO of the shell execution
     * @return                        rcConstants.failed or the name of the requested command
     */
    public String runDpN(String command, String expectedOk){

// Transition name or rcConstants.failed
        String result;

// Check to see if dp communication object is created and if this agent is coda2 component agent
        if((mydp==null) && (!_coda2component.equalsIgnoreCase("true"))){
            System.out.println(myName+"-ERROR:  inconsistent cool configuration.IPC=dpsh and " +
                    "component is not defined as coda2 component.");
            return rcConstants.failed;
        }

// Actual execution of the command
        try{
            result = mydp.sendCmd(command);
        } catch(rcException e){
            e.printStackTrace();
            String s = myName+": DP communication error.";
            informPlexuses(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
            return rcConstants.failed;
        }

// See if the execution of the main command (not the check command) is ok.
        if(expectedOk !=null){
            if((result!=null) && (!result.equalsIgnoreCase(expectedOk))) result = rcConstants.failed;
            else result = command;
        } else result = command;

// N.B. we return _requestedTransition, because the we will be getting the true status in the callback,
// where the client is sending continuosly the statistics information.
        return result;
    }


    /**
     * This method executes the Shell command using the global fields of this class,
     * assigned at the constructor.
     *
     * @param command                 Shell command name beeing executed
     * @param expectedOk              This string will be compared with the
     *                                captured IO of the shell execution
     * @return                        rcConstants.failed or the name of the current request transition
     */
    public String runShell(String command, String expectedOk){

        String res = null;
        String result;
        StringTokenizer st;
        List<String> l;
        BufferedReader myReader;

// Unix shell command execution IO stream, returned by the SysExe object.
        InputStream iostream;

// Actual execution of the command
        l = new ArrayList<String>();

// Tokenize the command.
        st = new StringTokenizer(command);

// Add the tokens into the list
        while(st.hasMoreTokens())l.add(st.nextToken());

// Create a string array
        String[] ta = new String[l.size()];

// Convert the list into the array of strings
        l.toArray(ta);
        String s =myName+"_Info: executing "+command+" ...";
        System.out.println(s);
        informPlexuses(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
        dalogMsg("info",s);


// Execute the unix command and get an IO inputstream of the stdio
        iostream = mySysExe.run_unix_stream(ta);
        if(iostream!=null){
            try {

// Create buffereader using the iostream.
                myReader = new BufferedReader(new InputStreamReader(iostream));

// Atention: Check command will return only one line summary.
                res =myReader.readLine();
                myReader.close();
            } catch (IOException e) {
                String ss = myName+": Shell communication error.";
                informPlexuses(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
                e.printStackTrace();
                return rcConstants.failed;
            }

// N.B. returned null iostream might be ok for some processes
// We do the following if expectedok is defined for the command
            if(expectedOk !=null){

// See if the execution of the main command (not the check command) is ok.
                if(!res.equalsIgnoreCase(expectedOk)) result = rcConstants.failed;
                else result = getRequestedTransition();
            } else result = getRequestedTransition();
        } else result = rcConstants.failed;
        s =myName+"_Info: completed "+command+": stdio => \n"+res;
        System.out.println(s);
        informPlexuses(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
        dalogMsg("info",s);
        return result;
    }


    /**
     * This method executes the Shell command using the global fields of this class,
     * assigned at the constructor.
     *
     * @param command                 Shell command name beeing executed
     * @param expectedOk              This string will be compared with the
     *                                captured IO of the shell execution
     * @return                        rcConstants.failed or the name of the requested command name
     */
    public String runShellN(String command, String expectedOk){

        String res = null;
        String result;
        StringTokenizer st;
        List<String> l;
        BufferedReader myReader;

// Unix shell command execution IO stream, returned by the SysExe object.
        InputStream iostream;

// Actual execution of the command
        l = new ArrayList<String>();

// Tokenize the command.
        st = new StringTokenizer(command);

// Add the tokens into the list
        while(st.hasMoreTokens())l.add(st.nextToken());

// Create a string array
        String[] ta = new String[l.size()];

// Convert the list into the array of strings
        l.toArray(ta);
        String s =myName+"_Info: executing "+command+" ...";
        System.out.println(s);
        informPlexuses(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
        dalogMsg("info",s);

// Execute the unix command and get an IO inputstream of the stdio
        iostream = mySysExe.run_unix_stream(ta);
        if(iostream!=null){
            try {

// Create buffereader using the iostream.
                myReader = new BufferedReader(new InputStreamReader(iostream));

// Atention: Check command will return only one line summary.
                res =myReader.readLine();
                myReader.close();
            } catch (IOException e) {
                String ss = myName+": Shell communication error.";
                informPlexuses(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
                e.printStackTrace();
                return rcConstants.failed;
            }

// N.B. returned null iostream might be ok for some processes
// We do the following if expectedok is defined for the command
            if(expectedOk !=null){

// See if the execution of the main command (not the check command) is ok.
                if(!res.equalsIgnoreCase(expectedOk)) result = rcConstants.failed;
                else result = command;
            } else result = command;
        } else result = rcConstants.failed;
        s =myName+"_Info: completed "+command+": stdio => \n"+res;
        System.out.println(s);
        informPlexuses(rcSubscription.RcGuiSubjWarning+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
        dalogMsg("info",s);
        return result;
    }


    /**
     * This method sends the status type information to all plexuses or user interfaces through cMsg (cMsg domain)ipc
     *
     * @param subject         Subject of the message. In some casses indicates the ceverity
     *                        of the message. for e.g. ERROR, WARNING, INFO, etc.
     * @param type :          Type of the message. In some casses it indicates the name of the transition
     *                        during which the message was generated.
     * @param info            Message content (String)
     */
    private void informPlexuses(String subject, String type, String info) {

        if((cMsgSrv!=null) && (cMsgSrv.isConnected())) {
            cMsgMessage msg = new cMsgMessage() ;
            msg.setSubject(subject);
            msg.setType(type);

            msg.setText(info);
            try {
                cMsgSrv.send(msg);
                cMsgSrv.flush(1);
            } catch (cMsgException ee) {
                try {
                    cMsgSrv.disconnect();
                    cMsgSrv = null;
                } catch (cMsgException e) {
                    cMsgSrv = null;
                    e.printStackTrace();
                }
                ee.printStackTrace();
            }
        }
    }


    /**
     * Method forwards the messages coming from ouside of the platform through
     * cMsg domain server to this agents representative clinet.
     *
     * @param msg                        cMsg message to be forwarded to the client
     */
    private void forward2client(cMsgMessage msg){

// Set send and get message true
        boolean sag;

// Message going to be forwarded to the client
        cMsgMessage fMsg;

// Message coming back as a response to the send and get from the client
        cMsgMessage rMsg;

// See if this is a send and get message
        sag = msg.isGetRequest();
        if((rcDomainSrv!=null) && (rcDomainSrv.isConnected())) {
            try {

// If it is send and get request
                if(sag){

// Create the forwarding message for the client
                    fMsg = new cMsgMessage();
                    fMsg.setSubject(msg.getSubject());
                    fMsg.setType(msg.getType());
                    rMsg = rcDomainSrv.sendAndGet(fMsg,1000);

// At this point we got back response from the client.
                    msg.makeResponse(rMsg);

// Forward the message to the user
                    if(cMsgSrv!=null)cMsgSrv.send(rMsg);
                } else {

// Otherwise we just pass this through RC domain rcDomainSrv to the client.
                    rcDomainSrv.send(msg);
                }
            } catch (cMsgException ee) {
                try {
                    rcDomainSrv.disconnect();
                    rcDomainSrv = null;
                } catch (cMsgException e) {
                    rcDomainSrv = null;
                    e.printStackTrace();
                }
                ee.printStackTrace();
            } catch (TimeoutException e) {
                e.printStackTrace();
            }
        }
    }


    /**
     * Method will analyze space separated string, representing coda statistics message.
     * i.e. event-number (int) event-rate (float) number-of-longs (long) data-rate (double)
     * Method also prepares the ACL message and sends it to the supervisor agent.
     *
     * @param content                     actual statistics string
     * @param status                      state of the component
     */
    private void analizeSTAMessage(String content, String status){

// Used for statistics
        StringBuffer sb;
        String statistics;

// CoolData object for the Medulla
        CoolData myData;

// If the message is null return
        if(content == null || status == null) return;

// if Medulla requested transaction is null return
        if(getRequestedTransition() == null) return;

// 0 the counter of the time since the last message was arrived.
        COUNT = 0;

// Tokenize the message text
        StringTokenizer stk = new StringTokenizer(content, " ");
        if(stk.hasMoreElements()) {

// Get the event number
            if(stk.hasMoreElements())event_number = Integer.parseInt(stk.nextToken());

// ... If daq is in the active state and the status is active report Medulla "active"
            Calendar cal;
            if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_go)){
                if (status.equalsIgnoreCase(rcConstants.CMD_go_ok)){

// Get the start time
                    cal = new GregorianCalendar();
                    runStartTime = cal.getTimeInMillis();

// 0 the average data rate
                    data_rate_a = 0.0;
                    num_of_reportings = 0;

// 0 the end event number left from the previous run
                    end_evt_num = 0;
                    if(!_myState_.equalsIgnoreCase(rcConstants.CMD_go)){

// Inform Medulla go succeeded
                        informStatus(getRequestedTransition(), getRequestedTransition(),null, Medulla_aid);
                        informStateSMS(getRequestedTransition());
                        _myState_ = rcConstants.CMD_go;
                    }
                }
                else
                    informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);
            }

// ... Postgo transition, i.e. we should be getting meaningfull information for event rate , data rate, etc.
            else if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_postgo))
            {

// Physical component status should be "active"
                if(status.equalsIgnoreCase(rcConstants.CMD_go_ok)) {
                    sb = new StringBuffer();
                    if(stk.hasMoreElements())event_rate     = Float.parseFloat(stk.nextToken());
                    if(stk.hasMoreElements())number_of_long = Long.parseLong(stk.nextToken());

// Get the data rate and make it kbyte /sec.
                    if(stk.hasMoreElements())data_rate      = (Double.parseDouble(stk.nextToken()))*4/1000;

// Calculate the average event rate and data rates
// Get the end time
                    cal = new GregorianCalendar();
                    long runEndTime = cal.getTimeInMillis();

// Calculate the overal time of the run
                    float obshiTime = (runEndTime - runStartTime) / 1000;

// Calculate the average values for the current agent-medulla reporting time interval
                    float event_rate_a = event_number / obshiTime;
                    data_rate_a = (data_rate + (data_rate_a*num_of_reportings))/(num_of_reportings += 1);

// Send message to the Medulla after calculating the time difference between last reporting
// N.B. independent of the reporting rate of the physical component, agent reports every
// rcConstants.AgentReportingPeriod time
                    cal = new GregorianCalendar();
                    long curTime = cal.getTimeInMillis();
                    if((curTime-prevTime)>=rcConstants.AgentReportingPeriod){
                        sb.append(event_number).append(" ")
                                .append(event_rate).append(" ")
                                .append(number_of_long).append(" ")
                                .append(data_rate).append(" ")
                                .append(event_rate_a).append(" ")
                                .append(data_rate_a).append(" ");

// Build the string for Medulla_aid
                        statistics = sb.toString();

// Send statistics data to the Medulla
// Compose the Data object, going to be sent to the Medulla.
                        myData = new CoolData();
                        myData.setData_Type("Integer Float Long Double Float Double");
                        myData.setData_Cardinality("multiple");
                        myData.setData_Semantics("Event_Number Event_Rate Number_of_longs Data_Rate Average_Event_Rate Average_Data_Rate");
                        myData.setData_Content(statistics);

// Send the result through the ACL message to the receiver (Medulla)
                        if(statistics !=null) informResult(myData, status,"statistics", Medulla_aid);
                        prevTime = curTime;
                        event_rate = 0;
                        data_rate = 0;
                    }
                } else informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);
            }

// ... Download transition
            else if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_download)){
                if(status.equalsIgnoreCase(rcConstants.CMD_download_ok)){
                    informStatus(getRequestedTransition(), getRequestedTransition(),null, Medulla_aid);
                    informStateSMS(getRequestedTransition());
                }
                else
                    informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);
            }

// ... Prestart transition
            else if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_prestart)){
                if(status.equalsIgnoreCase(rcConstants.CMD_prestart_ok)){
                    informStatus(getRequestedTransition(), getRequestedTransition(),null, Medulla_aid);
                    informStateSMS(getRequestedTransition());
                    _myState_ = rcConstants.CMD_prestart;
                }
                else
                    informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);
            }

// ... End transition
            else if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_end)){
                if (status.equalsIgnoreCase(rcConstants.CMD_end_ok)) {
                    informStatus(getRequestedTransition(), getRequestedTransition(),null, Medulla_aid);
                    informStateSMS(getRequestedTransition());
                    _myState_ = rcConstants.CMD_end;
                }
                else
                    informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);

// Send final statistics info to the supervisor agent.
                sb = new StringBuffer();
                if(stk.hasMoreElements())event_rate     = Float.parseFloat(stk.nextToken());
                if(stk.hasMoreElements())number_of_long = Long.parseLong(stk.nextToken());

// Get the data rate and make it kbyte /sec.
                if(stk.hasMoreElements())data_rate      = (Double.parseDouble(stk.nextToken()))*4/1000;

// Calculate the average event rate and data rates
// Get the end time
                cal = new GregorianCalendar();
                long runEndTime = cal.getTimeInMillis();

// Calculate the overal time of the run
                float obshiTime = (runEndTime - runStartTime) / 1000;

// Calculate the average values for the current agent-medulla reporting time interval
                float event_rate_a = event_number / obshiTime;

// Negative number means that sighn bit is used ( assuming 32 bit word)
                if(number_of_long<0) {
                    long ll = 1;
                    number_of_long = ((ll<<32) | (Math.abs(number_of_long)));
                }
                double data_rate_a = ((number_of_long*4)/1000) / obshiTime;

// Send message to the Medulla after calculating the time difference between last reporting
// N.B. independent of the reporting rate of the physical component, agent reports every
// rcConstants.AgentReportingPeriod time
                cal = new GregorianCalendar();
                long curTime = cal.getTimeInMillis();
                if((curTime-prevTime)>=rcConstants.AgentReportingPeriod) {
                    sb.append(event_number).append(" ")
                            .append(event_rate).append(" ")
                            .append(number_of_long).append(" ")
                            .append(data_rate).append(" ")
                            .append(event_rate_a).append(" ")
                            .append(data_rate_a).append(" ");

// Build the string for Medulla_aid
                    statistics = sb.toString();

// Send statistics data to the Medulla_aid
// Compose the Data object, going to be sent to the Medulla.
                    myData = new CoolData();
                    myData.setData_Type("Integer Float Long Double Float Double");
                    myData.setData_Cardinality("multiple");
                    myData.setData_Semantics("Event_Number Event_Rate Number_of_longs Data_Rate Average_Event_Rate Average_Data_Rate");
                    myData.setData_Content(statistics);

// Send the result through the ACL message to the receiver (Medulla)
                    if(statistics !=null) informResult(myData, status,"statistics", Medulla_aid);
                    prevTime = curTime;
                    event_rate = 0;
                    data_rate = 0;
                }
            }

// ... Reset transition
            else if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_reset)){
                if (status.equalsIgnoreCase(rcConstants.CMD_reset_ok)){
                    informStatus(getRequestedTransition(), getRequestedTransition(),null, Medulla_aid);
                    informStateSMS(getRequestedTransition());
                }
                else
                    informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);
            }

// ... Pause transition
            else if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_pause)){
                if (status.equalsIgnoreCase(rcConstants.CMD_pause_ok)) {
                    informStatus(getRequestedTransition(), getRequestedTransition(),null, Medulla_aid);
                    informStateSMS(getRequestedTransition());
                }
                else
                    informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);
            }

// ... Resume transition
            else if (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_resume)){
                if (status.equalsIgnoreCase(rcConstants.CMD_resume_ok)){
                    informStatus(getRequestedTransition(), getRequestedTransition(),null, Medulla_aid);
                    informStateSMS(getRequestedTransition());

// organizeTransition will assign _requestedTransition to resume. We need to force it to postgo.
                    setRequestedTransition(rcConstants.CMD_postgo);
                }

// If requested transition from Medulla_aid is NOT coda known transitions send failed
                else
                    informStatus(rcConstants.failed,getRequestedTransition(),null, Medulla_aid);
            }
        }
    }


    /**
     * Method parses TS message string and sends ACL message to the supervisor agent.
     *
     * @param content                     space separated string of the TS message
     */
    private void analizeTSMessage(String content){

// CoolData object for the medulla
        CoolData myData;
        if(content==null) return;

// Tokenize the message text
        StringTokenizer stk = new StringTokenizer(content, " ");
        if(stk.hasMoreElements()) {
            if(getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_postgo)){

// Read trigger supervisor messages: first is the livetime,
// and the next 12 are event reates in the TS trigger input bits.

                String lt,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12;
                if(stk.hasMoreElements())lt  = stk.nextToken(); else lt = "0";
                if(stk.hasMoreElements())b1 = stk.nextToken(); else b1 = "0";
                if(stk.hasMoreElements())b2 = stk.nextToken(); else b2 = "0";
                if(stk.hasMoreElements())b3 = stk.nextToken(); else b3 = "0";
                if(stk.hasMoreElements())b4 = stk.nextToken(); else b4 = "0";
                if(stk.hasMoreElements())b5 = stk.nextToken(); else b5 = "0";
                if(stk.hasMoreElements())b6 = stk.nextToken(); else b6 = "0";
                if(stk.hasMoreElements())b7 = stk.nextToken(); else b7 = "0";
                if(stk.hasMoreElements())b8 = stk.nextToken(); else b8 = "0";
                if(stk.hasMoreElements())b9 = stk.nextToken(); else b9 = "0";
                if(stk.hasMoreElements())b10 = stk.nextToken(); else b10 = "0";
                if(stk.hasMoreElements())b11 = stk.nextToken(); else b11 = "0";
                if(stk.hasMoreElements())b12 = stk.nextToken(); else b12 = "0";

// Send ts data to the Medulla_aid. We can send this info directly to Plexus, but it is
// better that Medulla_aid will be the contact point for all user interfaces.
// Compose the Data object
                try{
                    myData = new CoolData();
                    myData.setData_Type("Integer Integer Integer Integer Integer Integer Integer Integer" +
                            "Integer Integer Integer Integer Integer");
                    myData.setData_Cardinality("multiple");
                    myData.setData_Semantics("Live_Time B1_Rate B2_Rate B3_Rate B4_Rate B5_Rate " +
                            "B6_Rate B7_Rate B8_Rate B9_Rate B10_Rate B11_Rate B12_Rate");
                    myData.setData_Content(lt+" "+b1+" "+b2+" "+b3+" "+b4+" "+b5+" "+b6+" "+b7+" "
                            +b8+" "+b9+" "+b10+" "+b11+" "+b12);

// Send the result through the ACL message to the receiver
                    informTsResult(myData, Medulla_aid);
                } catch (NumberFormatException e){
                    e.printStackTrace();
                }
            }
        }
    }





/// ... Inner Class ...

    /**
     * This class is responsible for communication with the actual process or the hardware (component) using
     * IPC (inner process communication) protocols: DP, tcpServer.Shell and cMsg
     * This class is designed mainly for organizing the CODA transition distribution to  the CODA components through
     * the mentioned IPC protocols. This will also organize monitoring processes, as well as scripting support.
     *
     * @author Vardan Gyurjyan
     *         Date: Jun 26, 2006
     *         Time: 10:00:30 PM
     */
    private class RunProcess extends Thread {

        /** keeps transition name ( download, prestart,etc.)*/
        private String _transition;

        /** AID of the process result receiver agent*/
        private AID _receiver = null;

        /** Data type of the result*/
        private String _resultDataType;

        /** Data cardinality of the result, i.e. the result is one number oer String or many.*/
        private String _resultDataCardinality;

        /** Data semantics of the result, i.e. what each naumber/data point means in the result.*/
        private String _resultDataSemantics;

        /** Command string of the process.*/
        private String _commandName;

        /** Parameter of the process command.*/
        private String _parameter;

        /** Command type, defins the IPC protocole. For e.g. dp, cmsg */
        private String _commandType;

        /** Process execution result string in case execution succeded.*/
        private String _expectedOk;

        /** How long we need to wait before the next execution of the loop.*/
        private int dl = 0;

        /** Process execution stating time*/
        private long startTime;

        /** Process execution end time.*/
        private long endTime;

        /** Following boolean variable are used to control the thread execution.*/
        volatile boolean shouldRun = false;
        volatile boolean shouldPause = false;

        /** This variable is set in case the process is one of the coda transitions.
         * N.B. Coda transition processes are defined by Coretx, not by cool*/
        volatile boolean isStateTransition = false;


        /**
         * Constructor. Initializes the global fields of this class.
         *
         * @param cm                  transition name
         * @param process             <clas>CoolProcess</clas> object from the <clas>CodaRunControlOntology</clas>.
         */
        public RunProcess( String cm, CoolProcess process ){

// Initialize the transition
            _transition = cm;

// See if this is a coda state transition
            if(process.getProcess_Isstatetransition()!=null) {
                if(process.getProcess_Isstatetransition().equalsIgnoreCase("true"))
                    isStateTransition = true;
            }

// Report to Medulla
            _receiver = Medulla_aid;

// Getting command related information from the passed CoolCommand ontology concept.
            _commandName = process.getProcess_Commandname();

// Get the command type.
            if(process.getProcess_Ipc()!=null)
                _commandType = process.getProcess_Ipc();

// If input parameter is specified
            if(process.getProcess_Inputdata()!=null){
                _parameter = process.getProcess_Inputdata().getData_Content();
            }

// if return data is specified
            if(process.getProcess_Outputdata()!=null){

// Get the  Data type
                if(process.getProcess_Outputdata().getData_Type()!=null)
                    _resultDataType = process.getProcess_Outputdata().getData_Type();

// Get the  Data cardinality
                if(process.getProcess_Outputdata().getData_Cardinality()!=null)
                    _resultDataCardinality =  process.getProcess_Outputdata().getData_Cardinality();

// Get the  Data semantics
                if(process.getProcess_Outputdata().getData_Semantics()!=null)
                    _resultDataSemantics =  process.getProcess_Outputdata().getData_Semantics();
            }

// Command loop parameters
            if(process.getProcess_Delay()!=null)
                dl = process.getProcess_Delay();

// Get expected_ok string set in the cool configuration file
            if((process.getProcess_Outputdata()!=null) &&
                    (process.getProcess_Outputdata().getData_Content()!=null))
                _expectedOk = process.getProcess_Outputdata().getData_Content();
        }


        /**
         * The run method of the class
         */
        public void run() {

// Boolean status of cMsg execution
            boolean stcm;

// This variable will keep calculated delay/wait time before the next execution of the process
            int actualDelay;

// Temporary String holds the process execution result.
            String result = null;

// Execution time of the process
            long execTime;

// Calendar object to calculate the actual process execution time by detecting the start and
// end time of the process execution.
            Calendar myCalendar;

// Negative values of the loop  will indicate that it is not a continuos process
            if(_loop >= 0) {
                for(int i = 0;i < _loop; i++) {

// ... Command type is cMsg.
                    if(_commandType.equalsIgnoreCase("cMsg")){

// "isstatetransition" is an ontology concept desighned to distinguish between coda basic transitions and others.
// Usualy types defines a coda state transition, like download, prestart,...
// however sometimes those transitions have parameters passed in the text of the message.                        
// ... Reset is sent through subject rc/control
                        if(_commandName.equalsIgnoreCase("reset")) {
                            stcm = runcMsg(myName, rcSubscription.SessionControlType +"/"+_commandName,"reset");

// Stop monitoring processes of the client
                            controlClient(rcSubscription.SessionControlType, rcConstants.CMD_StopReporting,"stopreporting");
                        }
                        else if(isStateTransition)
                            stcm = runcMsg(myName, rcSubscription.RunTransitionType +"/"+_commandName, _parameter);
                        else
                            stcm = runcMsg(myName, rcSubscription.SessionControlType +"/"+rcConstants.CMD_Exec,_commandName);
                        if(stcm){

// If sending the cMsg succeedded send the inform status = transition to
// the Medulla if it is NOT the basic coda transitions
                            if(!_transition.equalsIgnoreCase("prestart") &&
                                    ((_transition.contains("pre")) || (_transition.contains("post")) ||
                                            ((_transition.equals("sm"))))) {
                                informStatus(_transition, _transition,null, _receiver);
                                informStateSMS(_transition);
                            }
                        }
                    }

// ... Command type is tcpServer/vxServer.
                    else if(_commandType.equalsIgnoreCase("tcp")){
                        if(_commandName.equalsIgnoreCase("exit")) {

// Stop monitoring processes of the client
                            controlClient(rcSubscription.SessionControlType, rcConstants.CMD_StopReporting,"stopreporting");
                        }
                        result = runTcp(_commandName, _expectedOk);
                    }

// ... Command type is Dp.
                    else if(_commandType.equalsIgnoreCase("dpsh")){
                        if(_commandName.equalsIgnoreCase("exit")) {

// Stop monitoring processes of the client
                            controlClient(rcSubscription.SessionControlType, rcConstants.CMD_StopReporting,"stopreporting");
                        }
                        result = runDp(_commandName, _expectedOk);
                    }

// ... Command type is Unix, i.e. the external process is unix process
                    else if (_commandType.equalsIgnoreCase("shell")) {
                        result = runShell(_commandName, _expectedOk);
                    }

// The stemcell agent does not understand other than cMsg, tcp ,shell and dp protocols
                    else {
                        System.err.println("Unknown command type: "+ _commandType);
                    }

// We sleep for next external process execution
                    try {
                        sleep(dl *1000);
                    } catch (Exception e) {e.printStackTrace();}
                }

// Sending the message to Medulla.
                if((result!= null) && (!result.equals(rcConstants.failed))){

// If the execution of the process was successful send the _result = transition to
// the Medulla if it is NOT the basic coda transitions.
// N.B. _result  will be set in the IPC specific methods (e.g. runDp())
                    if(!_transition.equalsIgnoreCase("prestart") &&
                            ((_transition.contains("pre")) || (_transition.contains("post"))) ||
                            ((_transition.equals("sm")))) {
                        informStatus(result, _transition, null, _receiver);
                        informStateSMS(_transition);
                    }
                }

// Cyclic process execution: loop<0
            } else {

// We report the transition back to thalamus to go on with the coda transitons without waiting
// for the results of the monitoring processes.
// Attention: loop<0 processes are asynchronous, and no associated check commands
                informStatus(_transition, _transition,null, _receiver);
                informStateSMS(_transition);

// Set the delay. This will later be set to the value equal to the execution time.
                actualDelay = dl *1000;

// Now this thread is busy serving this cyclic process execution
                while(shouldRun) {
                    if(!shouldPause) {

// .. Command type is cMsg.
                        if(_commandType.equalsIgnoreCase("cMsg")){

// If cMsg communication fails exit this thread.
// N.B. cyclic processes can not be coda state transitions
                            if(isStateTransition)shouldRun = false;
                            if(!runcMsg(myName, rcSubscription.SessionControlType +"/"+rcConstants.CMD_Exec, _commandName)) shouldRun = false;
                        }

// .. tcpServer IPC command
                        else if(_commandType.equalsIgnoreCase("tcp")) {

// Get the time when the execution starts
                            myCalendar = new GregorianCalendar();
                            startTime = myCalendar.getTimeInMillis();

// Actual execution
                            result = runTcp(_commandName, _expectedOk);

// If we got communication problem, we need to stop this thread.
                            if((result ==null) && (result.equalsIgnoreCase(rcConstants.failed)))shouldRun = false;

// Get the time when the foreign process execution is over.
                            myCalendar = new GregorianCalendar();
                            endTime = myCalendar.getTimeInMillis();
                        }

// .. DP IPC command
                        else if(_commandType.equalsIgnoreCase("dpsh")) {

// Get the time when the execution starts
                            myCalendar = new GregorianCalendar();
                            startTime = myCalendar.getTimeInMillis();

// Actual execution
                            result = runDp(_commandName, _expectedOk);

// We got communication problem, we need to stop the thread.
                            if((result == null) && (result.equalsIgnoreCase(rcConstants.failed))) shouldRun = false;

// Get the time when the foreign process execution is over.
                            myCalendar = new GregorianCalendar();
                            endTime = myCalendar.getTimeInMillis();
                        }

// .. Unix command
                        else if (_commandType.equalsIgnoreCase("shell")) {

// Get the time when the execution starts
                            myCalendar = new GregorianCalendar();
                            startTime = myCalendar.getTimeInMillis();

// Actual execution
                            result = runShell(_commandName, _expectedOk);

// We got communication problem, we need to stop the thread.
                            if((result == null) && (result.equalsIgnoreCase(rcConstants.failed))) shouldRun = false;

// Get the time when the foreign process execution is over.
                            myCalendar = new GregorianCalendar();
                            endTime = myCalendar.getTimeInMillis();

// Unknown communication protocole
                        } else  System.err.println("Unknown command type: "+ _commandType);

// Report the result to the agent receiver. Attention: for cyclic monitoring processes
// we do not send the status of the execution, but only the result of the execution, which
// is stdout or stderr of the process.
// Compose the Data object
                        CoolData myData = new CoolData();
                        myData.setData_Type(_resultDataType);
                        myData.setData_Cardinality(_resultDataCardinality);
                        myData.setData_Semantics(_resultDataSemantics);
                        myData.setData_Content(result);

// Send the result through the ACL message to the receiver.
                        if(_receiver != null) informResult(myData, result, _commandName, _receiver);

// Calculate the process execution time. We add 1sec. relaxation time.
                        execTime = (endTime-startTime)+1000;

// If requested delay of the monitoring cycle is less then actual process execution time + 1sec.,
// then delay will be set to be equal to the execution time + 1sec relaxation time.
                        if(((dl *1000) <= execTime) && (execTime<rcConstants.LONG_EXEC_TIME)) {
                            actualDelay=(int)execTime;
                        }

// Otherwise it will stay the required monitoring cycle delay.
                        else actualDelay = dl *1000;
                    } // end of shouldPause

// Implementation of the loop delay
                    try{
                        sleep(actualDelay);
                    } catch (Exception e) {e.printStackTrace();}
                } // end of shouldRun
            } // end of continuos loop
        } // end of run method
    } // End of the RunProcess class



    /**
     * Callback class for handling messages of the type coda/info/*
     * from the outside user.
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2006
     *         Time: 10:00:30 PM
     */

    private class CodaInfoCallback extends cMsgCallbackAdapter{

        /**
         * Callback method to handle cMsg messages coming from the outside of the platform for coda objects
         * info/* type
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {
            forward2client(msg);
        }
    }



    /**
     * Callback class for handling messages of the type session/control/*
     * from the outside user.
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2006
     *         Time: 10:00:30 PM
     */

    private class SessionControlCallback extends cMsgCallbackAdapter{

        /**
         * Callback method to handle cMsg messages coming from the outside of the platform for session objects
         * control/* type
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {
            forward2client(msg);
        }
    }



    /**
     * Callback class for handling messages of the type run/control/*
     * from the outside user.
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2006
     *         Time: 10:00:30 PM
     */

    private class RunControlCallback extends cMsgCallbackAdapter{

        /**
         * Callback method to handle cMsg messages coming from the outside of the platform for run objects
         * control/* type
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {
            forward2client(msg);
        }
    }



    /**
     * Callback class for handling messages of the type run/transition/*
     * from the outside user.
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2006
     *         Time: 10:00:30 PM
     */

    private class RunTransitionCallback extends cMsgCallbackAdapter{

        /**
         * Callback method to handle cMsg messages coming from the outside of the platform for run objects
         * transition/* type
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {
            forward2client(msg);
        }
    }



    /**
     * Callback class for handling messages :subject = null type = null
     * from the representing physical component (client).
     * N.B. This callback is for handling physical client messages, which are not used within the AFECS platform
     * and will be passed the the outside subscribers through cMsg domain.
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2006
     *         Time: 10:00:30 PM
     */

    private class DontCareCallback extends cMsgCallbackAdapter{

        /**
         * Callback method to handle cMsg messages coming from the client
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {

            if((cMsgSrv!=null) && (cMsgSrv.isConnected())) {

// We just passe this through cMsg domain to the user
                try {
                    cMsgSrv.send(msg);
                } catch (cMsgException ee) {
                    try {
                        cMsgSrv.disconnect();
                        cMsgSrv = null;
                    } catch (cMsgException e) {
                        cMsgSrv = null;
                        e.printStackTrace();
                    }
                    ee.printStackTrace();
                }
            }
        }
    }



    /**
     * Callback class for handling messages of the type run/transition/*
     * from the representing physical component (client).
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2006
     *         Time: 10:00:30 PM
     */

    private class ResponseCallback extends cMsgCallbackAdapter{

        /**
         * Callback method to handle cMsg messages coming from the the client
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {

            if((cMsgSrv!=null) && (cMsgSrv.isConnected())) {
                try {
                    cMsgSrv.send(msg);
                } catch (cMsgException ee) {
                    try {
                        cMsgSrv.disconnect();
                        cMsgSrv = null;
                    } catch (cMsgException e) {
                        cMsgSrv = null;
                        e.printStackTrace();
                    }
                    ee.printStackTrace();
                }
            }
        }
    }



    /**
     * Class handles real world physical client statistics messages.
     */
    private class RunStatisticsXMLCallback extends cMsgCallbackAdapter {

        /**
         * Callback method for P2P, rc domain messages coming from the client.
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {

// Get the text of the message
            String monitor = msg.getText();

// If the message is null return
            if(monitor == null) return;

            synchronized(handler){

// Parsing XML string
                try{
                    SAXParserFactory factory = SAXParserFactory.newInstance();
                    factory.setValidating(false);
                    factory.newSAXParser().parse(new InputSource(new StringReader(monitor)), handler);
                }  catch (SAXException e){
                    System.out.println("NOT VALID XML");
                    e.printStackTrace();
                } catch(ParserConfigurationException e) {
                    System.out.println("CONFIGURATION EX.");
                    e.printStackTrace();
                }   catch (IOException e){
                    System.out.println("IO EX.");
                    e.printStackTrace();
                }
            }
        }
    }


    /**
     * Class handles real world physical client dalog messages.
     */
    private class DalogCallback extends cMsgCallbackAdapter {

        private String previousText;
        /**
         * Callback method for P2P, rc domain messages coming from the client.
         *
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {

            String codaClass = null, text;
            cMsgMessage nm = new cMsgMessage();
            List<Integer> sever = new ArrayList<Integer>();
            int severity;

// The subject is the name of the physical component which
// is the same as its representative agent name.
// To construct the subject we need the type of the component, which we will get from the
// message sent from the physical client.
            try {
                if(msg.getPayloadItem(rcSubscription.dalogMessageComponentClass)!=null)
                    codaClass = msg.getPayloadItem(rcSubscription.dalogMessageComponentClass).getString();
            } catch (cMsgException e) {
                e.printStackTrace();
            }

// Get the severity
            severity = msg.getUserInt();
            if(rcSubscription.expid!=null && _session!=null && _runtype!=null && codaClass!=null && myName!=null){

                nm.setSubject(rcSubscription.expid+"/"+_session+"/"+_runtype+"/"+codaClass+"/"+myName);
                nm.setType("dalog/"+severity);

// get the message text
                text = msg.getText();

//            System.out.println(myName+"-Debug: subject = "+rcSubscription.expid+"/"+_session+"/"+_runtype+"/"+myType+"/"+myName+
//                    " type = "+"dalog/"+severity+" text = "+text);

// Check if this is the duplicate message from the client then return
                if(!text.equals(previousText)) {

                    // Do not send in case we need to filter the messages based on the severity level
// Parse the DALOG_FILTER set in the cool configuration file and get the severity
// levels that need to be filtered, i.e. not send
                    if(DALOG_FILTER!=null){
                        try{
                            StringTokenizer st1 = new StringTokenizer(DALOG_FILTER,",");
                            while(st1.hasMoreTokens()) {
                                String token1 = st1.nextToken();
                                if(token1.contains("-")){
                                    StringTokenizer st2 = new StringTokenizer(st1.nextToken(),"-");
                                    while(st2.hasMoreTokens()){
                                        sever.add(new Integer(st2.nextToken()));

                                    }
                                }
                                else sever.add(new Integer(token1));
                            }
                        }catch (NumberFormatException e){
                            e.printStackTrace();
                        }
                        if(!sever.isEmpty()){
                            for (Integer aSeverity : sever) if (aSeverity == severity) return;
                        }
                    }

// Set the text of the new message
                    previousText = text;
                    try {
// build outgoing message from the agent
                    nm.copyPayload(msg);
                    nm.setText(text);

                        String nseverity=null;
                               if(msg.getPayloadItem("severity")!=null)
                                   nseverity = msg.getPayloadItem("severity").getString();
                        String nstate=null;
                                if(msg.getPayloadItem("state")!=null)
                                    nstate= msg.getPayloadItem("state").getString();

                        cMsgPayloadItem newseverityitem = new cMsgPayloadItem("severity",nseverity+"."+severity);
                        cMsgPayloadItem newstateitem = new cMsgPayloadItem("state",nstate);
                        cMsgPayloadItem p_expid = new cMsgPayloadItem("EXPID",rcSubscription.expid);
                        cMsgPayloadItem p_session = new cMsgPayloadItem("session",_session);
                        cMsgPayloadItem p_config = new cMsgPayloadItem("config",_runtype);
                        cMsgPayloadItem p_codaName = new cMsgPayloadItem("codaName",myName);
                        nm.addPayloadItem(newseverityitem);
                        nm.addPayloadItem(newstateitem);
                        nm.addPayloadItem(p_expid);
                        nm.addPayloadItem(p_session);
                        nm.addPayloadItem(p_config);
                        nm.addPayloadItem(p_codaName);

                        cMsgSrv.send(nm);
                        cMsgSrv.flush(1);
                    } catch (cMsgException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
    }



    /**
     * Inner class extends SAX parser and does parsing of the rcstatistics XML messages.
     */
    private class RcStatusHandler extends DefaultHandler{
        String st, lt;
        String tag;
        String filename;
        String codaclass;

        public void startElement(String namespaceURI, String localName,
                                 String qName, Attributes atts) {
            {
                tag = qName;
                int attrCount = atts.getLength();
                if(attrCount>0) {
//                    System.out.println("Attributes:");
                    for(int i = 0 ; i<attrCount ; i++) {
//                        System.out.println("  Name : " + atts.getQName(i));
//                        System.out.println("  Type : " + atts.getType(i));
//                        System.out.println("  Value: " + atts.getValue(i));
                        if(atts.getQName(i).equalsIgnoreCase("state")) st=atts.getValue(i);
                        else if(atts.getQName(i).equalsIgnoreCase("livetime")) {
                            lt = atts.getValue(i);
                            if ((getRequestedTransition()!=null) && (_agentType.equalsIgnoreCase(rcConstants.tsType)))analizeTSMessage(lt);
                        }
                        else if(atts.getQName(i).equalsIgnoreCase("fileName")){
                            filename = atts.getValue(i);
                        }
                        else if(atts.getQName(i).equalsIgnoreCase("codaClass")){
                            codaclass = atts.getValue(i);
                        }
                    }
                }
            }
            if(filename !=null && (codaclass.equalsIgnoreCase("ER") ||
                    codaclass.equalsIgnoreCase("CDEB")||
                    codaclass.equalsIgnoreCase("EB"))){
            informPlexuses(rcSubscription.RcGuiSubjInfo+"_"+_session, rcSubscription.RcGuiTypeFileName,filename);
            }
        }


        public void characters ( char[] ch, int start, int len ) {
            String text = new String( ch, start, len );
            String text1 = text.trim();
            if( text1.length() > 0 ) {
                if((tag.equalsIgnoreCase("runStatistics")) && (st!=null) && (getRequestedTransition()!=null)) analizeSTAMessage(text1,st);
            }
        }
    }



    /**
     * Lanches a thread that will count the time before the next cMsg message will arive.
     * In case message is not arriving it will send warning ACL messages to Medulla and Plexus agents.
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2006
     *         Time: 10:00:30 PM
     */
    private class messageWatcher extends Thread {

        // Thread control variable
        public volatile boolean shouldrun = true;

        /**
         * Run method of the Thread
         */
        public void run(){
            while(shouldrun){

// Increment the count and sleep for a second
                COUNT++;
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

// The reset button has beed clicked. Reset the count and continue.
                if(RESETTIME){
                    COUNT=0;
                    continue;
                }

// Build a info string for Medulla
                String txt = myName+" is not reporting "+COUNT+" seconds ";

// If we are in the active state and component is not responding for n sec. send message to thalamus
// which  will broadcast to all plexuses to print on their consols.
                if((getRequestedTransition() !=null) &&
                        (getRequestedTransition().equalsIgnoreCase(rcConstants.CMD_postgo)
                                && (COUNT !=0) && ((COUNT%rcConstants.ACTIVE_BUT_NOTREPORTING)==0) )) {

// Send warning message to Medulla
                    informStatus(txt,getRequestedTransition(),"Warning", Medulla_aid);

// Send szero statistics data to the Medulla. This will force 0-ing the histograms on the plexuses
// Compose the Data object. This assumes that we know the statistics data structure.
                    CoolData myData = new CoolData();
                    myData.setData_Type("Integer Float Long Double");
                    myData.setData_Cardinality("multiple");
                    myData.setData_Semantics("Event_Number Event_Rate Number_of_longs Data_Rate");
                    myData.setData_Content("0 0.0 0.0 0.0");

// Send the result through the ACL message to the receiver
                    informResult(myData, rcConstants.failed,"statistics", Medulla_aid);

// If it is not reset transition and we are in postgo ( i.e. active state)
// see if the component is not reporting END_REQUEST_TIME seconds.
                    if(COUNT > rcConstants.END_REQUEST_TIME) {
                        String s = myName+" agent is suggesting to end the run ... ";
                        informPlexuses(rcSubscription.RcGuiSubjError+"_"+_session, rcSubscription.RcGuiTypeStemCell,s);
                        dalogMsg("info",s);

// This will also force Medulla to put this component into the failed component list.
                        informStatus(txt,getRequestedTransition(),"Error", Medulla_aid);

// Let us not exit this thread, but o the COUNT instead.
                        COUNT=0;
                    }
                }
            }
        }
    } // end messageWatch class


}
