/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.agents;

import org.jlab.coda.runcontrol.util.*;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.*;
import rca.domain.FIPAAgentManagement.DFAgentDescription;
import rca.domain.FIPAAgentManagement.ServiceDescription;
import rca.domain.DFService;
import rca.domain.FIPAException;
import rca.domain.FIPANames;
import rca.content.lang.Codec;
import rca.content.lang.sl.SLCodec;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.content.onto.OntologyException;
import rca.lang.acl.ACLMessage;
import rca.util.leap.List;
import rca.util.leap.ArrayList;
import rca.util.leap.Iterator;
import rca.core.AID;

import java.util.LinkedHashMap;


/**
 * AFECS control system supervisor deputy/assistent agent. Incapsulates coda state machine.
 * Object is configured using the ACL messages sent by <code>Cortex<code>.
 *
 * Author: gurjyan
 * Date: Jun 9, 2006
 * Time: 2:00:52 PM
 *
 */
public class Thalamus extends CAgent{

    /** The name of this agent */
    private String myName;

    /** Reporting interval to the sherif (Medulla)*/
    private int REPINTERVAL = 2000;

    /** Object that transforms a message's content to and from a specific language
     * representation. The language in this case is SL0. */
    private Codec  codec = new SLCodec();

    /** Verbosity level defined in the rcConstants class.*/
    private int verbose = rcConstants.debugError;

    /** List of StatusQuant objects containing component statuses*/
    private LinkedHashMap<String,StatusQuant> reportedStates =  new LinkedHashMap<String,StatusQuant>();

    /** List of DataQuant objects containing component statistics messages*/
    private LinkedHashMap<String,DataQuant> reportedData = new LinkedHashMap<String,DataQuant>();

    /** Lists holding information from the last update*/
    private final List StatusList = new ArrayList();
    private final List DataList = new ArrayList();
    private  List prevStatusList = new ArrayList();
    private  List prevDataList = new ArrayList();

    /**
     * This method is for application specific startup code. When this method is
     * called the agent has been already registered with the Agent Platform AMS
     * and is able to send and receive messages. However, no behaviour scheduling is
     * active yet. This method is essential to at least add a behaviour to the
     * agent, in order for it to be able to do anything.
     */
    protected void setup() {

// Register the supported content language
        getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);

// Register the supported ontologi
        getContentManager().registerOntology(CodaRunControlOntology.getInstance());

// This is the given name when the agent was started by the Hepatic.
        myName = getLocalName();

// Register with the DF
        DFAgentDescription dfd = new DFAgentDescription();
        ServiceDescription sd = new ServiceDescription();
        sd.setType("Thalamus");
        sd.setName(myName);
        sd.setOwnership("JLABCODA");
        dfd.setName(getAID());
        dfd.addServices(sd);
        try {
            DFService.register(this,dfd);
        }
        catch (FIPAException e) {
            System.err.println(getLocalName()+" registration with DF unsucceeded. Reason: "+e.getMessage());
            doDelete();
        }

// Adds the responder behaviour, which will be scheduled for the execution in the agent thread.
// This is the behaviour where we process request information from other agents. The actual
// processing method is defined in this class but called from the GenResponderBehaviour class
        addBehaviour(new GenRespondBehaviour(this, verbose));

// Limit agent internal queue size
        this.setQueueSize(2000);

// Start reporting thread
        Report2Sherif rep = new Report2Sherif();
        rep.start();
    }


    /**
     * This is the implementatioin of the abstract method defined in the CAgent class.
     * This method is going to be called byt the BaseicRequestBehaviour object to process
     * the iform messages coming back to this agent as a result of the initiating the request
     * to some agent on the platfrom.
     *
     * @param msg             ACLMessage object passed from the GenRequestBehaviour object.
     */
    protected void processInform(ACLMessage msg) {
    }


    /**
     * This method is mainly for handling requests from the Cortex and component representative agents.
     * This is the implementation of the abstract method defined in the CAgent class.
     * This method is going to be called by the BaseicResponderBehaviour object to process
     * the request to this agent.
     * N.B. this is the case when requests are coming through agent communication channel.
     *
     * @param request                   ACLMessage request object passed from the GenResponderBehaviour object.
     *                                  Here we fill it and return to the GenResponderBehaviour object.
     * @param a                         Action Object
     * @return                          ACLMessage object
     */
    protected ACLMessage processRequest(ACLMessage request, Action a) {

// This method actualy will start a separate thread to carry out external unix or vxworks executable functions.
// Create an acl message objects
        ACLMessage reply = request.createReply();

// Temporary message content string
        String tmpStr = "NoContent";


// Component agents are sending back status info to Thalamus
        if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                "CodaRunControl.AcceptStatus")) {

// Get the ontology action object
            AcceptStatus myAcceptStatus = (AcceptStatus)a.getAction();
            StatusQuant sq = new StatusQuant();
            sq.setStatusquant_Producer(myAcceptStatus.getAcceptstatus_Producer());
            sq.setStatusquant_Severity(myAcceptStatus.getAcceptstatus_Severity());
            sq.setStatusquant_Status(myAcceptStatus.getAcceptstatus_Status());
            sq.setStatusquant_Timestamp(myAcceptStatus.getAcceptstatus_Timestamp());
            sq.setStatusquant_Transition(myAcceptStatus.getAcceptstatus_Transition());
            statusUpdate(myAcceptStatus.getAcceptstatus_Producer().getProducer_Component().getComponent_Name(),sq);
        }

// Component agents are sending back data (string) to Thalamus
        else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                "CodaRunControl.AcceptData")) {
            AcceptData myAcceptData = (AcceptData)a.getAction();
            DataQuant dq = new DataQuant();
            dq.setDataquant_Data(myAcceptData.getAcceptdata_Data());
            dq.setDataquant_Process(myAcceptData.getAcceptdata_Process());
            dq.setDataquant_Processstate(myAcceptData.getAcceptdata_Processstate());
            dq.setDataquant_Producer(myAcceptData.getAcceptdata_Producer());
            dq.setDataquant_Timestamp(myAcceptData.getAcceptdata_Timestamp());
            dataUpdate(myAcceptData.getAcceptdata_Producer().getProducer_Component().getComponent_Name(),dq);
        }
        try{

// Construct done  action and send to the initiator
            Done d = new Done();
            d.setAction(a);
            getContentManager().fillContent(reply, d);
            reply.setPerformative(ACLMessage.INFORM);
            reply.setContent(tmpStr);
        }
        catch (Codec.CodecException e) {
            System.err.println(this.getLocalName()+" Fill/extract content unsucceeded. Reason:" + e.getMessage());
        }
        catch (OntologyException e){
            System.err.println(this.getLocalName()+" Ontology exception. Reason:" + e.getMessage());
        }
        return reply;
    }


    /**
     * Sends the <clas>AcceptGroupStatus</clas> action ontology object to the Medulla,
     * including the list of <clas>StatusQuant</clas> objects
     *
     * @param receiver AID of the medulla ( gran supervisor)
     */
    public void informStatus(AID receiver) {

        if(reportedStates.isEmpty())return;

// Iterate over the reportedStates hashmap and fill the List
        getStatusList();

// Compare current and precious reportings. If nothing is changed do not report to the Sherif.
        if(compareStatusIfEqual()) return;

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(myName,receiver.getLocalName());
            AcceptGroupStatus myaccept = new AcceptGroupStatus();
            myaccept.setAcceptgroupstatus_Status(StatusList);

// Create an action transport object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(myaccept);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
    }


    /**
     * Sends the <clas>AcceptGroupData</clas> action ontology object to the Medulla,
     * including the list of <clas>DataQuant</clas> objects
     *
     * @param receiver AID of the medulla ( gran supervisor)
     */
    public void informData(AID receiver) {

// See if this agent has internaly blocked messages, and do not send
// any new messages untill there is room for new messages.
        if(this.countBlockedMessages()<=10){
        if(reportedData.isEmpty())return;

// Iterate over the reporteddata hashmap and fill the List
        getDataList();

// Compare current and precious reportings. If nothing is changed do not report to the Sherif.
        if(compareDataIfEqual()) return;

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(myName,receiver.getLocalName());
            AcceptGroupData myaccept = new AcceptGroupData();
            myaccept.setAcceptgroupdata_Data(DataList);

// Create an action transport object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(myaccept);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, verbose));
        }
    }


    /** Add/update <em>reportedStatus</em> hashmap
     *
     * @param k key
     * @param v value
     */
    public synchronized void statusUpdate(String k, StatusQuant v){
        reportedStates.put(k,v);
    }


    /**
     * Reads the <em>reportedStatus</em> hashmap and fill the StatusList
     */
    public synchronized void getStatusList(){
        StatusList.clear();

// Iterate over the reportedStates hashmap and fill the StatusList
        for (String s : reportedStates.keySet()) {
            StatusList.add(reportedStates.get(s));
        }
    }


    /** Add/update <em>reportedData</em> hashmap
     *
     * @param k key
     * @param v value
     * */
    public synchronized void dataUpdate(String k, DataQuant v){
        reportedData.put(k,v);
    }


    /**
     * Reads the reportedData hashmap and fill the DataList
     */
    public synchronized  void getDataList(){
        DataList.clear();

// Iterate over the reportedData hashmap and fill the DataList
        for (String s : reportedData.keySet()) {
            DataList.add(reportedData.get(s));
        }
    }


    /** Compares two lists containing StatusQuant objects
     *
     * @return status
     * */
    public synchronized boolean compareStatusIfEqual(){
        boolean status = true;
        StatusQuant ctmp, ptmp;
        boolean notequal = false;

        if(StatusList.size()!=prevStatusList.size())notequal = true;
        else{
            Iterator cit = StatusList.iterator();
            Iterator pit = prevStatusList.iterator();
            while(cit.hasNext() && pit.hasNext()){
                ctmp = (StatusQuant)cit.next();
                ptmp = (StatusQuant)pit.next();
                if(!(ctmp.getStatusquant_Status().equals(ptmp.getStatusquant_Status()))) {
                    notequal = true;
                    break;
                }
            }
        }
        if(notequal){
            prevStatusList.clear();
           Iterator cit = StatusList.iterator();
             while(cit.hasNext()){
                 prevStatusList.add(cit.next());
             }
            status = false;
        }
        return status;
    }

    
    /** Compares two lists and creats the new list containing only changed <clas>StatusQuant</clas> objects
     *
     * @param current                         current list
     * @param previous                        previous list
     * @return l                              list containing only changed objects
     * */
    public synchronized List newReportedStatusList(List current, List previous){
        StatusQuant ctmp, ptmp;

        List l = new ArrayList();

        if(current.size()!=previous.size())return current;
        else{
            Iterator cit = current.iterator();
            Iterator pit = previous.iterator();
            while(cit.hasNext() && pit.hasNext()){
                ctmp = (StatusQuant)cit.next();
                ptmp = (StatusQuant)pit.next();
                if(!(ctmp.getStatusquant_Status().equals(ptmp.getStatusquant_Status()))) l.add(ctmp);
            }
        }
        return l;
    }


    /**
     * Compares two lists containing <clas>DataQuant</clas> objects
     *
     * @return status
     * */
    public synchronized boolean compareDataIfEqual(){
        boolean status = true;
        DataQuant ctmp, ptmp;
        boolean notequal = false;

        if(DataList.size()!=prevDataList.size())notequal = true;
        else{
            Iterator cit = DataList.iterator();
            Iterator pit = prevDataList.iterator();
            while(cit.hasNext() && pit.hasNext()){
                ctmp = (DataQuant)cit.next();
                ptmp = (DataQuant)pit.next();
                if(!(ctmp.getDataquant_Data().getData_Content().equals( ptmp.getDataquant_Data().getData_Content()))){
                    notequal = true;
                    break;
                }
            }
        }
        if(notequal){
            prevDataList.clear();
           Iterator cit = DataList.iterator();
             while(cit.hasNext()){
                 prevDataList.add(cit.next());
             }
            status = false;
        }
        return status;
    }


    /**
     * Compares two lists and creats the new list containing only changed <clas>DataQuant</clas> objects
     *
     * @param current                        current list
     * @param previous                       previous list
     * @return l                             list containig only changed objects
     * */
    public synchronized List newReportedDataList(List current, List previous){
        DataQuant ctmp, ptmp;

        List l = new ArrayList();

        if(current.size()!=previous.size())return current;
        else{
            Iterator cit = current.iterator();
            Iterator pit = previous.iterator();
            while(cit.hasNext() && pit.hasNext()){
                ctmp = (DataQuant)cit.next();
                ptmp = (DataQuant)pit.next();
                if(!(ctmp.getDataquant_Data().equals(ptmp.getDataquant_Data()))) l.add(ctmp);
            }
        }
        return l;
    }


// ... Inner Class ...

    /**
     * Inner class that periodically reports to the grand supervisor (Medulla)
     */
    private class Report2Sherif extends Thread {
        public boolean active = true;

        String sherif_name = myName.substring(0,myName.lastIndexOf("_dp_"));
        AID sherif_aid = new AID(sherif_name, AID.ISLOCALNAME);

        public void run(){
            while(active){
                informStatus(sherif_aid);
                informData(sherif_aid);
                try {
                    Thread.sleep(REPINTERVAL);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }
}