/*---------------------------------------------------------------------------*
*  Copyright (c) 2008        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.agents;

import org.jlab.coda.cMsg.*;
import org.jlab.coda.runcontrol.ontologies.CodaPlatform.CodaPlatformAdminOntology;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.*;
import org.jlab.coda.runcontrol.util.rcConstants;
import org.jlab.coda.runcontrol.util.rcSubscription;
import org.snmp4j.CommunityTarget;
import org.snmp4j.PDU;
import org.snmp4j.Snmp;
import org.snmp4j.TransportMapping;
import org.snmp4j.event.ResponseEvent;
import org.snmp4j.event.ResponseListener;
import org.snmp4j.mp.MPv3;
import org.snmp4j.mp.SnmpConstants;
import org.snmp4j.security.SecurityModels;
import org.snmp4j.security.SecurityProtocols;
import org.snmp4j.security.USM;
import org.snmp4j.smi.*;
import org.snmp4j.transport.DefaultUdpTransportMapping;
import rca.content.lang.Codec;
import rca.content.lang.sl.SLCodec;
import rca.content.onto.OntologyException;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.core.AID;
import rca.domain.DFService;
import rca.domain.FIPAAgentManagement.DFAgentDescription;
import rca.domain.FIPAAgentManagement.ServiceDescription;
import rca.domain.FIPAException;
import rca.domain.FIPANames;
import rca.lang.acl.ACLMessage;
import rca.util.leap.List;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.StringTokenizer;

/**
 * Class will incapsulate monitoring details of the Wienere VME crate.
 * SNMP is used for communication. Followind OID's will be monitored
 * periodically. SNMP node IP address and the monitoring period
 * will be configured at the differentiation stage of the AFECS
 * control system.
 * N.B. This agent has fixed/predefined states, due to the fact that
 *      SNMP control of the wiener crate is not supported.
 *      The states are: on, off, error, warning.
 * N.B. 1 means on/ok, 0 means off/failed
 *
 * @author: gurjyan
 * @date: Mar 12, 2008
 * @time: 3:46:29 PM
 */
@SuppressWarnings({"deprecation"})
public class WienerCrateAgent extends CAgent {

    /** List containing the names of the SMS agents*/
    private java.util.List<String> _mySMS = new ArrayList<String>();

    /** Defines the content language, used for inter-agent communications*/
    private Codec  codec = new SLCodec();

    /** Internal field showing the agents registration status with the DF of the AFECS platform.*/
    private boolean       _RegisteredFlag    = false;

    /** Crate monitoring thread*/
    CrateMonitor _crateMonitoringThread;

    /** The type of this agent is wnc (wiener crate).*/
    private String _agentType;

    /** The name of the agent, for e.g. crate1, crate2, etc.*/
    private String        _agentName;

    /** Session of the agent. The default this agent is session agnostic/independent*/
    private String        _agentSession = "Universal";

    /** Runtype of the agent. */
    private String        _agentRuntype;

    /** Priority of the agent. */
    private int        _agentPriority;

    /** Wiener crate status*/
    private volatile String        _agentState = rcConstants.WienerUnknown;

    /** cMsg domain server (Hepatic) connection object */
    private cMsg          cMsgSrv;

    /** Subdomain (where this agent belongs to) supervisor name*/
    private AID mySupervisor;

    /** reporting interval set in cool*/
    private int myReportingInterval;

    /** IP address of the representing wiener crate*/
    private String myIPAddress;

    /** on/off status of wiener crate*/
    private volatile int crateIsOn;

    /** channel names*/
    private volatile String wn0="";
    private volatile String wn1="";
    private volatile String wn2="";
    private volatile String wn3="";
    private volatile String wn4="";

    /** channel set voltage*/
    private volatile float wsv0;
    private volatile float wsv1;
    private volatile float wsv2;
    private volatile float wsv3;
    private volatile float wsv4;

    /** channel output voltage*/
    private volatile float wv0;
    private volatile float wv1;
    private volatile float wv2;
    private volatile float wv3;
    private volatile float wv4;

    /** channel output current*/
    private volatile float wi0;
    private volatile float wi1;
    private volatile float wi2;
    private volatile float wi3;
    private volatile float wi4;

    /** power module temperature status*/
    private volatile int wt0;
    private volatile int wt1;
    private volatile int wt2;
    private volatile int wt3;
    private volatile int wt4;

    /** Channel on status*/
    private volatile int wcs0;
    private volatile int wcs1;
    private volatile int wcs2;
    private volatile int wcs3;
    private volatile int wcs4;

    /** time in sec the fan tray was on*/
    private volatile int activeSec;

    /** crate fan speed*/
    private volatile int ftspeed;

    /** Currently supported OID's*/
    /** Crate status 0-off 1-on*/
    private static String crateStatus ="1.3.6.1.4.1.19947.1.1.1.0";

    /** The short name of the Channel 1  */
    private static String nameChannel1 ="1.3.6.1.4.1.19947.1.3.2.1.2.1";
    /** The short name of the Channel 2  */
    private static String nameChannel2 ="1.3.6.1.4.1.19947.1.3.2.1.2.2";
    /** The short name of the Channel 3  */
    private static String nameChannel3 ="1.3.6.1.4.1.19947.1.3.2.1.2.3";
    /** The short name of the Channel 4  */
    private static String nameChannel4 ="1.3.6.1.4.1.19947.1.3.2.1.2.4";
    /** The short name of the Channel 5  */
    private static String nameChannel5 ="1.3.6.1.4.1.19947.1.3.2.1.2.6";

    /** Units in V*/
    /** The measured voltage of the Channel 1  */
    private static String voltageChannel1 ="1.3.6.1.4.1.19947.1.3.2.1.5.1";
    /** The measured voltage of the Channel 2  */
    private static String voltageChannel2 ="1.3.6.1.4.1.19947.1.3.2.1.5.2";
    /** The measured voltage of the Channel 3  */
    private static String voltageChannel3 ="1.3.6.1.4.1.19947.1.3.2.1.5.3";
    /** The measured voltage of the Channel 4  */
    private static String voltageChannel4 ="1.3.6.1.4.1.19947.1.3.2.1.5.4";
    /** The measured voltage of the Channel 5  */
    private static String voltageChannel5 ="1.3.6.1.4.1.19947.1.3.2.1.5.6";

    /** Units in A*/
    /** The measured current of the Channel 1  */
    private static String currentChannel1 ="1.3.6.1.4.1.19947.1.3.2.1.7.1";
    /** The measured current of the Channel 2  */
    private static String currentChannel2 ="1.3.6.1.4.1.19947.1.3.2.1.7.2";
    /** The measured current of the Channel 3  */
    private static String currentChannel3 ="1.3.6.1.4.1.19947.1.3.2.1.7.3";
    /** The measured current of the Channel 4  */
    private static String currentChannel4 ="1.3.6.1.4.1.19947.1.3.2.1.7.4";
    /** The measured current of the Channel 5  */
    private static String currentChannel5 ="1.3.6.1.4.1.19947.1.3.2.1.7.6";

    /** Ok(-128) failure(127)*/
    /** The measured temperature of the power module 1  */
    private static String temperatureChannel1 ="1.3.6.1.4.1.19947.1.3.2.1.8.1";
    /** The measured temperature of the power module 2  */
    private static String temperatureChannel2 ="1.3.6.1.4.1.19947.1.3.2.1.8.2";
    /** The measured temperature of the power module 3  */
    private static String temperatureChannel3 ="1.3.6.1.4.1.19947.1.3.2.1.8.3";
    /** The measured temperature of the power module 4  */
    private static String temperatureChannel4 ="1.3.6.1.4.1.19947.1.3.2.1.8.4";
    /** The measured temperature of the power module 5  */
    private static String temperatureChannel5 ="1.3.6.1.4.1.19947.1.3.2.1.8.6";

    /** Switch on (1) off(0) channels*/
    /** Channel 1  */
    private static String controlChannel1 ="1.3.6.1.4.1.19947.1.3.2.1.9.1";
    /** Channel 2  */
    private static String controlChannel2 ="1.3.6.1.4.1.19947.1.3.2.1.9.2";
    /** Channel 3  */
    private static String controlChannel3 ="1.3.6.1.4.1.19947.1.3.2.1.9.3";
    /** Channel 4  */
    private static String controlChannel4 ="1.3.6.1.4.1.19947.1.3.2.1.9.4";
    /** Channel 5  */
    private static String controlChannel5 ="1.3.6.1.4.1.19947.1.3.2.1.9.6";

    /** The time in sec. for how long the fan tray was on*/
    private static volatile String timeFanTrayOn ="1.3.6.1.4.1.19947.1.7.3.0";

    /** The nominal fan rotation speed*/
    private static volatile String fanRotationSpeed ="1.3.6.1.4.1.19947.1.7.6.0";

    /** SNMP response listener object*/
    private SNMPResponseListener listener = new SNMPResponseListener();



    /**
     * Setup method of the agent. It organizes the
     * ontology registration, and agent behaviour scheduling.
     */
    protected void setup(){

// This is the given name when the agent was started by the Hepatic.
        _agentName = getLocalName();

// Register supported languages
        getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);

// Register the AFECS platform administration ontology for Hepatic conversations
        getContentManager().registerOntology(CodaPlatformAdminOntology.getInstance());

// Register the coda run control ontology for Cortex and Medulla conversations
        getContentManager().registerOntology(CodaRunControlOntology.getInstance());

// Add the responder behaviour.
// This is the behaviour where we process request information from the other agents in the platform.
// The actual processing method is defined in this class but called from the GenResponderBehaviour object
        addBehaviour(new GenRespondBehaviour(this, 0));

// DF registration
        if(!_RegisteredFlag){
            DFAgentDescription dfd = new DFAgentDescription();
            dfd.setName(getAID());
            ServiceDescription sd = new ServiceDescription();
            sd.setType("WienerCrate");
            sd.setName(_agentName);
            sd.setOwnership("JLABCODA");
            dfd.addServices(sd);
            try {
                DFService.register(this,dfd);
                _RegisteredFlag = true;
            } catch (FIPAException e) {
                System.err.println(getLocalName()+" registration with DF not succeeded. Reason: "+e.getMessage());
                _RegisteredFlag = false;
            }
        }

    }

    /**
     * This method is called at the agent shutdown
     */
    protected void takeDown()
    {
        try { DFService.deregister(this); }
        catch (Exception e) {e.printStackTrace();}
    }


    /**
     * Method connects to the cMsg Server started by the Hepatic. This will allow this agent
     * to respond directly to the cMsg requests from outside of the platform.
     */
    protected void connectTocMsg(){

        // Connect to the cMsg server
        try {
            if((cMsgSrv!=null) && (cMsgSrv.isConnected())) {
                // We are connected do nothing
            }

            // Start a new connection
            else {

                // Disconnect first if there is cMsgSrv object
                if(cMsgSrv!=null) cMsgSrv.disconnect();
                cMsgSrv = new cMsg(rcSubscription.PlatformcMsgServerUDL, _agentName,
                        rcSubscription.PlatformcMsgServerDescription);
                cMsgSrv.connect();

                // Enable message reception
                cMsgSrv.start();
            }
        } catch (cMsgException ee) {
            try {
                cMsgSrv.disconnect();
                cMsgSrv = null;
            } catch (cMsgException e) {
                cMsgSrv = null;
                e.printStackTrace();
            }
            ee.printStackTrace();
        }
    }

    // TESTING ........
//    public WienerCrateAgent(){
//        _crateMonitoringThread = new CrateMonitor("129.57.29.202",3);
//        _crateMonitoringThread.start();

//    }

//    public static void main(String args[]){
//        WienerCrateAgent myagent = new WienerCrateAgent();
//    }
// END OF THESTING  .............

    /**
     *  Sends a SNMP requests to all interested OID's of the wienere crate,
     *  using asynchronous SNMPv1 request.
     * @param oid  of the SNMP channel
     */
    public void SNMPSetInt(String ip, String oid, int val) {
        /** SNMP communication object*/
        Snmp snmp = null;

        /** SNMP address of the wienere crate*/
        Address targetAddress;

        ResponseEvent response = null;

        targetAddress = GenericAddress.parse("udp:"+ ip +"/161");

        TransportMapping transport;
        try {
            transport = new DefaultUdpTransportMapping();
            snmp = new Snmp(transport);
            USM usm = new USM(SecurityProtocols.getInstance(),
                    new OctetString(MPv3.createLocalEngineID()), 0);
            SecurityModels.getInstance().addSecurityModel(usm);
            transport.listen();
        } catch (IOException e) {
            e.printStackTrace();
        }

        // Setting up target
        CommunityTarget target = new CommunityTarget();
        target.setCommunity(new OctetString("guru"));
        target.setAddress(targetAddress);
        target.setRetries(3);
        target.setTimeout(1500);
        target.setVersion(SnmpConstants.version2c);

        // Creating PDU
        PDU pdu = new PDU();

        // Adding variable
        VariableBinding vb = new VariableBinding(new OID(oid));
//        Variable variable = new Integer32(Integer.parseInt(val));
        Variable variable = new Integer32(val);
        vb.setVariable(variable);

        pdu.add(vb);
        pdu.setType(PDU.SET);
        try {
            if (snmp!=null) response = snmp.send(pdu, target);
        } catch (IOException e) {
            e.printStackTrace();
        }
        try {
            System.out.println(response.getResponse().getErrorStatusText());
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * This method gets called whenever the ACL message (inter-agent communication) is reached to this agent.
     * This is the implementatioin of the abstract method defined in the <clas>CAgent</clas>, which is
     * going to be called by the <clas>BaseicResponderBehaviour</clas> object to process the request to
     * this agent. Most of the ACL message request are coming from the supervisor agent
     * <clas>Medulla</clas> and/or <clas>Cortex</clas>.
     *
     * @param request      ACLMessage request object, passed from the GenResponderBehaviour object.
     * @return             ACLMessage object
     */
    protected ACLMessage processRequest(ACLMessage request, Action a) {

        // Reference to the CoolComponent object.
        CoolComponent component;

        // Reference to the CoolOption object. Not used for the moment. May be used for passing MIB's
        CoolOption option;
        String tmpStr = null;

        // Create an agent communication reply object
        ACLMessage reply = request.createReply();
        try {

            // This the messages are usually coming from the Cortex.
            if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.Differentiate")) {
                // Get the differentiate ontology object
                Differentiate myDifferentiate = (Differentiate)a.getAction();

                // Intialization of the variables
//                option          = myDifferentiate.getDifferentiate_Option();
                component       = myDifferentiate.getDifferentiate_Component();
                differentiate(component);
            }
            // This the messages are coming from the Supervisor agent.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.CoolActionMoveTo")) {
                // Get the ontology object
                CoolActionMoveTo move2 = (CoolActionMoveTo)a.getAction();
                CoolState st = move2.getActionMoveto_State();
                move2State(st);
            }
// ... ResultNotify.
// This request tells this agent to report its state not only to Medulla but also tho the SMS agent.
            else if (a.getAction().getClass().getName().equalsIgnoreCase("org.jlab.coda.runcontrol.ontologies."+
                    "CodaRunControl.ResultNotify")) {
                ResultNotify rnt = (ResultNotify)a.getAction();
                if(rnt.getResultNotify_Component()!=null && rnt.getResultNotify_Component().getComponent_Name()!=null)
                    _mySMS.add(rnt.getResultNotify_Component().getComponent_Name());
            }


            // Create and send done action message (ACL message)  to the client agent.
            Done d = new Done();
            d.setAction(a);
            getContentManager().fillContent(reply, d);
        } catch (Codec.CodecException e) {
            e.printStackTrace();
        } catch (OntologyException e) {
            e.printStackTrace();
        }
        reply.setPerformative(ACLMessage.INFORM);

        // tmpStr contains information for e.g. runtype , etc.
        reply.setContent(tmpStr);
        return reply;
    }

    /**
     * This is the implementatioin of the abstract method defined in the CAgent class.
     * This method is going to be called by the BaseicRequestBehaviour object to process
     * the iform messages coming back to this agent as a result of initiating the request
     * to some agent on the platfrom.
     *
     * @param msg      ACLMessage object passed from the GenRequestBehaviour object.
     */
    protected void processInform(ACLMessage msg) {
    }


    /**
     * Configures this agent, based on the deffinitions in the cool file.
     * @param component Ontology object describing this agent
     */
    private void differentiate(CoolComponent component){

        // get the agent name defined in the cool file
        _agentName      = component.getComponent_Name();

        // get agent type
        _agentType = component.getComponent_Type();

        // set the session for this agent. Most of the time wnc type agents are session agnostic.
        if(component.getComponent_Session()!=null){
            _agentSession  = component.getComponent_Session();
        }
        if(component.getComponent_Runtype()!=null){
            _agentRuntype  = component.getComponent_Runtype();
        }

        if(component.getComponent_Priority()!=null){
            _agentPriority  = component.getComponent_Priority();
        }

        // Report to the wiener crate domain supervisor defined by the cool,
        // through setComponent_Reoprtstocomponent ontology action
        if(component.getComponent_Reportsto()!=null){
            String supervisorName = _agentSession+"_"+component.getComponent_Reportsto();
            mySupervisor = new AID(supervisorName,AID.ISLOCALNAME);
            _mySMS.add(supervisorName);
        }

        // get reporting interval (reporting periodicity)
        try{
            if(component.getComponent_Reportinginterval()!=null){
                myReportingInterval = Integer.parseInt(component.getComponent_Reportinginterval());
            }
        } catch(NumberFormatException e){ e.printStackTrace();}

        // get the rerpesenting winner crate IP address
        if(component.getComponent_Representsipaddress()!=null){
            myIPAddress = component.getComponent_Representsipaddress();
        }


// DF registration
        if(!_RegisteredFlag){
            DFAgentDescription dfd = new DFAgentDescription();
            dfd.setName(getAID());
            ServiceDescription sd = new ServiceDescription();
            sd.setType("WienerCrate");
            sd.setName(_agentName);
            sd.setOwnership("JLABCODA");
            dfd.addServices(sd);
            try {
                DFService.register(this,dfd);
                _RegisteredFlag = true;
            } catch (FIPAException e) {
                System.err.println(getLocalName()+" registration with DF not succeeded. Reason: "+e.getMessage());
                _RegisteredFlag = false;
            }
        }

//        System.out.println(_agentName+" ............... in the differentiate");
        // Connect to the platform cMsg server
        connectTocMsg();

        // Subscribe Control messages directly from the Wiener GUI.
// ... run/control
        cMsgCallbackInterface wncontrol = new WncControlCallback();
        try {
            cMsgSrv.subscribe(_agentName, rcSubscription.WNCControlType, wncontrol, null);
        } catch (cMsgException e) {
            e.printStackTrace();
        }

        // start monitoring thread
        if(_crateMonitoringThread!=null){
            _crateMonitoringThread.shouldrun=false;
        }
        // Start the monitoring thread
        // The first parameter is the phisical component (actual wiener crate) IP address.
        if(myReportingInterval!=0){
            _crateMonitoringThread = new CrateMonitor(myIPAddress, myReportingInterval);
            _crateMonitoringThread.start();
        }
    }

    /**
     * Transitions to the state
     * Only default (non coll described) agent states are supported (i.e. active and pasive).
     * N.B. active basic state is overiden by the supervisor agent, and added process details.
     * @param s CoolState object including process description necessary to achieve to requested state
     */
    private void move2State(CoolState s){
        CoolState tmpst;
        String tmpstname = s.getState_Name();
        List tmpprs = s.getState_Achievedthrough();
        CoolProcess tmppr;

        if(tmpprs!=null) tmppr = (CoolProcess)tmpprs.get(0);

        if(tmpstname.equalsIgnoreCase(rcConstants.AgentStatePasive)){
            _crateMonitoringThread.shouldrun = false;
            SNMPSetInt(myIPAddress,crateStatus,0);
        } else if(tmpstname.equalsIgnoreCase(rcConstants.AgentStateActive)){

            if(_crateMonitoringThread==null){
                // Start the monitoring thread
                // The first parameter is the phisical component (actual wiener crate) IP address.
                _crateMonitoringThread = new CrateMonitor(myIPAddress, myReportingInterval);
                _crateMonitoringThread.start();
            }
            SNMPSetInt(myIPAddress,crateStatus,1);
        }
    }

    /**
     * Informs status to receiver (Medulla most of the time).
     * This method sends to the receiver an <clas>AcceptStatus</clas> coda ontology ACL message,
     * describing the producer, i.e. the process that produces information and timestamp.
     *
     * @param status               The state of the component
     * @param severity             The severity of the message
     * @param receiver             The receiver of the message
     */
    public void informStatus(String status, String severity, AID receiver) {

// Create the acl message object
        ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
        try{
            getContentManager().lookupOntology(CodaRunControlOntology.NAME);

// Preparing the message
            msg.addReceiver(receiver);
            msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            msg.setOntology(CodaRunControlOntology.NAME);
            msg.addUserDefinedParameter(_agentName,receiver.getLocalName());

// Create a codacomponent ontology concept object
            CoolComponent cm = new CoolComponent();
            cm.setComponent_Name(_agentName);
            cm.setComponent_Type(_agentType);
            cm.setComponent_Session(_agentSession);

// Create a Producer ontology concept object
            Producer pr = new Producer();
            pr.setProducer_Component(cm);

// Create a coda timestamp object
            CodaTimeStamp tm = new CodaTimeStamp();
            tm.setTimestamp_Value(new Date());

// Create a acceptStatus ontology action object
            AcceptStatus myaccept = new AcceptStatus();
            myaccept.setAcceptstatus_Producer(pr);
            myaccept.setAcceptstatus_Timestamp(tm);
            myaccept.setAcceptstatus_Status(status);
            if(severity!=null)myaccept.setAcceptstatus_Severity(severity);

// Create an action transport object
            Action a = new Action();
            a.setActor(receiver);
            a.setAction(myaccept);

// Fill the message content
            getContentManager().fillContent(msg, a);
        } catch(Exception e) { e.printStackTrace(); }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
        addBehaviour(new GenRequestBehaviour(this, msg, 0));
    }

    /**
     * Informs state to receiver (In this case this is the SMS agent).
     * This method sends to the receiver a <clas>AcceptState</clas> coda ontology ACL message, describing the producer
     * i.e. the components that produces information and timestamp.
     *
     * @param stateName               The state of the component
     * @param severity                of the component
     */
    public void informStateSMS(String stateName, String severity) {
        String tmps;
        AID tmpaid;
        if(_mySMS!=null && !_mySMS.isEmpty()){
            for (String a_mySMS : _mySMS) {
                tmps = a_mySMS;

// Create the acl message object
                ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
                try {
                    getContentManager().lookupOntology(CodaRunControlOntology.NAME);
                    tmpaid = new AID(tmps, AID.ISLOCALNAME);

// Preparing the message
                    msg.addReceiver(tmpaid);
                    msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                    msg.setOntology(CodaRunControlOntology.NAME);
                    msg.addUserDefinedParameter(_agentName, tmpaid.getLocalName());

// Create a codacomponent ontology concept object
                    CoolComponent cm = new CoolComponent();
                    cm.setComponent_Name(_agentName);
                    cm.setComponent_Type(_agentType);
                    cm.setComponent_Priority(_agentPriority);
                    cm.setComponent_Runtype(_agentRuntype);
                    cm.setComponent_Session(_agentSession);

// Create a Producer ontology concept object
                    Producer pr = new Producer();
                    pr.setProducer_Component(cm);

// Create a coda timestamp object
                    CodaTimeStamp tm = new CodaTimeStamp();
                    tm.setTimestamp_Value(new Date());

// Create coolState object
                    CoolState myState = new CoolState();
                    myState.setState_Name(stateName);
                    myState.setState_Isfor(cm);

// Create a acceptStatus ontology action object
                    AcceptState myaccept = new AcceptState();
                    myaccept.setAcceptstate_Producer(pr);
                    myaccept.setAcceptstate_Timestamp(tm);
                    myaccept.setAcceptstate_State(myState);
                    myaccept.setAcceptstate_Severity(severity);

// Create an action transport object
                    Action a = new Action();
                    a.setActor(tmpaid);
                    a.setAction(myaccept);

// Fill the message content
                    getContentManager().fillContent(msg, a);
                } catch (Exception e) {
                    e.printStackTrace();
                }

// Create and add behaviour to this agent behaviour queue, which will be scheduled for delivery.
                addBehaviour(new GenRequestBehaviour(this, msg,0));
            }
        }
    }

    public void knel(int msec){
        try {
            Thread.sleep(msec);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

    }
    /**
     * Sending dalog message
     * @param severity      severity of the message (int 1-16)
     * @param message       content of the message
     */
    private void dalogMsg(int severity, String message){
        String subject;
        String type = null;
        cMsgPayloadItem sever = null;

//        System.out.println("========== >>> "+_agentType);

        if(cMsgSrv!=null && cMsgSrv.isConnected() && rcSubscription.expid!=null && _agentType!=null){
            cMsgMessage msg = new cMsgMessage();
//            if(_agentSession.equals("Universal)")){
//                subject = rcSubscription.expid+"/"+_agentType+"/"+_agentName;
//            } else {
            subject = rcSubscription.expid+"/"+_agentSession+"/"+_agentRuntype+"/"+_agentType+"/"+_agentName;

//            }

            int[] iwcs = new int[5];
            iwcs[0]=wcs0;
            iwcs[1]=wcs0;
            iwcs[2]=wcs0;
            iwcs[3]=wcs0;
            iwcs[4]=wcs0;

            int[] iwt = new int[5];
            iwt[0]=wt0;
            iwt[1]=wt1;
            iwt[2]=wt2;
            iwt[3]=wt3;
            iwt[4]=wt4;

            float[] fwv = new float[5];
            fwv[0]=wv0;
            fwv[1]=wv1;
            fwv[2]=wv2;
            fwv[3]=wv3;
            fwv[4]=wv4;

            float[] fwi = new float[5];
            fwi[0]=wi0;
            fwi[1]=wi1;
            fwi[2]=wi2;
            fwi[3]=wi3;
            fwi[4]=wi4;

            String[]wn = new String[5];
            wn[0]=wn0;
            wn[1]=wn1;
            wn[2]=wn2;
            wn[3]=wn3;
            wn[4]=wn4;

            try {
                // payload items
                cMsgPayloadItem state = new cMsgPayloadItem("state",_agentState);
                cMsgPayloadItem wienerC = new cMsgPayloadItem("WienerCrate",crateIsOn);
                cMsgPayloadItem wienerS = new cMsgPayloadItem("WienerStatus",iwcs);
                cMsgPayloadItem wienerV = new cMsgPayloadItem("WienerVoltage",fwv);
                cMsgPayloadItem wienerI = new cMsgPayloadItem("WienerCurrent",fwi);
                cMsgPayloadItem wienerT = new cMsgPayloadItem("WienerTemperature",iwt);
                cMsgPayloadItem wienerF = new cMsgPayloadItem("WienerFanSpeed",ftspeed);
                cMsgPayloadItem wienerU = new cMsgPayloadItem("WienerUpTime",activeSec);
                cMsgPayloadItem wienerN = new cMsgPayloadItem("WienerModule",wn);
                cMsgPayloadItem p_expid = new cMsgPayloadItem("EXPID",rcSubscription.expid);
                cMsgPayloadItem p_session = new cMsgPayloadItem("session",_agentSession);
                cMsgPayloadItem p_config = new cMsgPayloadItem("config",_agentRuntype);
                cMsgPayloadItem p_type = new cMsgPayloadItem("codaClass",_agentType);
                cMsgPayloadItem p_codaName = new cMsgPayloadItem("codaName",_agentName);


//                System.out.println(" crate status    = "+crateIsOn);
//                System.out.println(" output voltages = "+wv[0]+" "+wv[1]+" "+wv[2]+" "+wv[3]+" "+wv[4]+ " [in V]");
//                System.out.println(" output currents = "+wi[0]+" "+wi[1]+" "+wi[2]+" "+wi[3]+" "+wi[4] +" [in A]");
//                System.out.println(" temperatures    = "+wt[0]+" "+wt[1]+" "+wt[2]+" "+wt[3]+" "+wt[4]);
//                System.out.println(" channel status  = "+wcs[0]+" "+wcs[1]+" "+wcs[2]+" "+wcs[3]+" "+wcs[4]);
//                System.out.println(" fan speed       = "+ftspeed+" rpm");
//                System.out.println(" uptime          = "+activeSec+" sec.");

                if(severity<5){
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","INFO."+severity);

                }
                else if(severity>=5 && severity<9) {
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","WARN."+severity);
                }
                else if(severity>=9 && severity<13) {
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","ERROR."+severity);
                }
                else if(severity>=13 && severity<17){
                    type = "dalog/"+severity;
                    sever = new cMsgPayloadItem("severity","ERRORS."+severity);
                }
                if(subject!=null && type !=null){

                    msg.setSubject(subject);
                    msg.setType(type);
                    msg.setText(message);
                    msg.addPayloadItem(sever);
                    msg.addPayloadItem(state);
                    msg.addPayloadItem(wienerC);
                    msg.addPayloadItem(wienerS);
                    msg.addPayloadItem(wienerV);
                    msg.addPayloadItem(wienerI);
                    msg.addPayloadItem(wienerT);
                    msg.addPayloadItem(wienerF);
                    msg.addPayloadItem(wienerU);
                    msg.addPayloadItem(wienerN);
                    msg.addPayloadItem(p_expid);
                    msg.addPayloadItem(p_session);
                    msg.addPayloadItem(p_config);
                    msg.addPayloadItem(p_type);
                    msg.addPayloadItem(p_codaName);
                    cMsgSrv.send(msg);
                    cMsgSrv.flush(1);
                }
            } catch (cMsgException e) {
                e.printStackTrace();
            }

        }

    }



    public class CrateMonitor extends Thread {
        /** SNMP communication object*/
        private  Snmp snmp;

        /** SNMP address of the wienere crate*/
        private Address targetAddress;

        /** Thread control variable */
        public volatile boolean shouldrun = true;

        /** Period of the monitoring in seconds*/
        private int period;


        /** Constructor
         *
         * @param ip   IP address of the wiener crate
         * @param p    period of the monitoring
         * */
        CrateMonitor(String ip, int p){

            period = p;
            targetAddress = GenericAddress.parse("udp:"+ ip +"/161");

            TransportMapping transport;
            try {
                transport = new DefaultUdpTransportMapping();
                snmp = new Snmp(transport);
                USM usm = new USM(SecurityProtocols.getInstance(),
                        new OctetString(MPv3.createLocalEngineID()), 0);
                SecurityModels.getInstance().addSecurityModel(usm);
                transport.listen();
            } catch (IOException e) {
                e.printStackTrace();
            }
            System.out.println(_agentName+"  running ... "+shouldrun);
        }


        /**
         * Run method of the Thread
         */
        public void run(){
            while(shouldrun){

// define the state of this agent and report
                defineStateAndsend();

// ask SNMP
                SNMPSend(crateStatus);
                knel(100);
                SNMPSend(nameChannel1);
                knel(100);
                SNMPSend(nameChannel2);
                knel(100);
                SNMPSend(nameChannel3);
                knel(100);
                SNMPSend(nameChannel4);
                knel(100);
                SNMPSend(nameChannel5);
                knel(100);

                SNMPSend(voltageChannel1);
                knel(100);
                SNMPSend(voltageChannel2);
                knel(100);
                SNMPSend(voltageChannel3);
                knel(100);
                SNMPSend(voltageChannel4);
                knel(100);
                SNMPSend(voltageChannel5);
                knel(100);

                SNMPSend(currentChannel1);
                knel(100);
                SNMPSend(currentChannel2);
                knel(100);
                SNMPSend(currentChannel3);
                knel(100);
                SNMPSend(currentChannel4);
                knel(100);
                SNMPSend(currentChannel5);
                knel(100);

                SNMPSend(temperatureChannel1);
                knel(100);
                SNMPSend(temperatureChannel2);
                knel(100);
                SNMPSend(temperatureChannel3);
                knel(100);
                SNMPSend(temperatureChannel4);
                knel(100);
                SNMPSend(temperatureChannel5);
                knel(100);

                SNMPSend(controlChannel1);
                knel(100);
                SNMPSend(controlChannel2);
                knel(100);
                SNMPSend(controlChannel3);
                knel(100);
                SNMPSend(controlChannel4);
                knel(100);
                SNMPSend(controlChannel5);
                knel(100);

                SNMPSend(timeFanTrayOn);
                knel(100);

                SNMPSend(fanRotationSpeed);
                knel(100);

                try {
                    Thread.sleep(period*1000);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

            }
        }

        public synchronized void defineStateAndsend(){

//            System.out.println("Start defining a state ");



            if(crateIsOn==1){

// Check channel statuses
                if((!wn0.equalsIgnoreCase("noSuchInstance") && wcs0==0) ||
                        (!wn1.equalsIgnoreCase("noSuchInstance") && wcs1==0) ||
                        (!wn2.equalsIgnoreCase("noSuchInstance") && wcs2==0) ||
                        (!wn3.equalsIgnoreCase("noSuchInstance") && wcs3==0) ){
                    _agentState = rcConstants.WienerError;
//                        informStatus(_agentState,"Error",mySupervisor);
                    informStateSMS(_agentState, "Error");
                    dalogMsg(13,"Power module channel status problem");

// Check channel temperature
                } else  if((!wn0.equalsIgnoreCase("noSuchInstance") && wt0==0) ||
                        (!wn1.equalsIgnoreCase("noSuchInstance") && wt1==0) ||
                        (!wn2.equalsIgnoreCase("noSuchInstance") && wt2==0) ||
                        (!wn4.equalsIgnoreCase("noSuchInstance") && wt3==0) ){
                    _agentState = rcConstants.WienerWarning;
//                        informStatus(_agentState,"Warning",mySupervisor);
                    informStateSMS(_agentState, "Warning");
                    dalogMsg(7,"Power module temperature problem");

// Check output voltages
                } else  if((!wn0.equalsIgnoreCase("noSuchInstance") && wcs0 ==0 && wv0<1) ||
                        (!wn1.equalsIgnoreCase("noSuchInstance") && wcs1 ==0 && wv1<1) ||
                        (!wn2.equalsIgnoreCase("noSuchInstance") && wcs2 ==0 && wv2<1) ||
                        (!wn4.equalsIgnoreCase("noSuchInstance") && wcs3 ==0 && wv3<1) ){
                    _agentState = rcConstants.WienerError;
//                        informStatus(_agentState,"Warning",mySupervisor);
                    informStateSMS(_agentState, "Error");
                    dalogMsg(10,"Output voltage is out of spec.");
// Crate is OK
                }else {
                    _agentState = rcConstants.WienerOn;
//                        informStatus(_agentState,"Info",mySupervisor);
                    informStateSMS(_agentState, "Info");
                    dalogMsg(1,_agentName+" wiener crate is on");
                }

// Crate is Off
            } else {
                _agentState = rcConstants.WienerOff;
//                    informStatus(_agentState,"Info",mySupervisor);
                informStateSMS(_agentState, "Warning");
//                System.out.println(" HEYYYY .... warning crate is OFAFA");
                dalogMsg(6,_agentName+" wiener crate is off");
            }

//            System.out.println("WC-DEBUG-Names  : "+wn0+"  "+wn1+"  "+wn2+"  "+wn3+"  "+wn4);
//            System.out.println("WC-DEBUG-CStatus: "+wcs0+"  "+wcs1+"  "+wcs2+"  "+wcs3+"  "+wcs4);
//            System.out.println("WC-DEBUG-V      : "+wv0+"  "+wv1+"  "+wv2+"  "+wv3+"  "+wv4);
//            System.out.println("WC-DEBUG-I      : "+wi0+"  "+wi1+"  "+wi2+"  "+wi3+"  "+wi4);
//            System.out.println("WC-DEBUG-T      : "+wt0+"  "+wt1+"  "+wt2+"  "+wt3+"  "+wt4);

        }

        /**
         *  Sends a SNMP requests to all interested OID's of the wienere crate,
         *  using asynchronous SNMPv1 request.
         * @param oid  of the SNMP channel
         */
        public void SNMPSend(String oid) {

            // Setting up target
            CommunityTarget target = new CommunityTarget();
            target.setCommunity(new OctetString("public"));
            target.setAddress(targetAddress);
            target.setRetries(3);
            target.setTimeout(1500);
            target.setVersion(SnmpConstants.version2c);

            // Creating PDU
            PDU pdu = new PDU();

            // Adding odu's
            pdu.add(new VariableBinding(new OID(oid)));
            pdu.setType(PDU.GET);
            try {
                snmp.sendPDU(pdu, target, null, listener);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }


    /**
     * SNMP response listener class.
     * After getting messages from a wiener vme crate it checks to see
     * which OID channel is the sender.
     * Based on the recieved information it defines the state of this agent.
     * Then it buields CoolData ontology object, including state information,
     * and sends the result to the supervisor agent.
     */
    public class SNMPResponseListener implements ResponseListener {
        PDU respdu;


        public synchronized void onResponse(ResponseEvent event) {

//            System.out.println(".... Got SNMP response");

            // Always cancel async request when response has been received
            // otherwise a memory leak is created! Not canceling a request
            // immediately can be useful when sending a request to a broadcast
            // address.
            ((Snmp)event.getSource()).cancel(event.getRequest(), this);
            respdu = event.getResponse();

            if(respdu!=null){

                // Get responding oid
                String roid = respdu.get(0).getOid().toString();

                //Checking if the crate is on or off
                if(roid.equals(crateStatus)) {
                    crateIsOn = respdu.get(0).getVariable().toInt();
                }

                // Checking name of the channels
                else if(roid.equals(nameChannel1)){
                    String value = respdu.get(0).getVariable().toString();
                    wn0=value;
                    if(!value.equalsIgnoreCase("noSuchInstance")){
                        wsv0=decodeValueName(value);
                    }
                }
                else if(roid.equals(nameChannel2)){
                    String value = respdu.get(0).getVariable().toString();
                    wn1=value;
                    if(!value.equalsIgnoreCase("noSuchInstance")){
                        wsv1=decodeValueName(value);
                    }
                }
                else if(roid.equals(nameChannel3)){
                    String value = respdu.get(0).getVariable().toString();
                    wn2=value;
                    if(!value.equalsIgnoreCase("noSuchInstance")){
                        wsv2=decodeValueName(value);
                    }
                }
                else if(roid.equals(nameChannel4)){
                    String value = respdu.get(0).getVariable().toString();
                    wn3=value;
                    if(!value.equalsIgnoreCase("noSuchInstance")){
                        wsv3=decodeValueName(value);
                    }
                }
                else if(roid.equals(nameChannel5)){
                    String value = respdu.get(0).getVariable().toString();
                    wn4=value;
                    if(!value.equalsIgnoreCase("noSuchInstance")){
                        wsv4=decodeValueName(value);
                    }
                }

                // Checking output voltages
                else if(roid.equals(voltageChannel1)){
                    String value = respdu.get(0).getVariable().toString();
                    wv0=decodeSnmpFloat(value);
                }
                else if(roid.equals(voltageChannel2)){
                    String value = respdu.get(0).getVariable().toString();
                    wv1=decodeSnmpFloat(value);
                }
                else if(roid.equals(voltageChannel3)){
                    String value = respdu.get(0).getVariable().toString();
                    wv2=decodeSnmpFloat(value);
                }
                else if(roid.equals(voltageChannel4)){
                    String value = respdu.get(0).getVariable().toString();
                    wv3=decodeSnmpFloat(value);
                }
                else if(roid.equals(voltageChannel5)){
                    String value = respdu.get(0).getVariable().toString();
                    wv4=decodeSnmpFloat(value);
                }

                // Checking output currents
                else if(roid.equals(currentChannel1)){
                    String value = respdu.get(0).getVariable().toString();
                    wi0=decodeSnmpFloat(value);
                }
                else if(roid.equals(currentChannel2)){
                    String value = respdu.get(0).getVariable().toString();
                    wi1=decodeSnmpFloat(value);
                }
                else if(roid.equals(currentChannel3)){
                    String value = respdu.get(0).getVariable().toString();
                    wi2=decodeSnmpFloat(value);
                }
                else if(roid.equals(currentChannel4)){
                    String value = respdu.get(0).getVariable().toString();
                    wi3=decodeSnmpFloat(value);
                }
                else if(roid.equals(currentChannel5)){
                    String value = respdu.get(0).getVariable().toString();
                    wi4=decodeSnmpFloat(value);
                }

                //Checking power module temperature status
                else if(roid.equals(temperatureChannel1)) {
                    int v = respdu.get(0).getVariable().toInt();
                    if (v==-128)wt0=1;
                    else if(v==127)wt0=0;
                }
                else if(roid.equals(temperatureChannel2)) {
                    int v = respdu.get(0).getVariable().toInt();
                    if (v==-128)wt1=1;
                    else if(v==127)wt1=0;
                }
                else if(roid.equals(temperatureChannel3)) {
                    int v = respdu.get(0).getVariable().toInt();
                    if (v==-128)wt2=1;
                    else if(v==127)wt2=0;
                }
                else if(roid.equals(temperatureChannel4)) {
                    int v = respdu.get(0).getVariable().toInt();
                    if (v==-128)wt3=1;
                    else if(v==127)wt3=0;
                }
                else if(roid.equals(temperatureChannel5)) {
                    int v = respdu.get(0).getVariable().toInt();
                    if (v==-128)wt4=1;
                    else if(v==127)wt4=0;
                }

                // Channel status
                else if(roid.equals(controlChannel1)) {
                    wcs0=respdu.get(0).getVariable().toInt();
                }
                else if(roid.equals(controlChannel2)) {
                    wcs1=respdu.get(0).getVariable().toInt();
                }
                else if(roid.equals(controlChannel3)) {
                    wcs2=respdu.get(0).getVariable().toInt();
                }
                else if(roid.equals(controlChannel4)) {
                    wcs3=respdu.get(0).getVariable().toInt();
                }
                else if(roid.equals(controlChannel5)) {
                    wcs4=respdu.get(0).getVariable().toInt();
                }

                // how long fan tray was running
                else if(roid.equals(timeFanTrayOn)) {
                    activeSec = respdu.get(0).getVariable().toInt();
                }

                // fan speed
                else if(roid.equals(fanRotationSpeed)) {
                    ftspeed = respdu.get(0).getVariable().toInt();
                }
            }


//            try {
//                Thread.sleep(myReportingInterval*1000);
//            } catch (InterruptedException e) {
//                e.printStackTrace();
//            }
        }



        /**
         * Converts SNMP opaque syntax (BER encoding) type into float number.
         *
         * @param value  SNMP BER opaque
         * @return  float number
         */
        private float decodeSnmpFloat(String value){
            StringBuffer hexstr=null;
            StringTokenizer  st = new StringTokenizer(value,":");
            if(st.countTokens()==7){
                hexstr=new StringBuffer();

                for(int i=0;i<=2;i++)st.nextToken();
                while(st.hasMoreTokens())hexstr.append(st.nextToken());
            }
            if(hexstr!=null){
                return hexToFloat(hexstr.toString());
            } else return 0.0f;

        }

        /**
         * parses name of the wiener powersuply channel name and gets the voltage value
         * @param name of the channel ( e.g. +3V3)
         * @return fv float value of the set voltage (e.g. 3.3)
         */
        public float decodeValueName(String name){
            float fv=0.0f;
            if(name.startsWith("+") || name.startsWith("-") ) {
                name = name.substring(1);
            }
            if(name.contains("V")){
                StringTokenizer st = new StringTokenizer(name,"V");
                if(st.countTokens()==1)name = st.nextToken();
                else if(st.countTokens()==2)name = st.nextToken()+"."+st.nextToken();
            }

            try{
                fv = Float.parseFloat(name);
            } catch(NumberFormatException e){e.printStackTrace();}
            return fv;
        }

        /**
         * Converts string to float value
         * @param str   hex string representing the value
         * @return  floatvalue
         */
        public float hexToFloat(String str){
            float floatVal;
            int decimalValue =(int)Long.parseLong(str,16);
            floatVal=Float.intBitsToFloat(decimalValue );
            return floatVal;

        }

    }

    /**
     * Callback class for handling messages of the type wnc/control*
     * coming from wiener GUI
     *
     * @author Vardan Gyurjyan
     *         Date: Oct 23, 2008
     *         Time: 10:00:30 PM
     */

    private class WncControlCallback extends cMsgCallbackAdapter{

        /**
         * @param msg                  cMsgMessage delivered to the callback
         * @param userObject           binary user object
         */
        public void callback(cMsgMessage msg, Object userObject) {
            String txt = msg.getText();

            if(txt!=null){
                if(txt.equalsIgnoreCase("on")){
                    SNMPSetInt(myIPAddress,crateStatus,1);
                }
                else if(txt.equalsIgnoreCase("off")){
                    SNMPSetInt(myIPAddress,crateStatus,0);
                }
            }
        }
    }


}
