/*----------------------------------------------------------------------------*
*  Copyright (c) 2002        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, 6A/602-13     *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5800             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package org.jlab.coda.runcontrol.util.coolparser;

import java.io.FileInputStream;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;

import org.jlab.coda.runcontrol.ontologies.CodaRunControl.*;
import org.jlab.coda.runcontrol.util.rcConstants;

import com.hp.hpl.jena.rdf.model.* ;
import com.hp.hpl.jena.rdql.* ;
import java.util.Date;
import java.util.StringTokenizer;

import rca.util.leap.List;
import rca.util.leap.ArrayList;
import rca.util.leap.Iterator;

//import rca.util.leap.List;

/**
 * Parser for the cool (control oriented ontology language) configuration file.
 *
 * @author vardan Gyurjyan
 */
public class CoolParser {

    /** COOL language deffinition (taxonomy) cool.rdfs file url*/
//    private static String COOL_CORE = "http://coda.jlab.org/COOL/cool#";
    private static String COOL_CORE = "http://COOLHOME/schema/cool#";

    /** Verbosity level defined in the rcConstants class. */
    private int verbose = rcConstants.debugInfo;

    /** Jena model*/
    private Model model;

    /** List of the included cool configuration file jena models*/
    private java.util.List<Model> includeModeles = new java.util.ArrayList<Model>();

    /** class variable keeps the name of the control system */
    public String COOL_CONTROL;

    /** class variable keeps the session name of the control system */
    public volatile String COOL_SESSION;

    /** Option object for main control, i.e. Thalamus/Medulla*/
    public CoolOption CodaOption = new CoolOption();
    public CoolOption EbOption = new CoolOption();
    public CoolOption ErOption = new CoolOption();
    public CoolOption RocOption = new CoolOption();
    public CoolOption SlcOption = new CoolOption();
    public CoolOption TsOption = new CoolOption();
    public CoolOption EtOption = new CoolOption();
    public CoolOption AnaOption = new CoolOption();

    /** class variable list to keep all configured components */
    public ArrayList CONTROL_COMPONENTS   = new ArrayList();

    /** keeps all subcontrol systems participating to this control */
    public ArrayList CONTROL_SUBSYSTEMS   = new ArrayList();

    /** keeps control system states*/
    public ArrayList CONTROL_STATES   = new ArrayList();

    /** keeps control system rules*/
    public ArrayList CONTROL_RULES   = new ArrayList();

    /** The name of the supervisor agent */
    public String SUPERVISOR;

    /** variable for splitmb. N.B> depricated for backwards compatibility */
    public int SPLITMB=2047;

    private String _coolhome = System.getenv("COOL_HOME");

    /** coolparser will log the results in this file*/
    private String logFileName = _coolhome+"/coolparser.log";
    private BufferedWriter out;


    /**
     * Constructor
     *
     * @param fileName   of the cool configuration file
     * @param streamline boolean variable indicating the mode of the parsing. if it is true then
     *                   constructor will start the chain of parsing of the entire file.
     * @param session    session of the run
     */
    public CoolParser(String session, String fileName, boolean streamline) {

        String fn = null;
        COOL_SESSION = session;

        StringTokenizer fnst = new StringTokenizer(fileName,"/");
        while(fnst.hasMoreTokens())fn=fnst.nextToken();

        assert fn != null;
        COOL_CONTROL = fn.substring(0,fn.lastIndexOf(".rdf"));

        try {
            out = new BufferedWriter(new FileWriter(logFileName));
        } catch (IOException e) {
            e.printStackTrace();
        }
        try {

// create the jena model
            model = ModelFactory.createDefaultModel();

// Call the read method of the model and pass the file input streem to it.
            model.read(new FileInputStream(fileName),"http://COOLHOME","RDF/XML") ;
            if(streamline){
                out.write(new Date().toString()+"\n");
                out.write("Parsing results for "+fileName+"\n");

                file2model();

                if(!includeModeles.isEmpty()){
                    // create union of the jena models
                    for(Model m:includeModeles){
                        model = model.union(m);
                    }
                }
// Start parsing by calling method to parse ControlProcess
                parseControl(model);
            }
            out.close();
        } catch (Exception ex)
        {
            System.out.println("Exception: "+ex) ;
            ex.printStackTrace(System.err) ;
        }
    }


    /**
     * Main method used for testing and debugging purposes.
     *
     * @param args   used to pass the cool configuration file name
     */
    public static void main(String[] args) {
        new CoolParser(args[0],args[1],true);
    }


    /**
     * Configuration can be modular ( consisting of multiple rdf files).
     * This method will parse the control and find a components.
     * The resource of the component contains the path to the rdf file/files including components definitions.
     * jena models will be created based on these files and will be stored in the global arraylist object.
     * Later this arraylist will be used to create a union model for final parsing.
     */
    public void file2model(){

        String subject = COOL_CORE+"Control";
        String predicate = "include";
        String tmp;

// empty includeModels list
        includeModeles.clear();

// Creating query string
        String tq = "SELECT ?x "+ "WHERE(<"+subject+
                ">, <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;
                tmp = x.toString().substring((x.toString().indexOf("COOLHOME"))+8);

                System.out.println("DB.... "+tmp);
// create the jena model
                Model m = ModelFactory.createDefaultModel();

// Call the read method of the model and pass the file input streem to it.
                m.read(new FileInputStream(_coolhome+tmp),"http://COOLHOME","RDF/XML") ;
                includeModeles.add(m);
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
    }

    /**
     * parseses the cool Control
     *
     * @param model  Jena Model
     */
    public void parseControl(Model model) {

// local arraylist holds the component objects from the ComponentGroup
        List groupComponents;

        try{
            if (verbose >= rcConstants.debugInfo){

                out.write("Debug - Starting to parse Control =>       "  +COOL_CONTROL +"\n") ;
                out.write("==================================================="  +"\n") ;
            }

            CONTROL_COMPONENTS = parseComponentList(COOL_CORE+"Control","hasComponent");
            CONTROL_STATES     = parseStateList(COOL_CORE+"Control","hasState");
            CONTROL_RULES      = parseRuleList(COOL_CORE+"Control","hasRule");
            CONTROL_SUBSYSTEMS = parseSubSystemList(COOL_CORE+"Control","hasSystem");

            if(CONTROL_RULES != null) dumprule(CONTROL_RULES);

// parse ComponentGroups
            groupComponents = parseComponentGroup(COOL_CORE+"Control","hasComponentGroup");

// add components in the groups into the CONTROL_COMPONENTS list
            Iterator it = groupComponents.iterator();
            while(it.hasNext())CONTROL_COMPONENTS.add((CoolComponent)it.next());

// Parse Options
            CodaOption = parseOption(COOL_CORE+"Control","hasCodaOption");
            if(CodaOption==null) out.write("Warning: no hasCodaOption is defined. \n");

            EbOption = parseOption(COOL_CORE+"Control","hasEbOption");
            if(EbOption==null) out.write("Warning: no hasEbOption is defined. \n");

            ErOption = parseOption(COOL_CORE+"Control","hasErOption");
            if(ErOption==null) out.write("Warning: no hasErOption is defined. \n");

            RocOption = parseOption(COOL_CORE+"Control","hasRocOption");
            if(RocOption==null) out.write("Warning: no hasRocOption is defined. \n");

            TsOption = parseOption(COOL_CORE+"Control","hasTsOption");
            if(TsOption==null) out.write("Warning: no hasTsOption is defined. \n");

            EtOption = parseOption(COOL_CORE+"Control","hasEtOption");
            if(EtOption==null) out.write("Warning: no hasEtOption is defined. \n");

            SlcOption = parseOption(COOL_CORE+"Control","hasSlcOption");
            if(SlcOption==null) out.write("Warning: no hasSlcOption is defined. \n");

            AnaOption = parseOption(COOL_CORE+"Control","hasAnaOption");
            if(AnaOption==null) out.write("Warning: no hasAnaOption is defined. \n");

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Control  =>       "  + COOL_CORE+"Control"+"\n") ;
                out.write("==================================================="  +"\n") ;
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

    }


    /**
     * Parsing the componentGroup of the control. Returns the list of the CoolComponent objects in the group.
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return cl               CoolComponent list
     */
    public ArrayList parseComponentGroup(Object subject, String predicate) {
        ArrayList cl= new ArrayList();
        String gnames, nameprefix = null, gnumbers = null, startnumst = null, endnumst = null;
        int startnum, endnum;
        CoolComponent comp, gcomp;
        StringTokenizer st;
// Creating query string
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;

                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Starting to parse ComponentGroup of the Control   =>       "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
// componentGroup names
                gnames = getValue(x,"hasCGName");
                if(gnames==null){
                    out.write("COOL-ERROR: ComponentGroup "+x+" does not have a names "+"\n");
                    System.out.println("COOL-ERROR: ComponentGroup "+x+" does not have a names");
                }

// hasSettingsOf parsing, i.e. common component settings
                comp = parseComponent(x,"hasSettingsOf");
                st = new StringTokenizer(gnames,"?");
                if(st.hasMoreTokens())nameprefix = st.nextToken();
                if(st.hasMoreTokens())gnumbers = st.nextToken();
                if(nameprefix!=null && gnumbers!=null){
                    st = new StringTokenizer(gnumbers,"-");
                    if(st.hasMoreTokens())startnumst = st.nextToken();
                    if(st.hasMoreTokens())endnumst = st.nextToken();
                    if(startnumst!=null && endnumst!=null){
                        startnum = Integer.parseInt(startnumst);
                        endnum = Integer.parseInt(endnumst);
                        for(int i= startnum; i<=endnum;i++){
                            gcomp = new CoolComponent();
                            gcomp.setComponent_ClientHost(comp.getComponent_ClientHost());
                            gcomp.setComponent_ClientPort(comp.getComponent_ClientPort());
                            gcomp.setComponent_Code(comp.getComponent_Code());
                            gcomp.setComponent_Controls(comp.getComponent_Controls());
                            gcomp.setComponent_Filtersdalog(comp.getComponent_Filtersdalog());
                            gcomp.setComponent_Ipc(comp.getComponent_Ipc());
                            gcomp.setComponent_Loginglevel(comp.getComponent_Loginglevel());
                            gcomp.setComponent_Node(comp.getComponent_Node());
                            gcomp.setComponent_Objecttype(comp.getComponent_Objecttype());
                            gcomp.setComponent_Option(comp.getComponent_Option());
                            gcomp.setComponent_Priority(comp.getComponent_Priority());
                            gcomp.setComponent_Processes(comp.getComponent_Processes());
                            gcomp.setComponent_Reportinginterval(comp.getComponent_Reportinginterval());
                            gcomp.setComponent_Reportsto(comp.getComponent_Reportsto());
                            gcomp.setComponent_Representscoda2client(comp.getComponent_Representscoda2client());
                            gcomp.setComponent_Representsipaddress(comp.getComponent_Representsipaddress());
                            gcomp.setComponent_Runtype(comp.getComponent_Runtype());
                            gcomp.setComponent_Session(comp.getComponent_Session());
                            gcomp.setComponent_Type(comp.getComponent_Type());
                            gcomp.setComponent_Name(nameprefix+String.valueOf(i));
                            cl.add(gcomp);
                        }
                    }
                }
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }

        return cl;
    }


    /**
     * Parsing the component of the control. Fills the global hashtable of the control components.
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return cl               CoolComponent list
     */

    public ArrayList parseComponentList(Object subject, String predicate) {

        CoolNode tmpn;
        CoolComponent cmp, smscmp;
        CoolOption op;
        List l,processlist,ruleslist,complist;
        ArrayList cl= new ArrayList();
        String tmps, componentName=null, componentSession;

// Creating query string
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;

// cretate component object
                cmp = new CoolComponent();

                if (verbose >= rcConstants.debugInfo){
                    out.write("Start parsing Component of the Control   =>       "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

// component name
                tmps = getValue(x,"hasName");
                if(tmps!=null)componentName = tmps;
                else {
                    out.write("COOL-ERROR: component "+x+" does not have a name "+"\n");
                    System.out.println("COOL-ERROR: component "+x+" does not have a name");
                }

//  Parse priority and add to CodaComponent object
                tmps = getValue(x,"hasPriority");
                if(tmps!=null) cmp.setComponent_Priority(Integer.valueOf(tmps));
                else {
                    out.write("COOL-ERROR: no priority defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no priority defined for "+x);
                }

//  Parse type and add to CodaComponent object
                tmps = getValue(x,"hasType");
                if(tmps!=null){
                    cmp.setComponent_Type(tmps);
                }
                else {
                    out.write("COOL-ERROR: no type defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no type defined for "+x);
                }

// component ipc
                tmps = getValue(x,"hasIpc");
                if(tmps!=null)cmp.setComponent_Ipc(tmps);
                else {
                    out.write("COOL-ERROR: no interprocess communication protocol defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no interprocess communication protocol defined for "+x);
                }

//  Parse hasCode and add to CodaComponent object
                tmps = getValue(x,"hasCode");
                if(tmps!=null)cmp.setComponent_Code(tmps);
                else out.write("COOL-Warning: no code defined"+"\n");

//  Parse session and add to CodaComponent object
                tmps = getValue(x,"hasSession");
                if(tmps!=null){
                    cmp.setComponent_Session(tmps);
                    componentSession = tmps;
                }
                else {
                    cmp.setComponent_Session(COOL_SESSION);
                    componentSession = COOL_SESSION;
                    out.write("COOL-Warning: component "+x+" is using defined global session *"+COOL_SESSION+"*\n");
                }

//  Parse runType and add to CodaComponent object
                tmps = getValue(x,"hasRunType");
                if(tmps!=null)cmp.setComponent_Runtype(tmps);
                else {
                    cmp.setComponent_Runtype(COOL_CONTROL);
                    out.write("COOL-Warning: component "+x+" is usning defined global runtype\n");
                }

//  Parse Node and add to CodaComponent object
                tmpn = parseNode(x, "hasNode");
                if(tmpn!=null)cmp.setComponent_Node(tmpn);
                else out.write("COOL-Warning: no node defined"+"\n");

//  Parse Process and add to CodaComponent object
                processlist = parseProcessList(x,"hasExternalProcess");
                if(processlist!=null && !processlist.isEmpty())cmp.setComponent_Processes(processlist);
                else out.write("COOL-Warning: no external processes defined"+"\n");

// Parse representsCoda2Client
                tmps = getValue(x,"representsCoda2Client");
                if(tmps!=null)cmp.setComponent_Representscoda2client(tmps);
                else {
                    out.write("COOL-WARNING representsCoda2Client is not defined.\n");
                }

//  logging level
                tmps = getValue(x,"hasLoggingLevel");
                if(tmps!=null) cmp.setComponent_Loginglevel(Integer.valueOf(tmps));
                else out.write("COOL-Warning: no loggingLevel defined"+"\n");

// representsIPAddress
                tmps = getValue(x,"representsIPAddress");
                if(tmps!=null)cmp.setComponent_Representsipaddress(tmps);
                else {
                    out.write("COOL-WARNING representsIPAddress is not defined.\n");
                }

// hasOption
                op = parseOption(x,"hasOption");
                if(op!=null)cmp.setComponent_Option(op);
                else {
                    out.write("COOL-WARNING hasOption is not defined.\n");
                }

// controlsComponent
                complist = getValueList(x,"controlsComponent");
                if(complist!=null && !complist.isEmpty()) {
                    cmp.setComponent_Controls(complist);
                }
                else out.write("COOL-Warning: no controlsComponent is defined"+"\n");

// representsObjectType
                tmps = getValue(x,"representsObjectType");
                if(tmps!=null)cmp.setComponent_Objecttype(tmps);
                else {
                    out.write("COOL-Warning: no representsObjectType is defined for "+x+"\n");
                }

//  setsReportingInterval
                tmps = getValue(x,"setsReportingInterval");
                if(tmps!=null) cmp.setComponent_Reportinginterval(tmps);
                else {
                    out.write("COOL-Warning: no setsReportingInterval is defined for "+x+"\n");
                }

//  filtersDalog
                tmps = getValue(x,"filtersDalog");
                if(tmps!=null)cmp.setComponent_Filtersdalog(tmps);
                else {
                    out.write("COOL-Warning: no filtersDalog is defined for "+x+"\n");
                }

// reportsTo
                tmps = getValue(x,"reportsTo");
                if(tmps!=null)cmp.setComponent_Reportsto(tmps);
                else {
                    out.write("COOL-Warning: no SMS agent name is defined for this agent");
                }

// construct and set component name
                if (componentSession!=null &&
                        ((cmp.getComponent_Type().equalsIgnoreCase(rcConstants.rcSupervisorType)) ||
                                (cmp.getComponent_Type().equalsIgnoreCase(rcConstants.statemachinesupervisorType)))){
                    SUPERVISOR = componentSession+"_"+COOL_CONTROL;
                    cmp.setComponent_Name(SUPERVISOR);
                } else{
                    cmp.setComponent_Name(componentName);
                }

// add to the components list.
                cl.add(cmp);
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the Component of the Control  =>      "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return cl;
    }


    /**
     * Parsing subcontrol systems of the control. Fills the global hashtable of the cool systems.
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return cl               list of CoolSystem objects
     */

    public ArrayList parseSubSystemList(Object subject, String predicate) {

        CoolSystem cmp;

        List l,processlist,ruleslist,complist;
        ArrayList cl= new ArrayList();
        String tmps;

// Creating query string
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;

// cretate component object
                cmp = new CoolSystem();

                if (verbose >= rcConstants.debugInfo){
                    out.write("Start parsing subSystem of the Control   =>       "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

// Sustem supervisor agent name
                tmps = getValue(x,"hasSMSAgent");
                if(tmps!=null)cmp.setSystem_Supervisor(tmps);
                else {
                    out.write("COOL-ERROR: Subcontrol system supervisor "+x+" does not have a name "+"\n");
                    System.out.println("COOL-ERROR: Subcontrol system supervisor  "+x+" does not have a name");
                }

// iscontrollable
                tmps = getValue(x,"isControllable");
                if(tmps!=null)cmp.setSystem_Iscontrollable(tmps);
                else {
                    out.write("COOL-WARNING: Subcontrol system supervisor "+x+" does not have iscontrollable concept set "+"\n");
                }

//  reporting interval
                tmps = getValue(x,"hasReportingInterval");
                if(tmps!=null) cmp.setSystem_Reporinginterval(Integer.valueOf(tmps));
                else {
                    out.write("COOL-ERROR: no reporting interval is defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no reporting interval is defined for "+x);
                }

// hasState
                l = parseStateList(x,"hasState");
                if(l != null && !l.isEmpty())cmp.setSystem_State(l);
                else out.write("COOL-Warning: no hasState is defined for subcontrol system.\n");

// add to the components list.
                cl.add(cmp);
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the Component of the Control  =>      "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return cl;
    }



    /**
     * Called by the component parsere and will parse state
     * machine specific algorithm, described in the cool rule ontology.
     *
     * @param  subject               object of the cool concept
     * @param predicate              property of the cool concept
     * @return algor                 the sequence of the state machine statements.
     */
    public ArrayList parseRuleList(Object subject, String predicate){
        CoolState tmpif;
        CoolState tmpwhile;
        String tmpelse;
        CoolState tmpelseif;

        List tmporif;
        List tmpandif;
        List tmpexecute;
        List tmpmoveto;


        java.util.List l;
        ArrayList algor = new ArrayList();

        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                "> <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;

                if (verbose >= rcConstants.debugInfo){
                    out.write("Start parsing a Rule =>       "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

                CoolRule tmprule = new CoolRule();

                // See if we have IF statement in the cool. N.B. ifCodaState is not supported (version 1)
                tmpif = parseIf(x);
                if(tmpif!=null){
                    tmprule.setRule_If(tmpif);
                }

                // Next parse orif
                tmporif = parseOrIf(x);
                if(tmporif!=null){
                    tmprule.setRule_Orif(tmporif);
                }

                // parse andif
                tmpandif = parseAndIf(x);
                if(tmpandif!=null){
                    tmprule.setRule_Andif(tmpandif);
                }

                // parse thenMoveTo
                tmpmoveto = parseMoveTo(x,"thenMoveTo");
                if(tmpmoveto!=null){
                    tmprule.setRule_Thenmoveto(tmpmoveto);
                }

                // parse thenExecute
                tmpexecute = parseExecute(x,"thenExecute");
                if(tmpexecute!=null){
                    tmprule.setRule_Thenexecute(tmpexecute);
                }


                // Now parse ELSEIF. N.B. else if can not have orif or andif (version 1)
                tmpelseif = parseElseIf(x);
                if(tmpelseif!=null){
                    tmprule.setRule_Elseif(tmpelseif);
                }

                // parse elseIfMoveTo
                tmpmoveto = parseMoveTo(x,"elseIfMoveTo");
                if(tmpmoveto!=null){
                    tmprule.setRule_Elseifmoveto(tmpmoveto);
                }

                // parse elseIfExecute
                tmpexecute = parseExecute(x,"elseIfExecute");
                if(tmpexecute!=null){
                    tmprule.setRule_Elseifexecute(tmpexecute);
                }

                // Now parse ELSE.
                tmpelse = parseElse(x);
                if(tmpelse!=null){
                    tmprule.setRule_Else(tmpelse);
                }

                // parse elseMoveTo
                tmpmoveto = parseMoveTo(x,"elseMoveTo");
                if(tmpmoveto!=null){
                    tmprule.setRule_Elsemoveto(tmpmoveto);
                }

                // parse elseExecute
                tmpexecute = parseExecute(x,"elseExecute");
                if(tmpexecute!=null){
                    tmprule.setRule_Elseexecute(tmpexecute);
                }

                // Now parse WHILE
                tmpwhile = parseWhile(x);
                if (tmpwhile != null){
                    tmprule.setRule_While(tmpwhile);
                }

                // parse whileMoveTo
                tmpmoveto = parseMoveTo(x,"whileMoveTo");
                if(tmpmoveto!=null){
                    tmprule.setRule_Whilemoveto(tmpmoveto);
                }

                // parse whileExecute
                tmpexecute = parseExecute(x,"whileExecute");
                if(tmpexecute!=null){
                    tmprule.setRule_Whileexecute(tmpexecute);
                }

                algor.add(tmprule);

                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the Rule =>      "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        if(algor.isEmpty()) return null;
        else return algor;
    }


    /**
     * get the processes
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return  pl              List of CoolPRocess objects
     */
    public List parseProcessList(Object subject, String predicate) {

        List pl = new ArrayList();
        CoolPublisher tmpub;
        CoolData tmpd;
        CoolProcess p;
        String tmps;

        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                "> <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;

// create the process object
                p = new CoolProcess();

                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Starting to parse Process   =>       "  +x+"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

//  parse isPartOfState
                tmps = getValue(x,"isPartOfState");
                if(tmps!=null)p.setProcess_Ispartofstate(tmps);
                else out.write("COOL-Warning: no isPartOfState defined"+"\n");

//  Parse Commandname
                tmps = getValue(x,"hasCommandName");
                if(tmps!=null)p.setProcess_Commandname(tmps);
                else {
                    out.write("COOL-ERROR: no command name is defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no command name is defined for "+x);
                }

//  Parse processipc
                tmps = getValue(x,"hasProcessIpc");
                if(tmps!=null)p.setProcess_Ipc(tmps);
                else {
                    out.write("COOL-ERROR: no process IPC is defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no process IPC is defined for "+x);
                }

// Parse Timeout
                tmps = getValue(x,"hasTimeOut");
                try{
                    if(tmps != null)p.setProcess_Timeout(Integer.parseInt(tmps));
                    else p.setProcess_Timeout(0);
                }catch( NumberFormatException e) {e.printStackTrace();}

// Parse Periodicity
                tmps = getValue(x,"hasPeriodicity");
                if(tmps!=null){
                    try{
                        p.setProcess_Periodicity(Integer.parseInt(tmps));
                    }catch( NumberFormatException e) {e.printStackTrace();}
                }
                else {
                    out.write("COOL-Warning: no periodicity is defined for "+x+"\n");
                }

// Parse Delay
                tmps = getValue(x,"hasDelay");
                if(tmps!=null){
                    try{
                        p.setProcess_Delay(Integer.parseInt(tmps));
                    }catch( NumberFormatException e) {e.printStackTrace();}
                }
                else {
                    out.write("COOL-Warning: no delay is defined for "+x+"\n");
                }

//  Parse Input Data
                tmpd = parseData(x, "hasInputData");
                if(tmpd!=null)p.setProcess_Inputdata(tmpd);
                else out.write("COOL-Warning: no Input Data defined"+"\n");

//  Parse Output Data
                tmpd = parseData(x, "hasOutputData");
                if(tmpd!=null)p.setProcess_Outputdata(tmpd);
                else out.write("COOL-Warning: no Output Data defined"+"\n");

//  Parse Subscribes To
                tmpub = parsePublisher(x, "subscribesTo");
                if(tmpub!=null)p.setProcess_Subscribesto(tmpub);
                else out.write("COOL-Warning: no Publisher defined"+"\n");


//  Parse isStateTransition
                tmps = getValue(x,"isStateTransition");
                if(tmps!=null)p.setProcess_Isstatetransition(tmps);
                else {
                    out.write("COOL-Warning: no isStateTransition is defined \n");
                }

// hasRequestId
                tmps = getValue(x,"hasRequestId");
                if(tmps!=null){
                    try{
                        p.setProcess_Requestid(Integer.parseInt(tmps));
                    }catch( NumberFormatException e) {e.printStackTrace();}
                }
                else {
                    out.write("COOL-Warning: no requestId is defined for "+x+"\n");
                }

// hasErrorIndex
                tmps = getValue(x,"hasErrorIndex");
                if(tmps!=null){
                    try{
                        p.setProcess_Errorindex(Integer.parseInt(tmps));
                    }catch( NumberFormatException e) {e.printStackTrace();}
                }
                else {
                    out.write("COOL-Warning: no errorIndex is defined for "+x+"\n");
                }

//  hasErrorStatus
                tmps = getValue(x,"hasErrorStatus");
                if(tmps!=null)p.setProcess_Errorstatus(tmps);
                else {
                    out.write("COOL-Warning: no errorStatus is defined \n");
                }

//  gets
                tmps = getValue(x,"gets");
                if(tmps!=null)p.setProcess_Gets(tmps);
                else {
                    out.write("COOL-Warning: no Gets is defined \n");
                }

//  sets
                tmps = getValue(x,"sets");
                if(tmps!=null)p.setProcess_Sets(tmps);
                else {
                    out.write("COOL-Warning: no Sets is defined \n");
                }

//  traps
                tmps = getValue(x,"traps");
                if(tmps!=null)p.setProcess_Traps(tmps);
                else {
                    out.write("COOL-Warning: no Traps is defined \n");
                }

                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the Process   =>      "  +x+"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
                pl.add(p);
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return pl;
    }

    /**
     * get the COOL process
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return  pl              CoolPRocess objects
     */
    public CoolProcess parseProcess(Object subject, String predicate) {
        List l = parseProcessList(subject,predicate);
        if(l==null) return null;
        else if (l.isEmpty()) return null;
        else return (CoolProcess)l.get(0);
    }

    /**
     * get the COOL component
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return  pl              CoolComponent objects
     */
    public CoolComponent parseComponent(Object subject, String predicate) {
        List l = parseComponentList(subject,predicate);
        if(l==null) return null;
        else if (l.isEmpty()) return null;
        else return (CoolComponent)l.get(0);
    }


    /**
     * get the COOL state
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return  st              CoolState objects
     */
    public CoolState parseState(Object subject, String predicate) {
        List l = parseStateList(subject,predicate);
        if(l==null) return null;
        else if (l.isEmpty()) return null;
        else return (CoolState)l.get(0);
    }

    /**
     * Method parses the cool State concept
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return states            List of the CoolState objects
     */
    public ArrayList parseStateList(Object subject, String predicate){

        CoolState tmpp;
        CoolComponent comp;
        List proc, depstates;
        ArrayList states = new ArrayList();
        String iscritical, name, tmps;

        /* Creating query string*/
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Start parsing the State for "+x.toString()+"       \n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

// Create the CoolPublisher object
                tmpp = new CoolState();

// hasStateName
                name = getValue(x,"hasStateName");
                if(name!=null)tmpp.setState_Name(name);
                else {
                    out.write("COOL-ERROR: no name is defined for the State of the "+x.toString()+"\n");
                    System.out.println("COOL-ERROR: no name is defined for the State of the"+x.toString());
                }

// isfor

                comp = parseComponent(x,"isFor");
                if(comp!=null){
                    tmpp.setState_Isfor(comp);
                }
                else {
                    out.write("COOL-Warning: no isFor is defined for the State\n");
                }


// achievedThrough
                proc = parseProcessList(x,"achevedThrough");
                if(proc.isEmpty()) {
                    out.write("COOL-WARNING: no achievedThrough defined for State"+"\n");
                }
                else  tmpp.setState_Achievedthrough(proc);

// isCritical
                iscritical = getValue(x,"isCritical");
                if(iscritical!=null)tmpp.setState_Iscritical(iscritical);
                else {
                    out.write("COOL-Warning: no isCritical is defined for the State\n");
                }

// dependsOn
                depstates = parseStateList(x,"dependsOn");
                if(depstates.isEmpty()) {
                    out.write("COOL-WARNING: no dependsOn defined for State"+"\n");
                }
                else  tmpp.setState_Dependson(depstates);

// severity
                tmps = getValue(x,"hasSeverity");
                if(tmps!=null){
                    try{
                        tmpp.setState_Severity(Integer.parseInt(tmps));
                    }catch( NumberFormatException e) {e.printStackTrace();}
                }
                else {
                    out.write("COOL-Warning: no severity is defined for "+x+"\n");
                }


                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the State      \n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
                states.add(tmpp);
            }
            results.close() ;
        } catch (Exception ex1) {
            ex1.printStackTrace(System.err) ;
        }
        return states;
    }


    /**
     * Method parses the Publisher, i.e. cMsg publish/subscribe sets for the Control
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return tmpp             CoolPublisher object
     */
    public CoolPublisher parsePublisher(Object subject, String predicate){

        CoolPublisher tmpp = null;
        String connectsto, sendsubjects, sendtypes, subscribesubjects, subscribetypes;

        /* Creating query string*/
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Starting to parse Publisher for   =>       \n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

// Create the CoolPublisher object
                tmpp = new CoolPublisher();

// cMsg udl
                connectsto = getValue(x,"hasUDL");
                if(connectsto==null) {
                    out.write("COOL-ERROR: no hasUDL defined"+"\n");
                    System.out.println("COOL-ERROR: no hasUDL defined");
                }
                else  tmpp.setPublisher_Udl(connectsto);

// cMsg sendsubject
                sendsubjects = getValue(x,"hasSendSubject");
                if(sendsubjects == null) out.write("COOL-Warning: no hasSendSubject defined"+"\n");
                else tmpp.setPublisher_Sendsubject(sendsubjects);

// cMsg sendtype
                sendtypes =getValue(x,"hasSendType");
                if(sendtypes == null) out.write("COOL-Warning: no hasSendType defined"+"\n");
                else tmpp.setPublisher_Sendtype(sendtypes);

// cMsg subscribesubject
                subscribesubjects = getValue(x,"hasSubscribeSubject");
                if(subscribesubjects == null)out.write("COOL-Warning: no hasSubscribeSubject defined"+"\n");
                else tmpp.setPublisher_Subscribesubject(subscribesubjects);

// cMsg subscribetype
                subscribetypes = getValue(x,"hasSubscribeType");
                if(subscribetypes == null) out.write("COOL-Warning: no hasSubscribeType defined"+"\n");
                else tmpp.setPublisher_Subscribetype(subscribetypes);

                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the publisher for  =>      \n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }
            results.close() ;
        } catch (Exception ex1) {
            ex1.printStackTrace(System.err) ;
        }
        return tmpp;
    }


    /**
     * Method parses the Option of the Control. Fills the global Option object.
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return co               CoolOption object
     */
    public CoolOption parseOption(Object subject, String predicate){

        CoolOption co = null;

        String tmp;
        CoolScheduler tmpsc;

// Creating query string
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";

        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Starting to parse Options of the Control   =>       "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

                co = new CoolOption();

// parse scheduler
                tmpsc = parseScheduler(x,"hasScheduler");
                if(tmpsc!=null)co.setOption_Scheduler(tmpsc);
                else out.write("COOL-Warning: no scheduler defined"+"\n");

// parse datafile of the control.
                tmp = getValue(x,"hasDataFile");
                if(tmp!=null)co.setOption_Datafile(tmp);
                else out.write("COOL-Warning: no datafile defined"+"\n");

// parse hascoda2component.
                tmp = getValue(x,"hasCoda2Component");
                if(tmp!=null)co.setOption_Hascoda2component(tmp);
                else out.write("COOL-Warning: no coda2 components defined"+"\n");

// parse hasAutoStart.
                tmp = getValue(x,"hasAutoStart");
                if(tmp!=null)co.setOption_Hasautostart(tmp);
                else out.write("COOL-Warning: no autostart defined for the component"+"\n");

// parse hasStartRun.
                tmp = getValue(x,"hasStartRun");
                if(tmp!=null)co.setOption_Hasstartrun(tmp);
                else out.write("COOL-Warning: no startrun defined for the component"+"\n");

// parse datalimit of the control.
                tmp = getValue(x,"hasDataLimit");
                try{
                    if(tmp!=null)co.setOption_Datalimit(Integer.parseInt(tmp));
                }catch( NumberFormatException e) {e.printStackTrace();}

// parse eventlimit of the control.
                tmp = getValue(x,"hasEventLimit");
                try{
                    if(tmp!=null){
                        co.setOption_Eventlimit(Integer.parseInt(tmp));
                        System.out.println("... coolparser evt limit ... "+tmp);
                    }
                }catch( NumberFormatException e) {e.printStackTrace();}

// hasDataFilePrefix
                tmp = getValue(x,"hasDataFilePrefix");
                if(tmp!=null)co.setOption_Datafileprefix(tmp);
                else out.write("COOL-Warning: no dataFilePrefix defined"+"\n");

// hasConfigFile
                tmp = getValue(x,"hasConfigFile");
                if(tmp!=null)co.setOption_Configfile(tmp);
                else out.write("COOL-Warning: no configFile defined"+"\n");

// hasConfigString
                tmp = getValue(x,"hasConfigString");
                if(tmp!=null)co.setOption_Configstring(tmp);
                else out.write("COOL-Warning: no configString defined"+"\n");

// configIsXml
                tmp = getValue(x,"configIsXml");
                if(tmp!=null)co.setOption_Configisxml(tmp);
                else out.write("COOL-Warning: no configIsXml defined"+"\n");

// configIsMib
                tmp = getValue(x,"configIsMib");
                if(tmp!=null)co.setOption_Configismib(tmp);
                else out.write("COOL-Warning: no configIsMib defined"+"\n");

                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the Options of the Control  =>      "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }
            results.close() ;
        } catch (Exception ex1) {
            ex1.printStackTrace(System.err) ;
        }
        return co;
    }


    /**
     * Method parses the Scheduler
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return tmpsc            Scheduler object
     */
    public CoolScheduler parseScheduler(Object subject, String predicate){
        List processes;
        List stops;
        List pauses ;
        CoolScheduler tmpsc = null;

// Creating query string
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Starting to parse Scheduler   =>  "+x+"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

// Create scheduler object
                tmpsc = new CoolScheduler();

// parsing the runconfig process
                processes = parseProcessList(x,"hasExternalRunConfigProcess");
                if(processes.isEmpty()) out.write("COOL-Warning: no externalRunConfigProcess defined"+"\n");
                else  tmpsc.setScheduler_ProcessList(processes);

// parsing hasstopafterevtcount
                stops = getValueList(x,"hasStopAfterEvtCount");
                if(stops.isEmpty()) out.write("COOL-Warning: no stopafterevtcount defined"+"\n");
                else tmpsc.setScheduler_StopafterList(stops);

// parsing haspauseafterevtcount
                pauses = getValueList(x,"hasPauseAfterEvtCount");
                if(pauses.isEmpty()) out.write("COOL-Warning: no pauseafterevtcount defined"+"\n");
                else tmpsc.setScheduler_PauseafterList(pauses);
                if((!stops.isEmpty()) && (!pauses.isEmpty())){
                    System.out.println("COOL-ERROR: Mixing stopAfter and pauseAfter concepts in the Scheduler is not supported. ");
                    out.write("COOL-ERROR: Mixing stopAfter and pauseAfter concepts in the Scheduler is not supported. ");
                }

                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the Scheduler  =>      \n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }

            results.close() ;
        } catch (Exception ex1) {
            ex1.printStackTrace(System.err) ;
        }
        return tmpsc;
    }


    /**
     * Parse the node
     *
     * @param subject           object of the cool concept
     * @param predicate         String of the cool predicate name
     * @return nd               Node object
     */
    public CoolNode parseNode(Object subject, String predicate) {

        String tmp;
        CoolNode nd = null;

        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                ">, <"+COOL_CORE+""+predicate+">,?x )";

        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Starting to parse Node of the Component   =>"+x+"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }

// create node object
                nd = new CoolNode();

//  Parse name and add to Node object
                tmp = getValue(x,"hasNodeName");
                if(tmp!=null)nd.setNode_Name(tmp);
                else out.write("COOL-Warning: no nodeName defined"+"\n");

//  Parse  IP address and add to Node object
                tmp = getValue(x,"hasIP");
                if(tmp!=null)nd.setNode_Ip(tmp);
                else out.write("COOL-Warning: no IP defined"+"\n");

//  Parse  local and add to Node object
                tmp = getValue(x,"isLocal");
                if(tmp!=null)nd.setNode_Local(tmp);

//  Parse  username and add to Node object
                tmp = getValue(x,"hasUserName");
                if(tmp!=null)nd.setNode_User(tmp);
                else out.write("COOL-Warning: no userName defined"+"\n");

//  Parse  password and add to Node object
                tmp = getValue(x,"hasPassword");
                if(tmp!=null)nd.setNode_Passwd(tmp);
                else out.write("COOL-Warning: no password defined"+"\n");
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing the Node of the Component  =>      "  +x  +"\n") ;
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return nd;
    }


    /**
     * Parse the data of the command
     *
     * @param subject            object of the cool concept
     * @param predicate          String: cool:predicate ( for example has timestamp, hasdatatype, etc.)
     * @return                   Data object
     */
    public CoolData parseData(Object subject, String predicate){
        String tmp;
        CoolData dt = null;
        String tq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                "> <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(tq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                Object x = res.get("x");
                if(x==null)break;
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Starting to parse the Data of the "+subject.toString()+"  =>  "  +x+"\n");
                    out.write("---------------------------------------------------"  +"\n") ;
                }

// Create data object
                dt = new CoolData();

//  Parse timestamp
                tmp = getValue(x,"hasTimeStamp");
                if(tmp!=null)dt.setData_Timestamp(tmp);
                else out.write("COOL-Warning: no timeStamp defined"+"\n");

//  Parse type
                tmp = getValue(x,"hasDataType");
                if(tmp!=null)dt.setData_Type(tmp);
                else {
                    out.write("COOL-ERROR: no dataType defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no dataType defined for "+x);
                }

//  Parse cardinality
                tmp = getValue(x,"hasCardinality");
                if(tmp!=null)dt.setData_Cardinality(tmp);
                else out.write("COOL-Warning: no cardinality defined"+"\n");

//  Parse semantics
                tmp = getValue(x,"hasSemantics");
                if(tmp!=null)dt.setData_Semantics(tmp);
                else out.write("COOL-Warning: no semantics defined"+"\n");

//  Parse content
                tmp = getValue(x,"hasContent");
                if(tmp!=null)dt.setData_Content(tmp);
                else {
                    out.write("COOL-ERROR: no content defined for "+x+"\n");
                    System.out.println("COOL-ERROR: no content defined for "+x);
                }
                if (verbose >= rcConstants.debugInfo){
                    out.write("Debug - Finished parsing Data of the "+subject.toString()+"  =>  "  +x+"\n");
                    out.write("---------------------------------------------------"  +"\n") ;
                }
            }
            results.close() ;
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return dt;
    }


    /**
     * Parse the Cool if concept
     *
     * @param subject           object of the cool concept
     * @return a                CoolState object
     */
    public CoolState parseIf(Object subject) {

        CoolState st = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool IF   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

//  Parse the state of the if statement
            st = parseState(subject,"if");
            if(st==null){
                out.write("COOL-WARNING: no cool if statement is defined for the component"+"\n");
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool IF =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }

    /**
     * Parse the Cool orif concept
     *
     * @param subject           object of the cool concept
     * @return orifs           List of Cool ORIf objects
     */
    public List parseOrIf(Object subject) {

        List st = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool ORIF   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

//  Parse the state of the if statement
            st = parseStateList(subject,"orif");
            if(st==null || st.isEmpty()){
                out.write("COOL-WARNING: no cool orif concept is defined for the component"+"\n");
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool ORIF =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }

    /**
     * Parse the Cool andif concept
     *
     * @param subject           object of the cool concept
     * @return andifs           List of Cool ANDIf objects
     */
    public List parseAndIf(Object subject) {

        List st = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool ANDIF   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

//  Parse the state of the if statement
            st = parseStateList(subject,"andif");
            if(st==null || st.isEmpty()){
                out.write("COOL-WARNING: no cool andif concept is defined for the component"+"\n");
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool ANDIF =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }


    /**
     * Parse the Cool if concept
     *
     * @param subject           object of the cool concept
     * @return a                CoolState object
     */
    public CoolState parseWhile(Object subject) {

        CoolState st = null;
        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool WHILE   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

//  Parse the state of the while statement
            st = parseState(subject,"while");
            if(st==null){
                out.write("COOL-WARNING: no cool while statement is defined for the component"+"\n");
            }
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool WHILE =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }

    /**
     * Parse the Cool endwhile concept
     *
     * @param subject           object of the cool concept
     * @return a                Cool ENDWHILE object
     */
    public CoolEndWhile parseEndwhile(Object subject) {

        String st;
        CoolEndWhile a = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool ENDWHILE   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

// create cool endwhile object
            a = new CoolEndWhile();

            st = getValue(subject,"endwhile");
            if(st!=null)a.setEndwhile_Value(st);
            else {
                out.write("COOL-Warning: no cool endwhile is defined for the component"+"\n");
                a=null;
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool ENDWHILE =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return a;
    }



    /**
     * Parse the Cool  moveTo concepts ( thenMoveTo, elseMoveTo, elseIfMoveTo)
     *
     * @param subject           object of the cool concept
     * @param moveto            subject of the cool concept
     * @return movetos           List of Cool MOVETO objects
     */
    public List parseMoveTo(Object subject, String moveto) {

        List st=null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool MOVETO   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

//  Parse the state of the moveto statement
            st = parseStateList(subject,moveto);
            if(st==null || st.isEmpty()){
                out.write("COOL-WARNING: no cool moveTo concept is defined for the component"+"\n");
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool MOVETO =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }


    /**
     * Parse the Cool  executeProcess concept
     *
     * @param subject           object of the cool concept
     * @param predicate        object of the cool predicate
     * @return execs           List of Cool Process objects
     */
    public List parseExecute(Object subject, String predicate) {

        List st=null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool executeProcess   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

//  Parse the state of the executeProcess statement
            st = parseProcessList(subject,predicate);
            if(st==null || st.isEmpty()){
                out.write("COOL-WARNING: no cool executePRocess concept is defined for the component"+"\n");
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool EXECUTEPROCESS =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }


    /**
     * Parse the Cool else if concept
     *
     * @param subject           object of the cool concept
     * @return a                Cool ELSEIf object
     */
    public CoolState parseElseIf(Object subject) {

        CoolState st = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool ELSEIF   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }


//  Parse the state of the if statement
            st = parseState(subject,"elseif");
            if(st==null){
                out.write("COOL-Warning: no cool elseif is defined for the component"+"\n");
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool ELSEIF =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }


    /**
     * Parse the Cool else concept
     *
     * @param subject           object of the cool concept
     * @return a                CoolState object
     */
    public String parseElse(Object subject) {

        String st = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool ELSE   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }


            st = getValue(subject,"else");
            if(st==null){
                out.write("COOL-Warning: no cool else is defined for the component"+"\n");
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool ELSE =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return st;
    }


    /**
     * Parse the Cool endif concept
     *
     * @param subject           object of the cool concept
     * @return a                Cool ENDIF object
     */
    public CoolEndIf parseEndif(Object subject) {

        String st;
        CoolEndIf a = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool ENDIF   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

// create cool endif object
            a = new CoolEndIf();

            st = getValue(subject,"endif");
            if(st!=null)a.setEndif_Value(st);
            else {
                out.write("COOL-Warning: no cool endif is defined for the component"+"\n");
                a=null;
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool ENDIF =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return a;
    }

    /**
     * Parse the Cool Loop concept
     *
     * @param subject           object of the cool concept
     * @return a                Cool LOOP object
     */
    public CoolLoop parseLoop(Object subject) {

        String st;
        CoolLoop a = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool LOOP   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

// create cool loop object
            a = new CoolLoop();

            st = getValue(subject,"loop");
            if(st!=null){
                try{
                    a.setLoop_Param(Integer.parseInt(st));
                } catch (NumberFormatException e){e.printStackTrace();}
            }
            else {
                out.write("COOL-Warning: no cool loop is defined for the component"+"\n");
                a=null;
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool LOOP =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return a;
    }

    /**
     * Parse the Cool endLoop concept
     *
     * @param subject           object of the cool concept
     * @return a                Cool ENDLOOP object
     */
    public CoolEndLoop parseEndloop(Object subject) {

        String st;
        CoolEndLoop a = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool ENDLOOP   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

// create cool endloop object
            a = new CoolEndLoop();

            st = getValue(subject,"endLoop");
            if(st!=null)a.setEndloop_Value(st);
            else {
                out.write("COOL-Warning: no cool endloop is defined for the component"+"\n");
                a=null;
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool ENDLOOP =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return a;
    }

    /**
     * Parse the Cool wait concept
     *
     * @param subject           object of the cool concept
     * @return a                Cool WAIT object
     */
    public CoolWait parseWait(Object subject) {

        String st;
        CoolWait a = null;

        try {
            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Starting to parse Cool WAIT   \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }

// create cool wait object
            a = new CoolWait();

            st = getValue(subject,"wait");
            if(st!=null)a.setWait_Sec(st);
            else {
                out.write("COOL-Warning: no cool wait is defined for the component"+"\n");
                a=null;
            }

            if (verbose >= rcConstants.debugInfo){
                out.write("Debug - Finished parsing Cool WAIT =>     \n") ;
                out.write("---------------------------------------------------"  +"\n") ;
            }
        } catch (Exception ex1)
        {
            ex1.printStackTrace(System.err) ;
        }
        return a;
    }

    /**
     *  Method returns the session
     * @return String session.
     */
    public String getSession(){
        Object y=null;

        /* core taxonomy of the cool is stored in this URL. Fix it, and make it not hardcoded */
        String tq3 =
                "SELECT ?x ?y WHERE(?x <"+COOL_CORE+"hasSession>, ?y )";

        try {
            Query query = new Query(tq3) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;

            /* query result can include one or many results, depending from the cool file of cource.
            * so, we just siply take the last result from the result set */
            if(results.hasNext())
            {
                ResultBinding res = (ResultBinding)results.next() ;
                y = res.get("y") ;
            }
            results.close() ;

        } catch (Exception ex)
        {
            System.out.println("Exception: "+ex) ;
            ex.printStackTrace(System.err) ;
        }
        if (y == null) {
            return null;
        } else return y.toString();
    }


    /**
     *  Method returns the runtype
     * @return String runtype.
     */
    public String getRunType(){
        Object y=null;
        String tq3 =
                "SELECT ?x ?y WHERE(?x <"+COOL_CORE+"hasRunType>, ?y )";

        try {
            Query query = new Query(tq3) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            if(results.hasNext())
            {
                ResultBinding res = (ResultBinding)results.next() ;
                y = res.get("y") ;
            }
            results.close() ;

        } catch (Exception ex)
        {
            System.out.println("Exception: "+ex) ;
            ex.printStackTrace(System.err) ;
        }
        if (y == null) {
            return null;
        } else return y.toString();
    }


    /**
     * Method creats a query from the string and executes it within the
     * jena model and returns the result as a string.
     *
     * @param subject            object of the cool concept
     * @param predicate          String: cool:predicate ( for example has timestamp, hasdatatype, etc.)
     * @return  result           string
     */
    public String getValue(Object subject, String predicate){
        Object x = null;
        String sq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                "> <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(sq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                x = res.get("x");
            }
            results.close() ;
        } catch (Exception ex)
        { ex.printStackTrace(System.err) ;}
        if (x == null) {
            return null;
        } else return x.toString();
    }


    /**
     * Method creats a query from the string and executes it within the
     * jena model and returns the result as a list of strings.
     *
     * @param subject            object of the cool concept
     * @param predicate          String: cool:predicate ( for example has timestamp, hasdatatype, etc.)
     * @return  result           string
     */
    public List getValueList(Object subject, String predicate){
        Object x;
        List l = new ArrayList();

        String sq = "SELECT ?x "+ "WHERE(<"+subject.toString()+
                "> <"+COOL_CORE+""+predicate+">,?x )";
        try {
            Query query = new Query(sq) ;
            query.setSource(model);
            QueryExecution qe = new QueryEngine(query) ;

            QueryResults results = qe.exec() ;
            for ( ; results.hasNext() ; )
            {
                ResultBinding res = (ResultBinding)results.next() ;
                x = res.get("x");
                if(x!=null) l.add(x.toString());
            }
            results.close() ;
        } catch (Exception ex)
        { ex.printStackTrace(System.err) ;}
        return l;
    }

    public void dumprule(List l){
        Iterator it, itt;
        CoolRule r;
        CoolProcess p;

        List t;
        itt = l.iterator(); while(itt.hasNext()){
        r = (CoolRule)itt.next();

        System.out.println("======== Rule ========");

//................. IF
        if(r.getRule_If()!=null){
            System.out.println("if state = "+r.getRule_If().getState_Name()+" of "+ r.getRule_If().getState_Isfor().getComponent_Name());
        }

        //orif
        if(r.getRule_Orif()!=null && !r.getRule_Orif().isEmpty()){
            t=r.getRule_Orif();


            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" orif state = "+s.getState_Name());
        }
        }

        // andif
        if(r.getRule_Andif()!=null && !r.getRule_Andif().isEmpty()){
            t=r.getRule_Andif();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" andif state = "+s.getState_Name());
        }
        }

        // thenmoveto
        if(r.getRule_Thenmoveto()!=null && !r.getRule_Thenmoveto().isEmpty()){
            t=r.getRule_Thenmoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" then moveto state = "+s.getState_Name());
        }
        }

        // thenexecute
        if(r.getRule_Thenexecute()!=null && !r.getRule_Thenexecute().isEmpty()){
            t=r.getRule_Thenexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" then execute process = "+p.getProcess_Commandname());
        }
        }

// .............. ELSEIF
        if(r.getRule_Elseif()!=null){
            System.out.println("elseif state = "+r.getRule_Elseif().getState_Name());
        }

        // elseifmoveto
        if(r.getRule_Elseifmoveto()!=null && !r.getRule_Elseifmoveto().isEmpty()){
            t=r.getRule_Elseifmoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" elseif moveto state = "+s.getState_Name());
        }
        }

        // elseifexecute
        if(r.getRule_Elseifexecute()!=null && !r.getRule_Elseifexecute().isEmpty()){
            t=r.getRule_Elseifexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" elseif execute process = "+p.getProcess_Commandname());
        }
        }


// .............. ELSE
        if(r.getRule_Else()!=null){
            System.out.println("else");
        }

        // elsemoveto
        if(r.getRule_Elsemoveto()!=null && !r.getRule_Elsemoveto().isEmpty()){
            t=r.getRule_Elsemoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" else moveto state = "+s.getState_Name());
        }
        }

        // elseexecute
        if(r.getRule_Elseexecute()!=null && !r.getRule_Elseexecute().isEmpty()){
            t=r.getRule_Elseexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" else execute process = "+p.getProcess_Commandname());
        }
        }

// .............. WHILE
        if(r.getRule_While()!=null){
            System.out.println("while state = "+r.getRule_While().getState_Name());
        }

        // whilemoveto
        if(r.getRule_Whilemoveto()!=null && !r.getRule_Whilemoveto().isEmpty()){
            t=r.getRule_Whilemoveto();
            it = t.iterator(); while(it.hasNext()){
            CoolState s = (CoolState)it.next();
            System.out.println(" while moveto state = "+s.getState_Name());
        }
        }

        // whileexecute
        if(r.getRule_Whileexecute()!=null && !r.getRule_Whileexecute().isEmpty()){
            t=r.getRule_Whileexecute();
            it = t.iterator(); while(it.hasNext()){
            p = (CoolProcess)it.next();
            System.out.println(" while execute process = "+p.getProcess_Commandname());
        }
        }
    }
    }
}
