/*---------------------------------------------------------------------------*
*  Copyright (c) 2003        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Carl Timmer                                                    *
*             timmer@jlab.org                   Jefferson Lab, MS-12H        *
*             Phone: (757) 269-5130             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5800             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package org.jlab.coda.runcontrol.util.datatypes;

import java.util.LinkedList;
import java.util.ListIterator;

public class ComponentMonitorData {
    // Static data
    /** Name of the CODA component which is also its agent's name. */
    private String componentName;
    /** Type of the CODA component (eg. ROC, EB). */
    private String componentType;
    /** Priority of the CODA component. */
    private int componentPriority;
    
    /** This class stores dynamic data and calculated results. */
    static class DataPoint {
        long eventNumber;
        long bytes;
        long time;

        /** dateRate in bytes/sec calculated by component. */
        double componentDataRate;
        /** eventRate in events/sec calculated by component. */
        double componentEventRate;
        /** eventSize in bytes/events = componentDataRate/componentEventRate. */
        double componentEventSize;

        /** Running avg of componentDataRate. */
        double avgComponentDataRate;
        /** Running avg of componentEventRate. */
        double avgComponentEventRate;
        /** Running avg of componentEventSize. */
        double avgComponentEventSize;

        double  dataRate;
        double  avgDataRate;
        double  eventRate;
        double  avgEventRate;
        double  eventSize;
        double  avgEventSize;
        
        public DataPoint(long e, long b, double eRate, double dRate, long t) {
            bytes = b;
            eventNumber = e;
            componentDataRate  = dRate;
            componentEventRate = eRate;
            componentEventSize = 1000.*dRate/eRate;
            time = t;
        }
    }

    // Component determined quantities
    /** Data rate in bytes/sec calculated by CODA component (in last 2 second period). */
    private double componentDataRate;
    /** Event rate in events/sec calculated by CODA component (in last 2 second period). */
    private double componentEventRate;
    /** Event size in bytes/event = componentDataRate/componentEventRate. */
    private double componentEventSize;

    // Calculated averages of component quantities
    /** Average of all data rates in bytes/sec from CODA component. */
    private double avgComponentDataRate;
    private double runningDataRateTotal;
    /** Average of all event rates in events/sec from CODA component. */
    private double avgComponentEventRate;
    private double runningEventRateTotal;
    /** Average of all event sizes in bytes/event from CODA component. */
    private double avgComponentEventSize;
    private double runningEventSizeTotal;

    private long   numberOfRatePoints;

    // Calculated quantities with time coming from component agents
    /** Data rate in bytes/sec between last 2 data entries (in last monitoring period). */
    private double dataRate;
    /** Data rate in bytes/sec averaged over entire monitoring period. */
    private double avgDataRate;
    /** Event rate in events/sec (Hz) between last 2 data entries (in last monitoring period). */
    private double eventRate;
    /** Event rate in events/sec (Hz) averaged over entire monitoring period. */
    private double avgEventRate;
    /** Average event size in bytes between last 2 data entries (in last monitoring period). */
    private double eventSize;
    /** Average event size in bytes over entire monitoring period. */
    private double avgEventSize;
    
    /** List of DataPoint objects used to store limited amounts of the most recent data. */
    LinkedList<Object> dataHistory;
    /** Maximumn dynamic data list size. */
    static final int DATA_HISTORY_SIZE = 100;

    // DataPoint objects are used to store the two most recently collected
    // monitor data to compute rates & averages.
    
    /** Object with the previously collected monitor data. */
    DataPoint oldData;
    /** Object with the most recently collected monitor data. */
    DataPoint newData;
        
    /** Number of data points added to this object with {@link #addData} */
    private long numberOfDataPoints;
    /** Time of latest data point. */
    private long lastTime;
    /** Time data taking started. */
    private long startingTime;
    /** EventNumber of latest data point. */
    private long lastEventNumber;
    /** First event number. With current setup it is always 0. */
    private long startingEventNumber;
    /** Total number of bytes thru component. */
    private long totalBytes;
    /** Total number of bytes at initialization. With current setup it is always 0. */
    private long startingTotalBytes;

    /** Constructor which sets which component for which data will be gathered. */
    public ComponentMonitorData(String name, String type, int priority) {
        this(name, type, priority, 0L);
    }

    /** Constructor which sets which component for which data will be gathered. */
    public ComponentMonitorData(String name, String type, int priority, long time) {
        componentName = name;
        componentType = type;
        componentPriority = priority;
        startingTime = time;
        dataHistory =  new LinkedList<Object>();
        lastTime = -1;
        totalBytes = -1;
        lastEventNumber = -1;
    }
    
    // Getters for Plexus consumable data
    
    /** Get component (and agent's) name. */
    public String getName()    {return componentName;}
    public String getType()    {return componentType;}
    public int  getPriority()  {return componentPriority;}
    public long getTime()      {return lastTime;}
    public long getStartTime() {return startingTime;}
    public int  getSize()      {return dataHistory.size();}
    public long getBytes()     {return totalBytes;}
    public long getEvents()    {return newData.eventNumber - startingEventNumber;}
    public LinkedList getHistory() {return dataHistory;}

    public double getComponentDataRate()         {return componentDataRate;}
    public double getAverageComponentDataRate()  {return avgComponentDataRate;}

    public double getComponentEventRate()        {return componentEventRate;}
    public double getAverageComponentEventRate() {return avgComponentEventRate;}

    public double getComponentEventSize()        {return componentEventSize;}
    public double getAverageComponentEventSize() {return avgComponentEventSize;}

    public double getDataRate()         {return dataRate;}
    public double getAverageDataRate()  {return avgDataRate;}
    
    public double getEventRate()        {return eventRate;}
    public double getAverageEventRate() {return avgEventRate;}
    
    public double getEventSize()        {return eventSize;}
    public double getAverageEventSize() {return avgEventSize;}

    // Since it's more efficient to calculate all the following quantities at once,
    // provide the means to do so.

    private double[] dataRates;
    private double[] eventRates;
    private double[] eventSizes;
    private double[] absoluteTimes;
    private double[] relativeTimes;
    private double[] componentDataRates;
    private double[] componentEventRates;
    private double[] componentEventSizes;

    public double[] getDataRates() {return dataRates;}
    public double[] getEventRates() {return eventRates;}
    public double[] getEventSizes() {return eventSizes;}
    public double[] getAbsoluteTimes() {return absoluteTimes;}
    public double[] getRelativeTimes() {return relativeTimes;}
    public double[] getComponentDataRates() {return componentDataRates;}
    public double[] getComponentEventRates() {return componentEventRates;}
    public double[] getComponentEventSizes() {return componentEventSizes;}

    public long getStartingTime() {return startingTime;}
    public void setStartingTime(long startingTime) {this.startingTime = startingTime;}


    /**
     * This method is for the efficient storage of all data types at once
     * for times after the first argument.
     * @param time data is collected for times after this absolute time
     * @param referenceTime time to which all relative times are referenced
     */
    public void findData(long time, long referenceTime) {
        if (dataHistory.size() < 1) {
            return;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) {
            dataRates = null;
            eventRates = null;
            eventSizes = null;
            absoluteTimes = null;
            relativeTimes = null;
            componentDataRates = null;
            componentEventRates = null;
            componentEventSizes = null;
            return;
        }

        dataRates = new double[size];
        eventRates = new double[size];
        eventSizes = new double[size];
        absoluteTimes = new double[size];
        relativeTimes = new double[size];
        componentDataRates = new double[size];
        componentEventRates = new double[size];
        componentEventSizes = new double[size];

        index = 0;
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            dataRates[index] = d.dataRate;
            eventRates[index] = d.eventRate;
            eventSizes[index] = d.eventSize;
            absoluteTimes[index] = (double) (d.time)/1000.0;
            relativeTimes[index] = (double) (d.time - referenceTime)/1000.0;
            componentDataRates[index] = d.componentDataRate;
            componentEventRates[index] = d.componentEventRate;
            componentEventSizes[index] = d.componentEventSize;
            index++;
        }
    }

    /**
     * Method to get all component-calculated data rates after time.
     * @param time data is returned for all times after this argument
     * @return
     */
    public double[] getComponentDataRates(long time) {
        if (dataHistory.size() < 1) {
            return null;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) return null;

        index = 0;
        double[] y = new double[size];
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            y[index++] = d.componentDataRate;
        }
        return y;
    }

    public double[] getComponentEventRates(long time) {
        if (dataHistory.size() < 1) {
            return null;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) return null;

        index = 0;
        double[] y = new double[size];
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            y[index++] = d.componentEventRate;
        }
        return y;
    }

    public double[] getComponentEventSizes(long time) {
        if (dataHistory.size() < 1) {
            return null;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) return null;

        index = 0;
        double[] y = new double[size];
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            y[index++] = d.componentEventSize;
        }
        return y;
    }

    public double[] getDataRates(long time) {
        if (dataHistory.size() < 1) {
            return null;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) return null;

        index = 0;
        double[] y = new double[size];
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            y[index++] = d.dataRate;
        }
        return y;
     }

    public double[] getEventRates(long time) {
        if (dataHistory.size() < 1) {
            return null;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) return null;

        index = 0;
        double[] y = new double[size];
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            y[index++] = d.eventRate;
        }
        return y;
    }
    
    public double[] getEventSizes(long time) {
        if (dataHistory.size() < 1) {
            return null;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) return null;

        index = 0;
        double[] y = new double[size];
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            y[index++] = d.eventSize;
        }
        return y;
    }

    /**
     * Get array of times in seconds relative to the reference time.
     * If the reference time is Zero, then times are absolute.
     * @param referenceTime time to which all relative times are referenced
     */
    public double[] getRelativeTimes(long time, long referenceTime) {
        if (dataHistory.size() < 1) {
            return null;
        }

        DataPoint d;
        int size = 0;
        int index = 0;

        // Find out how many data points occur after "time"
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) {
                index++;
                continue;
            }
            size = dataHistory.size() - index;
            break;
        }

        if (size < 1) return null;

        index = 0;
        double[] x = new double[size];
        for (ListIterator l = dataHistory.listIterator(0); l.hasNext(); ) {
            d = (DataPoint) l.next();
            if (d.time < time) continue;
            x[index++] = (double) (d.time - referenceTime)/1000.0;
        }
        return x;
    }


    /** Calculate various data and event rates. */
    public void computeRates() {
        // Need at least 2 data entries to calculate a rate
        if (oldData == null) {
            return;
        }
                
        // Calculate instantaneous quantities (requires only last 2 entries)
        long deltaTime = (newData.time - oldData.time); // milliseconds
        if (deltaTime < 0) {
            System.out.println("\ncomputeRates: got NEGATIVE delta time: "
                    + newData.time + " - " + oldData.time + " = " + deltaTime + "\n");
        }
        long events = newData.eventNumber - oldData.eventNumber;
        if (events < 0) {
            System.out.println("\ncomputeRates: got NEGATIVE number of events: "
                    + newData.eventNumber + " - " + oldData.eventNumber + " = " + events + "\n");
        }
        dataRate  = (double)(newData.bytes - oldData.bytes)/deltaTime; // KBytes/sec
        eventRate = (double)events*1000./deltaTime; // events/sec
//System.out.println("  CompMonData " + componentName + ": DeltaTime = " + deltaTime + ", events = " + events);
        if (events < 1) {
            eventSize = 0.;
        }
        else {
            eventSize = (double)(newData.bytes - oldData.bytes)/events; // bytes/event
        }
//System.out.println("  Rates: data  = " + dataRate + ", comp = " + componentDataRate);
//System.out.println("         event = " + eventRate + ", comp = " + componentEventRate);
//System.out.println("         event size  = " + eventSize + ", comp = " + componentEventSize);
        
        // Calculate quantities averaged over the whole monitoring period
        deltaTime = newData.time - startingTime; // milliseconds
//System.out.println("  CompMonData " + componentName + ": TotalTime = " + deltaTime + ", totalBytes = " + (totalBytes-startingTotalBytes));
        events = newData.eventNumber - startingEventNumber;
        avgDataRate  = (double) (totalBytes-startingTotalBytes)/deltaTime; // KBytes/sec
        avgEventRate = (double) events*1000./deltaTime; // events/sec
        if (events < 1) {
            eventSize = 0.;
        }
        else {
            avgEventSize = (double) (totalBytes-startingTotalBytes)/events; // bytes/event
        }
//System.out.println("  Avg  : data  = " + avgDataRate + ", comp = " + avgComponentDataRate);
//System.out.println("         event = " + avgEventRate + ", comp = " + avgComponentEventRate);
//System.out.println("         event size  = " + avgEventSize + ", comp = " + avgComponentEventSize);

        // Store calculated values
        newData.dataRate = dataRate;
        newData.eventRate = eventRate;
        newData.eventSize = eventSize;
        newData.avgDataRate = avgDataRate;
        newData.avgEventRate = avgEventRate;
        newData.avgEventSize = avgEventSize;
    }


    /**
     * Add the latest, unique dynamic data.
     *
     * @param eventNumber event number at this time reported by component
     * @param bytes total number of bytes to go thru the component at this time reported by component
     * @param eRate instantaneous event rate in events/sec reported by component
     * @param dRate instantaneous data rate in bytes/sec reported by component
     * @param time time when the measurement was taken by agent
     */
    public void addData(long eventNumber, long bytes, double eRate, double dRate, long time) {
        // Use the arrival of the first data point for the
        // starting time if the actual time is not available.
        if ((numberOfDataPoints < 1) && (startingTime < 1)) {
            startingTime = time;
        }

//System.out.println(componentName + " -> "+eventNumber);
        boolean timeToReturn = false;
        // Do NOT add duplicate or nonsense data
        if (time <= lastTime) {
//System.out.println("  Reject earlier data:");
            timeToReturn = true;
        }
        if (eventNumber <= lastEventNumber) {
//System.out.println("  Allow data with <= event #");
            //System.out.println("  Old time = " + lastTime + ", bytes = " + totalBytes + ", event# = " + lastEventNumber);
            //System.out.println("  New time = " + time + ", bytes = " + bytes + ", event# = " + eventNumber);
            //timeToReturn = true;
        }
        if (bytes <= totalBytes) {
//System.out.println("  Allow data with <= byte total");
            //timeToReturn = true;
        }
        if (timeToReturn) return;

        // convert to KBytes/sec
        dRate = dRate/1000.;
        // Store last data point
        oldData = newData;
        // Make new data point
        newData = new DataPoint(eventNumber, bytes, eRate, dRate, time);

        // Make sure the list does NOT grow over its limit
        if (dataHistory.size() >= DATA_HISTORY_SIZE) {
            dataHistory.removeFirst();
        }
        dataHistory.add(newData);

        // Keep track of stuff
        lastTime = time;
        totalBytes = bytes;
        lastEventNumber = eventNumber;
        numberOfDataPoints++;

        // Keep track of component determined quantities
        componentDataRate  = dRate;
        componentEventRate = eRate;
        // don't divide by small number
        if (eRate < 1.e-2) {componentEventSize = 0.;}
        else {componentEventSize = 1000.*dRate/eRate;}

        runningDataRateTotal  += dRate;
        runningEventRateTotal += eRate;
        runningEventSizeTotal += componentEventSize;

        // The first data rates are = 0, so don't figure that into the average
        if (numberOfDataPoints > 1) {
            numberOfRatePoints++;
        }
        else if (!(dRate < 1.e-10)) {
            numberOfRatePoints++;
        }
        avgComponentDataRate   = runningDataRateTotal/numberOfRatePoints;
        avgComponentEventRate  = runningEventRateTotal/numberOfRatePoints;
        avgComponentEventSize  = runningEventSizeTotal/numberOfRatePoints;
//System.out.println("dRate = " + dRate + ", runningTotal = " + runningDataRateTotal +
//        ", avg DataRate = " + avgComponentDataRate + ", # pts = " + numberOfDataPoints);
        newData.avgComponentDataRate  = avgComponentDataRate;
        newData.avgComponentEventRate = avgComponentEventRate;
        newData.avgComponentEventSize = avgComponentEventSize;

        computeRates();
    }
    
}

