/*----------------------------------------------------------------------------*
*  Copyright (c) 2003        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, 6A/602-13     *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5800             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.util.datatypes;

import org.jlab.coda.runcontrol.ontologies.CodaRunControl.Producer;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.CodaTimeStamp;
import org.jlab.coda.runcontrol.ontologies.CodaRunControl.CoolData;

import java.util.Set;
import java.util.Vector;
import java.util.Hashtable;
import java.io.Serializable;

/**
 * Utility class implementing serializable, which is used by the agent to 
 * send data through the msg content. 
 */
public class DaqMonitorData implements Serializable {
    private static final long serialVersionUID = 7526472295622776147L;

    /** hash table containing the name of the agent and status */
    public Hashtable<String, String>  _status;

    /** hash table containing process name and the vector including 
        data, producer and codaTimestamp object.*/
    public Hashtable<String, Vector> _monitorProcesses;

    /** 
     * constructor
     *
     * @param status HashMap key = String component Name; value = String status
     * @param monitorProcesses HashMap key = String process name; value = vector
     * containing Data, Producer and CodaTimeStamp objects
     */
    public DaqMonitorData(Hashtable<String, String> status, Hashtable<String, Vector> monitorProcesses) {
        _status = status;
        _monitorProcesses = monitorProcesses;
    }                                                                                                           

    /** 
     * Returns the set of active monitoring processes in the runControl Agent platform
     *
     * @return set of the String process names
     */
    public Set Processes() {
        return _monitorProcesses.keySet();
    }

    /** 
     *Checks if the component is reporting the status information
     *
     * @param component name: String
     * @return true if  component agent is reporting its status
     */
    public boolean containsStatusComponent (String component) {
        if(_status.containsKey(component)) return true;
        else return false;
    }

    /** 
     * returns the status of the component
     *
     * @param component name : String
     * @return status : String
     */
    public String getStatus(String component){
        return  (String)_status.get(component);
    }

    /** 
     * Returns the producer object of the monitored process
     *
     * @param processName : String
     * @return  Producer: CodaRunControl Ontology
     */
    public Producer getProducer(String processName){
        Vector tmpvec = (Vector)_monitorProcesses.get(processName);
        return (Producer)tmpvec.elementAt(0);
    }
    /** 
     * Returns CodaTimeStamp object of the monitored process
     *
     * @param processName : String
     * @return  CodaTimeStamp : CodaRunControl Ontology
     */
    public CodaTimeStamp getTimeStamp(String processName){
        Vector tmpvec = (Vector)_monitorProcesses.get(processName);
        return (CodaTimeStamp)tmpvec.elementAt(1);
    }

    /** 
     * Returns Data object of the monitored process
     *
     * @param processName : String
     * @return  Data : CodaRunControl Ontology
     */
    public CoolData getData(String processName){
        Vector tmpvec = (Vector)_monitorProcesses.get(processName);
        return (CoolData)tmpvec.elementAt(2);
    }
}
