package org.jlab.coda.runcontrol.util.db;

import java.util.*;

/**
 * Contains all the common data that all transition routines
 * find necessary to carry out their functions. It's a central
 * repository of database-obtained and/or other data that is created
 * or filled-in during the "configure" transition.
 *
 * @author  Carl Timmer
 */
public class CommonData {
    /** Data that we're storing for a particular runType from database. */
    private RunTypeData runTypeData;
    
    /** Name of the runType whose data we're storing. */
    private String runType;
    
    /**
     * The retriever is generic so as to hide multiple
     * possible types of data retrievers - one for databases,
     * one for agents, etc.
     */
    private DataRetriever retriever;
       
    /** The single instance of this class. */
    private static CommonData instance = new CommonData();
        

    /** Private constructor. Use {@link #getInstance} to obtain object. */
    private CommonData () {}

    /**
     * This method returns the single CommonData object.
     * @return the CommonData object
     */
    public static CommonData getInstance() {
	return instance;
    }
    
    /**
     * Get data from the database (specified earlier in RunControl object).
     *
     * @param rType name of the runType (or experiment configuration)
     * @exception org.jlab.coda.runcontrol.util.db.RCException
     *            If the database driver cannot be registered,
     *            the connection to the database cannot be made, or
     *            there's a problem with SQl communications
     */
    synchronized public void getData(String rType) throws RCException {
	if (retriever == null) {
	    throw new RCException("Need to define \"retriever\" with setRetriever method");
	}
	runType = rType;
	runTypeData = retriever.getData(rType);
    }
    
    
    public String toString() {
	RunComponent rcomp;
	StringBuffer buf = new StringBuffer(2000);
	
        buf.append("Configuration is " + runType + " runType \n");
	
        if (runTypeData.allComponents.size() > 0) {
	    buf.append("\n  Components:\n");
	    for (Iterator i=runTypeData.allComponents.values().iterator() ; i.hasNext(); ) {
		rcomp = (RunComponent) i.next();
                buf.append("    " + rcomp.getName() +
		           ", of type " +  rcomp.getType() +
		           ", on host " +  rcomp.getHost() +
			   ", and port " + rcomp.getPort() +
			   "\n");
                buf.append("    in " + rcomp.getState() +
		           " state, with pid " +  rcomp.getPid() +
			   ", and inUse = " + rcomp.getIsInUse() + 
			   "\n\n");
            }
	}
	
	return buf.toString();
    }
    
    // getters & setters
    public RunTypeData getRunTypeData() { return runTypeData; }
    public void setRetriever(DataRetriever retriever) { this.retriever = retriever; }
    
    public void   setRunType(String runType) { this.runType = runType; }
    public String getRunType() { return runType; }
    
    public HashMap getAllComponents() { return runTypeData.allComponents; }
    public HashMap getRocComponents() { return runTypeData.rocComponents; }
    public HashMap getEbComponents()  { return runTypeData.ebComponents;  }
    public HashMap getErComponents()  { return runTypeData.erComponents;  }
    public HashMap getTsComponents()  { return runTypeData.tsComponents;  }
    public HashMap getScripts()       { return runTypeData.scripts;  }

}
