/*---------------------------------------------------------------------------*
*  Copyright (c) 2003        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Carl Timmer                                                    *
*             timmer@jlab.org                   Jefferson Lab, MS-12H        *
*             Phone: (757) 269-5130             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5800             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package org.jlab.coda.runcontrol.util.db;

import java.sql.*;

/**
  * Create a single instance (singleton) of a JDBC database
  * "Connection" object. Class methods are thread-safe.
  *
  * @author  Carl Timmer
  */
public class DatabaseConnection {
    /** Examples of existing database drivers. */
    private static final String[] drivers = {
	     "org.hsqldb.jdbcDriver",
	     "org.hsqldb.jdbcDriver",
	     "com.imaginary.sql.msql.MsqlDriver",
	     "sun.jdbc.odbc.JdbcOdbcDriver",
	     "org.gjt.mm.mysql.Driver" };

    /** Examples of existing database urls. */
    private static final String[] urls = {
	     "jdbc:hsqldb:",
	     "jdbc:hsqldb:hsql://localhost:10999",
	     "jdbc:msql://ankaa.jlab.org:8101/vg",
	     "jdbc:odbc:rcdb",
	     "jdbc:mysql://ankaa.jlab.org:3306/vg" };
    
    /** Connection to database server. */
    private static Connection connection;
    
    // The next "particular" values are used only once to
    // establish the database connection object.
    
    /** The particular database to connect to. */
    private static String url;
    /** The particular database driver. */
    private static String driver;
    /** The particular database user name. */
    private static String user = "dummy";
    /** The particular database password. */
    private static String pswd = "";
    
    /** The single instance of this class. */
    private static DatabaseConnection instance;
    
    /**
     * Gets the url of the connected database.
     * @return <code>null</code> if there is no connection to a database, or
     *         the url of the connected database
     */
    public static String getUrl() {return url;}
    
    /**
     * Gets the driver of the connected database.
     * @return <code>null</code> if there is no connection to a database, or
     *         the driver of the connected database
     */
    public static String getDriver() {return driver;}
    
    /**
     * Gets the user name used when connecting to the database.
     * @return <code>null</code> if there is no connection to a database, or
     *         the user of the connected database
     */
    public static String getUser() {return user;}
    
    /**
     * Gets the password used when connecting to the database.
     * @return <code>null</code> if there is no connection to a database, or
     *         the password of the connected database
     */
    public static String getPassword() {return pswd;}
    
    /**
     * Gets the password used when connecting to the database.
     * @return <code>null</code> if there is no connection to a database, or
     *         the password of the connected database
     */
    public static boolean connected() {
        if (instance == null) return false;
	return true;
    } 

    /**
     * This method returns the database connection object. There is
     * only one such object - a singleton.
     *
     */
    synchronized public static Connection getConnection() throws DBException {
	if (instance == null) {
	    throw new DBException("No database connection has been made, " +
	                           "call \"createConnection\" first");
	}
	return connection;
    }
    
    /**
     * This method closes the database connection.
     */
    synchronized public static void closeConnection() {
	if (instance == null) {
	    return;
	}
	
	try {connection.close();}
	catch (SQLException ex) {}
	instance = null;
	
	url = null;
	driver = null;
	user = "dummy";
	pswd = "";
	
	return;
    }
    
    /**
     * Create a connection to the database specified by the arguments.
     * Only allow this method to be run if the single instance of the
     * database connection has not been made yet. Once the connection
     * is made, it can be closed and later recreated.
     *
     * @param _url URL for the database, i.e. jdbc:msql://mizar.jlab.org:8101/ghexp
     * @param _driver jdbc driver suitable for associated URL, i.e. com.imaginary.sql.msql.MsqlDriver
     * @param _user user name for the database, i.e. any string for msql
     * @param _pswd password name for the database, i.e. blank string for msql
     * @return <code>true</code> if a new connection to the database was successfully made, or
     *         <code>false</code> if the database connection already exists
     * @exception org.jlab.coda.runcontrol.util.db.DBException
     *            If the database driver cannot be registered or
     *            the connection to the database cannot be made.
    */
    synchronized public static boolean createConnection(String _url, String _driver,
                                                        String _user, String _pswd)
							throws DBException {
	if (instance != null) {
	  return false;
	}
	
	url    = _url;
	driver = _driver;
	user   = _user;
	pswd   = _pswd;
	
        instance = new DatabaseConnection();
	
	return true;
    }


    /**
     * This method creates the database connection object.
     *
     * @exception org.jlab.coda.runcontrol.util.db.DBException
     *            If the database driver cannot be registered or
     *            the connection to the database cannot be made.
     */
    private DatabaseConnection() throws DBException {

	// Register the Driver
	try {
	    Class.forName(driver).newInstance();
	}
	catch (InstantiationException e) {
	    throw (new DBException(e.getMessage()));
	}
	catch (IllegalAccessException e) {
	    throw (new DBException(e.getMessage()));
	}
	catch (ClassNotFoundException e) {
	    throw (new DBException(e.getMessage()));
	}

	// Make connections to the database
	try {
	    connection = DriverManager.getConnection(url, user, pswd);
	}
	catch (SQLException e) {
	    throw (new DBException(e.getMessage()));
	}
    }

}
