/*----------------------------------------------------------------------------*
*  Copyright (c) 2004        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-12H        *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5800             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package org.jlab.coda.runcontrol.util.dp;

import org.jlab.coda.runcontrol.util.db.DatabaseConnection;
import org.jlab.coda.runcontrol.util.db.DBException;
import org.jlab.coda.runcontrol.util.rcException;

import java.io.*;
import java.net.Socket;
import java.net.UnknownHostException;
import java.net.InetAddress;
import java.sql.*;

/**
 * Java DP client, simplified version of the DPCmd
 *
 * @author Graham Heyes and Vardan Gyurjyan
 *
 */
public class JdpClient {

    /** Host name of the component*/
    private String Host = null;

    /** Port number of the component*/
    private int Port = 0;

    /** coda database connection*/
    private Connection con;

    /** SQL statement object*/
    private Statement stmt;

    /** response from the component*/
    private InputStream response;

    /** request to the component*/
    private OutputStream request;

    /** communication socket to the component*/
    private Socket vxsock;

    /** reply string*/
    private String strReply;

    /**
     * Constructor
     *
     * @param component component name
     */
    public JdpClient(String component) throws rcException {

        /* Open connection to database*/
        String url = System.getenv("CODA_DB_URL");
        String driver = System.getenv("CODA_DB_DRIVER");
        String usr = System.getenv("CODA_DB_USER");
        String passwd = System.getenv("CODA_DB_PASSWORD");

        if(usr==null)usr=" ";
        if(passwd==null)passwd=" ";
        try {
            DatabaseConnection.createConnection(url,driver,usr,passwd);
        } catch (DBException e) {
            e.printStackTrace();
            throw new rcException("Jdp_Error: Can not create DB connection. Check DB related env. variable settings.");
        }

        if(con == null){

// get the singelton DB connection object. If the previous DB connection is broken try 3 time to close
// the existing DB connection and open the new one.
            boolean kt=false;
            int i=0;
            do{
                // Get a db connection
                try {
                    con = DatabaseConnection.getConnection();
                }
                catch (DBException ex) {
                    System.out.println(ex.getMessage());
                    kt=true;
                    i++;
// close existing DB connection
                    DatabaseConnection.closeConnection();

// open new connection
                    try {
                        DatabaseConnection.createConnection(url,driver,usr,passwd);
                    } catch (DBException e) {
                        e.printStackTrace();
                        throw new rcException("Jdp_Error: Can not create DB connection. Check DB related env. variable settings.");
                    }
                }

            }  while(kt && i<3);

            if(i>=3){
                throw new rcException("Jdp_Error: Can not get DB connection.");
            }
        }

        /* get hostname and portnumber */
        try{
            Statement stmt = con.createStatement();
            ResultSet rs;

            rs = stmt.executeQuery("SELECT * FROM process WHERE name = '"+component+"'");
            while(rs.next()) {
//                port =Integer.parseInt(rs.getString("inuse").trim());
                Port =rs.getInt("port");
//                System.out.println(Port);
            }

            rs = stmt.executeQuery("SELECT * FROM process WHERE name = '"+component+"'");
            while(rs.next()) {
                Host =rs.getString("host").trim();
//                System.out.println(Host);
            }

            rs.close();
            stmt.close();
            if((Port == 0) && (Host == null))throw new rcException("Jdp_Error: Component is not found in the DB");
        }
        catch (SQLException ex) {
            ex.printStackTrace();
            throw new rcException("Jdp_Error: Getting host and port from DB.");
        }

    }

    /**
     * Opens the socket connection to the component.
     * Sends the command to the component.
     *
     * @param command name
     * @return reply string from the execution
     */
    public String sendCmd ( String command) throws rcException {

        try{
            vxsock = new Socket(Host,Port);
            vxsock.setTcpNoDelay(true);


            InetAddress remhost = vxsock.getInetAddress();
//                System.out.println("Connected to remote host " +remhost);

            int servport = vxsock.getPort();
//                System.out.println("Server listens on the port "+servport);

            int localport = vxsock.getLocalPort();
//                System.out.println("my port is "+localport);

            InetAddress localaddress = vxsock.getLocalAddress();
//                System.out.println("my local address is "+localaddress);

            /*open the output stream on the socket*/
            request = vxsock.getOutputStream();

            /*open the input stream on the socket*/
            response = vxsock.getInputStream();

            int i;            // index to a char buffer
            StringBuffer str = new StringBuffer(); // buffer being indexed
            char c;			  // one char at a time...

            try {
                c = 0;
                for(i=0; c!='\n'; i++) {
                    c = (char)response.read();

                    if (c > 0) {
                        str.append(c);
                    } else {

                        // we get here if read returns a -1, i.e. s is closed at the other end
                        throw new rcException("Jdp_Error: Connect DPin socket unexpectedly closed");
                    }
                }
            } catch (InterruptedIOException e) {

                // get here if there was a timeout
                System.out.println("DPask connect timeout");
                throw new rcException("Jdp_Error: Connect timeout on DPin socket");
            } catch (Exception e) {

                // get here if anything else went wrong...
                System.out.println("Another exception " + e.getMessage());
                throw new rcException("Jdp_Error: Connect exception on DPin socket");
            }

            // construct the DP message
            // Send string formed by "el" followed by our command.
            // the "e" tells the server that this is a command
            String cmd = "el";

            sendDP(cmd.concat(command));

            // Read back the reply
            strReply = recvDP();

            // Anything starting with x is an exception
            if (strReply.startsWith("x")) {
                throw new rcException(strReply.substring(2));
            }

        } catch (UnknownHostException e) {
                    throw new rcException();
        }
        catch (IOException e){
            throw new rcException();
        }
        finally {
            try {
                request.close();
                response.close();
                vxsock.close();
            }
            catch(IOException e){
                e.printStackTrace();
            }
            catch(NullPointerException ee){
                ee.printStackTrace();
            }
        }
        return strReply.substring(2);
    }



    /**
     * Reads the string from the dp socket input stream of the connection to the component
     *
     * @return String result from the socket input stream
     * @throws IOException
     */
    private String recvDP() throws IOException {

        StringBuffer str = new StringBuffer();

        int i,j;
        int ch1 = -1;
        int ch2 = -1;
        int ch3 = -1;
        int ch4 = -1;

        // this looks odd but it gets around byte ordering problems see previous method
        try {
            ch1 = response.read();
            ch2 = response.read();
            ch3 = response.read();
            ch4 = response.read();

            // assemble the message length
            j = (ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0);

            if ((ch1 == -1) || (ch2 == -1) || (ch3 == -1) || (ch4 == -1))  {
                throw new IOException("read on closed socket");
            }

            // get the message
            for(i=0; i<j; i++) {
                char c = (char)response.read();
                str.append(c);
            }
        } catch (InterruptedIOException e) {
            throw new IOException("read timeout on DPin socket");
        }
        return str.toString();

    }

    /**
     * Sends the string to the component over the dp socket output stream.
     *
     * @param str string to be sent
     * @throws IOException
     */
    private void sendDP(String str)
            throws IOException {

        // This code gets around byte ordering problems since
        // the DP protocol expects a 4 byte length followed by the data.
        // (The data consists of a two byte code followed by a clear text string
        // and is not affected by byte ordering).
        int i,j,k;

        j = k = str.length();

        byte buf[] = new byte[j+4];

        buf[0] = (byte) ((k >>> 24) & 0xFF);
        buf[1] = (byte) ((k >>> 16) & 0xFF);
        buf[2] = (byte) ((k >>>  8) & 0xFF);
        buf[3] = (byte) ((k >>>  0) & 0xFF);

        for(i=0; i<j; i++) {
            buf[4+i] = (byte)str.charAt(i);
        }

//        vxsock.setSoTimeout(timeout);

        try {
            request.write(buf);
        } catch (InterruptedIOException e) {
            System.out.println("DPask write timeout");
            throw new IOException("write timeout on DPin socket");
        }
    }
}
