/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.content;

import rca.lang.acl.ACLMessage;
import rca.core.AID;
import rca.util.leap.Iterator;

/**
 * Utility class that allow using an <code>ACLMessage</code> object
 * as an ontological agent action.
 * Original author Giovanni Caire - TILAB
 */
public class OntoACLMessage extends ACLMessage implements AgentAction {

  /**
   * Construct an ontological ACL message whose performative
   * is ACLMessage.NOT_UNDERSTOOD
   */
    public OntoACLMessage() {
        super(ACLMessage.NOT_UNDERSTOOD);
    }

  /**
   * Construct an ontological ACL message with a given 
   * performative
   * @param performative the performative of this ACL message. 
   * @see ACLMessage#ACLMessage(int)
   */
    public OntoACLMessage(int performative) {
        super(performative);
    }

    /**
     * Create an ontological ACL message that wraps an existing
     * <code>ACLMessage</code>.
     * @param msg the <code>ACLMessage</code>to be wrapped. If
     * <code>msg</code>
     * is already an ontological ACL message no new object is
     * created and <code>msg</code> is returned with the sender
     * and receivers properly wrapped if necessary.
     */
    public static OntoACLMessage wrap(ACLMessage msg) {
        OntoACLMessage wrapper = null;
        if (msg != null) {
            if (msg instanceof OntoACLMessage) {
                wrapper = (OntoACLMessage) msg;
            }
            else {
                wrapper = new OntoACLMessage(msg.getPerformative());
                // This automatically performs the wrapping
                wrapper.setSender(msg.getSender());
                Iterator it = msg.getAllReceiver();
                while (it.hasNext()) {
                    // This automatically performs the wrapping
                    wrapper.addReceiver((AID) it.next());
                }

                it = msg.getAllReplyTo();
                while (it.hasNext()) {
                    // This automatically performs the wrapping
                    wrapper.addReplyTo((AID) it.next());
                }

            wrapper.setLanguage(msg.getLanguage());
            wrapper.setOntology(msg.getOntology());
            wrapper.setProtocol(msg.getProtocol());
            wrapper.setInReplyTo(msg.getInReplyTo());
          wrapper.setReplyWith(msg.getReplyWith());
            wrapper.setConversationId(msg.getConversationId());
            wrapper.setReplyByDate(msg.getReplyByDate());
            if (msg.hasByteSequenceContent()) {
                wrapper.setByteSequenceContent(msg.getByteSequenceContent());
            }
            else {
                wrapper.setContent(msg.getContent());
            }
            wrapper.setEncoding(msg.getEncoding());

            //FIXME: Message Envelope is missing
            }
        }
        return wrapper;
    }

    /**
     * This method is redefined so that the sender AID is automatically
     * wrapped into an OntoAID
     */
    public void setSender(AID aid) {
        super.setSender(OntoAID.wrap(aid));
    }

    /**
     * This method is redefined so that the receiver AID is automatically
     * wrapped into an OntoAID
     */
    public void addReceiver(AID aid) {
        super.addReceiver(OntoAID.wrap(aid));
    }

    /**
     * This method is redefined so that the replyTo AID is automatically
     * wrapped into an OntoAID
     */
    public void addReplyTo(AID aid) {
        super.addReplyTo(OntoAID.wrap(aid));
    }

    // FIXME: clone method should be redefined too

}

