/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.content.abs;

import rca.content.schema.*;

/**
 * An Abstract descriptor that can hold an Identifying
 * Referential Expression (IRE).
 * Note that an IRE is both a content element (as in the case of
 * a QUERY-REF communicative act) and a Term (as in the case of
 * (== (X) (iota ?x P(?x))
 * Original author Paola Turci, Federico Bergenti - Universita` di Parma
 */
public class AbsIRE extends AbsObjectImpl implements AbsContentElement, AbsTerm {

    private boolean isAContentExpression = false;
    /**
     * Construct an Abstract descriptor to hold a IRE of
     * the proper type (e.g. ANY, IOTA, ALL...).
     * @param typeName The name of the type of the IRE held by 
     * this abstract descriptor.
     */
    public AbsIRE(String typeName) {
        super(typeName);
    }

    /**
     * Sets the variable of this IRE.
     * @param variable The abstract descriptor holding the variable.
     */
    public void setVariable(AbsVariable variable) {
        set(IRESchema.VARIABLE, variable);
    }

    /**
     * Sets the sequence of variables of this IRE.
     * @param variables The abstract descriptor holding the sequence of variables.
     */
    public void setVariables(AbsAggregate variables) {
        set(IRESchema.VARIABLE, variables);
    }

    /**
     * Sets the variable term of this IRE.
     * @param t The abstract descriptor holding the variable or sequence of variables.
     */
    public void setTerm(AbsTerm t) {
        if (t instanceof AbsVariable) {
            setVariable((AbsVariable) t);
        }
        else if (t instanceof AbsAggregate) {
            setVariables((AbsAggregate) t);
        }
        else {
            throw new IllegalArgumentException("Invalid term "+t+" for an AbsIRE");
        }
    }

    /**
     * Sets the proposition of this IRE.
     * @param proposition The abstract descriptor holding the proposition.
     */
    public void setProposition(AbsPredicate proposition) {
        set(IRESchema.PROPOSITION, proposition);
    }

    /**
     * Gets the variable of this IRE.
     * @return the abstract descriptor holding the variable of this IRE.
     */
    public AbsVariable getVariable() {
        return (AbsVariable) getAbsObject(IRESchema.VARIABLE);
    }

    /**
     * Gets the sequence of variables of this IRE.
     * @return the abstract descriptor holding the sequence of variables of this IRE.
     */
    public AbsAggregate getVariables() {
        return (AbsAggregate) getAbsObject(IRESchema.VARIABLE);
    }

    /**
     * Gets the variable term of this IRE.
     * @return the abstract descriptor holding the variable term of this IRE.
     */
    public AbsTerm getTerm() {
        return (AbsTerm) getAbsObject(IRESchema.VARIABLE);
    }

    /**
     * Gets the proposition of this IRE.
     * @return the abstract descriptor holding the proposition of this IRE.
     */
    public AbsPredicate getProposition() {
        return (AbsPredicate) getAbsObject(IRESchema.PROPOSITION);
    }

    /**
     * Redefine the <code>isGrounded()</code> method in order to 
     * always return <code>false</code>. Infact an IRE always
     * includes a variable.
     */
    public boolean isGrounded() {
            return false;
    }

    // Easy way to access the Java class representing AbsIRE.
         // Useful in MIDP where XXX.class is not available
        private static Class absIREClass = null;
         public static Class getJavaClass() {
             if (absIREClass == null) {
                 try {
                     absIREClass = Class.forName("rca.content.abs.AbsIRE");
                 }
                 catch (Exception e) {
                     // Should never happen
                     e.printStackTrace();
                 }
             }
             return absIREClass;
         }


      /**
       * @see AbsContent.isAContentExpression
       */
      public boolean isAContentExpression() {
          return isAContentExpression;
      }

      /**
       * @see AbsContent.setIsAContentExpression(boolean flag)
       */
      public void setIsAContentExpression(boolean flag) {
          isAContentExpression = flag;
      }
}

