/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.content.onto;
//#MIDP_EXCLUDE_FILE

import rca.content.abs.AbsObject;
import rca.content.abs.AbsPrimitive;
import rca.content.abs.AbsConcept;
import rca.content.schema.PrimitiveSchema;
import rca.content.schema.ConceptSchema;

import org.apache.commons.codec.binary.Base64;
import java.io.*;

/**
   This ontology allows dealing with Java Serializable objects as
   if they were instances of ontological elements.
   <br>
   <b>NOT available in MIDP</b>
   <br>
   Original author Giovanni Caire - TILAB
 */
public class SerializableOntology extends Ontology {
  // The singleton instance of this ontology
  private static final SerializableOntology theInstance = new SerializableOntology();
  
  private static final String SERIALIZABLE = "serializable";
  private static final String SERIALIZABLE_VALUE = "value";
  
  /**
   * Returns the singleton instance of the <code>SerializableOntology</code>.
   * @return the singleton instance of the <code>SerializableOntology</code>
   */
  public static Ontology getInstance() {
    return theInstance;
  }
  
  /**
     Construct a SerializableOntology object
   */
  private SerializableOntology() {
    super("Serializable-ontology", (Ontology) null, null);
  	try {
	    // Add the primitive schema for binary data
	    PrimitiveSchema stringSchema = (PrimitiveSchema) BasicOntology.getInstance().getSchema(BasicOntology.STRING);
	    add(stringSchema);
	    // Add the schema for a generic Serializable object
	    ConceptSchema serializableSchema = new ConceptSchema(SERIALIZABLE);
	    serializableSchema.add(SERIALIZABLE_VALUE, stringSchema);
	    add(serializableSchema);
  	}
  	catch (Exception e) {
  		// Should never happen
  		e.printStackTrace();
  	}
  }
            
	//#APIDOC_EXCLUDE_BEGIN
  /**
   */
  protected Object toObject(AbsObject abs, String lcType, Ontology globalOnto)
    			throws UnknownSchemaException, UngroundedException, OntologyException {
		if (SERIALIZABLE.equals(abs.getTypeName())) {
			try {
				AbsPrimitive absValue = (AbsPrimitive) abs.getAbsObject(SERIALIZABLE_VALUE);
				String stringValue = absValue.getString();
				byte[] value = Base64.decodeBase64(stringValue.getBytes("US-ASCII"));
	      ObjectInput in = new ObjectInputStream(new ByteArrayInputStream(value));
	      return in.readObject();
			}
			catch (Throwable t) {
				throw new OntologyException("Error in object deserialization.", t);
			}
		}
		else {
			throw new OntologyException("Abs-object "+abs+" is not serializable");
		}
	}
  	
  /**
   */
  protected AbsObject fromObject(Object obj, Ontology globalOnto) 
  			throws UnknownSchemaException, OntologyException {
  				
    // If obj is already an abstract descriptor --> just return it
  	if (obj instanceof AbsObject) {
  		return (AbsObject) obj;
  	}
  	
  	if (obj instanceof Serializable) {
  		try {
    		ByteArrayOutputStream baos = new ByteArrayOutputStream();
    		ObjectOutputStream out = new ObjectOutputStream(baos);
    		out.writeObject(obj);
    		AbsConcept absSerializable = new AbsConcept(SERIALIZABLE);
				String stringValue = new String(Base64.encodeBase64(baos.toByteArray()), "US-ASCII");
    		absSerializable.set(SERIALIZABLE_VALUE, stringValue);
    		return absSerializable;
  		}
			catch (Throwable t) {
				throw new OntologyException("Error in object deserialization.", t);
			}
		}
		else {
			throw new OntologyException("Object "+obj+" is not serializable");
		}
  } 
	//#APIDOC_EXCLUDE_END
}
