/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.content.schema;

import rca.content.abs.*;
import rca.content.onto.*;

/**
 * The class to be used to define schemas of agent actions in 
 * an ontology.
 * Note that an AgentActionSchema should also be a ConceptSchema, but
 * this inheritance relation is cut as Java does not support
 * multiple inheritance. As a consequence in practice it will 
 * not be possible to define e.g. a ConceptSchema with a slot 
 * whose value must be instances of a certain type of agent-action 
 * even if in theory this should be
 * possible as a ConceptSchema can have slots of type term and 
 * an agent-action is a concept and therefore a term.
 * Original author Federico Bergenti - Universita` di Parma
 */
public class AgentActionSchema extends ConceptSchema {
    public static final String         BASE_NAME = "AgentAction";
    private static AgentActionSchema baseSchema = new AgentActionSchema();

    /**
     * Construct a schema that vinculates an entity to be a generic
     * agent action
     */
    private AgentActionSchema() {
        super(BASE_NAME);
    }

    /**
     * Creates an <code>AgentActionSchema</code> with a given type-name.
     * @param typeName The name of this <code>AgentActionSchema</code>.
     */
    public AgentActionSchema(String typeName) {
        super(typeName);
    }

    /**
     * Retrieve the generic base schema for all agent actions.
     * @return the generic base schema for all agent actions.
     */
    public static ObjectSchema getBaseSchema() {
        return baseSchema;
    } 

    /**
     * Add a mandatory slot of type PredicateSchema to this schema. 
     * @param name The name of the slot.
     * @param slotSchema The schema of the slot.
     */
    public void add(String name, PredicateSchema slotSchema) {
        super.add(name, slotSchema);
    } 

    /**
     * Add a slot of type PredicateSchema to this schema. 
     * @param name The name of the slot.
     * @param slotSchema The schema of the slot.
     * @param optionality The optionality, i.e. <code>OPTIONAL</code> 
     * or <code>MANDATORY</code>
     */
    public void add(String name, PredicateSchema slotSchema, int optionality) {
        super.add(name, slotSchema, optionality);
    } 

    /**
     * Creates an Abstract descriptor to hold an agent action of
     * the proper type.
     */
    public AbsObject newInstance() throws OntologyException {
        return new AbsAgentAction(getTypeName());
    } 

		/**
	     Check whether a given abstract descriptor complies with this 
	     schema.
	     @param abs The abstract descriptor to be checked
	     @throws OntologyException If the abstract descriptor does not 
	     complies with this schema
	   */
  	public void validate(AbsObject abs, Ontology onto) throws OntologyException {
			// Check the type of the abstract descriptor
  		if (!(abs instanceof AbsAgentAction)) {
				throw new OntologyException(abs+" is not an AbsAgentAction");
			}
			
			// Check the slots
			validateSlots(abs, onto);
  	}
  	
  	/**
  	   Return true if 
  	   - s is the base schema for the XXXSchema class this schema is
  	     an instance of (e.g. s is ConceptSchema.getBaseSchema() and this 
  	     schema is an instance of ConceptSchema)
  	   - s is the base schema for a super-class of the XXXSchema class
  	     this schema is an instance of (e.g. s is TermSchema.getBaseSchema()
  	     and this schema is an instance of ConceptSchema.
  	   Moreover, as AgentActionSchema extends GenericActionSchema, but should
  	   also extend ConceptSchema (this is not possible in practice as
  	   Java does not support multiple inheritance), this method
  	   returns true also in the case that s is equals to, or is an
  	   ancestor of, ConceptSchema.getBaseSchema() (i.e. TermSchema.getBaseSchema()
  	   descends from s)
  	 */
  	protected boolean descendsFrom(ObjectSchema s) {
			if (s != null) {
	  	 	if (s.equals(getBaseSchema())) {
		  		return true;
  			}
  			if (super.descendsFrom(s)) {
  				return true;
  			}
  			return ContentElementSchema.getBaseSchema().descendsFrom(s);
			}
			else {
				return false;
			}
  	}
}
