/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.content.schema;

import rca.content.abs.*;
import rca.content.onto.*;
import rca.util.leap.Iterator;

/**
 * Original author Federico Bergenti - Universita` di Parma
 */
public class ContentElementListSchema extends ContentElementSchema {
    public static final String BASE_NAME = "ContentElementList";
    private static ContentElementListSchema baseSchema = new ContentElementListSchema();

    /**
     * Construct a schema that vinculates an entity to be a content element 
     * list. Note that there are no different types of content element 
     * list as it 
     * happens for concepts (e.g. Person, Address...), IREs (e.g. IOTA,
     * ANY, ALL...) and the like. Therefore there is no ContentElementListSchema
     * constructor that takes a String parameter.
     */
    private ContentElementListSchema() {
        super(BASE_NAME);
    }

    /**
     * Retrieve the generic base schema for all content element lists.
     * @return the generic base schema for all content element lists.
     */
    public static ObjectSchema getBaseSchema() {
        return baseSchema;
    } 

    /**
     * Creates an Abstract descriptor to hold a content element list 
     */
    public AbsObject newInstance() throws OntologyException {
        return new AbsContentElementList();
    } 
    
		/**
	     Check whether a given abstract descriptor complies with this 
	     schema.
	     @param abs The abstract descriptor to be checked
	     @throws OntologyException If the abstract descriptor does not 
	     complies with this schema
	   */
  	public void validate(AbsObject abs, Ontology onto) throws OntologyException {
  		// Check the type of the abstract descriptor
  		if (!(abs instanceof AbsContentElementList)) {
  			throw new OntologyException(abs+" is not an AbsContentElementList");
  		}
  		
  		// Validate the elements in the content element list against 
  		// their schemas.
  		// Note that there is no need to check that these schemas are
  		// compliant with ContentElementSchema.getBaseSchema() because the
  		// AbsContentElementList class already forces that.
  		AbsContentElementList list = (AbsContentElementList) abs;
  		Iterator it = list.iterator();
  		while (it.hasNext()) {
  			AbsContentElement el = (AbsContentElement) it.next();
  			ObjectSchema s = onto.getSchema(el.getTypeName());
  			s.validate(el, onto);
  		}
  	}
  	
  	/**
  	   Return true if 
  	   - s is the base schema for the XXXSchema class this schema is
  	     an instance of (e.g. s is ConceptSchema.getBaseSchema() and this 
  	     schema is an instance of ConceptSchema)
  	   - s is the base schema for a super-class of the XXXSchema class
  	     this schema is an instance of (e.g. s is TermSchema.getBaseSchema()
  	     and this schema is an instance of ConceptSchema)
  	 */
  	protected boolean descendsFrom(ObjectSchema s) {
  		if (s != null) {
  			if (s.equals(getBaseSchema())) {
	  			return true;
  			}
  			return super.descendsFrom(s);
  		}
  		else {
  			return false;
  		}
  	}
}
