/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.content.schema;

import rca.content.abs.*;
import rca.content.onto.*;

/**
 * This class represent the schema of primitive entities in
 * an ontology.
 * Original author Federico Bergenti - Universita` di Parma
 */
public class PrimitiveSchema extends TermSchema {
    public static final String         BASE_NAME = "Primitive";
    private static PrimitiveSchema baseSchema = new PrimitiveSchema();

    /**
     * Construct a schema that vinculates an entity to be a generic
     * primitive element
     */
    private PrimitiveSchema() {
        super(BASE_NAME);
    }

    /**
     * Creates a <code>PrimitiveSchema</code> with a given type-name.
     *
     * @param typeName The name of this <code>PrimitiveSchema</code>.
     */
    public PrimitiveSchema(String typeName) {
        super(typeName);
    }

    /**
     * Retrieve the generic base schema for all primitives.
     *
     * @return the generic base schema for all primitives.
     */
    public static ObjectSchema getBaseSchema() {
        return baseSchema;
    } 

    /**
     * Creates an Abstract descriptor to hold a primitive of
     * the proper type.
     */
    public AbsObject newInstance() throws OntologyException {
        return new AbsPrimitive(getTypeName());
    } 

		/**
	     Check whether a given abstract descriptor complies with this 
	     schema.
	     @param abs The abstract descriptor to be checked
	     @throws OntologyException If the abstract descriptor does not 
	     complies with this schema
	   */
  	public void validate(AbsObject abs, Ontology onto) throws OntologyException {
			// Check the type of the abstract descriptor
  		if (!(abs instanceof AbsPrimitive)) {
				throw new OntologyException(abs+" is not an AbsPrimitive");
			}
  	}
  	
  	/**
  	   Return true if 
  	   - s is the base schema for the XXXSchema class this schema is
  	     an instance of (e.g. s is ConceptSchema.getBaseSchema() and this 
  	     schema is an instance of ConceptSchema)
  	   - s is the base schema for a super-class of the XXXSchema class
  	     this schema is an instance of (e.g. s is TermSchema.getBaseSchema()
  	     and this schema is an instance of ConceptSchema)
  	 */
  	protected boolean descendsFrom(ObjectSchema s) {
  		if (s != null) {
  			if (s.equals(getBaseSchema())) {
	  			return true;
  			}
  			return super.descendsFrom(s);
  		}
  		else {
  			return false;
  		}
  	}
}

