/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#APIDOC_EXCLUDE_FILE

import rca.util.leap.Iterator;
import rca.util.leap.List;
import rca.util.leap.ArrayList;

import rca.util.Logger;

import rca.lang.acl.ACLMessage;
import rca.core.behaviours.Behaviour;

import rca.core.messaging.GenericMessage;
import rca.core.management.AgentManagementSlice;

import rca.domain.FIPANames;
import rca.domain.FIPAAgentManagement.InternalError;
import rca.domain.JADEAgentManagement.JADEManagementOntology;

import rca.mtp.MTPDescriptor;
import rca.mtp.TransportAddress;

import rca.security.JADESecurityException;
import rca.security.CredentialsHelper;
import rca.security.JADEPrincipal;
import rca.security.Credentials;


/**
 This class is a concrete implementation of the rca agent
 container, providing runtime support to rca agents.
 
 This class cannot be instantiated from applications. Instead, the
 <code>Runtime.createAgentContainer(Profile p)</code> method must be called.
 
 @see Runtime#createAgentContainer(Profile)
 
 Original author Giovanni Rimassa - Universita' di Parma
 Original author Jerome Picault - Motorola Labs
 Original author Giovanni Caire - TILAB
 @version $Date: 2006-01-20 15:01:50 +0100 (Fri, 20 Jan 2006) $ $Revision: 5853 $
 
 */
class AgentContainerImpl implements AgentContainer, AgentToolkit {
	
	public static final String ENABLE_MONITOR = "jade_core_AgentContainerImpl_enablemonitor";
	public static final String MONITOR_AGENT_NAME = "monitor-%C";
	public static final String MONITOR_AGENT_CLASS = "rca.core.ContainerMonitorAgent";
	
	private Logger myLogger = Logger.getMyLogger(this.getClass().getName());
	
	// Local agents, indexed by agent name
	protected LADT localAgents;
	
	// The Profile defining the configuration of this Container
	protected Profile myProfile;
	
	// The Command Processor through which all the vertical commands in this container will pass
	protected CommandProcessor myCommandProcessor;
	
	//#MIDP_EXCLUDE_BEGIN
	// The agent platform this container belongs to
	protected MainContainerImpl myMainContainer; // FIXME: It should go away
	//#MIDP_EXCLUDE_END
	
	// The IMTP manager, used to access IMTP-dependent functionalities
	protected IMTPManager myIMTPManager;
	
	// The platform Service Manager
	private ServiceManager myServiceManager;
	
	// The platform Service Finder
	private ServiceFinder myServiceFinder;
	
	// The Object managing Thread resources in this container
	private ResourceManager myResourceManager;
	
	protected ContainerID myID;
	protected NodeDescriptor myNodeDescriptor;
	
	// These are only used at bootstrap-time to initialize the local
	// NodeDescriptor. Further modifications take no effect
	protected JADEPrincipal ownerPrincipal;
	protected Credentials ownerCredentials;
	
	private AID theAMS;
	private AID theDefaultDF;
	
	
	
	// Default constructor
	AgentContainerImpl() {
	}
	
	// Package scoped constructor, so that only the Runtime
	// class can actually create a new Agent Container.
	AgentContainerImpl(Profile p) {
		myProfile = p;
		localAgents = new LADT(16);
	}
	
	//#MIDP_EXCLUDE_BEGIN
	/////////////////////////////////////////////////
	// Support for the in-process interface section
	/////////////////////////////////////////////////
	rca.wrapper.AgentContainer getContainerController() {
		return getContainerController(myNodeDescriptor.getOwnerPrincipal(), myNodeDescriptor.getOwnerCredentials());
	}
	
	public rca.wrapper.AgentContainer getContainerController(JADEPrincipal principal, Credentials credentials) {
		return new rca.wrapper.AgentContainer(getContainerProxy(principal, credentials), this, getPlatformID());
	}
	
	/**
	 Return a proxy that allows making requests to the local container
	 as if they were received from the main. This allows dealing
	 uniformly with local and remote requests.
	 Local requests occurs at bootstrap and following calls to the
	 in-process interface.
	 */
	private rca.wrapper.ContainerProxy getContainerProxy(final JADEPrincipal principal, final Credentials credentials) {
		return new rca.wrapper.ContainerProxy() {
			GenericCommand dummyCmd = new GenericCommand(null, null, null);
			
			{
				dummyCmd.setPrincipal(principal);
				dummyCmd.setCredentials(credentials);
			}
			
			public void createAgent(AID id, String className, Object[] args) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				AgentManagementSlice target = (AgentManagementSlice) getProxyToLocalSlice(AgentManagementSlice.NAME);
				target.createAgent(id, className, args, principal, null, AgentManagementSlice.CREATE_ONLY, dummyCmd);
			}
			
			public void killContainer() throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				AgentManagementSlice target = (AgentManagementSlice) getProxyToLocalSlice(AgentManagementSlice.NAME);
				// FIXME: set Principal and Credentials
				target.exitContainer();
			}
			
			public MTPDescriptor installMTP(String address, String className) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				rca.core.messaging.MessagingSlice target = (rca.core.messaging.MessagingSlice) getProxyToLocalSlice(rca.core.messaging.MessagingSlice.NAME);
				// FIXME: set Principal and Credentials
				return target.installMTP(address, className);
			}
			
			public void uninstallMTP(String address) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				rca.core.messaging.MessagingSlice target = (rca.core.messaging.MessagingSlice) getProxyToLocalSlice(rca.core.messaging.MessagingSlice.NAME);
				// FIXME: set Principal and Credentials
				target.uninstallMTP(address);
			}
			
			public void suspendAgent(AID id) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				rca.core.management.AgentManagementSlice target = (rca.core.management.AgentManagementSlice) getProxyToLocalSlice(rca.core.management.AgentManagementSlice.NAME);
				// FIXME: set Principal and Credentials
				target.changeAgentState(id, Agent.AP_SUSPENDED);
			}
			
			public void activateAgent(AID id) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				rca.core.management.AgentManagementSlice target = (rca.core.management.AgentManagementSlice) getProxyToLocalSlice(rca.core.management.AgentManagementSlice.NAME);
				// FIXME: set Principal and Credentials
				target.changeAgentState(id, Agent.AP_ACTIVE);
			}
			
			public void killAgent(AID id) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				rca.core.management.AgentManagementSlice target = (rca.core.management.AgentManagementSlice) getProxyToLocalSlice(rca.core.management.AgentManagementSlice.NAME);
				target.killAgent(id, dummyCmd);
			}
			
			public void moveAgent(AID id, Location where) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				rca.core.mobility.AgentMobilitySlice target = (rca.core.mobility.AgentMobilitySlice) getProxyToLocalSlice(rca.core.mobility.AgentMobilitySlice.NAME);
				// FIXME: set Principal and Credentials
				target.moveAgent(id, where);
			}
			
			public void cloneAgent(AID id, Location where, String newName) throws Throwable {
				// Do as if it was a remote call from the main to allows
				// security checks to take place if needed
				rca.core.mobility.AgentMobilitySlice target = (rca.core.mobility.AgentMobilitySlice) getProxyToLocalSlice(rca.core.mobility.AgentMobilitySlice.NAME);
				// FIXME: set Principal and Credentials
				target.copyAgent(id, where, newName);
			}
			
			private SliceProxy getProxyToLocalSlice(String serviceName) throws Throwable {
				Service svc = myServiceFinder.findService(serviceName);
				return (SliceProxy) myIMTPManager.createSliceProxy(serviceName, svc.getHorizontalInterface(), myIMTPManager.getLocalNode());
			}
		};
	}
	////////////////////////////////////////////////////////
	// END of support for the in-process interface section
	////////////////////////////////////////////////////////
	//#MIDP_EXCLUDE_END
	
	/**
	 Issue an INFORM_CREATED vertical command.
	 Note that the Principal, if any, is that of the
	 owner of the newly born agent, while the Credentials, if any, are a set
	 additional initial credentials to be attached to the newly born agent. 
	 The SecurityService, if active,
	 will create a new Principal for the newly born agent and will initialize its 
	 credentials as the union of the initial credential and the ownership 
	 certificate.
	 */
	public void initAgent(
			AID agentID, Agent instance,
			JADEPrincipal ownerPrincipal, Credentials initialCredentials)
	throws NameClashException, IMTPException, NotFoundException, JADESecurityException {
		
		//replaces wildcards
		agentID.setName(JADEManagementOntology.adjustAgentName(agentID.getName(), new String[] {myID.getName()}));
		
		// Setting the AID and toolkit here is redundant, but
		// allows services to retrieve their agent helper correctly
		// when processing the INFORM_CREATED command.
		instance.setAID(agentID);
		instance.setToolkit(this);
		
		GenericCommand cmd = new GenericCommand(rca.core.management.AgentManagementSlice.INFORM_CREATED, rca.core.management.AgentManagementSlice.NAME, null);
		cmd.addParam(agentID);
		cmd.addParam(instance);
		cmd.addParam( ownerPrincipal );
		cmd.addParam( initialCredentials );
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof NameClashException) {
				throw ((NameClashException) ret);
			}
			else if (ret instanceof IMTPException) {
				throw ((IMTPException) ret);
			}
			else if (ret instanceof NotFoundException) {
				throw ((NotFoundException) ret);
			}
			else if (ret instanceof JADESecurityException) {
				throw ((JADESecurityException) ret);
			}
			else if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	
	public NodeDescriptor getNodeDescriptor() {
		return myNodeDescriptor;
	}
	
	protected void init() throws IMTPException, ProfileException {
		myCommandProcessor = myProfile.getCommandProcessor();
		
		// Create and initialize the IMTPManager
		myIMTPManager = myProfile.getIMTPManager();
		myIMTPManager.initialize(myProfile);
		
		// Get the Service Manager and the Service Finder
		myServiceManager = myProfile.getServiceManager();
		myServiceFinder = myProfile.getServiceFinder();
		
		// Attach CommandProcessor and ServiceManager to the local node
		BaseNode localNode = (BaseNode) myIMTPManager.getLocalNode();
		localNode.setCommandProcessor(myCommandProcessor);
		localNode.setServiceManager(myServiceManager);
		
		//#MIDP_EXCLUDE_BEGIN
		myMainContainer = myProfile.getMain();
		//#MIDP_EXCLUDE_END
		
		// This string will be used to build the GUID for every agent on
		// this platform.
		AID.setPlatformID(myServiceManager.getPlatformName());
		
		// Build the Agent IDs for the AMS and for the Default DF.
		theAMS = new AID(FIPANames.AMS, AID.ISLOCALNAME);
		theDefaultDF = new AID(FIPANames.DEFAULT_DF, AID.ISLOCALNAME);
		
		// Create the ResourceManager
		myResourceManager = myProfile.getResourceManager();
		
		// Initialize the Container ID
		TransportAddress addr = (TransportAddress) myIMTPManager.getLocalAddresses().get(0);
		myID = new ContainerID(myProfile.getParameter(Profile.CONTAINER_NAME, PlatformManager.NO_NAME), addr);
		myNodeDescriptor = new NodeDescriptor(myID, myIMTPManager.getLocalNode());
	}
	
	/**
	 Add the node to the platform with the basic services
	 */
	protected void startNode() throws IMTPException, ProfileException, ServiceException, JADESecurityException, NotFoundException {
		// Initialize all services (without activating them)
		List services = new ArrayList();
		ServiceDescriptor dsc = startService("rca.core.management.AgentManagementService", false);
		dsc.setMandatory(true);
		services.add(dsc);
		//#MIDP_EXCLUDE_BEGIN
		dsc = startService("rca.core.messaging.MessagingService", false);
		//#MIDP_EXCLUDE_END
		/*#MIDP_INCLUDE_BEGIN
		 dsc = startService("rca.core.messaging.LightMessagingService", false);
		 #MIDP_INCLUDE_END*/
		dsc.setMandatory(true);
		services.add(dsc);
		
		List l = myProfile.getSpecifiers(Profile.SERVICES);
		myProfile.setSpecifiers(Profile.SERVICES, l); // Avoid parsing services twice
		Iterator serviceSpecifiers = l.iterator();
		while(serviceSpecifiers.hasNext()) {
			Specifier s = (Specifier) serviceSpecifiers.next();
			String serviceClass = s.getClassName();
			boolean isMandatory = false;
			if ( s.getArgs() != null ) {
				isMandatory = CaseInsensitiveString.equalsIgnoreCase( (String) s.getArgs()[0], "true" );
			}
			try {
				dsc = startService(serviceClass, false);
				dsc.setMandatory(isMandatory);
				services.add(dsc);
			} 
			catch (ServiceException se) {
				if (isMandatory) {
					throw se;
				}
				else {
					myLogger.log(Logger.WARNING,"Exception initializing service " + serviceClass + " : " + se.toString());
					se.printStackTrace();
				}
			}
		}
		
		// Register with the platform
		ServiceDescriptor[] descriptors = new ServiceDescriptor[services.size()];
		for (int i = 0; i < descriptors.length; ++i) {
			descriptors[i] = (ServiceDescriptor) services.get(i);
		}
		// This call performs the real connection to the platform and can modify the 
		// name of this container
		myServiceManager.addNode(myNodeDescriptor, descriptors);
		
		// coda all services
		for (int i = 0; i < descriptors.length; ++i) {
			
			ServiceDescriptor currentServDesc = descriptors[i];
			try {
				currentServDesc.getService().boot(myProfile);
			} 
			catch(Throwable t) {
				if ( currentServDesc.isMandatory() ) {
					throw new ServiceException("An error occurred during service booting", t);
				}
				else {
					myLogger.log(Logger.WARNING,"Exception booting service " + currentServDesc.getName() + " : " + t.toString());
					t.printStackTrace();
				}
			}
		}
		
		//#MIDP_EXCLUDE_BEGIN
		// If we are the master main container --> start the AMS and DF
		if(myMainContainer != null) {
			boolean startThem = !myProfile.getBooleanProperty(Profile.LOCAL_SERVICE_MANAGER, false);
			myMainContainer.initSystemAgents(this, startThem);
		}
		//#MIDP_EXCLUDE_END
	}
	
	boolean joinPlatform() {
		//#J2ME_EXCLUDE_BEGIN
		// Redirect output if the -output option is specified
		String output = myProfile.getParameter("output", null);
		if (output != null) {
			try {
				rca.util.PerDayFileLogger fl = new rca.util.PerDayFileLogger(output);
				rca.util.PrintStreamSplitter pss = new rca.util.PrintStreamSplitter(System.out, fl);
				System.setOut(pss);
				System.setErr(pss);
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
		//#J2ME_EXCLUDE_END
		
		try {
			// Perform the initial setup from the profile
			init();
			
			// Connect the local node to the platform and activate all the services
			startNode();
		}
		catch (IMTPException imtpe) {
			myLogger.log(Logger.SEVERE,"Communication failure while joining agent platform: " + imtpe.getMessage());
			imtpe.printStackTrace();
			endContainer();
			cleanIMTPManager();
			return false;
		}
		catch (JADESecurityException ae) {
			myLogger.log(Logger.SEVERE,"Authentication or authorization failure while joining agent platform.");
			ae.printStackTrace();
			endContainer();
			cleanIMTPManager();
			return false;
		}
		catch (Exception e) {
			myLogger.log(Logger.SEVERE,"Some problem occurred while joining agent platform.");
			e.printStackTrace();
			endContainer();
			cleanIMTPManager();
			return false;
		}
		
		// Create and activate agents that must be launched at bootstrap
		startBootstrapAgents();
		
//		myLogger.log(Logger.INFO, "--------------------------------------\nAgent container " + myID + " is ready.\n--------------------------------------------");
		return true;
	}
	
	private void cleanIMTPManager() {
		// In case container startup failed, we clean IMTPManager resources. 
		// This is important when the JVM is not killed on rca termination.
		if (myIMTPManager != null) {
			myIMTPManager.shutDown();
		}
	}
	
	private void startBootstrapAgents() {
		try {
			List l = myProfile.getSpecifiers(Profile.AGENTS);
			Iterator agentSpecifiers = l.iterator();
			while(agentSpecifiers.hasNext()) {
				Specifier s = (Specifier) agentSpecifiers.next();
				if (s.getName() != null) {
					AID agentID = new AID(s.getName(), AID.ISLOCALNAME);
					
					try {
						//#MIDP_EXCLUDE_BEGING
						getContainerProxy(myNodeDescriptor.getOwnerPrincipal(), myNodeDescriptor.getOwnerCredentials()).createAgent(agentID, s.getClassName(), s.getArgs());
						//#MIDP_EXCLUDE_END
						/*#MIDP_INCLUDE_BEGIN
						 String serviceName = rca.core.management.AgentManagementSlice.NAME;
						 Service svc = myServiceFinder.findService(serviceName);
						 rca.core.management.AgentManagementSlice target = (rca.core.management.AgentManagementSlice) myIMTPManager.createSliceProxy(serviceName, svc.getHorizontalInterface(), myIMTPManager.getLocalNode());
						 GenericCommand dummyCmd = new GenericCommand(null, null, null);
						 dummyCmd.setPrincipal(myNodeDescriptor.getOwnerPrincipal());
						 dummyCmd.setCredentials(myNodeDescriptor.getOwnerCredentials());
						 target.createAgent(agentID, s.getClassName(), s.getArgs(), myNodeDescriptor.getOwnerPrincipal(), null, target.CREATE_ONLY, dummyCmd);
						 #MIDP_INCLUDE_END*/
					}
					catch (Throwable t) {
						myLogger.log(Logger.SEVERE,"Cannot create agent "+s.getName()+": "+t.getMessage());
					}
				}
				else {
					myLogger.log(Logger.WARNING,"Cannot create an agent with no name. Class was "+s.getClassName());
				}              	
			}
			
			// Now activate all agents (this call starts their embedded threads)
			AID[] allLocalNames = localAgents.keys();
			for (int i = 0; i < allLocalNames.length; i++) {
				AID id = allLocalNames[i];
				
				if(!id.equals(theAMS) && !id.equals(theDefaultDF)) {
					try {
						powerUpLocalAgent(id);
					}
					catch (NotFoundException nfe) {
						// Should never happen
						nfe.printStackTrace();
					}
				}
			}
			
			//#J2ME_EXCLUDE_BEGIN
			// If the -rca_core_AgentContainerImpl_enablemonitor option is specified activate a ContainerMonitorAgent
			if (myProfile.getBooleanProperty(ENABLE_MONITOR, false)) {
				AID monitorId = new AID(MONITOR_AGENT_NAME, AID.ISLOCALNAME);
				try {
					getContainerProxy(myNodeDescriptor.getOwnerPrincipal(), myNodeDescriptor.getOwnerCredentials()).createAgent(monitorId, MONITOR_AGENT_CLASS, new Object[]{this, localAgents});
					powerUpLocalAgent(monitorId);
				}
				catch (Throwable t) {
					myLogger.log(Logger.WARNING, "Cannot strat ContainerMonitor agent. Check that the misc add-on is in the classpath.");
				}
			}
			//#J2ME_EXCLUDE_END
		}
		catch (ProfileException pe) {
			myLogger.log(Logger.WARNING, "Error reading initial agents. "+pe);
			pe.printStackTrace();
		}
	}
	
	public void shutDown() {
		// Remove all non-system agents
		Agent[] allLocalAgents = localAgents.values();
		
		for(int i = 0; i < allLocalAgents.length; i++) {
			// Kill agent and wait for its termination
			Agent a = allLocalAgents[i];
			
			// Skip the Default DF and the AMS
			AID id = a.getAID();
			if(id.equals(getAMS()) || id.equals(getDefaultDF()))
				continue;
			
			//System.out.println("Killing agent "+a.getLocalName());
			//System.out.flush();
			a.doDelete();
			//System.out.println("Done. Waiting for its termination...");
			//System.out.flush();
			a.join();
			//System.out.println("Agent "+a.getLocalName()+" terminated");
			//System.out.flush();
			a.resetToolkit();
		}
		
		try {
			myServiceManager.removeNode(myNodeDescriptor);
			myIMTPManager.shutDown();
		}
		catch(IMTPException imtpe) {
			imtpe.printStackTrace();
		}
		catch(ServiceException se) {
			se.printStackTrace();
		}
		
		// Releases Thread resources
		myResourceManager.releaseResources();
		
		// Notify the rca Runtime that the container has terminated execution
		endContainer();
	}
	
	// calls Runtime.instance().endContainer()
	// with the security priviledges of AgentContainerImpl
	// no matter priviledges of who originaltely triggered this action
	private void endContainer() {
		try {
			Runtime.instance().endContainer();
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	
	////////////////////////////////////////////
	// AgentToolkit interface implementation
	////////////////////////////////////////////
	
	public Location here() {
		return myID;
	}
	
	/**
	 Issue a SEND_MESSAGE VerticalCommand for each receiver
	 */
	public void handleSend(ACLMessage msg, AID sender, boolean needClone) {
		Iterator it = msg.getAllIntendedReceiver();
		// If there are multiple receivers the message must always be cloned
		// since the MessageManager will modify it. If there is a single 
		// receiver we clone it or not depending on the needClone parameter
		boolean isFirst = true;
		while (it.hasNext()) {
			AID receiver = (AID)it.next();
			if (isFirst) {
				needClone = needClone || it.hasNext();
				isFirst = false;
			}
			GenericCommand cmd = new GenericCommand(rca.core.messaging.MessagingSlice.SEND_MESSAGE, rca.core.messaging.MessagingSlice.NAME, null);
			cmd.addParam(sender);
			ACLMessage toBeSent = null;
			if (needClone) {
				toBeSent = (ACLMessage) msg.clone();
			}
			else {
				toBeSent = msg;
			}
			isFirst = false;
			GenericMessage gmsg = new GenericMessage(toBeSent);
			cmd.addParam(gmsg);
			cmd.addParam(receiver);
			// Set the credentials of the sender
			initCredentials(cmd, sender);
			Object ret = myCommandProcessor.processOutgoing(cmd);
			if (ret != null) {
				if (ret instanceof Throwable) {
					// The SEND_MESSAGE VerticalCommand was blocked by some Filter 
					// before reaching the Messaging Souce Sink --> Issue
					// a NOTIFY_FAILURE VerticalCommand to notify the sender
					cmd = new GenericCommand(rca.core.messaging.MessagingSlice.NOTIFY_FAILURE, rca.core.messaging.MessagingSlice.NAME, null);
					cmd.addParam(gmsg);
					cmd.addParam(receiver);
					cmd.addParam(new InternalError("Message blocked: "+ret));
					ret = myCommandProcessor.processOutgoing(cmd);
					if (ret != null) {
						if (ret instanceof Throwable) {
							((Throwable) ret).printStackTrace();
						}
					}
				}
			}
		}
		
	}
	
	//#MIDP_EXCLUDE_BEGIN
	// FIXME: to be removed
	public void handlePosted(AID agentID, ACLMessage msg) {
		GenericCommand cmd = new GenericCommand(rca.core.event.NotificationSlice.NOTIFY_POSTED, rca.core.event.NotificationSlice.NAME, null);
		cmd.addParam(msg);
		cmd.addParam(agentID);
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	/**
	 Issue a NOTIFY_RECEIVED VerticalCommand
	 */
	public void handleReceived(AID agentID, ACLMessage msg) {
		GenericCommand cmd = new GenericCommand(rca.core.event.NotificationSlice.NOTIFY_RECEIVED, rca.core.event.NotificationSlice.NAME, null);
		cmd.addParam(msg);
		cmd.addParam(agentID);
		// No security check is meaningful on this action --> don't even set the Credentials
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
		
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	public void handleBehaviourAdded(AID agentID, Behaviour b) {
		GenericCommand cmd = new GenericCommand(rca.core.event.NotificationSlice.NOTIFY_BEHAVIOUR_ADDED, rca.core.event.NotificationSlice.NAME, null);
		cmd.addParam(agentID);
		cmd.addParam(b);
		// No security check is meaningful on this action --> don't even set the Credentials
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	public void handleBehaviourRemoved(AID agentID, Behaviour b) {
		GenericCommand cmd = new GenericCommand(rca.core.event.NotificationSlice.NOTIFY_BEHAVIOUR_REMOVED, rca.core.event.NotificationSlice.NAME, null);
		cmd.addParam(agentID);
		cmd.addParam(b);
		// No security check is meaningful on this action --> don't even set the Credentials
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	public void handleChangeBehaviourState(AID agentID, Behaviour b, String from, String to) {
		GenericCommand cmd = new GenericCommand(rca.core.event.NotificationSlice.NOTIFY_CHANGED_BEHAVIOUR_STATE, rca.core.event.NotificationSlice.NAME, null);
		cmd.addParam(agentID);
		cmd.addParam(b);
		cmd.addParam(from);
		cmd.addParam(to);
		// No security check is meaningful on this action --> don't even set the Credentials
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	// FIXME: to be removed
	public void handleChangedAgentPrincipal(AID agentID, JADEPrincipal oldPrincipal, Credentials creds) {
		
		/***
		 
		 myNotificationManager.fireEvent(NotificationManager.CHANGED_AGENT_PRINCIPAL,
		 new Object[]{agentID, oldPrincipal, (AgentPrincipal)certs.getIdentityCertificate().getSubject()});
		 try {
		 myPlatform.changedAgentPrincipal(agentID, certs);
		 }
		 catch (IMTPException re) {
		 re.printStackTrace();
		 }
		 catch (NotFoundException nfe) {
		 nfe.printStackTrace();
		 }
		 
		 ***/
	}
	//#MIDP_EXCLUDE_END
	
	
	public void handleChangedAgentState(AID agentID, int oldState, int newState) {
		AgentState from = AgentState.getInstance(oldState);
		AgentState to = AgentState.getInstance(newState);
		
		GenericCommand cmd = new GenericCommand(rca.core.management.AgentManagementSlice.INFORM_STATE_CHANGED, rca.core.management.AgentManagementSlice.NAME, null);
		cmd.addParam(agentID);
		cmd.addParam(from);
		cmd.addParam(to);
		// No security check is meaningful on this action --> don't even set the Credentials
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	
	public void handleEnd(AID agentID) {
		GenericCommand cmd = new GenericCommand(rca.core.management.AgentManagementSlice.INFORM_KILLED, rca.core.management.AgentManagementSlice.NAME, null);
		cmd.addParam(agentID);
		// Set the credentials of the terminating agent
		initCredentials(cmd, agentID);
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	
	/* FIXME: Needed due to the Persistence Service being an add-on
	 public void handleSave(AID agentID, String repository) throws ServiceException, NotFoundException, IMTPException {
	 GenericCommand cmd = new GenericCommand("Save-Myself", "rca.core.persistence.Persistence", null);
	 cmd.addParam(agentID);
	 cmd.addParam(repository);
	 // Set the credentials of the agent to be saved
	  initCredentials(cmd, agentID);
	  
	  Object ret = myCommandProcessor.processOutgoing(cmd);
	  if (ret != null) {
	  if (ret instanceof ServiceException) {
	  throw ((ServiceException) ret);
	  }
	  else if (ret instanceof NotFoundException) {
	  throw ((NotFoundException) ret);
	  }
	  else if (ret instanceof IMTPException) {
	  throw ((IMTPException) ret);
	  }
	  else if (ret instanceof Throwable) {
	  ((Throwable) ret).printStackTrace();
	  }
	  }
	  }*/
	
	// FIXME: Needed due to the Persistence Service being an add-on
	/*public void handleReload(AID agentID, String repository) throws ServiceException, NotFoundException, IMTPException {
	 GenericCommand cmd = new GenericCommand("Reload-Myself", "rca.core.persistence.Persistence", null);
	 cmd.addParam(agentID);
	 cmd.addParam(repository);
	 // Set the credentials of the agent to be reloaded
	  initCredentials(cmd, agentID);
	  
	  Object ret = myCommandProcessor.processOutgoing(cmd);
	  if (ret != null) {
	  if (ret instanceof ServiceException) {
	  throw ((ServiceException) ret);
	  }
	  else if (ret instanceof NotFoundException) {
	  throw ((NotFoundException) ret);
	  }
	  else if (ret instanceof IMTPException) {
	  throw ((IMTPException) ret);
	  }
	  else if (ret instanceof Throwable) {
	  ((Throwable) ret).printStackTrace();
	  }
	  }
	  }*/
	
	// FIXME: Needed due to the Persistence Service being an add-on
	/*public void handleFreeze(AID agentID, String repository, ContainerID bufferContainer) throws ServiceException, NotFoundException, IMTPException {
	 GenericCommand cmd = new GenericCommand("Freeze-Myself", "rca.core.persistence.Persistence", null);
	 cmd.addParam(agentID);
	 cmd.addParam(repository);
	 cmd.addParam(bufferContainer);
	 // Set the credentials of the agent to be frozen
	  initCredentials(cmd, agentID);
	  
	  Object ret = myCommandProcessor.processOutgoing(cmd);
	  if (ret != null) {
	  if (ret instanceof ServiceException) {
	  throw ((ServiceException) ret);
	  }
	  else if (ret instanceof NotFoundException) {
	  throw ((NotFoundException) ret);
	  }
	  else if (ret instanceof IMTPException) {
	  throw ((IMTPException) ret);
	  }
	  else if (ret instanceof Throwable) {
	  ((Throwable) ret).printStackTrace();
	  }
	  }
	  }*/
	
	public void setPlatformAddresses(AID id) {
		GenericCommand cmd = new GenericCommand(rca.core.messaging.MessagingSlice.SET_PLATFORM_ADDRESSES, rca.core.messaging.MessagingSlice.NAME, null);
		cmd.addParam(id);
		// No security check is meaningful on this action --> don't even set the Credentials
		
		Object ret = myCommandProcessor.processOutgoing(cmd);
		if (ret != null) {
			if (ret instanceof Throwable) {
				((Throwable) ret).printStackTrace();
			}
		}
	}
	
	public AID getAMS() {
		return (AID)theAMS.clone();
	}
	
	public AID getDefaultDF() {
		return (AID)theDefaultDF.clone();
	}
	
	public String getProperty(String key, String aDefault) {
		return myProfile.getParameter(key, aDefault);
	}
	
	public ServiceHelper getHelper(Agent a, String serviceName) throws ServiceException {
		try {
			
			// Retrieve the service
			Service s = myServiceFinder.findService(serviceName);
			if(s == null) {
				throw new ServiceNotActiveException(serviceName);
			}
			
			return s.getHelper( a );
		}
		catch (IMTPException imtpe) {
			throw new ServiceException(" ServiceHelper could not be created for: " + serviceName, imtpe);
		}
	}
	
	
	// Private and package scoped methods
	
	
	
	/**
	 */
	public String getPlatformID() {
		return AID.getPlatformID();
	}
	
	public Agent addLocalAgent(AID id, Agent a) {
		
		a.setToolkit(this);
		return localAgents.put(id, a);
	}
	
	public void powerUpLocalAgent(AID agentID) throws NotFoundException {
		Agent instance = localAgents.acquire(agentID);
		if (instance == null) {
			throw new NotFoundException("powerUpLocalAgent() failed to find agent "+agentID.getName());
		}
		Thread t = myResourceManager.getThread(ResourceManager.USER_AGENTS, agentID.getLocalName(), instance);
		instance.powerUp(agentID, t);
		localAgents.release(agentID);
	}
	
	public void removeLocalAgent(AID id) {
		localAgents.remove(id);
	}
	
	public Agent acquireLocalAgent(AID id) {
		return localAgents.acquire(id);
	}
	
	public void releaseLocalAgent(AID id) {
		localAgents.release(id);
	}
	
	public AID[] agentNames() {
		return localAgents.keys();
	}
	
	//#MIDP_EXCLUDE_BEGIN
	public void fillListFromMessageQueue(List messages, Agent a) {
		MessageQueue mq = a.getMessageQueue();
		
		synchronized(mq) {
			Iterator i = mq.iterator();
			while (i.hasNext()) {
				messages.add(i.next());
			}
		}
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	public void fillListFromReadyBehaviours(List behaviours, Agent a) {
		
		Scheduler s = a.getScheduler();
		
		// (Mutual exclusion with Scheduler.add(), remove()...)
		synchronized (s) {
			Iterator it = s.readyBehaviours.iterator();
			while (it.hasNext()) {
				Behaviour b = (Behaviour) it.next();
				behaviours.add(new BehaviourID(b));
			}
			
		}
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	public void fillListFromBlockedBehaviours(List behaviours, Agent a) {
		
		Scheduler s = a.getScheduler();
		
		// (Mutual exclusion with Scheduler.add(), remove()...)
		synchronized (s) {
			Iterator it = s.blockedBehaviours.iterator();
			while (it.hasNext()) {
				Behaviour b = (Behaviour) it.next();
				behaviours.add(new BehaviourID(b));
			}
		}
	}
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	/*public void commitMigration(Agent instance) {
	 instance.doGone();
	 localAgents.remove(instance.getAID());
	 }*/
	//#MIDP_EXCLUDE_END
	
	//#MIDP_EXCLUDE_BEGIN
	/*public void abortMigration(Agent instance) {
	 instance.doExecute();
	 }*/
	//#MIDP_EXCLUDE_END
	
	public void addAddressToLocalAgents(String address) {
		Agent[] allLocalAgents = localAgents.values();
		
		// Add the address to the AIDs of all local agents
		for(int j = 0; j < allLocalAgents.length; j++) {
			allLocalAgents[j].addPlatformAddress(address);
		}
		
		// Add the new addresses to the AMS and Default DF AIDs
		theAMS.addAddresses(address);
		theDefaultDF.addAddresses(address);
	}
	
	public void removeAddressFromLocalAgents(String address) {
		Agent[] allLocalAgents = localAgents.values();
		
		// Remove the address from the AIDs of all local agents
		for(int j = 0; j < allLocalAgents.length; j++) {
			allLocalAgents[j].removePlatformAddress(address);
		}
		
		// Remove the address from the AIDs of the AMS and the Default DF
		theAMS.removeAddresses(address);
		theDefaultDF.removeAddresses(address);
	}
	
	public boolean postMessageToLocalAgent(ACLMessage msg, AID receiverID) {
		Agent receiver = localAgents.acquire(receiverID);
		if(receiver == null) {
			return false;
		}
		receiver.postMessage(msg);
		localAgents.release(receiverID);
		
		return true;
	}
	
	// Tells whether the given AID refers to an agent of this platform
	// or not.
	public boolean livesHere(AID id) {
		String hap = id.getHap();
		return CaseInsensitiveString.equalsIgnoreCase(hap, AID.getPlatformID());
	}
	
	public ContainerID getID() {
		return myID;
	}
	
	public MainContainer getMain() {
		//#MIDP_EXCLUDE_BEGIN
		return myMainContainer;
		//#MIDP_EXCLUDE_END
		/*#MIDP_INCLUDE_BEGIN
		 return null;
		 #MIDP_INCLUDE_END*/
	}
	
	public ServiceManager getServiceManager() {
		return myServiceManager;
	}
	
	public ServiceFinder getServiceFinder() {
		return myServiceFinder;
	}
	
	// Utility method to start a kernel service
	protected ServiceDescriptor startService(String name, boolean activateIt) throws ServiceException {
		
		try {
			Class svcClass = Class.forName(name);
			Service svc = (Service)svcClass.newInstance();
			svc.init(this, myProfile);
			ServiceDescriptor dsc = new ServiceDescriptor(svc.getName(), svc);
			
			if (activateIt) {
				myServiceManager.activateService(dsc);
				svc.boot(myProfile);
			}
			return dsc;
		}
		catch(ServiceException se) {
			// Let it through
			throw se;
		}
		catch(Throwable t) {
			throw new ServiceException("An error occurred during service activation", t);
		}
	}
	
	protected void stopService(String name) throws ServiceException {
		try {
			myServiceManager.deactivateService(name);
		}
		catch(ServiceException se) {
			// Let it through
			throw se;
		}
		catch(Throwable t) {
			throw new ServiceException("An error occurred during service deactivation", t);
		}
	}
	
	public void becomeLeader() {
		//#MIDP_EXCLUDE_BEGIN
		try {
			myMainContainer.startSystemAgents(this);
		}
		catch(Exception e) {
			e.printStackTrace();
		}
		//#MIDP_EXCLUDE_END
	}
	
	
	//#ALL_EXCLUDE_BEGIN
	//FIXME: These methods have been added to support
	// PlatformListener registration from the In-process-interface
	// with minimum effort. They will possibly be removed in a
	// future (more general) implementation
	public void addPlatformListener(AgentManager.Listener l) throws ClassCastException {
		AgentManager m = (AgentManager) myMainContainer;
		m.addListener(l);
	}
	
	public void removePlatformListener(AgentManager.Listener l) throws ClassCastException {
		AgentManager m = (AgentManager) myMainContainer;
		m.removeListener(l);
	}
	//#ALL_EXCLUDE_END
	
	private void initCredentials(Command cmd, AID id) {
		//#MIDP_EXCLUDE_BEGIN
		Agent agent = localAgents.acquire(id);
		if (agent != null) {
			try {
				CredentialsHelper ch = (CredentialsHelper) agent.getHelper("rca.core.security.Security");
				cmd.setPrincipal(ch.getPrincipal());
				cmd.setCredentials(ch.getCredentials());
			}
			catch (ServiceException se) {
				// The security plug-in is not there. Just ignore it
			}
		}
		localAgents.release(id);
		//#MIDP_EXCLUDE_END
	}
	
}
