/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#APIDOC_EXCLUDE_FILE
//#MIDP_EXCLUDE_FILE

import rca.util.leap.List;

import rca.core.ContainerID;

import rca.core.event.PlatformListener;
import rca.core.event.MTPListener;

import rca.mtp.MTPException;
import rca.mtp.MTPDescriptor;

import rca.security.JADESecurityException;
import rca.security.Credentials;
import rca.security.JADEPrincipal;

import rca.domain.FIPAAgentManagement.AMSAgentDescription;
import rca.domain.FIPAAgentManagement.AlreadyRegistered;
import rca.domain.FIPAAgentManagement.NotRegistered;

/**
Original author Giovanni Rimassa - Universita' di Parma
@version $Date: 2005-09-07 12:37:04 +0200 (Wed, 07 Sep 2005) $ $Revision: 5776 $
*/

/**
  This interface provides Agent Life Cycle management services to the
  platform AMS.
  */
public interface AgentManager {

  /**
     This callback interface is implemented by the AMS in order to be
     notified of significant platform-level events (e.g. container
     added or removed, agents birth or death, mtp configuration changes, etc.).
   */
  public static interface Listener extends PlatformListener, MTPListener {
  }

  // Listeners related methods
  void addListener(Listener l);
  void removeListener(Listener l);

  // Platform information retrieval methods
  ContainerID[] containerIDs();
  AID[] agentNames();
  List containerMTPs(ContainerID cid) throws NotFoundException;
  List containerAgents(ContainerID cid) throws NotFoundException;

  void addTool(AID tool);
  void removeTool(AID tool);
  AID[] agentTools();
  
  ContainerID getContainerID(AID agentID) throws NotFoundException;
  AMSAgentDescription getAMSDescription(AID agentID) throws NotFoundException;
    
  // rca actions method
    /**
     * Create an agent on the given container. If the container-name is null
     * then the agent is created on the main-container.
     * @throws NotFoundException if the passed container does not exist
     * @throws UnreachableException if the container is unreachable 
     * @throws JADESecurityException if this action is not authorized
     **/
  void create(String agentName, String className, Object arguments[], ContainerID cid, JADEPrincipal owner, Credentials initialCredentials, JADEPrincipal requesterPrincipal, Credentials requesterCredentials) throws UnreachableException, JADESecurityException, NotFoundException, NameClashException;
  void kill(AID agentID, JADEPrincipal requesterPrincipal, Credentials requesterCredentials) throws NotFoundException, UnreachableException, JADESecurityException;
  void suspend(AID agentID) throws NotFoundException, UnreachableException, JADESecurityException;
  void activate(AID agentID) throws NotFoundException, UnreachableException, JADESecurityException;
  void wait(AID agentID, String password) throws NotFoundException, UnreachableException;
  void wake(AID agentID, String password) throws NotFoundException, UnreachableException;

  void move(AID agentID, Location where) throws NotFoundException, UnreachableException, JADESecurityException;
  void copy(AID agentID, Location where, String newAgentName) throws NotFoundException, NameClashException, UnreachableException, JADESecurityException;

  void shutdownPlatform(JADEPrincipal requesterPrincipal, Credentials requesterCredentials) throws JADESecurityException;
  void killContainer(ContainerID cid, JADEPrincipal requesterPrincipal, Credentials requesterCredentials) throws NotFoundException, UnreachableException, JADESecurityException;
  MTPDescriptor installMTP(String address, ContainerID cid, String className) throws NotFoundException, UnreachableException, MTPException;
  void uninstallMTP(String address, ContainerID cid) throws NotFoundException, UnreachableException, MTPException;

  void take(AID agentID, String username, byte[] password) throws NotFoundException, UnreachableException, JADESecurityException;

  void sniffOn(AID snifferName, List toBeSniffed) throws NotFoundException, UnreachableException;
  void sniffOff(AID snifferName, List toBeSniffed) throws NotFoundException, UnreachableException;
  void debugOn(AID debuggerName, List toBeDebugged) throws NotFoundException, UnreachableException;
  void debugOff(AID debuggerName, List toBeDebugged) throws NotFoundException, UnreachableException;

  // FIPA actions method
  void amsRegister(AMSAgentDescription dsc) throws AlreadyRegistered, JADESecurityException;
  void amsDeregister(AMSAgentDescription dsc) throws NotRegistered, JADESecurityException;
  void amsModify(AMSAgentDescription dsc) throws NotRegistered, NotFoundException, UnreachableException, JADESecurityException; 
  List amsSearch(AMSAgentDescription template, long maxResults);
}

