/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#APIDOC_EXCLUDE_FILE

import rca.util.leap.Map;
import rca.util.leap.HashMap;
import rca.util.leap.Serializable;
import rca.util.Logger;

/**
   This class provides a partial implementation of the
   <code>Node</code> interface. Concrete IMTPs will have to provide a
   full implementation of the <code>Node</code> interface, possibly by
   subclassing this class.

   Original author Giovanni Rimassa - FRAMeTech s.r.l.

 */
public abstract class BaseNode implements Node, Serializable {

	private Logger myLogger;
	
    public BaseNode(String name, boolean hasPM) {
	myName = name;
	hasLocalPM = hasPM;
	localSlices = new HashMap(5);
	myLogger = Logger.getMyLogger(getClass().getName());
    }

    public void setName(String name) {
	myName = name;
    }

    public String getName() {
	return myName;
    }

    public boolean hasPlatformManager() {
	return hasLocalPM;
    }

    public void exportSlice(String serviceName, Service.Slice localSlice) {
	localSlices.put(serviceName, localSlice);
    }

    public void unexportSlice(String serviceName) {
	localSlices.remove(serviceName);
    }

    protected Service.Slice getSlice(String serviceName) {
	return (Service.Slice)localSlices.get(serviceName);
    }


    /**
       Provides an IMTP independent implementation for the
       horizontal command serving mechanism. IMTP dependent implementations 
       of the accept() method should invoke this method.
       @param cmd The horizontal command to process.
       @return The object that is the result of processing the command.
       @throws ServiceException If the service the command belongs to
       is not present on this node.
     */
    public Object serveHorizontalCommand(HorizontalCommand cmd) throws ServiceException {

	String serviceName = cmd.getService();
	String commandName = cmd.getName();
	Object[] commandParams = cmd.getParams();
	
	if (myLogger.isLoggable(Logger.FINE)) {
		myLogger.log(Logger.FINE, "Node "+myName+" serving incoming H-Command "+commandName+" of Service "+serviceName);
	}

	// Look up in the local slices table and find the slice to dispatch to
	Service.Slice slice = getSlice(serviceName);

	if(slice != null) {
		Object ret = null;
    VerticalCommand vCmd = slice.serve(cmd);

    if(vCmd != null) {
		  vCmd.setPrincipal(cmd.getPrincipal());
		  vCmd.setCredentials(cmd.getCredentials());
			// Hand it to the command processor
			if (myLogger.isLoggable(Logger.FINE)) {
				myLogger.log(Logger.FINE, "Node "+myName+" issuing incoming V-Command "+vCmd.getName()+" of Service "+vCmd.getService());
			}
			serveVerticalCommand(vCmd);
			ret = vCmd.getReturnValue();
    }
    else {
			ret = cmd.getReturnValue();
    }
    
    if (ret != null) {
			if (myLogger.isLoggable(Logger.FINE)) {
				myLogger.log(Logger.FINE, "Node "+myName+" return value for incoming H-Command "+commandName+" of Service "+serviceName+" = "+ret);
			}
    }
    return ret;
	}
	else {
		String s = new String("Node "+myName+": Service "+serviceName+" Unknown. Command = "+commandName);
	  throw new ServiceException("-- "+s+" --");
	}
    }

    public void setCommandProcessor(CommandProcessor cp) {
	processor = cp;
    }

    public void setServiceManager(ServiceManager mgr) {
    	myServiceManager = mgr;
    }
    
    public void platformManagerDead(String deadPMAddr, String notifyingPMAddr) throws IMTPException {
    	if (deadPMAddr.equals(notifyingPMAddr)) {
    		// This is a PlatformManager that recovered from a fault
    		((ServiceManagerImpl) myServiceManager).reattach(notifyingPMAddr);
    	}
    	else {
	    	myServiceManager.addAddress(notifyingPMAddr);
	    	myServiceManager.removeAddress(deadPMAddr);
    	}
    }
    
    /**
       Serves an incoming vertical command, locally. This method is
       invoked if a new <code>VerticalCommand</code> object is
       generated by a slice targetted by a former
       <code>HorizontalCommand</code>, which happens if the
       <code>Slice.serve()</code> yields a non-null result.

       This method makes it so that the newly created vertical command
       is handed to the command processor to first pass through all
       incoming filters and then to be dispatched to its proper
       incoming command sink.
       @param cmd The vertical command to process.
       @return The object that is the result of processing the command.
       @throws ServiceException If some problem occurs.
     */
    private Object serveVerticalCommand(VerticalCommand cmd) throws ServiceException {
	if(processor == null) {
	    throw new ServiceException("No command processor for node <" + getName() + ">");
	}

	return processor.processIncoming(cmd);
    }

    protected transient ServiceManager myServiceManager;
    private transient CommandProcessor processor;
    
    // The name of this node
    private String myName;
    
    // True if a local copy of the Platform Manager is deployed at this Node
    private boolean hasLocalPM = false;

    // A map, indexed by service name, of all the local slices of this
    // node. This map is used to dispatch incoming commands to the
    // service they belong to.
    private transient Map localSlices;
}
