/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#MIDP_EXCLUDE_FILE

import rca.mtp.MTPDescriptor;

import rca.util.leap.Iterator;
import rca.util.leap.List;
import rca.util.leap.LinkedList;
import rca.util.leap.Map;
import rca.util.leap.HashMap;

import rca.util.Logger;

class ContainerTable {

  // Initial size of containers hash table
  private static final int CONTAINERS_SIZE = 10;

  private Logger myLogger = Logger.getMyLogger(getClass().getName());
  
  private static class Entry {
    private List mtps = new LinkedList();

    public Entry() {
    }

    public void addMTP(MTPDescriptor mtp) {
      mtps.add(mtp);
    }

    public void removeMTP(MTPDescriptor mtp) {
      mtps.remove(mtp);
    }

    public List getMTPs() {
      return mtps;
    }

  } // End of Entry class


  private Map entries = new HashMap(CONTAINERS_SIZE);

  public synchronized void addContainer(ContainerID cid) {
    Entry e = new Entry();
    entries.put(cid, e);
  }

  public synchronized void addMTP(ContainerID cid, MTPDescriptor mtp) throws NotFoundException {
    Entry e = (Entry)entries.get(cid);
    if(e == null)
      throw new NotFoundException("No container named " + cid.getName() + " was found.");
    List l = e.getMTPs();
    l.add(mtp);
  }

  public synchronized void removeContainer(ContainerID cid) {
    entries.remove(cid);
    notifyAll();
  }

  public synchronized void removeMTP(ContainerID cid, MTPDescriptor mtp) throws NotFoundException {
    Entry e = (Entry)entries.get(cid);
    if(e == null)
      throw new NotFoundException("No container named " + cid.getName() + " was found.");
    List l = e.getMTPs();
    l.remove(mtp);
  }

  public synchronized List getMTPs(ContainerID cid) throws NotFoundException {
    Entry e = (Entry)entries.get(cid);
    if(e == null)
      throw new NotFoundException("No container named " + cid.getName() + " was found.");
    return e.getMTPs();
  }

  public int size() {
    return entries.size();
  }

  public synchronized ContainerID[] names() {
    ContainerID[] result = new ContainerID[entries.size()];
    Iterator it = entries.keySet().iterator();
    int i = 0;
    while(it.hasNext()) {
      result[i++] = (ContainerID) it.next();
    }
    return result;
  }

  synchronized boolean waitForRemoval(ContainerID cid, long timeout) {
  	try {
	    while(entries.containsKey(cid)) {
	      wait(timeout);
	      if (entries.containsKey(cid)) {
	      	myLogger.log(Logger.WARNING, "Container "+cid.getName()+" did not terminate when requested to do so.");
	      	return false;
	      }
		  }
  	}
  	catch (InterruptedException ie) {
    	myLogger.log(Logger.WARNING, "Interrupted while waiting for container "+cid.getName()+" termination");
    	return false;
  	}
  	return true;
  }

  synchronized boolean waitUntilEmpty(long timeout) {
		long time = System.currentTimeMillis();
		long deadline = time + timeout;
  	try {
	    while(!entries.isEmpty()) {
	    	if (timeout > 0 && time >= deadline) {
	    		// Timeout expired
	      	myLogger.log(Logger.WARNING, "Some entries still present in container table");
	    		break;
	    	}
	      wait(deadline - time);
	      time = System.currentTimeMillis();
		  }
  	}
  	catch (InterruptedException ie) {
    	myLogger.log(Logger.WARNING, "Interrupted while waiting for container table to be empty");
  	}
  	return entries.isEmpty();  
  }

}
