/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#APIDOC_EXCLUDE_FILE


import rca.util.leap.List;
import rca.util.leap.LinkedList;


/**

   A generic implementation of the <code>Command</code> interface,
   operating at the meta-level to provide a generic transformation of
   method invocations.
   Generic commands can be used both as vertical and horizontal commands.
   Objects of this class can also be recycled to build <i>command
   pools</i>, since every information they hold can be rewritten.

   Original author Giovanni Rimassa - FRAMeTech s.r.l.
*/
public class GenericCommand implements HorizontalCommand, VerticalCommand  {

    /**
       Creates a new generic command, with the given name and
       belonging to the given service and interaction.

       @param name The name of this command.
       @param svc The name of the service this command belongs to.
       @param interaction The identifier of the service interation this command belongs to.
    */
    public GenericCommand(String name, String service, String interaction) {
	myName = name;
	myService = service;
	myInteraction = interaction;

	params = new LinkedList();
    }


    public String getName() {
	return myName;
    }

    public String getService() {
	return myService;
    }

    public String getInteraction() {
	return myInteraction;
    }

    /**
       Add a new parameter to this command object.
       @param obj The parameter to add. The actual type of
       <code>obj</code> must be such that can be delivered over the
       network by the concrete platform IMTP.

       @see rca.core.IMTPManager
    */
    public void addParam(Object obj) {
	params.add(obj);
    }

    /**
       Remove a parameter from this command object.
       @param obj The parameter to remove.
    */
    public void removeParam(Object obj) {
	params.remove(obj);
    }
    
    public void clear() {
    	params.clear();
    }
    
    public void setReturnValue(Object rv) {
	returnValue = rv;
    }

    public Object[] getParams() {
	return params.toArray();
    }

    public Object getReturnValue() {
	return returnValue;
    }

     public rca.security.JADEPrincipal getPrincipal() {
       return principal;
     } // end getPrincipal
     public void setPrincipal( rca.security.JADEPrincipal p) {
       // principal can be set only one time
       // following setPrincipal are ignored
       if (principal==null) {
         this.principal = p;
       }
     } // end setPrincipal
 
     public rca.security.Credentials getCredentials() {
       return creds;
     } // end getCredentials
     public void setCredentials( rca.security.Credentials creds) {
       // credentials can be set only one time
       // following setCredentials are ignored
       if (this.creds==null) {
         this.creds = creds;
       }
     } // end setCredentials
 
 
 
    private rca.security.JADEPrincipal principal = null;
    private rca.security.Credentials creds;

    private final String myName;
    private final String myService;
    private final String myInteraction;

    private final List params;
    private Object returnValue;

}
