/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#APIDOC_EXCLUDE_FILE

import rca.security.JADESecurityException;
import rca.util.leap.Serializable;

import java.io.InterruptedIOException;

/**
   Common bas class for all agent life cycle states.
   Note that a single LifeCycle object may represent more than
   one agent state.
   Original author Giovanni Caire - TILAB
 */
public abstract class LifeCycle implements Serializable {
	protected Agent myAgent;
	protected int myState;
	
	public LifeCycle(int s) {
		myState = s;
	}
	
	/**
	   This is only called by the Agent class.
	 */
	void setAgent(Agent a) {
		myAgent = a;
	}
	
	/**
	   This method is invoked when the agent thread starts
	 */	   
	public void init() {
	}
	

	/**
	   This method actually implements what the agent has to do in this 
	   LifeCycle state
	 */
	public void execute() throws JADESecurityException, InterruptedException, InterruptedIOException {
	}
	
	/**
	   This method is invoked when the agent thread termiantes
	 */	   
	public void end() {
	}
	
	/**
	   Specifies whether or not this is a termination state
	 */	   
	public boolean alive() {
		return true;
	}
	
	public int getState() {
		return myState;
	}
	
	/**
	   This method is invoked as soon as we enter this LifeCycle state.
	   When this method is executed this LifeCycle object is the current
	   agent LifeCycle.
	 */
	public void transitionFrom(LifeCycle from) {	
	}
	
	/**
	   This method is invoked just before leaving this LifeCycle state.
	   When this method is executed this LifeCycle object is still the 
	   current agent LifeCycle.
	   The boolean return value is used to prevent transitions 
	   to certain states. In that case the life cycle state does not 
	   change.
	 */
	public boolean transitionTo(LifeCycle to) {
		return false;
	}
	
	/**
	 * Specifies whether or not the agent should react to incoming messages when in this 
	 * LifeCycle state.  
	 */
	public boolean isMessageAware() {
		return false;
	}
	
	public boolean equals(Object obj) {
		if (obj instanceof LifeCycle) {
			return myState == ((LifeCycle) obj).myState;
		}
		return false;
	}
}

