/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#APIDOC_EXCLUDE_FILE
import rca.util.leap.Serializable;

/**

   The <code>ServiceDescriptor</code> class serves as a meta-level
   description of a kernel-level service.
   Instances of this class contain a <code>Service</code> object,
   along with its name and properties, and are used in service
   management operations, as well as in agent-level introspection of
   platform-level entities.

   Original author Giovanni Rimassa - FRAMeTech s.r.l.

   @see rca.core.Service

*/
public class ServiceDescriptor implements Serializable {

    /**
       Builds a new service descriptor, describing the given service
       with the given name and properties.

       @param sn The name of the described service.
       @param svc The described <code>Service</code> object.
    */
    public ServiceDescriptor(String sn, Service svc) {
	myName = sn;
	myService = svc;
	myIsMandatory = false;
    }

    /**
       Builds an uninitialized service descriptor.

       @see rca.core.ServiceDescriptor#setName(String sn)
       @see rca.core.ServiceDescriptor#setService(Service svc)
    */
    public ServiceDescriptor() {
	this(null, null);
    }

    /**
       Change the name (if any) of the described service.

       @param sn The name to assign to the described service.
    */
    public void setName(String sn) {
	myName = sn;
    }

    /**
       Retrieve the name (if any) of the described service.

       @return The name of the described service, or <code>null</code>
       if no name was set.
    */
    public String getName() {
	return myName;
    }

    /**
       Change the described service (if any).

       @param svc The <code>Service</code> object that is to be
       described by this service descriptor.
    */
    public void setService(Service svc) {
	myService = svc;
    }

    /**
       Retrieve the described service.

       @return The <code>Service</code> object described by this
       service descriptor, or <code>null</code> if no service was set.
    */
    public Service getService() {
	return myService;
    }
    
    public void setMandatory(boolean isMandatory) {
    	myIsMandatory = isMandatory;
    }
    
    public boolean isMandatory() {
    	return myIsMandatory;
    }

    private String myName;
    private boolean myIsMandatory;
    private transient Service myService;

	//#MIDP_EXCLUDE_BEGIN
  private String serviceClass;
  private void writeObject(java.io.ObjectOutputStream out) throws java.io.IOException {
		if (myService != null) {
			serviceClass = myService.getClass().getName();
		}
    out.defaultWriteObject();
  }

  private void readObject(java.io.ObjectInputStream in) throws java.io.IOException, ClassNotFoundException {
    in.defaultReadObject();
		if (serviceClass != null) {
  		try {
  			myService = (Service) Class.forName(serviceClass).newInstance();
  		}
  		catch (ClassNotFoundException cnfe) {
  			throw cnfe;
  		}
  		catch (Throwable t) {
  			throw new java.io.IOException("Can't create service "+serviceClass+". "+t.getMessage());
  		}
		}			
  }
	//#MIDP_EXCLUDE_END
}
