/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core;

//#APIDOC_EXCLUDE_FILE


import java.util.Vector;

/**
This class implements the rca internal timing system. It should not
be used by application developers.
Original author Giovanni Rimassa - Universita' di Parma
@version $Date: 2004-11-23 10:08:25 +0100 (Tue, 23 Nov 2004) $ $Revision: 5447 $
*/

public class TimerDispatcher implements Runnable {
	// The singleton TimerDispatcher
	private static TimerDispatcher theDispatcher;
	
  private Thread myThread = null;
  private Vector timers = new Vector();
  //private SortedSet timers = new SortedSetImpl();
  private boolean active;

  void setThread(Thread t) {
    if(myThread == null) {
      myThread = t;
    }
  }

  public synchronized Timer add(Timer t) {
  	if (myThread == null) {
  		myThread = new Thread(this);
  		start();
  	}
  	while (!addTimer(t)) {
  		t.setExpirationTime(t.expirationTime()+1);
  	}
    // If this is the first timer, wake up the dispatcher thread
    if(timers.firstElement() == t) {
      notifyAll();
    }
    return t;
  }

  public synchronized void remove(Timer t) {
    timers.removeElement(t);
  }

  public void run() {
    try {
    	while (active) {
	    	Timer t = null;
    		synchronized(this) {
    			while (active) {
	    			long timeToWait = 0;
	    			if (!timers.isEmpty()) {
	    				t = (Timer) timers.firstElement();
	    				if (t.isExpired()) {
    						timers.removeElement(t);
	    					break;
	    				}
	    				else {
		      			timeToWait = t.expirationTime() - System.currentTimeMillis();
		      			if(timeToWait <= 0) {
		      				// Avoid wait(0), that means 'for ever'
									timeToWait = 1;
		      			}
	    				}
	    			}
	    			wait(timeToWait);
    			}
    		}
    		// This check just avoids NullPointerException on termination
    		if (active) {
	    		t.fire();
    		}
    	} 
    }
    catch(InterruptedException ie) {
      // Do nothing, but just return, since this is a shutdown.
    }
    timers.removeAllElements();
  }

  void start() {
  	synchronized(myThread) {
  		active = true;
      myThread.start();
    }
  }

  void stop() {
  	if (myThread != null) {
	    synchronized(myThread) {
	      if(Thread.currentThread().equals(myThread)) {
					System.out.println("Deadlock avoidance: TimerDispatcher thread calling stop on itself!");
	      }
	      else {
					active = false;
					synchronized (this) {
						notifyAll();
					}
					try {
		  			myThread.join();
					}
					catch (InterruptedException ignore) {
		  			// Do nothing
					}
	      }
	      myThread = null;
	    }
  	}
  }

  public static TimerDispatcher getTimerDispatcher() {
  	if (theDispatcher == null) {
  		theDispatcher = new TimerDispatcher();
  	}
  	return theDispatcher;
  }
  
  static void setTimerDispatcher(TimerDispatcher td) {
  	theDispatcher = td;
  }
  
  private boolean addTimer(Timer t) {
  	if (!timers.contains(t)) {
  		int size = timers.size();
  		for (int i = 0; i < size; i++) {
  			Timer t1 = (Timer) timers.elementAt(i);
  			if (t.expirationTime() < t1.expirationTime()) {
  				timers.insertElementAt(t, i);
  				return true;
  			}
  		}
  		timers.addElement(t);
  		return true;
  	}
  	return false;
  }
}
