/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core.event;


//#J2ME_EXCLUDE_FILE


import rca.core.HorizontalCommand;
import rca.core.VerticalCommand;
import rca.core.Service;
import rca.core.BaseService;
import rca.core.Sink;
import rca.core.Filter;
import rca.core.Node;

import rca.core.Agent;
import rca.core.AgentContainer;
import rca.core.MainContainer;
import rca.core.AID;
import rca.core.ContainerID;
import rca.core.BehaviourID;
import rca.core.AgentState;
import rca.core.Channel;
import rca.core.Profile;
import rca.core.ProfileException;
import rca.core.ServiceException;
import rca.core.IMTPException;
import rca.core.NotFoundException;

import rca.core.messaging.GenericMessage;

import rca.core.behaviours.Behaviour;

import rca.lang.acl.ACLMessage;
import rca.security.JADEPrincipal;
import rca.tools.ToolNotifier;

import rca.util.SynchList;
import rca.util.leap.Iterator;
import rca.util.leap.Map;
import rca.util.leap.HashMap;
import rca.util.leap.List;
import rca.util.leap.LinkedList;


/**
 
 The rca service to manage the event notification subsystem installed
 on the platform.
 
 Original author Giovanni Rimassa - FRAMeTech s.r.l.
 
 */
public class NotificationService extends BaseService {
	
	
	/**
	 The name of this service.
	 */
	static final String NAME = "rca.core.event.Notification";
	
	private static final String[] OWNED_COMMANDS = new String[] {
		NotificationSlice.SNIFF_ON,
		NotificationSlice.SNIFF_OFF,
		NotificationSlice.DEBUG_ON,
		NotificationSlice.DEBUG_OFF,
		NotificationSlice.NOTIFY_POSTED,
		NotificationSlice.NOTIFY_RECEIVED,
		NotificationSlice.NOTIFY_CHANGED_AGENT_STATE,
		NotificationSlice.NOTIFY_CHANGED_AGENT_PRINCIPAL,
		NotificationSlice.NOTIFY_BEHAVIOUR_ADDED,
		NotificationSlice.NOTIFY_BEHAVIOUR_REMOVED,
		NotificationSlice.NOTIFY_CHANGED_BEHAVIOUR_STATE	
	};
	
	
	// The special name of an auxiliary thread used to avoid deadlock when debugging the AMS
	private final static String AMS_DEBUG_HELPER = "AMS-debug-helper";
	
	public void init(AgentContainer ac, Profile p) throws ProfileException {
		super.init(ac, p);
		
		myContainer = ac;
		
		// Create a local slice
		localSlice = new ServiceComponent();
		
	}
	
	public String getName() {
		return NotificationSlice.NAME;
	}
	
	public Class getHorizontalInterface() {
		return NotificationSlice.class;
	}
	
	public Slice getLocalSlice() {
		return localSlice;
	}
	
	public Filter getCommandFilter(boolean direction) {
		if(direction == Filter.OUTGOING) {
			return localSlice;
		}
		else {
			return null;
		}
	}
	
	public Sink getCommandSink(boolean side) {
		return null;
	}
	
	public String[] getOwnedCommands() {
		return OWNED_COMMANDS;
	}
	
	
	/**
	 Inner mix-in class for this service: this class receives
	 commands through its <code>Filter</code> interface and serves
	 them, coordinating with remote parts of this service through
	 the <code>NotificationSlice</code> interface (that extends the
	 <code>Service.Slice</code> interface).
	 */
	private class ServiceComponent extends Filter implements NotificationSlice {
		
		
		// Implementation of the Filter interface
		
		public boolean accept(VerticalCommand cmd) {
			
			try {
				String name = cmd.getName();
				if(name.equals(SNIFF_ON)) {
					handleSniffOn(cmd);
				}
				if(name.equals(SNIFF_OFF)) {
					handleSniffOff(cmd);
				}
				else if(name.equals(DEBUG_ON)) {
					handleDebugOn(cmd);
				}
				else if(name.equals(DEBUG_OFF)) {
					handleDebugOff(cmd);
				}
				else if(name.equals(rca.core.messaging.MessagingSlice.SEND_MESSAGE)) {
					handleSendMessage(cmd);
				}
				else if(name.equals(NOTIFY_POSTED)) {
					handleNotifyPosted(cmd);
				}
				else if(name.equals(NOTIFY_RECEIVED)) {
					handleNotifyReceived(cmd);
				}
				else if(name.equals(rca.core.management.AgentManagementSlice.INFORM_STATE_CHANGED)) {
					handleNotifyChangedAgentState(cmd);
				}
				else if(name.equals(NOTIFY_CHANGED_AGENT_PRINCIPAL)) {
					handleNotifyChangedAgentPrincipal(cmd);
				}
				else if(name.equals(NOTIFY_BEHAVIOUR_ADDED)) {
					handleNotifyAddedBehaviour(cmd);
				}
				else if(name.equals(NOTIFY_BEHAVIOUR_REMOVED)) {
					handleNotifyRemovedBehaviour(cmd);
				}
				else if(name.equals(NOTIFY_CHANGED_BEHAVIOUR_STATE)) {
					handleNotifyChangedBehaviourState(cmd);
				}
			}
			catch(Throwable t) {
				cmd.setReturnValue(t);
			}
			
			// Never veto a command
			return true;
		}
		
		public void setBlocking(boolean newState) {
			// Do nothing. Blocking and Skipping not supported
		}
		
		public boolean isBlocking() {
			return false; // Blocking and Skipping not implemented
		}
		
		public void setSkipping(boolean newState) {
			// Do nothing. Blocking and Skipping not supported
		}
		
		public boolean isSkipping() {
			return false; // Blocking and Skipping not implemented
		}
		
		
		// Implementation of the Service.Slice interface
		
		public Service getService() {
			return NotificationService.this;
		}
		
		public Node getNode() throws ServiceException {
			try {
				return NotificationService.this.getLocalNode();
			}
			catch(IMTPException imtpe) {
				throw new ServiceException("Problem in contacting the IMTP Manager", imtpe);
			}
		}
		
		public VerticalCommand serve(HorizontalCommand cmd) {
			try {
				String cmdName = cmd.getName();
				Object[] params = cmd.getParams();
				
				if(cmdName.equals(H_SNIFFON)) {
					AID snifferName = (AID)params[0];
					AID targetName = (AID)params[1];
					
					sniffOn(snifferName, targetName);
				}
				else if(cmdName.equals(H_SNIFFOFF)) {
					AID snifferName = (AID)params[0];
					AID targetName = (AID)params[1];
					
					sniffOff(snifferName, targetName);
				}
				else if(cmdName.equals(H_DEBUGON)) {
					AID introspectorName = (AID)params[0];
					AID targetName = (AID)params[1];
					
					debugOn(introspectorName, targetName);
				}
				else if(cmdName.equals(H_DEBUGOFF)) {
					AID introspectorName = (AID)params[0];
					AID targetName = (AID)params[1];
					
					debugOff(introspectorName, targetName);
				}
			}
			catch(Throwable t) {
				cmd.setReturnValue(t);
			}
			
			if(cmd instanceof VerticalCommand) {
				return (VerticalCommand)cmd;
			}
			else {
				return null;
			}
		}
		
		// Implementation of the service-specific horizontal interface NotificationSlice
		
		public void sniffOn(AID snifferName, AID targetName) throws IMTPException {
			
			ToolNotifier tn = findNotifier(snifferName);
			if(tn == null) { // Need a new notifier 
				tn = new ToolNotifier(snifferName);
				AID id = new AID(snifferName.getLocalName() + "-on-" + myID().getName(), AID.ISLOCALNAME);
				try {
					myContainer.initAgent(id, tn, null, null); // FIXME: Modify to use a proper owner Principal
					myContainer.powerUpLocalAgent(id);
					addMessageListener(tn);
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
			tn.addObservedAgent(targetName);
			
		}
		
		public void sniffOff(AID snifferName, AID targetName) throws IMTPException {
			ToolNotifier tn = findNotifier(snifferName);
			if(tn != null) { 
				tn.removeObservedAgent(targetName);
				if(tn.isEmpty()) {
					removeMessageListener(tn);
					tn.doDelete();
				}
			}
		}
		
		public void debugOn(AID introspectorName, AID targetName) throws IMTPException {
			
			// AMS debug enabling must be done by a separated Thread to avoid
			// deadlock with ToolNotifier startup
			if (targetName.equals(myContainer.getAMS()) && !(Thread.currentThread().getName().equals(AMS_DEBUG_HELPER))) {
				final AID in = introspectorName;
				final AID tn = targetName;
				Thread helper = new Thread(new Runnable() {
					public void run() {
						try {
							debugOn(in, tn);
						}
						catch(IMTPException imtpe) {
							imtpe.printStackTrace();
						}
					}
				});
				helper.setName(AMS_DEBUG_HELPER);
				helper.start();
				return;
			}
			
			// Get the ToolNotifier for the indicated debugger (or create a new one
			// if not yet there)
			ToolNotifier tn = findNotifier(introspectorName);
			if(tn == null) { // Need a new notifier
				tn = new ToolNotifier(introspectorName);
				AID id = new AID(introspectorName.getLocalName() + "-on-" + myID().getName(), AID.ISLOCALNAME);
				try {
					myContainer.initAgent(id, tn, null, null); // FIXME: Modify to use a proper owner Principal
					myContainer.powerUpLocalAgent(id);
					if (targetName.equals(myContainer.getAMS())) {
						// If we are debugging the AMS, let's wait for the ToolNotifier 
						// be ready to avoid deadlock problems. Note also that in 
						// this case this code is executed by the helper thread and not
						// by the AMS thread
						tn.waitUntilStarted();
					}
					else {
						try {Thread.sleep(1000);} catch (Exception e) {};
					}
					addMessageListener(tn);
					addAgentListener(tn);
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
			tn.addObservedAgent(targetName);
			
			// Update the map of debuggers currently debugging the targetName agent
			synchronized (debuggers) {
				List l = (List) debuggers.get(targetName);
				if (l == null) {
					l = new LinkedList();
					debuggers.put(targetName, l);
				}
				if (!l.contains(introspectorName)) {
					l.add(introspectorName);
				}
			}
			
			Agent a = myContainer.acquireLocalAgent(targetName);
			
			// Activate generation of behaviour-related events on the
			// target agent
			a.setGenerateBehaviourEvents(true);
			
			//  Retrieve the current agent state
			AgentState as = a.getAgentState();
			
			// Retrieve the list of pending ACL messages
			List messages = new LinkedList();
			myContainer.fillListFromMessageQueue(messages, a);
			
			// Retrieve the list of ready and blocked agent behaviour IDs
			List readyBehaviours = new LinkedList();
			myContainer.fillListFromReadyBehaviours(readyBehaviours, a);
			List blockedBehaviours = new LinkedList();
			myContainer.fillListFromBlockedBehaviours(blockedBehaviours, a);
			
			myContainer.releaseLocalAgent(targetName);
			
			
			// Notify all the needed events	
			fireChangedAgentState(targetName, as, as);
			
			Iterator itReady = readyBehaviours.iterator();
			while(itReady.hasNext()) {
				BehaviourID bid = (BehaviourID)itReady.next();
				AgentEvent ev = new AgentEvent(AgentEvent.ADDED_BEHAVIOUR, targetName, bid, myContainer.getID());
				tn.addedBehaviour(ev);
			}
			
			Iterator itBlocked = blockedBehaviours.iterator();
			while(itBlocked.hasNext()) {
				BehaviourID bid = (BehaviourID)itBlocked.next();
				AgentEvent ev = new AgentEvent(AgentEvent.ADDED_BEHAVIOUR, targetName, bid, myContainer.getID());
				tn.addedBehaviour(ev);
				ev = new AgentEvent(AgentEvent.CHANGED_BEHAVIOUR_STATE, targetName, bid, Behaviour.STATE_READY, Behaviour.STATE_BLOCKED, myContainer.getID());
				tn.changedBehaviourState(ev);
			}
			
			Iterator itMessages = messages.iterator();
			while(itMessages.hasNext()) {
				ACLMessage msg = (ACLMessage)itMessages.next();
				MessageEvent ev = new MessageEvent(MessageEvent.POSTED_MESSAGE, msg, null, targetName, myContainer.getID());
				tn.postedMessage(ev);
			}
			
		}
		
		public void debugOff(AID introspectorName, AID targetName) throws IMTPException {
			
			ToolNotifier tn = findNotifier(introspectorName);
			if(tn != null) { 
				tn.removeObservedAgent(targetName);
				if(tn.isEmpty()) {
					removeMessageListener(tn);
					removeAgentListener(tn);
					tn.doDelete();
				}
			}
			
			boolean resetGenerateBehaviourEvents = true;
			synchronized (debuggers) {
				List l = (List) debuggers.get(targetName);
				if (l != null) {
					l.remove(introspectorName);
					if (l.size() > 0) {
						// There is still at least 1 debugger debugging the agent 
						// Do not stop generation of behaviour events
						resetGenerateBehaviourEvents = false;
					}
					else {
						debuggers.remove(targetName);
					}
				}
			}
			
			if (resetGenerateBehaviourEvents) {
				Agent a = myContainer.acquireLocalAgent(targetName);
				if (a != null) {
					a.setGenerateBehaviourEvents(false);
				}
				myContainer.releaseLocalAgent(targetName);
			}
			
		}
		
		
		
		// Vertical command handler methods
		
		private void handleSniffOn(VerticalCommand cmd) throws IMTPException, ServiceException, NotFoundException {
			Object[] params = cmd.getParams();
			AID sniffer = (AID)params[0];
			List targets = (List)params[1];
			
			MainContainer impl = myContainer.getMain();
			if(impl != null) {
				
				// Activate sniffing each element of the list
				Iterator it = targets.iterator();
				while(it.hasNext()) {
					
					AID target = (AID)it.next();
					ContainerID cid = impl.getContainerID(target);
					
					NotificationSlice slice = (NotificationSlice)getSlice(cid.getName());
					try {
						slice.sniffOn(sniffer, target);
					}
					catch(IMTPException imtpe) {
						// Try to get a newer slice and repeat...
						slice = (NotificationSlice)getFreshSlice(cid.getName());
						slice.sniffOn(sniffer, target);
					}
					
				}
			}
			else {
				// Do nothing for now, but could also route the command to the main slice, thus enabling e.g. AMS replication
			}
		}
		
		private void handleSniffOff(VerticalCommand cmd) throws IMTPException, ServiceException, NotFoundException {
			Object[] params = cmd.getParams();
			AID sniffer = (AID)params[0];
			List targets = (List)params[1];
			
			MainContainer impl = myContainer.getMain();
			if(impl != null) {
				
				// Deactivate sniffing each element of the list
				Iterator it = targets.iterator();
				while(it.hasNext()) {
					
					AID target = (AID)it.next();
					ContainerID cid = impl.getContainerID(target);
					
					NotificationSlice slice = (NotificationSlice)getSlice(cid.getName());
					try {
						slice.sniffOff(sniffer, target);
					}
					catch(IMTPException imtpe) {
						// Try to get a newer slice and repeat...
						slice = (NotificationSlice)getFreshSlice(cid.getName());
						slice.sniffOff(sniffer, target);
					}
					
				}
			}
			else {
				// Do nothing for now, but could also route the command to the main slice, thus enabling e.g. AMS replication
			}
		}
		
		private void handleDebugOn(VerticalCommand cmd) throws IMTPException, ServiceException, NotFoundException {
			Object[] params = cmd.getParams();
			AID introspector = (AID)params[0];
			List targets = (List)params[1];
			
			MainContainer impl = myContainer.getMain();
			if(impl != null) {
				
				// Activate debugging each element of the list
				Iterator it = targets.iterator();
				while(it.hasNext()) {
					
					AID target = (AID)it.next();
					ContainerID cid = impl.getContainerID(target);
					
					NotificationSlice slice = (NotificationSlice)getSlice(cid.getName());
					try {
						slice.debugOn(introspector, target);
					}
					catch(IMTPException imtpe) {
						// Try to get a newer slice and repeat...
						slice = (NotificationSlice)getFreshSlice(cid.getName());
						slice.debugOn(introspector, target);
					}
					
				}
			}
			else {
				// Do nothing for now, but could also route the command to the main slice, thus enabling e.g. AMS replication
			}
		}
		
		private void handleDebugOff(VerticalCommand cmd) throws IMTPException, ServiceException, NotFoundException {
			Object[] params = cmd.getParams();
			AID introspector = (AID)params[0];
			List targets = (List)params[1];
			
			MainContainer impl = myContainer.getMain();
			if(impl != null) {
				
				// Deactivate debugging each element of the list
				Iterator it = targets.iterator();
				while(it.hasNext()) {
					
					AID target = (AID)it.next();
					ContainerID cid = impl.getContainerID(target);
					
					NotificationSlice slice = (NotificationSlice)getSlice(cid.getName());
					try {
						slice.debugOff(introspector, target);
					}
					catch(IMTPException imtpe) {
						// Try to get a newer slice and repeat...
						slice = (NotificationSlice)getFreshSlice(cid.getName());
						slice.debugOff(introspector, target);
					}
					
				}
			}
			else {
				// Do nothing for now, but could also route the command to the main slice, thus enabling e.g. AMS replication
			}
		}
		
		private void handleSendMessage(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			AID sender = (AID)params[0];
			ACLMessage msg = ((GenericMessage)params[1]).getACLMessage();
			AID receiver = (AID)params[2];
			
			fireSentMessage(msg, sender, receiver);
		}
		
		private void handleNotifyPosted(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			ACLMessage msg = (ACLMessage)params[0];
			AID receiver = (AID)params[1];
			
			firePostedMessage(msg, receiver);
		}
		
		private void handleNotifyReceived(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			ACLMessage msg = (ACLMessage)params[0];
			AID receiver = (AID)params[1];
			
			fireReceivedMessage(msg, receiver);
		}
		
		private void handleNotifyChangedAgentState(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			AID id = (AID)params[0];
			AgentState from = (AgentState)params[1];
			AgentState to = (AgentState)params[2];
			
			fireChangedAgentState(id, from, to);
		}
		
		private void handleNotifyChangedAgentPrincipal(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			AID id = (AID)params[0];
			JADEPrincipal from = (JADEPrincipal)params[1];
			JADEPrincipal to = (JADEPrincipal)params[2];
			
			fireChangedAgentPrincipal(id, from, to);
		}
		
		private void handleNotifyAddedBehaviour(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			AID id = (AID)params[0];
			Behaviour b = (Behaviour)params[1];
			
			fireAddedBehaviour(id, b);
		}
		
		private void handleNotifyRemovedBehaviour(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			AID id = (AID)params[0];
			Behaviour b = (Behaviour)params[1];
			
			fireRemovedBehaviour(id, b);
		}
		
		private void handleNotifyChangedBehaviourState(VerticalCommand cmd) {
			Object[] params = cmd.getParams();
			AID id = (AID)params[0];
			Behaviour b = (Behaviour)params[1];
			String from = (String)params[2];
			String to = (String)params[3];
			
			fireChangedBehaviourState(id, b, from, to);
		}
		
		
	} // End of ServiceComponent class
	
	
	
	// Event dispatching methods
	private void fireSentMessage(ACLMessage msg, AID sender, AID receiver) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = messageListeners.startScanning();
		if (l != null) {
			MessageEvent ev = new MessageEvent(MessageEvent.SENT_MESSAGE, msg, sender, receiver, myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				MessageListener ml = (MessageListener) it.next();
				ml.sentMessage(ev);
			}
			messageListeners.stopScanning();
		}	
	}
	
	private void firePostedMessage(ACLMessage msg, AID receiver) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = messageListeners.startScanning();
		if (l != null) {
			MessageEvent ev = new MessageEvent(MessageEvent.POSTED_MESSAGE, msg, null, receiver, myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				MessageListener ml = (MessageListener) it.next();
				ml.postedMessage(ev);
			}
			messageListeners.stopScanning();
		}
	}
	
	private void fireReceivedMessage(ACLMessage msg, AID receiver) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = messageListeners.startScanning();
		if (l != null) {
			MessageEvent ev = new MessageEvent(MessageEvent.RECEIVED_MESSAGE, msg, null, receiver, myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				MessageListener ml = (MessageListener) it.next();
				ml.receivedMessage(ev);
			}
			messageListeners.stopScanning();
		}
	}
	
	private void fireRoutedMessage(ACLMessage msg, Channel from, Channel to) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = messageListeners.startScanning();
		if (l != null) {
			MessageEvent ev = new MessageEvent(MessageEvent.ROUTED_MESSAGE, msg, from, to, myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				MessageListener ml = (MessageListener) it.next();
				ml.routedMessage(ev);
			}
			messageListeners.stopScanning();
		}
	}
	
	private void fireAddedBehaviour(AID agentID, Behaviour b) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = agentListeners.startScanning();
		if (l != null) {
			AgentEvent ev = null;
			if (b == b.root()) {
				// The behaviour has been added to the Agent
				ev = new AgentEvent(AgentEvent.ADDED_BEHAVIOUR, agentID, new BehaviourID(b), myID());
			}
			else {
				// The behaviour is actually a new child that has been added to a CompositeBehaviour
				//FIXME: TO be done
				//ev = new AgentEvent(AgentEvent.ADDED_SUB_BEHAVIOUR, agentID, new BehaviourID(b.getParent()), new BehaviourID(b), myID());
			}
			
			Iterator it = l.iterator();
			while (it.hasNext()) {
				AgentListener al = (AgentListener) it.next();
				al.addedBehaviour(ev);
			}
			agentListeners.stopScanning();
		}
	}
	
	private void fireRemovedBehaviour(AID agentID, Behaviour b) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = agentListeners.startScanning();
		if (l != null) {
			AgentEvent ev = new AgentEvent(AgentEvent.REMOVED_BEHAVIOUR, agentID, new BehaviourID(b), myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				AgentListener al = (AgentListener) it.next();
				al.removedBehaviour(ev);
			}
			agentListeners.stopScanning();
		}
	}
	
	private void fireChangedBehaviourState(AID agentID, Behaviour b, String from, String to) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = agentListeners.startScanning();
		if (l != null) {
			AgentEvent ev = new AgentEvent(AgentEvent.CHANGED_BEHAVIOUR_STATE, agentID, new BehaviourID(b), from, to, myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				AgentListener al = (AgentListener) it.next();
				al.changedBehaviourState(ev);
			}
			agentListeners.stopScanning();
		}
	}
	
	
	
	
	//__SECURITY__BEGIN
	private void fireChangedAgentPrincipal(AID agentID, JADEPrincipal from, JADEPrincipal to) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = agentListeners.startScanning();
		if (l != null) {
			AgentEvent ev = new AgentEvent(AgentEvent.CHANGED_AGENT_PRINCIPAL, agentID, from, to, myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				AgentListener al = (AgentListener) it.next();
				al.changedAgentPrincipal(ev);
			}
			agentListeners.stopScanning();
		}
	}
	//__SECURITY__END
	
	private void fireChangedAgentState(AID agentID, AgentState from, AgentState to) {
		// NOTE: A normal synchronized block could create deadlock problems
		// as it prevents concurrent scannings of the listeners list.
		List l = agentListeners.startScanning();
		if (l != null) {
			AgentEvent ev = new AgentEvent(AgentEvent.CHANGED_AGENT_STATE, agentID, from, to, myID());
			Iterator it = l.iterator();
			while (it.hasNext()) {
				AgentListener al = (AgentListener) it.next();
				al.changedAgentState(ev);
			}
			agentListeners.stopScanning();
		}
	}
	
	private ToolNotifier findNotifier(AID observerName) {
		ToolNotifier tn = null;
		// Note that if a ToolNotifier exists it must be among the messageListeners
		// --> There is no need to search it also among the agentListeners.
		List l = messageListeners.startScanning();
		if (l != null) {
			Iterator it = l.iterator();
			while(it.hasNext()) {
				Object obj = it.next();
				if(obj instanceof ToolNotifier) {
					ToolNotifier tni = (ToolNotifier) obj;
					AID id = tni.getObserver();
					if(id.equals(observerName)) {
						tn = tni;
						break;
					}
				}
			}
			messageListeners.stopScanning();
		}
		
		if(tn != null && tn.getState() == Agent.AP_DELETED) { // A formerly dead notifier
			removeMessageListener(tn);
			removeAgentListener(tn);
			tn = null;
		}
		return tn;
	}
	
	private void addMessageListener(MessageListener ml) {
		List l = messageListeners.startModifying();
		l.add(ml);
		messageListeners.stopModifying();
	}
	
	private void removeMessageListener(MessageListener ml) {
		List l = messageListeners.startModifying();
		l.remove(ml);
		messageListeners.stopModifying();    
	}
	
	private void addAgentListener(AgentListener al) {
		List l = agentListeners.startModifying();
		l.add(al);
		agentListeners.stopModifying();    
	}
	
	private void removeAgentListener(AgentListener al) {
		List l = agentListeners.startModifying();
		l.remove(al);
		agentListeners.stopModifying();    
	}
	
	private ContainerID myID() {
		return (ContainerID) myContainer.here();
	}
	
	
	// The concrete agent container, providing access to LADT, etc.
	private AgentContainer myContainer;
	
	// The local slice for this service
	private ServiceComponent localSlice;
	
	// The list of all listeners of ACL messaging related events (uses RW-locking)
	private SynchList messageListeners = new SynchList();
	
	// The list of all listeners of agent life cycle events (uses RW-locking)
	private SynchList agentListeners = new SynchList();
	
	// This maps a debugged agent into the list of debuggers that are 
	// currently debugging it. It is used to know when an agent is no longer
	// debugged by any debugger.
	private Map debuggers = new HashMap();
	
}
