/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core.management;

import rca.core.Service;
import rca.core.AID;
import rca.core.ContainerID;
import rca.core.IMTPException;
import rca.core.NameClashException;
import rca.core.NotFoundException;
import rca.core.Command;

import rca.security.Credentials;
import rca.security.JADEPrincipal;
import rca.security.JADESecurityException;


/**

   The horizontal interface for the rca kernel-level service managing
   the basic agent life cycle: creation, destruction, suspension and
   resumption.

   Original author Giovanni Rimassa - FRAMeTech s.r.l.

*/
public interface AgentManagementSlice extends Service.Slice {

    /**
       The name of this service.
    */
    static final String NAME = "rca.core.management.AgentManagement";

    /**
       This command name represents the <code>create-agent</code>
       action. The target agent identifier in this command is set to
       <code>null</code>, because no agent exists yet.
       This command object represents only the <i>first half</i> of
       the complete agent creation process. Even if this command is
       accepted by the kernel, there is no guarantee that the
       requested creation will ever happen. Only when the
       <code>InformCreated</code> command is issued can one assume
       that the agent creation has taken place.
    */
    static final String REQUEST_CREATE = "Request-Create";

    /**
       This command name represents the <code>kill-agent</code>
       action.
       This command object represents only the <i>first half</i> of
       the complete agent destruction process. Even if this command is
       accepted by the kernel, there is no guarantee that the
       requested destruction will ever happen. Only when the
       <code>InformKilled</code> command is issued can one assume that
       the agent destruction has taken place.
    */
    static final String REQUEST_KILL = "Request-Kill";

    /**
       This command name represents all agent management actions requesting
       a change in the life cycle state of their target agent
       (suspend, resume, etc.).
       This command object represents only the <i>first half</i> of
       the complete agent state change process. Even if this command
       is accepted by the kernel, there is no guarantee that the
       requested state change will ever happen. Only when the
       <code>InformStateChanged</code> command is issued can one
       assume that the state change has taken place.
    */
    static final String REQUEST_STATE_CHANGE = "Request-State-Change";

    /**
       This command is issued by an agent that has just been created,
       and causes rca runtime to actually start up the agent thread.
       The agent creation can be the outcome of a previously issued
       <code>RequestCreate</code> command. In that case, this command
       represents only the <i>second half</i> of the complete agent
       creation process.
    */
    static final String INFORM_CREATED = "Inform-Created";

    /**
       This command is issued by an agent that has just been destroyed
       and whose thread is terminating.
       The agent destruction can either be an autonomous move of the
       agent or the outcome of a previously issued
       <code>RequestKill</code> command. In the second case, this
       command represents only the <i>second half</i> of the complete
       agent destruction process.

    */
    static final String INFORM_KILLED = "Inform-Killed";

    /**
       This command is issued by an agent that has just changed its
       life-cycle state.
       The agent state change can either be an autonomous move of the
       agent or the outcome of a previously issued
       <code>RequestStateChange</code> command. In that case, this
       command represents only the <i>second half</i> of the complete
       agent state tansition process.
    */
    static final String INFORM_STATE_CHANGED = "Inform-State-Changed";

    /**
       This command name represents the <code>kill-container</code>
       action.
    */
    static final String KILL_CONTAINER = "Kill-Container";

    /**
       This command name represents the action of adding a new tool to
       the platform.
    */
    static final String ADD_TOOL = "Add-Tool";

    /**
       This command name represents the action of removing an
       existing tool from the platform.
    */
    static final String REMOVE_TOOL = "Remove-Tool";



    static final boolean CREATE_AND_START = true;
    static final boolean CREATE_ONLY = false;


    // Constants for the names of horizontal commands associated to methods
    static final String H_CREATEAGENT = "1";
    static final String H_KILLAGENT = "2";
    static final String H_CHANGEAGENTSTATE = "3";
    static final String H_BORNAGENT = "4";
    static final String H_DEADAGENT = "5";
    static final String H_SUSPENDEDAGENT = "6";
    static final String H_RESUMEDAGENT = "7";
    static final String H_EXITCONTAINER = "8";

    void createAgent(AID agentID, String className, Object arguments[], JADEPrincipal owner, Credentials initialCredentials, boolean startIt, Command sourceCmd) throws IMTPException, NotFoundException, NameClashException, JADESecurityException;
    void killAgent(AID agentID, Command sourceCmd) throws IMTPException, NotFoundException;
    void changeAgentState(AID agentID, int newState) throws IMTPException, NotFoundException;

    void bornAgent(AID name, ContainerID cid, Command sourceCmd) throws IMTPException, NameClashException, NotFoundException, JADESecurityException;
    void deadAgent(AID name, Command sourceCmd) throws IMTPException, NotFoundException;
    void suspendedAgent(AID name) throws IMTPException, NotFoundException;
    void resumedAgent(AID name) throws IMTPException, NotFoundException;

    void exitContainer() throws IMTPException, NotFoundException;

}
