/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core.messaging;

//#MIDP_EXCLUDE_FILE

import rca.core.VerticalCommand;
import rca.core.Filter;
import rca.core.AID;
import rca.domain.FIPAAgentManagement.Envelope;
import rca.lang.acl.ACLMessage;
import rca.lang.acl.ACLCodec;
import rca.util.leap.Map;
import rca.util.leap.Iterator;
import rca.lang.acl.LEAPACLCodec;


/**
 * Class that filters incoming commands related to encoding of ACL messages.
 *
 * Original author Jerome Picault - Motorola Labs
 * Original author Nicolas Lhuillier - Motorola Labs
 * @version $Date: 2005-09-16 15:54:46 +0200 (Fri, 16 Sep 2005) $ $Revision: 5780 $
 */
public class IncomingEncodingFilter extends Filter {

  private Map messageEncodings;
  
  public IncomingEncodingFilter(Map m){
    messageEncodings = m;
    setPreferredPosition(50);
  }

  /**
   * Receive a command object for processing.
   *
   * @param cmd A <code>VerticalCommand</code> describing what operation has
   * been requested from previous layers (that can be the actual
   * prime source of the command or previous filters in the chain).
   */
  public boolean accept(VerticalCommand cmd) {
    String name = cmd.getName();
    Object[] params = cmd.getParams();
    if(name.equals(MessagingSlice.SEND_MESSAGE)) {
      GenericMessage gmsg = (GenericMessage)params[1];

      // The command always contains a non-null ACLMessage (for the purpose
      // of notification of failures), but it contains the real ACLMessage
      // when the payload is null
      if(gmsg.getPayload()==null){
        // If a real ACLMessage is present, just do nothing!
        return true;
      }
      else {
        Envelope env = gmsg.getEnvelope();
        byte[] payload = gmsg.getPayload();
        ACLMessage msg ;
        try{
          msg = decodeMessage(env,payload);
          msg.setEnvelope(env);
          
          if (env!=null){
          // If the 'sender' AID has no addresses, replace it with the
          // 'from' envelope slot
          AID sender = msg.getSender();
          if(sender == null) {
				    System.err.println("ERROR: Trying to dispatch a message with a null sender.");
				    System.err.println("Aborting send operation...");
				    return true;
          }
          Iterator itSender = sender.getAllAddresses();
          if(!itSender.hasNext())
				    msg.setSender(env.getFrom());
          }
          ((GenericMessage)params[1]).update(msg,null,null);
        } catch (MessagingService.UnknownACLEncodingException ee){
        	ee.printStackTrace();
        	cmd.setReturnValue(ee);
        	return false;
        } catch (ACLCodec.CodecException ce){
          ce.printStackTrace();
        	cmd.setReturnValue(ce);
        	return false;
        }
      }
    }
    return true;
  }


  /**
   * Decodes an endoded ACL message according to the acl-representation 
   * described in the envelope. 
   * @param env the Envelope of the message
   * @param payload the encoded message
   * @return the decoded <code>ACLMessage</code>
   */
  public ACLMessage decodeMessage(Envelope env, byte[] payload) throws MessagingService.UnknownACLEncodingException, ACLCodec.CodecException{
    String enc;
    if (env!=null){
      enc = env.getAclRepresentation();
    } else {
      // no envelope means inter-container communication; use LEAP codec
      enc = LEAPACLCodec.NAME.toLowerCase();
    }
    if (enc != null) { // A Codec was selected
	    ACLCodec codec =(ACLCodec)messageEncodings.get(enc.toLowerCase());
	    if (codec != null) {
    		// Supported Codec
    		String charset;
        if ((env == null) ||
            ((charset = env.getPayloadEncoding()) == null)) {
          charset = ACLCodec.DEFAULT_CHARSET;
        }
        return codec.decode(payload,charset);
	    } 
      else {
    		// Unsupported Codec
    		//FIXME: find the best according to the supported, the MTP (and the receivers Codec)
    		throw new MessagingService.UnknownACLEncodingException("Unknown ACL encoding: " + enc + ".");
	    }
    }
    else {
	    // no codec indicated. 
	    //FIXME: find the better according to the supported Codec, the MTP (and the receiver codec)
	    throw new MessagingService.UnknownACLEncodingException("No ACL encoding set.");
    }
  }

} // End of IncomingEncodingFilter class
