/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core.messaging;

import rca.core.Node;
import rca.core.SliceProxy;
import rca.core.GenericCommand;
import rca.core.AID;
import rca.core.ContainerID;
import rca.core.IMTPException;
import rca.core.ServiceException;
import rca.core.NotFoundException;
import rca.security.JADESecurityException;

import rca.domain.FIPAAgentManagement.Envelope;

import rca.mtp.MTPDescriptor;
import rca.mtp.MTPException;


/**
   The remote proxy for the rca kernel-level service managing
   the message passing subsystem installed in the platform.

   Original author Giovanni Rimassa - FRAMeTech s.r.l.
*/
public class MessagingProxy extends SliceProxy implements MessagingSlice {


    public void dispatchLocally(AID senderID, GenericMessage msg, AID receiverID) throws IMTPException, NotFoundException, JADESecurityException {
	try {
	    GenericCommand cmd = new GenericCommand(H_DISPATCHLOCALLY, NAME, null);
      cmd.addParam(senderID);
	    cmd.addParam(msg);
	    cmd.addParam(receiverID);
	    cmd.setPrincipal(msg.getSenderPrincipal());
	    cmd.setCredentials(msg.getSenderCredentials());

	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof NotFoundException) {
		    throw (NotFoundException)result;
		}
		else if(result instanceof JADESecurityException) {
		    throw (JADESecurityException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public void routeOut(Envelope env, byte[] payload,AID receiverID, String address) throws IMTPException, MTPException {
	try {
	    GenericCommand cmd = new GenericCommand(H_ROUTEOUT, NAME, null);
	    cmd.addParam(env);
      cmd.addParam(payload);
	    cmd.addParam(receiverID);
	    cmd.addParam(address);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof MTPException) {
		    throw (MTPException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public ContainerID getAgentLocation(AID agentID) throws IMTPException, NotFoundException {
	try {
	    GenericCommand cmd = new GenericCommand(H_GETAGENTLOCATION, NAME, null);
	    cmd.addParam(agentID);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof NotFoundException) {
		    throw (NotFoundException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	    return (ContainerID)result;
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public MTPDescriptor installMTP(String address, String className) throws IMTPException, ServiceException, MTPException {
	try {
	    GenericCommand cmd = new GenericCommand(H_INSTALLMTP, NAME, null);
	    cmd.addParam(address);
	    cmd.addParam(className);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof MTPException) {
		    throw (MTPException)result;
		}
		else if(result instanceof ServiceException) {
		    throw (ServiceException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	    return (MTPDescriptor)result;
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public void uninstallMTP(String address) throws IMTPException, ServiceException, NotFoundException, MTPException {
	try {
	    GenericCommand cmd = new GenericCommand(H_UNINSTALLMTP, NAME, null);
	    cmd.addParam(address);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof NotFoundException) {
		    throw (NotFoundException)result;
		}
		else if(result instanceof ServiceException) {
		    throw (ServiceException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public void newMTP(MTPDescriptor mtp, ContainerID cid) throws IMTPException, ServiceException {
	try {
	    GenericCommand cmd = new GenericCommand(H_NEWMTP, NAME, null);
	    cmd.addParam(mtp);
	    cmd.addParam(cid);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof ServiceException) {
		    throw (ServiceException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public void deadMTP(MTPDescriptor mtp, ContainerID cid) throws IMTPException, ServiceException {
	try {
	    GenericCommand cmd = new GenericCommand(H_DEADMTP, NAME, null);
	    cmd.addParam(mtp);
	    cmd.addParam(cid);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof ServiceException) {
		    throw (ServiceException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public void addRoute(MTPDescriptor mtp, String sliceName) throws IMTPException, ServiceException {
	try {
	    GenericCommand cmd = new GenericCommand(H_ADDROUTE, NAME, null);
	    cmd.addParam(mtp);
	    cmd.addParam(sliceName);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof ServiceException) {
		    throw (ServiceException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

    public void removeRoute(MTPDescriptor mtp, String sliceName) throws IMTPException, ServiceException {
	try {
	    GenericCommand cmd = new GenericCommand(H_REMOVEROUTE, NAME, null);
	    cmd.addParam(mtp);
	    cmd.addParam(sliceName);


	    Node n = getNode();
	    Object result = n.accept(cmd);
	    if((result != null) && (result instanceof Throwable)) {
		if(result instanceof IMTPException) {
		    throw (IMTPException)result;
		}
		else if(result instanceof ServiceException) {
		    throw (ServiceException)result;
		}
		else {
		    throw new IMTPException("An undeclared exception was thrown", (Throwable)result);
		}
	    }
	}
	catch(ServiceException se) {
	    throw new IMTPException("Unable to access remote node", se);
	}
    }

}
