/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.core.messaging;

//#J2ME_EXCLUDE_FILE


import java.io.IOException;

import rca.core.AID;
import rca.core.Profile;


import rca.util.leap.List;
import rca.util.leap.LinkedList;
import rca.util.leap.Map;
import rca.util.leap.HashMap;
import rca.util.leap.Iterator;


/**
 * This class supports the ACL persistent delivery service, managing
 * actual ACL messages storage, scheduled message delivery and other
 * utility tasks related to the service.
 *
 * Original author  Giovanni Rimassa - FRAMeTech s.r.l.
 * Original author  Nicolas Lhuillier - Motorola
 *
 */
class PersistentDeliveryManager {

  public static synchronized PersistentDeliveryManager instance(Profile p, MessageManager.Channel ch) {
    if(theInstance == null) {
	    theInstance = new PersistentDeliveryManager();
	    theInstance.initialize(p, ch);
    }

    return theInstance;
  }

  // How often to check for expired deliveries
  private static final long DEFAULT_SENDFAILUREPERIOD = 60*1000; // One minute

  // Default storage class
  private static final String DEFAULT_STORAGE = "rca.core.messaging.PersistentDeliveryManager$DummyStorage";

  private static class DeliveryItem {

    public DeliveryItem(GenericMessage msg, AID id, MessageManager.Channel ch, String sid) {
	    toDeliver = msg;
	    receiver = id;
	    channel = ch;
      storeName = sid;
    }
    
    public GenericMessage getMessage() {
	    return toDeliver;
    }

    public AID getReceiver() {
	    return receiver;
    }

    public MessageManager.Channel getChannel() {
	    return channel;
    }

    public String getStoreName() {
	    return storeName;
    }
    
    private GenericMessage toDeliver;
    private AID receiver;
    private MessageManager.Channel channel;
    private String storeName;


  } // End of DeliveryItem class



  private class ExpirationChecker implements Runnable {

    public ExpirationChecker(long t) {
	    period = t;
	    myThread = new Thread(this, "Persistent Delivery Service -- Expiration Checker Thread");
    }

    public void run() {
      while(active) {
        try {
          Thread.sleep(period);
          synchronized(pendingMessages) {
            // Try to send all stored messages...
            // If the receiver still not exists and the due date has elapsed
            // the sender will get back a FAILURE
            Object[] keys = pendingMessages.keySet().toArray();
            for(int i = 0; i < keys.length; i++) {
              flushMessages((AID) keys[i]);
            }
          }
        }
        catch (InterruptedException ie) {
          // Just do nothing
        }
      }
    }

    public void start() {
	    active = true;
	    myThread.start();
    }

    public void stop() {
	    active = false;
	    myThread.interrupt();
    }

    private boolean active = false;
    private long period;
    private Thread myThread;

  } // End of ExpirationChecker class


  public static class DummyStorage implements MessageStorage {
    
    public void init(Profile p) {
      // Do nothing
    }

    public String store(GenericMessage msg, AID receiver) throws IOException {
      // Do nothing
      return null;
    }

    public void delete(String storeName, AID receiver) throws IOException {
	    // Do nothing
    }

    public void loadAll(LoadListener il) throws IOException {
	    // Do nothing
    }

  } // End of DummyStorage class


  public void initialize(Profile p, MessageManager.Channel ch) {

    users = 0;
    myMessageManager = MessageManager.instance(p);
    deliveryChannel = ch;

    try {
      // Choose the persistent storage method
      String storageClass = p.getParameter(PersistentDeliveryService.PERSISTENT_DELIVERY_STORAGEMETHOD,DEFAULT_STORAGE);
      storage = (MessageStorage)Class.forName(storageClass).newInstance();
      storage.init(p);
      
      // Load all data persisted from previous sessions
	    storage.loadAll(new MessageStorage.LoadListener() {
          public void loadStarted(String storeName) {
            System.out.println("--> Load BEGIN <--");
          }

          public void itemLoaded(String storeName, GenericMessage msg, AID receiver) {

            // Put the item into the pending messages table
            synchronized(pendingMessages) {
              List msgs = (List)pendingMessages.get(receiver);
              if(msgs == null) {
                msgs = new LinkedList();
                pendingMessages.put(receiver, msgs);
              }

              DeliveryItem item = new DeliveryItem(msg, receiver, deliveryChannel, storeName);
              msgs.add(item);
            }

            System.out.println("Message for <" + receiver.getLocalName() + ">");
          }

          public void loadEnded(String storeName) {
            System.out.println("--> Load END <--");
          }

        });
    }
    catch(IOException ioe) {
	    ioe.printStackTrace();
    }
    catch(Exception e) {
	    e.printStackTrace();
    }

    sendFailurePeriod = DEFAULT_SENDFAILUREPERIOD;
    String s = p.getParameter(PersistentDeliveryService.PERSISTENT_DELIVERY_SENDFAILUREPERIOD, null);
    if(s != null) {
	    try {
        sendFailurePeriod = Long.parseLong(s);
	    }
	    catch(NumberFormatException nfe) {
        // Do nothing: the default value will be used...
	    }
    }
  }

  public void storeMessage(String storeName, GenericMessage msg, AID receiver) throws IOException {

    // Store the ACL message and its receiver for later re-delivery...
    synchronized(pendingMessages) {
	    List msgs = (List)pendingMessages.get(receiver);
	    if(msgs == null) {
        msgs = new LinkedList();
        pendingMessages.put(receiver, msgs);
	    }

	    String tmpName = storage.store(msg, receiver);
	    msgs.add(new DeliveryItem(msg, receiver, deliveryChannel, tmpName));
    }

  }

  public int flushMessages(AID receiver) {

    // Send messages for this agent, if any...
    int cnt = 0;
    List l = null;
    synchronized(pendingMessages) {
	    l = (List)pendingMessages.remove(receiver);
    }

    if(l != null) {
	    Iterator it = l.iterator();
	    while(it.hasNext()) {
        DeliveryItem item = (DeliveryItem)it.next();
        retry(item);
        cnt++;
	    }
    }
    return cnt;
  }

  public synchronized void start() {

    if(users == 0) {
	    failureSender = new ExpirationChecker(sendFailurePeriod);
	    failureSender.start();
    }

    users++;

  }

  public synchronized void stop() {
    users--;

    if(users == 0) {
	    failureSender.stop();
    }
  }


  // A shared instance to have a single thread pool
  private static PersistentDeliveryManager theInstance; // FIXME: Maybe a table, indexed by a profile subset, would be better?


  private PersistentDeliveryManager() {
  }

  private void retry(DeliveryItem item) {
    // Remove the message from the storage
    try {
      storage.delete(item.getStoreName(), item.getReceiver());
    }
    catch(IOException ioe) {
      ioe.printStackTrace();
    }
    // Deliver it
    myMessageManager.deliver(item.getMessage(), item.getReceiver(), item.getChannel());
  }


  // The component managing asynchronous message delivery and retries
  private MessageManager myMessageManager;

  // The actual channel over which messages will be sent
  private MessageManager.Channel deliveryChannel;

  // How often pending messages due date will be checked (the
  // message will be sent out if expired)
  private long sendFailurePeriod;

  // How many containers are sharing this active component
  private long users;

  // The table of undelivered messages to send
  private Map pendingMessages = new HashMap();

  // The active object that periodically checks the due date of ACL
  // messages and sends them after it expired
  private ExpirationChecker failureSender;

  // The component performing the actual storage and retrieval from
  // a persistent support
  private MessageStorage storage;
}
