/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.core.messaging;

//#MIDP_EXCLUDE_FILE

import rca.util.leap.Iterator;
import rca.util.leap.List;
import rca.util.leap.ArrayList;
import rca.util.leap.Map;
import rca.util.leap.HashMap;

import rca.core.AID;
import rca.core.CaseInsensitiveString;
import rca.core.IMTPException;
import rca.core.Runtime;

import rca.domain.FIPAAgentManagement.Envelope;
import rca.domain.FIPAAgentManagement.Property;

import rca.mtp.OutChannel;
import rca.mtp.MTP;
import rca.mtp.MTPException;
import rca.mtp.MTPDescriptor;

class RoutingTable {
  
  public interface OutPort {
    void route(Envelope env, byte[] payload, AID receiver, String address) throws MTPException;
  }


  // This class wraps an MTP installed on a remote container, using
  // RMI to forward the deliver() operation
  private static class OutViaSlice implements OutPort {

    private final String sliceName;
    private final MessagingSlice slice;

    public OutViaSlice(String sn, MessagingSlice ms) {
	sliceName = sn;
	slice = ms;
    }

    public void route(Envelope env, byte[] payload, AID receiver, String address) throws MTPException {
      try {
        slice.routeOut(env, payload, receiver, address);
      }
      catch(IMTPException imtpe) {
	throw new MTPException("Container unreachable during routing", imtpe);
      }
    }

    public boolean equals(Object o) {
      try {
	OutViaSlice rhs = (OutViaSlice)o;
	String sn = rhs.sliceName;
	if(sliceName.equals(sn))
	  return true;
	else
	  return false;
      }
      catch(ClassCastException cce) {
	return false;
      }
    }
    
  } // End of OutViaContainer class


  // This class wraps an MTP installed locally, using the ACC to encode
  // the message into an MTP payload.
  private static class OutViaMTP implements OutPort {

    private final OutChannel myChannel;
    private String platformInfo;

    public OutViaMTP(OutChannel proto, String platformInfo) {
      myChannel = proto;
      this.platformInfo = platformInfo;
    }

    public void route(Envelope env, byte[] payload, AID receiver, String address) throws MTPException {
      if (platformInfo != null) {
      	env.addProperties(new Property(MessagingService.PLATFORM_IDENTIFIER, platformInfo));
      	env.addProperties(new Property(MessagingService.MTP_IDENTIFIER, myChannel.getClass().getName()));
      }
    	myChannel.deliver(address, env, payload);
    }

    public boolean equals(Object o) {
      try {
	OutViaMTP rhs = (OutViaMTP)o;
	OutChannel ch = rhs.myChannel;
	if(myChannel.equals(ch))
	  return true;
	else
	  return false;
      }
      catch(ClassCastException cce) {
	return false;
      }
    }
  }

  private static final boolean LOCAL = true;
  private static final boolean REMOTE = false;

  private static class OutPortList {

    private final List local = new ArrayList(1);
    private final List remote = new ArrayList(1);

    public void add(OutPort port, boolean location) {
      if(location == LOCAL) {
	local.add(port);
      }
      else {
	remote.add(port);
      }
    }

    public void remove(OutPort port) {

      local.remove(port);
      remote.remove(port);

    }

    public OutPort get() {
      // Look first in the local list
      if(!local.isEmpty())
	return (OutPort)local.get(0);
      // Then look in the remote list
      else
	if(!remote.isEmpty())
	  return (OutPort)remote.get(0);
      return null;
    }

    public boolean isEmpty() {
      return local.isEmpty() && remote.isEmpty();
    }

    public String size() {
      return "[ local: " + local.size() + "  remote: " + remote.size() + " ]";
    }

  } // End of OutPortList class


  private final Map inPorts = new HashMap(2);
  private final Map outPorts = new HashMap(2);
    private static final int EXPECTED_PLATFORMADDRESSES_SIZE = 2;
  private final List platformAddresses = new ArrayList(EXPECTED_PLATFORMADDRESSES_SIZE);
  private String platformInfo = null;
  	
  public RoutingTable(boolean attachPlatformInfo) {
  	if (attachPlatformInfo) {
  		platformInfo = Runtime.instance().getVersionInfo() + " (" + System.getProperty("java.version") + ", " + System.getProperty("os.name") + " " + System.getProperty("os.version") + ")";
  	}
  }

  /**
     Adds a new locally installed MTP for the URL named
     <code>url</code>.
   */
  public synchronized void addLocalMTP(String url, MTP proto) {
    
      CaseInsensitiveString urlTmp = new CaseInsensitiveString(url);  
      // url = url.toLowerCase();
    // A local MTP can receive messages
    inPorts.put(urlTmp, proto);

    // A local MTP can also send messages, over all supported protocols
    OutPort out = new OutViaMTP(proto, platformInfo);
    String[] protoNames = proto.getSupportedProtocols();
    for(int i = 0; i < protoNames.length; i++) {
      addOutPort(protoNames[i], out, LOCAL);
    }

    // The new MTP is a valid address for the platform
    platformAddresses.add(url);

    /*
      java.util.Iterator it = outPorts.keySet().iterator();
      while(it.hasNext()) {
      String name = (String)it.next();
      OutPortList l = (OutPortList)outPorts.get(name);
      System.out.println("<" + name + "> ==> " + l.size());
      }
    */

  }

  /**
     Removes a locally installed MTP for the URL named
     <code>url</code>.
   */
  public synchronized MTP removeLocalMTP(String url) {
      // url = url.toLowerCase();
      CaseInsensitiveString urlTmp = new CaseInsensitiveString(url);
    // A local MTP appears both in the input and output port tables
    MTP proto = (MTP)inPorts.remove(urlTmp);
    if(proto != null) {
      // Remove all outgoing ports associated with this MTP
      String[] protoNames = proto.getSupportedProtocols();
      for(int i = 0; i < protoNames.length; i++) {
	OutPort out = new OutViaMTP(proto, platformInfo);
	removeOutPort(protoNames[i], out);
      }
    }

    // The MTP address is not a platform address anymore
    platformAddresses.remove(url);

    /*
    java.util.Iterator it = outPorts.keySet().iterator();
    while(it.hasNext()) {
      String name = (String)it.next();
      OutPortList l = (OutPortList)outPorts.get(name);
      System.out.println("<" + name + "> ==> " + l.size());
    }
    */

    return proto;
  }

  public synchronized void addRemoteMTP(MTPDescriptor mtp, String sliceName, MessagingSlice where) {

    // A remote MTP can be used only for outgoing messages, through an
    // OutPort that routes messages through a container
    OutPort out = new OutViaSlice(sliceName, where);
    String[] protoNames = mtp.getSupportedProtocols();
    for(int i = 0; i < protoNames.length; i++) {
      addOutPort(protoNames[i], out, REMOTE);
    }

    // Remote MTPs are valid platform addresses
    String[] mtpAddrs = mtp.getAddresses();
    platformAddresses.add(mtpAddrs[0]); 
  }

  /**
     Removes the MTP for the URL named <code>name</code>.
   */
  public synchronized void removeRemoteMTP(MTPDescriptor mtp, String sliceName, MessagingSlice where) {

    OutPort ch = new OutViaSlice(sliceName, where);
    String[] protoNames = mtp.getSupportedProtocols();
    for(int i = 0; i < protoNames.length; i++) {
      removeOutPort(protoNames[i], ch);
    }

    // Remote MTPs are valid platform addresses
    String[] mtpAddrs = mtp.getAddresses();
    platformAddresses.remove(mtpAddrs[0]);
  }

  /**
     Retrieves an outgoing channel object suitable for
     reaching the address <code>url</code>.
   */
  public synchronized OutPort lookup(String url) {
      //url = url.toLowerCase();
    String proto = extractProto(url);
    CaseInsensitiveString protoTmp = new CaseInsensitiveString(proto);
    OutPortList l = (OutPortList)outPorts.get(protoTmp);
    if(l != null) {
      return l.get();
    }
    else {
      return null;
    }
  }

  public synchronized Iterator getAddresses() {
    return platformAddresses.iterator();
  }

  private void addOutPort(String proto, OutPort port, boolean location) {
      //proto = proto.toLowerCase();
      CaseInsensitiveString protoTmp = new CaseInsensitiveString(proto);
    OutPortList l = (OutPortList)outPorts.get(protoTmp);
    if(l != null)
      l.add(port, location);
    else {
      l = new OutPortList();
      l.add(port, location);
      outPorts.put(protoTmp, l);
    }
  }

  private void removeOutPort(String proto, OutPort port) {
      //proto = proto.toLowerCase();
      CaseInsensitiveString protoTmp = new CaseInsensitiveString(proto);
    OutPortList l = (OutPortList)outPorts.get(protoTmp);
    if(l != null) {
      l.remove(port);
    }
  }

  private String extractProto(String address) {
    int colonPos = address.indexOf(':');
    if(colonPos == -1)
      return null;
    return address.substring(0, colonPos);
  }

}

