/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core.nodeMonitoring;

//#APIDOC_EXCLUDE_FILE
//#MIDP_EXCLUDE_FILE

import rca.core.Node;
import rca.core.NodeFailureMonitor;
import rca.core.NodeEventListener;
import rca.core.IMTPException;

import rca.util.Logger;


/**
 *  The <code>BlockingNodeFailureMonitor</code> class detects node failures and
 *  notifies its registered listener using blocking calls to the ping() 
 * remote method of the monitored Node.
 *
 * Original author Giovanni Rimassa - FRAMeTech s.r.l.
 * Original author Roland Mungenast - Profactor
 * @see rca.core.NodeFailureMonitor
 */
public class BlockingNodeFailureMonitor extends NodeFailureMonitor
  implements Runnable {

    private boolean nodeExited = false;
    private boolean stopped = false;
    
    private Logger myLogger = Logger.getMyLogger(getClass().getName());

    public void start(Node n, NodeEventListener nel) {
    	super.start(n, nel);
    	
      Thread thread = new Thread(this);
      thread.setName(target.getName()+"-failure-monitor");
      thread.start();
    }
    
    public void run() {
      fireNodeAdded();
      while(!nodeExited && !stopped) {
        try {
          nodeExited = target.ping(true); // Hang on this call
          if(myLogger.isLoggable(Logger.INFO)) {
            myLogger.log(Logger.INFO,"PING from node " + target.getName() + " returned [" + (nodeExited ? "EXIT]" : "GO ON]"));
          }
        }
        catch(IMTPException imtpe1) { // Connection down
          if(myLogger.isLoggable(Logger.INFO)) {
            myLogger.log(Logger.INFO,"PING from node " + target.getName() + " exited with exception. "+imtpe1.getMessage());
          }
          
          if(!stopped) {
            fireNodeUnreachable();
          }
          try {
            target.ping(false); // Try a non blocking ping to check
    
            if(myLogger.isLoggable(Logger.INFO))
              myLogger.log(Logger.INFO,"PING from node " + target.getName() + " returned OK");
            
            if(!stopped) {
              fireNodeReachable();
            }
          }
          catch(IMTPException imtpe2) { // Object down
            nodeExited = true;
          }
        }
        catch(Throwable t) {
          t.printStackTrace();
        }
      } // END of while
      
      // If we reach this point without being explicitly stopped the node is no longer active
      if(!stopped) {
        fireNodeRemoved();
      }
    }
 
    public void stop() {
      try {
      stopped = true;
      target.interrupt();
      
      } catch(IMTPException imtpe) {
        if(myLogger.isLoggable(Logger.INFO))
          myLogger.log(Logger.INFO,"-- The node <" + target.getName() + "> is already dead --" );
        // Ignore it: the node must be dead already...
      }
    }

}
