/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.core.nodeMonitoring;

// Take care that the DOTNET build file (dotnet.xml) uses this file (it is copied just after the preprocessor excluded it)
//#J2ME_EXCLUDE_FILE
//#APIDOC_EXCLUDE_FILE

import rca.core.Node;
import rca.core.NodeFailureMonitor;
import rca.core.NodeEventListener;

import rca.util.Logger;
import rca.util.leap.Collection;


/**
 * The <code>UDPNodeFailureMonitor</code> class detects node failures and
 * notifies its registered listener using the UDP protocol.
 * 
 * Original author Roland Mungenast - Profactor
 * @since rca 3.3
 * @see rca.core.NodeFailureMonitor
 */
class UDPNodeFailureMonitor extends NodeFailureMonitor {
  
  /**
   * The target is connected. That means that it is sending
   * ping messages regulary.
   */
  public static final int STATE_CONNECTED = 0;
  
  /**
   * The target is unreachable. The target doesn't
   * send ping messages any more.
   */
  public static final int STATE_UNREACHABLE = 1;
  
  /**
   * The final state. The target isn't
   * monitored anymore.
   */
  public static final int STATE_FINAL = 2;
  

  private long deadline = -1;
  private long lastPing = -1;
  private int state = -1;
  private long key = 0;
  private UDPMonitorServer server;
  private UDPNodeMonitoringService service;
  private Logger logger = Logger.getMyLogger(this.getClass().getName());
  
  
  /**
   * Constructor
   */
  UDPNodeFailureMonitor(UDPMonitorServer s, UDPNodeMonitoringService svc) {
  	server = s;
  	service = svc;
  }

  /**
   * Start monitoring a given node
   * @param n target node to monitor
   * @param nel listener to inform about new events
   */
  public void start(Node n, NodeEventListener nel) {
  	super.start(n, nel);
    setState(STATE_CONNECTED);
    // Request the node to be monitored to start sending UDP packets
    key = System.currentTimeMillis();
    service.activateUDP(target, key);
    server.register(this);
  }

  public void stop() {
    server.deregister(this);
    // Request the node to be monitored to stop sending UDP packets
    service.deactivateUDP(target, key);
  }
  
  /**
   * Returns all child nodes of the targeted node
   * @return a <code>Collection</code> of <code>Node</code> instances
   */
  public Collection getChildNodes() {
    return childNodes.values();
  }
  
  /**
   * Returns the time when the last ping message has been received 
   * from the targeted node
   * 
   * @return the difference, measured in milliseconds, 
   * between the current time and midnight, January 1, 1970 UTC.
   */
  public long getLastPing() {
    return lastPing;
  }
  
  /**
   * Returns the deadline until which the monitor will 
   * wait for a new ping message.
   * 
   * @return the difference, measured in milliseconds, 
   * between the current time and midnight, January 1, 1970 UTC.
   */
  public long getDeadline() {
    return deadline;
  }
  
  /**
   * Returns the current state.
   */
  public int getState() {
    return state;
  }
  
  /**
   * Sets the current state.
   */
  void setState(int newState) {
    
    if (logger.isLoggable(Logger.FINEST)) {
      logger.log(Logger.FINEST, "Transition to state " + newState + 
          " for node '" + target.getName() + "'");
    }
    
    // --> CONNECTED
    if (state == -1 && newState == STATE_CONNECTED) {
      fireNodeAdded();
    
    // CONNECTED --> UNREACHABLE
    } else if (state == STATE_CONNECTED && newState == STATE_UNREACHABLE) {
      fireNodeUnreachable();
    
    // UNREACHABLE --> CONNECTED
    } else if (state == STATE_UNREACHABLE && newState == STATE_CONNECTED) {
      fireNodeReachable();
    
    // REMOVED
    } else if (newState == STATE_FINAL) {
      fireNodeRemoved();
      server.deregister(this);
    }
    
    state = newState;
  }
  
  /**
   * Sets the time when the last ping message has been received 
   * from the targeted node
   * 
   * @param time the difference, measured in milliseconds, 
   * between the current time and midnight, January 1, 1970 UTC.
   */
  void setLastPing(long time) {
    lastPing = time;
  }
  
  /**
   * Sets the time until which the monitor will 
   * wait for a new ping message.
   * 
   * @param time the difference, measured in milliseconds, 
   * between the current time and midnight, January 1, 1970 UTC.
   */
  void setDeadline(long time) {
    deadline = time;
  }
  
}
