/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain;

//#MIDP_EXCLUDE_FILE

import rca.content.Concept;
import rca.content.Predicate;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Result;
import rca.content.onto.basic.Done;

import rca.lang.acl.ACLMessage;
import rca.lang.acl.MessageTemplate;

import rca.domain.FIPAAgentManagement.*;

import rca.util.leap.ArrayList;
import rca.util.leap.List;

import rca.security.JADESecurityException;

/**
   This behaviour serves the actions of the FIPA management ontology 
   supported by the AMS.
   Extends RequestManagementBehaviour and implements performAction() to 
   i) call the method of the AMS corresponding to the requested 
   action and ii) prepare the result notification depending on 
   whether a result should be returned (RESULT or DONE).
   Note that all actions in the FIPAManagementOntology produce a 
   "synchronous" effect --> there is no need to delay the notification 
   at a later time as for some actions in the JADEManagementOntology.
   Original author Tiziana Trucco - Tilab
   Original author Giovanni Caire - Tilab
  @version $Date: 2004-05-19 18:01:19 +0200 (Wed, 19 May 2004) $ $Revision: 5083 $
*/
class AMSFipaAgentManagementBehaviour extends RequestManagementBehaviour{

	private ams theAMS;
	
  protected AMSFipaAgentManagementBehaviour(ams a, MessageTemplate mt) {
		super(a,mt);
		theAMS = a;
  }
	
  /**
     Call the proper method of the ams and prepare the notification 
     message
   */
  protected ACLMessage performAction(Action slAction, ACLMessage request) throws JADESecurityException, FIPAException {
  	Concept action = slAction.getAction();
  	List resultItems = null;
  	
  	// REGISTER
  	if (action instanceof Register) {
  		theAMS.registerAction((Register) action, request.getSender());
  	}
  	// DEREGISTER
  	else if (action instanceof Deregister) {
  		theAMS.deregisterAction((Deregister) action, request.getSender());
  	}
  	// MODIFY
  	else if (action instanceof Modify) {
  		theAMS.modifyAction((Modify) action, request.getSender());
  	}
  	// SEARCH
  	else if (action instanceof Search) {
  		resultItems = theAMS.searchAction((Search) action, request.getSender());
  	}
  	// GET_DESCRIPTION
  	else if (action instanceof GetDescription) {
  		APDescription dsc = theAMS.getDescriptionAction(request.getSender());
  		resultItems = new ArrayList();
  		resultItems.add(dsc);
  	}
  	else {
  		throw new UnsupportedFunction();
  	}
  	
  	// Prepare the notification
  	ACLMessage notification = request.createReply();
  	notification.setPerformative(ACLMessage.INFORM);
  	Predicate p = null;
  	if (resultItems != null) {
  		// The action produced a result
  		p = new Result(slAction, resultItems);
  	}
  	else {
  		p = new Done(slAction);
  	}
  	try {
	  	theAMS.getContentManager().fillContent(notification, p);
  	}
  	catch (Exception e) {
  		// Should never happen
  		e.printStackTrace();
  	} 
  	return notification;
  }  
}
