/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain;

//#MIDP_EXCLUDE_FILE

import rca.content.Concept;
import rca.content.Predicate;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Done;
import rca.content.onto.basic.Result;
import rca.lang.acl.ACLMessage;
import rca.lang.acl.MessageTemplate;
import rca.domain.FIPAAgentManagement.*;
import rca.security.JADESecurityException;

/**
   This behaviour serves the actions of the FIPA management ontology 
   supported by the DF.
   It extends RequestManagementBehaviour and implements performAction() to 
   i) call the method of the DF corresponding to the requested 
   action and ii) prepare the result notification.
   Original author Tiziana Trucco - TILAB
   Original author Fabio Bellifemine - TILAB
   Original author Giovanni Caire - TILAB
 */
class DFFipaAgentManagementBehaviour extends RequestManagementBehaviour {
  
  private df theDF;

  protected DFFipaAgentManagementBehaviour(df a, MessageTemplate mt){
		super(a, mt);
		theDF = a;
  }

  /**
     Call the proper method of the DF and prepare the notification 
     message
   */
  protected ACLMessage performAction(Action slAction, ACLMessage request) throws JADESecurityException, FIPAException {
  	Concept action = slAction.getAction();
  	Object result = null;
  	boolean asynchNotificationRequired = false;
  	
  	// REGISTER
  	if (action instanceof Register) {
  		theDF.registerAction((Register) action, request.getSender());
  	}
  	// DEREGISTER
  	else if (action instanceof Deregister) {
  		theDF.deregisterAction((Deregister) action, request.getSender());
  	}
  	// MODIFY
  	else if (action instanceof Modify) {
  		theDF.modifyAction((Modify) action, request.getSender());
  	}
  	// SEARCH
  	else if (action instanceof Search) {
  		result = theDF.searchAction((Search) action, request.getSender());
  		if (result == null) {
  			asynchNotificationRequired = true;
  		}
  	}
  	else {
  		throw new UnsupportedFunction();
  	}
  	
  	if (!asynchNotificationRequired) {
  		// The requested action has been completed. Prepare the notification
	  	ACLMessage notification = request.createReply();
	  	notification.setPerformative(ACLMessage.INFORM);
	  	Predicate p = null;
	  	if (result != null) {
	  		// The action produced a result
	  		p = new Result(slAction, result);
	  	}
	  	else {
	  		p = new Done(slAction);
	  		
	  		/*
			 *  WebService Integration Gateway (WSIG) requirements
			 *  added by Whitestein Technologies AG,
			 *  Contributor(s): Jozef Nagy (jna at whitestein.com)
			 */
	  		// Also inform the WebService Integration Gateway
	  		theDF.addGatewayIfNeeded( request, notification );
			/*
			 *  End: WSIG requirements
			 */
	  	}
	  	try {
		  	theDF.getContentManager().fillContent(notification, p);
	  	}
	  	catch (Exception e) {
	  		// Should never happen
	  		e.printStackTrace();
	  	}
	  	return notification;
  	}
		else {
			// The requested action is being processed by a Behaviour. Store
			// the request message for later retrieval.
  		theDF.storePendingRequest(action, request);
  		return null;
  	}
  }
}
