/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain;

//#MIDP_EXCLUDE_FILE

import rca.core.behaviours.CyclicBehaviour;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Result;
import rca.content.onto.OntologyException;
import rca.content.lang.Codec.CodecException;
import rca.lang.acl.ACLMessage;
import rca.lang.acl.MessageTemplate;
import rca.lang.acl.ConversationList;
import rca.domain.FIPAAgentManagement.*;
import rca.domain.KBManagement.KBIterator;
import rca.proto.SSIteratedAchieveREResponder;
import rca.util.leap.List;
import rca.util.leap.ArrayList;

/**
   This behaviour serves search requests according to the iterated-fipa-request
   protocol. 
   Original author Giovanni Caire - TILAB
 */
class DFIteratedSearchManagementBehaviour extends CyclicBehaviour {
  
  private df theDF;
  private MessageTemplate template;
  private ConversationList conversations;
  
  public DFIteratedSearchManagementBehaviour(df theDF, MessageTemplate mt) {
		super(theDF);
		this.theDF = theDF;
		conversations = new ConversationList(theDF);
		template = MessageTemplate.and(mt, conversations.getMessageTemplate());
  }

  public void action() {
  	ACLMessage msg = theDF.receive(template);
  	if (msg != null) {
  		theDF.addBehaviour(new SSIteratedAchieveREResponder(theDF, msg) {
  			private KBIterator iterator = null;
  			private String myConversationId = null;
  			
  			protected ACLMessage handleRequest(ACLMessage request) throws RefuseException, FailureException, NotUnderstoodException {
  				ACLMessage reply = null;
  				
  				try {
	  				// Parse the request content
  					Action aExpr = (Action) theDF.getContentManager().extractContent(request);
  					Search search = (Search) aExpr.getAction();
  					SearchConstraints constraints = search.getConstraints();
  					int maxResult = theDF.getActualMaxResults(constraints);
  					
  					// If this is the first REQUEST, create the Iterator
	  				if (iterator == null) {
	  					iterator = theDF.iteratedSearchAction(search, request.getSender());
	  				}
	  				
  					// Get the requested number of results 
	  				List ll = new ArrayList();
	  				for (int i = 0; i < maxResult; ++i) {
	  					if (iterator.hasNext()) {
	  						ll.add(iterator.next());
	  					}
	  					else {
	  						iterator.close();
	  						// FIXME: properly handle recursive search
	  						closeSessionOnNextReply();
	  					}
	  				}
	  				
	  				// Fill the reply
	  				reply = request.createReply();
				  	try {
				  		Result result = new Result(aExpr, ll);
					  	theDF.getContentManager().fillContent(reply, result);
					  	reply.setPerformative(ACLMessage.INFORM);
				  	}
				  	catch (Exception e) {
				  		// Unexpected error encoding the reply
				  		e.printStackTrace();
				  		throw new FailureException(ExceptionVocabulary.INTERNALERROR+" \""+e.getMessage()+"\"");
				  	}
  				}
					catch (OntologyException oe) {
						throw new NotUnderstoodException(ExceptionVocabulary.UNRECOGNISEDVALUE+" content");
					}	
					catch (CodecException ce) {
						throw new NotUnderstoodException(ExceptionVocabulary.UNRECOGNISEDVALUE+" content");
					}					
					catch (Throwable t) {
			  		// Unexpected error
						t.printStackTrace();
			  		throw new FailureException(ExceptionVocabulary.INTERNALERROR+" \""+t.getMessage()+"\"");
					}
					
			  	return reply;
  			}
  			
  			protected void handleCancel(ACLMessage cancel) {
  				if (iterator != null) {
						iterator.close();
  				}
  			}
  			
  			/**
  			   Avoid conflicts between the main MessageTemplate and that used 
  			   internally by this IteratedFipaRequestResponder.
  			 */
  			protected void afterReply(ACLMessage reply) {
  				if (reply != null) {
  					myConversationId = reply.getConversationId();
				  	conversations.registerConversation(myConversationId);
  				}
			  }
			  
  			/**
  			   When the protocol terminates, deregister the conversation
  			 */
			  public int onEnd() {
			  	conversations.deregisterConversation(myConversationId);
			  	return super.onEnd();
			  }
  		} );
  	}
  	else {
  		block();
  	}
  }
}

