/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain;

//#MIDP_EXCLUDE_FILE
//#APIDOC_EXCLUDE_FILE

import rca.util.leap.Iterator;
import rca.util.leap.ArrayList;

import rca.core.AID;
import rca.domain.FIPAAgentManagement.DFAgentDescription;
import rca.domain.FIPAAgentManagement.Property;
import rca.domain.FIPAAgentManagement.ServiceDescription;
import rca.domain.KBManagement.*;

/**
 * Original author Elisabetta Cortese - TILab
 *
 */
public class DFMemKB extends MemKB {

  boolean entriesToDelete = false; // gets true if there's at least one entry to delete for the method clean
  
  /**
   * Constructor
   * @param maxResultLimit rca internal limit for maximum number of search results
   */
	public DFMemKB(int maxResultLimit){
		super(maxResultLimit);
		clean();
	}


  protected Object insert(Object name, Object fact) {
    DFAgentDescription desc = (DFAgentDescription)fact;
    if (desc.getLeaseTime() != null ) {
        entriesToDelete = true;
    }
  	return super.insert(name, fact);
  }

	/**
	   Scan the facts and remove those whose lease time has expired.
	 */
	protected void clean(){
	    
	  if (entriesToDelete) { 
	    ArrayList toBeRemoved = new ArrayList();
		  Iterator iter = facts.values().iterator();		
		  while(iter.hasNext()){
			  DFAgentDescription dfd = (DFAgentDescription) iter.next();
			  if(dfd.checkLeaseTimeExpired()) {
				  toBeRemoved.add(dfd.getName());
			  }
		  }
		  iter = toBeRemoved.iterator();
		  while (iter.hasNext()) {
			  facts.remove((AID) iter.next());
		  }
	  }
	}

	// match
	public final boolean match(Object template, Object fact) {
		return compare(template, fact);
	}
	
	public static final boolean compare(Object template, Object fact) {

		try {
		  DFAgentDescription templateDesc = (DFAgentDescription)template;
		  DFAgentDescription factDesc = (DFAgentDescription)fact;
	    // We must not return facts whose lease time has expired (no 
		  // matter if they match)
		  if(factDesc.checkLeaseTimeExpired())
		  	return false;
		  
		  // Match name
		  AID id1 = templateDesc.getName();
		  if(id1 != null) {
		    AID id2 = factDesc.getName();
		    if((id2 == null) || (!matchAID(id1, id2)))
		      return false;
		  }
	
		  // Match protocol set
		  Iterator itTemplate = templateDesc.getAllProtocols();
		  while(itTemplate.hasNext()) {
		    String templateProto = (String)itTemplate.next();
		    boolean found = false;
		    Iterator itFact = factDesc.getAllProtocols();
		    while(!found && itFact.hasNext()) {
		      String factProto = (String)itFact.next();
		      found = templateProto.equalsIgnoreCase(factProto);
		    }
		    if(!found)
		      return false;
		  }
	
		  // Match ontologies set
		  itTemplate = templateDesc.getAllOntologies();
		  while(itTemplate.hasNext()) {
		    String templateOnto = (String)itTemplate.next();
		    boolean found = false;
		    Iterator itFact = factDesc.getAllOntologies();
		    while(!found && itFact.hasNext()) {
		      String factOnto = (String)itFact.next();
		      found = templateOnto.equalsIgnoreCase(factOnto);
		    }
		    if(!found)
		      return false;
		  }
	
		  // Match languages set
		  itTemplate = templateDesc.getAllLanguages();
		  while(itTemplate.hasNext()) {
		    String templateLang = (String)itTemplate.next();
		    boolean found = false;
		    Iterator itFact = factDesc.getAllLanguages();
		    while(!found && itFact.hasNext()) {
		      String factLang = (String)itFact.next();
		      found = templateLang.equalsIgnoreCase(factLang);
		    }
		    if(!found)
		      return false;
		  }
	
		  // Match services set
		  itTemplate = templateDesc.getAllServices();
		  while(itTemplate.hasNext()) {
		    ServiceDescription templateSvc = (ServiceDescription)itTemplate.next();
		    boolean found = false;
		    Iterator itFact = factDesc.getAllServices();
		    while(!found && itFact.hasNext()) {
		      ServiceDescription factSvc = (ServiceDescription)itFact.next();
		      found = compareServiceDesc(templateSvc, factSvc);
		    }
		    if(!found)
		      return false;
		  }
	
		  return true;
		}
		catch(ClassCastException cce) {
		  return false;
		}
	}

	  // Helper method to compare two Service Description objects
	  public static final boolean compareServiceDesc(ServiceDescription template, ServiceDescription fact) {
	
	    // Match name
	    String n1 = template.getName();
	    if(n1 != null) {
	      String n2 = fact.getName();
	      if((n2 == null) || (!n1.equalsIgnoreCase(n2)))
		return false;
	    }
	
	    // Match type
	    String t1 = template.getType();
	    if(t1 != null) {
	      String t2 = fact.getType();
	      if((t2 == null) || (!t1.equalsIgnoreCase(t2)))
		return false;
	    }
	
	    // Match ownership
	    String o1 = template.getOwnership();
	    if(o1 != null) {
	      String o2 = fact.getOwnership();
	      if((o2 == null) || (!o1.equalsIgnoreCase(o2)))
		return false;
	    }
	
	    // Match ontologies set
	    Iterator itTemplate = template.getAllOntologies();
	    while(itTemplate.hasNext()) {
	      String templateOnto = (String)itTemplate.next();
	      boolean found = false;
	      Iterator itFact = fact.getAllOntologies();
	      while(!found && itFact.hasNext()) {
		String factOnto = (String)itFact.next();
		found = templateOnto.equalsIgnoreCase(factOnto);
	      }
	      if(!found)
		return false;
	    }
	
	    // Match languages set
	    itTemplate = template.getAllLanguages();
	    while(itTemplate.hasNext()) {
	      String templateLang = (String)itTemplate.next();
	      boolean found = false;
	      Iterator itFact = fact.getAllLanguages();
	      while(!found && itFact.hasNext()) {
		String factLang = (String)itFact.next();
		found = templateLang.equalsIgnoreCase(factLang);
	      }
	      if(!found)
		return false;
	    }
	
	    // Match protocols set
	    itTemplate = template.getAllProtocols();
	    while(itTemplate.hasNext()) {
	      String templateProto = (String)itTemplate.next();
	      boolean found = false;
	      Iterator itFact = fact.getAllProtocols();
	      while(!found && itFact.hasNext()) {
		String factProto = (String)itFact.next();
		found = templateProto.equalsIgnoreCase(factProto);
	      }
	      if(!found)
		return false;
	    }
	
	    // Match properties set
	    itTemplate = template.getAllProperties();
	    while(itTemplate.hasNext()) {
	      Property templateProp = (Property)itTemplate.next();
	      boolean found = false;
	      Iterator itFact = fact.getAllProperties();
	      while(!found && itFact.hasNext()) {
					Property factProp = (Property)itFact.next();
					if (templateProp.getName().equals(factProp.getName())) {
						// The property name matches. Check the value
						Object templateValue = templateProp.getValue(); 
						if (templateValue == null) {
							found = true;
						}
						else if (templateValue instanceof String) {
		    			found = ((String)templateValue).equalsIgnoreCase(factProp.getValue().toString());
						}
						else {
		    			found = templateValue.equals(factProp.getValue());
						}
	      	}
	      }
	      if(!found)
					return false;
	    }
	
	    return true;
	  }
}
