/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain;

import rca.content.Predicate;

import rca.lang.acl.ACLMessage;

/**
This class represents a generic FIPAException, i.e. one of
NotUnderstood,Failure,Refuse, as defined in
<code>rca.domain.FIPAAgentManagement</code>.  It has two
constructors, one based on an ACLMessage, and the second based on its
content, i.e. the exception message.
Original author Giovanni Rimassa - Universita' di Parma
@version $Date: 2003-11-18 17:26:38 +0100 (Tue, 18 Nov 2003) $ $Revision: 4565 $
*/

public class FIPAException extends Exception implements Predicate {

    //#APIDOC_EXCLUDE_BEGIN

    /**
  @serial
  */
  protected ACLMessage msg; // can be accessed by subclasses

    //#APIDOC_EXCLUDE_END


  /**
  @serial
  */
  private String content;

  /**
   * Constructs a generic <code>FIPAException</code>. The ACL message
   * performative is defaulted to <code>not-understood</code>.
   * @param message is the content of the ACLMessage
  **/
  public FIPAException(String message) {
    super();
    content = message;
  }

  /**
   * Constructs a <code>FIPAException</code> from the given ACL
   * message.
   * @param message is the ACL message representing this exception
   **/
  public FIPAException(ACLMessage message) {
    this(message.getContent());
    msg=(ACLMessage)message.clone();
  }

  /**
   * Retrieve the ACL message whose content is represented by this
   * exception.
   * @return the ACLMessage representing this exception
   **/
  public ACLMessage getACLMessage() {
    if (msg == null) {
      msg = new ACLMessage(ACLMessage.NOT_UNDERSTOOD);
      msg.setContent(getMessage());
    }
    return msg;
  }

  /**
   * Set the content of the ACL message representing this exception
   * @param message is the content
   **/
  protected void setMessage(String message) {
    content=message;
    if (msg!=null)
      msg.setContent(message);
  }

  /**
   * Get the content of the ACL message representing this exception
   * @return A string representing the message content that describes
   * this FIPA exception.
   **/
  public String getMessage() {
    return content;
  }

}



