/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain;

import rca.core.AID;
import rca.core.Agent;
import rca.lang.acl.ACLMessage;
import rca.lang.acl.MessageTemplate;

/*import rca.content.ContentManager;
import rca.content.lang.StringCodec;
import rca.content.lang.sl.SLCodec;
import rca.content.onto.Ontology;
import rca.content.onto.basic.Action;
import rca.content.onto.basic.Result;
*/
import rca.domain.FIPANames;
import rca.domain.FIPAAgentManagement.FIPAManagementVocabulary;

/**
 * This class provides a set of basic and static methods to perform the FIPA Agent Management actions.
 * However, developers should use <code>DFService</code> and <code>AMSService</code>
 * which provide specialized methods to communicate with the DF and the AMS.
 * Original author Fabio Bellifemine - CSELT S.p.A.
 * @version $Date: 2005-11-16 10:24:28 +0100 (Wed, 16 Nov 2005) $ $Revision: 5828 $  
 **/
public class FIPAService {

  /**
   * create a REQUEST message with the following slots:
   * <code> (REQUEST :sender sender.getAID() :receiver receiver
   * :protocol fipa-request :language FIPA-SL0 :ontology fipa-agent-management
   * :reply-with xxx :conversation-id xxx) </code>
   * where <code>xxx</code> are unique words generated on the basis of
   * the sender's name and the current time.
   * @param sender is the Agent sending the message
   * @param receiver is the AID of the receiver agent
   * @return an ACLMessage object 
   */
  static ACLMessage createRequestMessage(Agent sender, AID receiver) {
    ACLMessage request = new ACLMessage(ACLMessage.REQUEST);
    request.setSender(sender.getAID());
    request.addReceiver(receiver);
    request.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
    request.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
    request.setOntology(FIPAManagementVocabulary.NAME);
    request.setReplyWith("rw"+sender.getName()+System.currentTimeMillis());
    request.setConversationId("conv"+sender.getName()+System.currentTimeMillis());
    return request;
  }



  /**
   * This method plays the initiator role in the Fipa-Request
   * interaction protocol and performs all the steps of the
   * protocol. The method uses the
   * <code>:reply-with</code>/<code>:in-reply-to</code> ACL message
   * slots as a mechanism to match the protocol replies.
   * Take care because the method blocks until all the response messages are received.
   * Under error conditions, or if the responder does not wish to respond, that
   * might block for ever the execution of the agent.
   * For this reason, the <code>FipaRequestInitiatorBehaviour</code> is the preferred
   * way to play the protocol.
   * @param a is the Agent playing the initiator role
   * @param request is the ACLMessage to be sent. Notice that all the
   * slots of the message must have already been filled by the
   * caller. If the <code>:reply-with</code> message slot is not set,
   * a default one will be generated automatically.
   * @return the INFORM message received in the final state of the protocol, if
   * the protocol succeeded, otherwise it throws an Exception
   */
  public static ACLMessage doFipaRequestClient(Agent a, ACLMessage request) throws FIPAException {
      return doFipaRequestClient(a, request, 0);
  }

  /**
   * This method plays the initiator role in the Fipa-Request
   * interaction protocol and performs all the steps of the protocol,
   * and additionally sets a conversation timeout. The method uses the
   * <code>:reply-with</code>/<code>:in-reply-to</code> ACL message
   * slots as a mechanism to match the protocol replies.  Under error
   * conditions, or if the responder does not wish to respond, that
   * might block the execution of the agent until the conversation
   * timeout expires.  For this reason, the
   * <code>FipaRequestInitiatorBehaviour</code> is the preferred way
   * to play the protocol.
   * @param a is the Agent playing the initiator role
   * @param request is the ACLMessage to be sent. Notice that all the
   * slots of the message must have already been filled by the
   * caller. If the <code>:reply-with</code> message slot is not set,
   * a default one will be generated automatically.
   * @param timeout The maximum time to wait for the conversation to finish, in milliseconds.
   * @return the INFORM message received in the final state of the protocol, if
   * the protocol succeeded, otherwise it throws an Exception
   */
  public static ACLMessage doFipaRequestClient(Agent a, ACLMessage request, long timeout) throws FIPAException {
      // If the request message does not have a ':reply-with' slot set
    if (request.getReplyWith() == null)
      request.setReplyWith("rw"+a.getLocalName()+System.currentTimeMillis());

        long sendTime = System.currentTimeMillis();
    a.send(request);
    MessageTemplate mt = MessageTemplate.MatchInReplyTo(request.getReplyWith());

        ACLMessage reply = a.blockingReceive(mt, timeout);

        if(reply != null) {
            if (reply.getPerformative() == ACLMessage.INFORM) {
                return reply;
            }
            if (reply.getPerformative() == ACLMessage.AGREE){
                // We received an AGREE --> Go back waiting for the INFORM unless
                // a timeout was set and it is expired.
                if (timeout > 0) {
                    long agreeTime = System.currentTimeMillis();
                    timeout -= (agreeTime - sendTime);
                    if (timeout <= 0) {
                        return null;
                    }
                }
                reply = a.blockingReceive(mt, timeout);
            }
            if(reply != null) {
                if (reply.getPerformative() == ACLMessage.INFORM){
                    return reply;
                }
                else {
                    // We received a REFUSE, NOT_UNDERSTOOD, FAILURE or OUT_OF_SEQUENCE --> ERROR
                    throw new FIPAException(reply);
                }
            }
        }
        // The timeout has expired
        return null;
  }

    /**
       Default constructor
    */
    public FIPAService() {
    }

}
