/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain.KBManagement;

//#J2ME_EXCLUDE_FILE

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

/**
 * This class provides an abstract implementation of the 
 * <code>KB</code> interface where information are stored in
 * a database.
 * 
 * Original author Roland Mungenast - Profactor
 */
public abstract class DBKB extends KB {
	
	/**
	 * Used database driver
	 */
	protected String driver = "sun.jdbc.odbc.JdbcOdbcDriver";
	
	/**
	 * Used database connection
	 */
	protected Connection conn = null;
	
	/**
	 * Specifies whether the KB should delete all existing tables for the DF at startup
	 */
	protected boolean cleanTables;
	
	/**
	 * Constructs a new <code>DFKB</code> and establishes a connection to the database
	 * at the given URL using the <code>sun.jdbc.odbc.JdbcOdbcDriver</code> driver.
	 * 
	 * @param maxResultLimit internal limit for the number of maximum search results.
	 * @param url database URL
	 * @param maxResultLimit rca internal limit for the maximum number of search results
	 * @param cleanTables specifies whether the KB should delete all existing tables for the DF at startup
	 * @throws SQLException an error occured while opening a connection to the database
	 */
	public DBKB(String url, int maxResultLimit, boolean cleanTables) throws SQLException {
		this(null, url, maxResultLimit, cleanTables);
	}
	
	/**
	 * Constructs a new <code>DFKB</code> and establishes a connection to the database.
	 * 
	 * @param maxResultLimit internal limit for the number of maximum search results.
	 * @param drv database driver
	 * @param url database URL
	 * @param user database user name
	 * @param passwd database password
	 * @param maxResultLimit rca internal limit for the maximum number of search results
	 * @param cleanTables specifies whether the KB should delete all existing tables for the DF at startup
	 * @throws SQLException an error occured while opening a connection to the database
	 */
	public DBKB(String drv, String url, int maxResultLimit, boolean cleanTables) throws SQLException {
		this(drv, url, null, null, maxResultLimit, cleanTables);
	}
	
	/**
	 * Constructs a new <code>DFKB</code> and establishes a connection to the database.
	 * 
	 * @param maxResultLimit internal limit for the number of maximum search results.
	 * @param drv database driver
	 * @param url database URL
	 * @param user database user name
	 * @param passwd database password
	 * @param maxResultLimit rca internal limit for the maximum number of search results
	 * @param cleanTables specifies whether the KB should delete all existing tables for the DF at startup
	 * @throws SQLException an error occured while opening a connection to the database
	 */
	public DBKB(String drv, String url, String user, String passwd, int maxResultLimit, boolean cleanTables) throws SQLException {
		super(maxResultLimit);
		this.cleanTables = cleanTables;
		loadDBDriver(drv);
		setDBConnection(drv, url, user, passwd);
		setup();
	}
	
	/**
	 * Loads an JDBC driver
	 * @param drv driver name or <code>null</code> </ br>
	 * if the default JDBC-ODBC driver should be used
	 * @throws SQLException if the driver cannot be loaded
	 */
	protected void loadDBDriver(String drv) throws SQLException {
		//  Load DB driver
		try {
			if(drv != null) {
				if(!drv.equals("null"))
					driver = drv;
			}
			Class.forName(driver).newInstance();
		}
		catch(Exception e){
			throw new SQLException("Error loading driver "+driver+". "+e);
		}
	}
	
	/**
	 * Establishes a new connection to the database and stores a reference in the
	 * local attribute <code>conn</code>
	 * @param drv database driver
	 * @param url database URL
	 * @param user database user
	 * @param passwd database password
	 * @throws SQLException if a database access error occurs
	 */
	protected void setDBConnection(String drv, String url, String user, String passwd) throws SQLException {
		// Connect to the DB
		if (user != null) {
			conn = DriverManager.getConnection(url, user, passwd);
		}
		else {
			conn = DriverManager.getConnection(url);
		}
	}
	
	/**
	 * This method is called by the constructor after a
	 * connection to the database has been established. 
	 */
	abstract protected void setup() throws SQLException;
}
