/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.domain;

//#MIDP_EXCLUDE_FILE

//#APIDOC_EXCLUDE_FILE


import rca.core.Agent;
import rca.core.CaseInsensitiveString;

import rca.content.onto.OntologyException;
import rca.content.lang.Codec.CodecException;
import rca.content.onto.basic.Action;

import rca.lang.acl.ACLMessage;
import rca.lang.acl.MessageTemplate;

import rca.proto.SimpleAchieveREResponder;

import rca.domain.FIPAException;
import rca.domain.FIPAAgentManagement.FailureException;
import rca.domain.FIPAAgentManagement.RefuseException;
import rca.domain.FIPAAgentManagement.UnsupportedValue;
import rca.domain.FIPAAgentManagement.NotUnderstoodException;
import rca.security.JADESecurityException;
import rca.domain.FIPAAgentManagement.ExceptionVocabulary;

import rca.util.Logger;

/**
   Base class for AMS and DF behaviours managing requests from agents.
   This class handles the FIPA-request protocol and in particular prepares
   the response taking into account all possible exceptions.
   The preparation of the result notification is delegated to subclasses as
   its form (RESULT or DONE) and sending time (i.e. whether it can be sent 
   immediately or must be delayed at a later time) depends on the specific 
   action.
   Original author Giovanni Caire - Tilab
*/
public abstract class RequestManagementBehaviour extends SimpleAchieveREResponder {
	private ACLMessage notification;
	private Logger myLogger;

  protected RequestManagementBehaviour(Agent a, MessageTemplate mt){
		super(a,mt);
		myLogger = Logger.getMyLogger(myAgent.getLocalName());
  }
 
  protected abstract ACLMessage performAction(Action slAction, ACLMessage request) throws JADESecurityException, FIPAException; 

  /**
   * @return null when the AGREE message can be skipper, the AGREE message
   * otherwise.
   */
  protected ACLMessage prepareResponse(ACLMessage request) throws NotUnderstoodException, RefuseException {
  	ACLMessage response = null;  
  	Throwable t = null;
		try{	
			// Check the language is SL0, SL1, SL2 or SL. 
	    isAnSLRequest(request);

	    // Extract the content
	    Action slAction = (Action) myAgent.getContentManager().extractContent(request);
	    
	    // Perform the action
	    notification = performAction(slAction, request);
	    
	    // Action OK 
		} 
		catch (OntologyException oe) {
			// Error decoding request --> NOT_UNDERSTOOD
		        response = request.createReply();
			response.setPerformative(ACLMessage.NOT_UNDERSTOOD);
			response.setContent("("+ExceptionVocabulary.UNRECOGNISEDVALUE+" content)");
			t = oe;
		}	
		catch (CodecException ce) {
			// Error decoding request --> NOT_UNDERSTOOD
		  response = request.createReply();
			response.setPerformative(ACLMessage.NOT_UNDERSTOOD);
			response.setContent("("+ExceptionVocabulary.UNRECOGNISEDVALUE+" content)");
			t = ce;
		}	
		catch (RefuseException re) {
			// RefuseException thrown during action execution --> REFUSE
		  response = request.createReply();
			response.setPerformative(ACLMessage.REFUSE);
			response.setContent(prepareErrorContent(request.getContent(), re.getMessage()));
			t = re;
		}	
		catch (FailureException fe) {
			// FailureException thrown during action execution --> FAILURE
			notification = request.createReply();
			notification.setPerformative(ACLMessage.FAILURE);
			notification.setContent(prepareErrorContent(request.getContent(), fe.getMessage()));
			t = fe;
		}	
		catch(FIPAException fe){
			// Malformed request --> NOT_UNDERSTOOD
		  response = request.createReply();
			response.setPerformative(ACLMessage.NOT_UNDERSTOOD);
			response.setContent("("+fe.getMessage()+")");
			t = fe;
		}
		catch(Throwable tr){
			tr.printStackTrace();
			// Generic error --> FAILURE
			notification = request.createReply();
			notification.setPerformative(ACLMessage.FAILURE);
			
			notification.setContent(prepareErrorContent(request.getContent(), ExceptionVocabulary.INTERNALERROR+" \""+t+"\""));
		}
		if (t != null) {
			if (myLogger.isLoggable(Logger.CONFIG)) {
				myLogger.log(Logger.CONFIG, "Error handling request: "+t);
				t.printStackTrace();
			}
		}
		return response;
  }
    
  /**
     Just return the (already prepared) notification message (if any).
   */
  protected ACLMessage prepareResultNotification(ACLMessage request, ACLMessage response) throws FailureException{	
		return notification;       
  }
    
  //to reset the action
  public void reset(){
		super.reset();
		notification = null;
  }

  private void isAnSLRequest(ACLMessage msg) throws FIPAException { 
		String language = msg.getLanguage();
		if ( (!CaseInsensitiveString.equalsIgnoreCase(FIPANames.ContentLanguage.FIPA_SL0, language)) &&
	  	   (!CaseInsensitiveString.equalsIgnoreCase(FIPANames.ContentLanguage.FIPA_SL1, language)) &&
	    	 (!CaseInsensitiveString.equalsIgnoreCase(FIPANames.ContentLanguage.FIPA_SL2, language)) &&
	     	 (!CaseInsensitiveString.equalsIgnoreCase(FIPANames.ContentLanguage.FIPA_SL, language))) {
			throw new UnsupportedValue("language");
    }
  }

  private String prepareErrorContent(String content, String e) {
  	String tmp = content.trim();
  	tmp = tmp.substring(1, tmp.length()-1);
  	return "("+tmp+" "+e+")";
  }
}
