/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/


package rca.domain.introspection;

//#APIDOC_EXCLUDE_FILE

import rca.core.AID;
import rca.core.BehaviourID;
import rca.domain.introspection.Event;

/**

   An introspection event, recording the change of the state of an
   agent behaviour within the platform.

   Original author Giovanni Rimassa -  Universita' di Parma
   @version $Date: 2005-02-16 18:18:28 +0100 (Wed, 16 Feb 2005) $ $Revision: 5552 $
*/

public class ChangedBehaviourState implements Event {

    /**
       A string constant for the name of this event.
    */ 
    public static final String NAME = "Changed-Behaviour-State";

    private AID agent;
    private BehaviourID behaviour;
    private String from;
    private String to;


    /**
       Default constructor. A default constructor is necessary for
       ontological classes.
    */
    public ChangedBehaviourState() {
    }

    /**
       Set the <code>agent</code> slot of this event.
       @param id The agent identifier of the agent whose behaviour
       state changed.
    */
    public void setAgent(AID id) {
	agent = id;
    }

    /**
       Retrieve the value of the <code>agent</code> slot of this
       event, containing the agent identifier of the agent whose
       behaviour state changed.
       @return The value of the <code>agent</code> slot, or
       <code>null</code> if no value was set.
    */
    public AID getAgent() {
	return agent;
    }

    /**
       Set the <code>behaviour</code> slot of this event.
       @param id The behaviour identifier of the behaviour whose state
       changed.
    */
    public void setBehaviour(BehaviourID id) {
	behaviour = id;
    }

    /**
       Retrieve the value of the <code>behaviour</code> slot of this
       event, containing the behaviour identifier of the behaviour
       whose state changed.
       @return The value of the <code>behaviour</code> slot, or
       <code>null</code> if no value was set.
    */
    public BehaviourID getBehaviour() {
	return behaviour;
    }

    /**
       Set the <code>from</code> slot of this event.
       @param s The name of the state the behaviour was in before this
       event occurred.
    */
    public void setFrom(String s) {
	from = s;
    }

    /**
       Retrieve the value of the <code>from</code> slot of this event,
       containing the name of the state the behaviour was in before
       this event occurred.
       @return The value of the <code>from</code> slot, or
       <code>null</code> if no value was set.
    */
    public String getFrom() {
	return from;
    }

    /**
       Set the <code>to</code> slot of this event.
       @param s The name of the state the behaviour was in after this
       event occurred.
    */
    public void setTo(String s) {
	to = s;
    }

    /**
       Retrieve the value of the <code>to</code> slot of this event,
       containing the name of the state the behaviour was in after
       this event occurred.
       @return The value of the <code>to</code> slot, or
       <code>null</code> if no value was set.
    */
    public String getTo() {
	return to;
    }

    /**
       Retrieve the name of this event.
       @return A constant value for the event name.
    */
    public String getName() {
	return NAME;
    }

    /**
       Retrieve an SL0-like string representation for this object.
    */
    public String toString() {
  	return new String(NAME+ " Name: "+ behaviour.getName() + " from: " + from + " to: " + to);
    }

}
