/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/
package rca.domain.introspection;

import rca.core.Agent;
import rca.core.behaviours.CyclicBehaviour;
import rca.lang.acl.*;
import rca.domain.FIPANames;
import rca.content.ContentManager;
import rca.content.onto.Ontology;
import rca.content.onto.OntologyException;
import rca.content.onto.basic.*;
import rca.content.lang.Codec;
import rca.content.lang.Codec.CodecException;
import rca.content.lang.sl.SLCodec;
import rca.util.leap.List;
import rca.util.leap.ArrayList;

import java.lang.reflect.*;

public class IntrospectionServer extends CyclicBehaviour {
    private Codec codec;
    private Ontology onto;
    private MessageTemplate template;

    private static Class serializableClass;

    static {
        try {
            serializableClass = Class.forName("java.io.Serializable");
        }
        catch (Exception e) {
        }
    }

    public IntrospectionServer(Agent a) {
        super(a);
    }

    public void onStart() {
        ContentManager cm = myAgent.getContentManager();

        onto = IntrospectionOntology.getInstance();
        cm.registerOntology(onto);

        codec = cm.lookupLanguage(FIPANames.ContentLanguage.FIPA_SL);
        if (codec == null) {
            codec = cm.lookupLanguage(FIPANames.ContentLanguage.FIPA_SL2);
            if (codec == null) {
                codec = cm.lookupLanguage(FIPANames.ContentLanguage.FIPA_SL1);
                if (codec == null) {
                    codec = cm.lookupLanguage(FIPANames.ContentLanguage.FIPA_SL0);
                }
            }
        }
        if (codec == null) {
            codec = new SLCodec();
            cm.registerLanguage(codec);
        }

        template = MessageTemplate.and(
                MessageTemplate.MatchOntology(onto.getName()),
                MessageTemplate.MatchPerformative(rca.lang.acl.ACLMessage.REQUEST) );
    }

    public void action() {
        rca.lang.acl.ACLMessage request = myAgent.receive(template);
        if (request != null) {
            try {
                ContentManager cm = myAgent.getContentManager();
                Action actionExpr = (Action) cm.extractContent(request);
                Object act = actionExpr.getAction();
                if (act instanceof GetKeys) {
                    serveGetKeys(request, actionExpr, (GetKeys) act);
                }
                else if (act instanceof GetValue) {
                    serveGetValue(request, actionExpr, (GetValue) act);
                }
                else {
                    serveUnknownAction(request, actionExpr, act);
                }
            }
            catch (OntologyException oe) {
                reply(request, rca.lang.acl.ACLMessage.NOT_UNDERSTOOD);
                oe.printStackTrace();
            }
            catch (CodecException ce) {
                reply(request, rca.lang.acl.ACLMessage.NOT_UNDERSTOOD);
                ce.printStackTrace();
            }
            catch (ValueEncodingException vee) {
                rca.lang.acl.ACLMessage msg = request.createReply();
                msg.setPerformative(rca.lang.acl.ACLMessage.FAILURE);
                msg.setContent("VALUE_NOT_ENCODABLE");
                myAgent.send(msg);
            }
            catch (Throwable t) {
                reply(request, rca.lang.acl.ACLMessage.FAILURE);
                t.printStackTrace();
            }
        }
        else {
            block();
        }
    }

    protected void reply(rca.lang.acl.ACLMessage request, int performative) {
        rca.lang.acl.ACLMessage msg = request.createReply();
        msg.setPerformative(performative);
        myAgent.send(msg);
    }

    protected void serveGetKeys(rca.lang.acl.ACLMessage request, Action aExpr, GetKeys action) throws Exception {
        List keys = new ArrayList();
        Method[] mm = myAgent.getClass().getMethods();
        for (int i = 0; i < mm.length; ++i) {
            Method method = mm[i];
            if (method.getName().startsWith("get") && method.getParameterTypes().length == 0) {
                Class retType = method.getReturnType();
                if (retType.isPrimitive() || (serializableClass != null && serializableClass.isAssignableFrom(retType))) {
                    String key = method.getName().substring(3);
                    keys.add(key);
                }
            }
        }
        Result r = new Result(aExpr, keys);
        rca.lang.acl.ACLMessage reply = request.createReply();
        myAgent.getContentManager().fillContent(reply, r);
        reply.setPerformative(rca.lang.acl.ACLMessage.INFORM);
        myAgent.send(reply);
    }

    protected void serveGetValue(rca.lang.acl.ACLMessage request, Action aExpr, GetValue action) throws Exception {
        Method method = myAgent.getClass().getMethod("get"+action.getKey(), new Class[]{});
        Object value = method.invoke(myAgent, (Object[]) null);
        if (value == null) {
            value = "null";
        }
        Result r = new Result(aExpr, value);
        rca.lang.acl.ACLMessage reply = request.createReply();
        try {
            myAgent.getContentManager().fillContent(reply, r);
            reply.setPerformative(rca.lang.acl.ACLMessage.INFORM);
            myAgent.send(reply);
        }
        catch (OntologyException oe) {
            throw new ValueEncodingException();
        }
        catch (CodecException ce) {
            throw new ValueEncodingException();
        }
    }

    protected void serveUnknownAction(rca.lang.acl.ACLMessage request, Action aExpr, Object action) {
        reply(request, rca.lang.acl.ACLMessage.REFUSE);
    }

    private class ValueEncodingException extends Exception {
    }
}
