/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.tools;

//#APIDOC_EXCLUDE_FILE

import rca.core.Agent;

import rca.domain.FIPANames;
import rca.domain.FIPAAgentManagement.*;
import rca.domain.JADEAgentManagement.*;
import rca.domain.introspection.*;

import rca.lang.acl.ACLMessage;

import rca.content.lang.sl.SLCodec;

/**

  This abstract class is the common ancestor of all rca tools (RMA,
  Sniffer, Introspector, etc.). It provides suitable behaviours to
  interact with the AMS, registering for interesting events and
  requesting actions when needed.

  Original author Giovanni Rimassa - Universita' di Parma
  @version $Date: 2004-09-22 15:04:40 +0200 (Wed, 22 Sep 2004) $ $Revision: 5372 $

 */
public abstract class ToolAgent extends Agent {

  private ACLMessage AMSSubscription = new ACLMessage(ACLMessage.SUBSCRIBE);
  private ACLMessage AMSCancellation = new ACLMessage(ACLMessage.CANCEL);
  private ACLMessage AMSRequest = new ACLMessage(ACLMessage.REQUEST);


  // This is left here for backward compatibility
  public static interface EventHandler extends AMSSubscriber.EventHandler {
  }
  
    /**
       This abstract behaviour is used to receive notifications from
       the AMS. 
    */
  protected abstract class AMSListenerBehaviour extends AMSSubscriber {
  	// Redefine the onStart() method not to automatically subscribe.
  	public void onStart() {
  	}
  } // End of AMSListenerBehaviour class

  /**
   * Default constructor.
   */
    public ToolAgent() {
    }

  /**
     This method is invoked just after the generic agent
     setup. Subclasses must use this method the same way ordinary
     agents use their <code>setup()</code> method.
   */
  protected void toolSetup() {

  }

  /**
     This method is invoked just before the generic agent
     takedown. Subclasses must use this method the same way ordinary
     agents use their <code>takeDown()</code> method.
   */
  protected void toolTakeDown() {

  }

    /**
       Retrieve the <code>subscribe</code> ACL message with which this
       tool agent subscribed with the AMS.
       @return The subscription ACL message.
    */
  protected ACLMessage getSubscribe() {
    return AMSSubscription;
  }

    /**
       Retrieve the <code>cancel</code> ACL message with which this
       tool agent removes its subscription with the AMS.
       @return The cancellation ACL message.
    */
  protected ACLMessage getCancel() {
    return AMSCancellation;
  }

    /**
       Retrieve the <code>request</code> ACL message with which this
       tool agent requests the AMS tool-specific actions.
       @return The request ACL message.
    */
  protected ACLMessage getRequest() {
    return AMSRequest;
  }

  public final void setup() {

    // Register the supported ontologies
    getContentManager().registerOntology(JADEManagementOntology.getInstance());
    getContentManager().registerOntology(IntrospectionOntology.getInstance());
    getContentManager().registerOntology(FIPAManagementOntology.getInstance());

    // register the supported languages
    SLCodec codec = new SLCodec();
    getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL0);
    getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL1);
    getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL2);
    getContentManager().registerLanguage(codec, FIPANames.ContentLanguage.FIPA_SL);

    // Fill ACL messages fields
    AMSSubscription.setSender(getAID());
    AMSSubscription.clearAllReceiver();
    AMSSubscription.addReceiver(getAMS());
    AMSSubscription.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
    AMSSubscription.setOntology(IntrospectionOntology.NAME);
    AMSSubscription.setReplyWith(AMSSubscriber.AMS_SUBSCRIPTION);
    AMSSubscription.setConversationId(getLocalName());

    String content = AMSSubscriber.PLATFORM_EVENTS;
    AMSSubscription.setContent(content);

    AMSCancellation.setSender(getAID());
    AMSCancellation.clearAllReceiver();
    AMSCancellation.addReceiver(getAMS());
    AMSCancellation.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
    AMSCancellation.setOntology(IntrospectionOntology.NAME);
    AMSCancellation.setReplyWith(AMSSubscriber.AMS_CANCELLATION);
    AMSCancellation.setConversationId(getLocalName());
    // No content is needed (cfr. FIPA 97 Part 2 page 26)

    AMSRequest.setSender(getAID());
    AMSRequest.clearAllReceiver();
    AMSRequest.addReceiver(getAMS());
    AMSRequest.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
    AMSRequest.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);

    // Call tool-specific setup
    toolSetup();

  }

  protected final void takeDown() {
    // Call tool-specific takedown
    toolTakeDown();
  }

  protected void afterClone() {
      refillContentManager();
  }

  protected void afterMove() {
      refillContentManager();
  }

  protected void afterLoad() {
      refillContentManager();
  }

  protected void afterThaw() {
      refillContentManager();
  }

  protected void afterReload() {
      refillContentManager();
  }

  protected void beforeSave() {
  }

  protected void beforeFreeze() {
  }

  protected void beforeReload() {
  }

  private void refillContentManager() {
    // Register the supported ontologies
    getContentManager().registerOntology(JADEManagementOntology.getInstance());
    getContentManager().registerOntology(IntrospectionOntology.getInstance());
    getContentManager().registerOntology(FIPAManagementOntology.getInstance());

    // register the supported languages
    getContentManager().registerLanguage(new SLCodec(), FIPANames.ContentLanguage.FIPA_SL0);

    AMSSubscription.setSender(getAID());
    AMSSubscription.clearAllReceiver();
    AMSSubscription.addReceiver(getAMS());
    AMSSubscription.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
    AMSSubscription.setOntology(IntrospectionOntology.NAME);
    AMSSubscription.setReplyWith(AMSSubscriber.AMS_SUBSCRIPTION);
    AMSSubscription.setConversationId(getLocalName());

    String content = AMSSubscriber.PLATFORM_EVENTS;
    AMSSubscription.setContent(content);

    AMSCancellation.setSender(getAID());
    AMSCancellation.clearAllReceiver();
    AMSCancellation.addReceiver(getAMS());
    AMSCancellation.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
    AMSCancellation.setOntology(IntrospectionOntology.NAME);
    AMSCancellation.setReplyWith(AMSSubscriber.AMS_CANCELLATION);
    AMSCancellation.setConversationId(getLocalName());

    AMSRequest.setSender(getAID());
    AMSRequest.clearAllReceiver();
    AMSRequest.addReceiver(getAMS());
    AMSRequest.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
    AMSRequest.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);

  }

}
