/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.util;
 
import java.util.Vector;
import rca.core.behaviours.Behaviour;

/**
 * This class implements a FIFO queue of objects that can be put and got
 * in a synchronized way. This is useful when an external thread, 
 * e.g. a GUI, has to communicate with an agent: The external thread
 * puts objects in the queue and the agent gets and processes them.
 * The queue can be associated to a <code>Behaviour</code>. This
 * Behaviour will be restarted each time an object is inserted in the 
 * queue. 
 * This class can be effectively used in combination with the 
 * <code>Event</code> class to support a synchronization between the 
 * external therad (posting the event in the <code>InputQueue</code>)
 * and the Agent thread (processing the event).
 * @see rca.util.Event
 * Original author Giovanni Caire - TILab
 */
public class InputQueue {
	private Vector queue = new Vector();
	private Behaviour myManager;

    /**
       Default constructor.
    */
    public InputQueue() {
    }

	/**
	   Associate this <code>InputQueue</code> object with the indicated
	   <code>Behaviour</code> so that it will be restarted each time
	   a new object is inserted.
	   @param b The <code>Behaviour</code> to associate.
	 */
	public synchronized void associate(Behaviour b) {

		myManager = b;
		// If there were objects already inserted --> restart the manager 
		// so that it can manages them
		if (myManager != null && queue.size() > 0) {
			myManager.restart();
		}
	}
	
	/**
	   Insert an object into the queue. If there is a <code>Behaviour</code>
	   associated to this <code>InputQueue</code> it will be restarted.
	   @param obj The object to insert.
	 */
	public synchronized void put(Object obj) {
		queue.addElement(obj);
		
		// Restart the manager behaviour (if any) so that it can manage
		// the object
		if (myManager != null) {
			myManager.restart();
		}
	}	
	
	/**
	   Extract the first object in the queue (if any).
	   @return The first object in the queue or <code>null</code> if
	   the queue is empty.
	 */
	public synchronized Object get() {
		Object obj = null;
		if (queue.size() > 0) {
			obj = queue.elementAt(0);
			queue.removeElementAt(0);
		}
		return obj;
	}			

    /**
       Remove all elements from this queue.
    */
        public synchronized void clear() {
	    queue.removeAllElements();
	}

}
