/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.util;
 
//#J2ME_EXCLUDE_FILE
//#APIDOC_EXCLUDE_FILE

import java.io.*;
import java.text.*;
import java.util.Date;

/**
   This class is an OutputStream whose output is stored in 
   in different files according to the day it is produced.
	 Existing files, if any, are not rewritten as output is appended
	 at the end of the file. 
   Original author Giovanni Caire - TILab
 */
public class PerDayFileLogger extends PrintStream {
  private final static long DAY = 24*60*60*1000;
  
	private String file;
	private long dayCnt;
	private static DateFormat df = new SimpleDateFormat("yyyyMMdd");

    /**
       Create a new day-based logger.
       @param file The name of the file to write logs to.
       @throws IOException If some filesystem operation fails.
    */
	public PerDayFileLogger(String file) throws IOException {
		super(new FileOutputStream(file+"."+df.format(new Date()), true), true);
		
		this.file = file;
		dayCnt = System.currentTimeMillis() / DAY;
	}
	
	private void checkDate() {
		long n = System.currentTimeMillis() / DAY;
		if (n > dayCnt) {
			dayCnt = n;
			try {
				out.close();
				out = new FileOutputStream(file+"."+df.format(new Date()), true);
			}
			catch (Exception e) {}
		}
	}

    /**
       Print a Java object to the proper file, replacing the file
       every new day.
       @param obj The Java object to print.
    */	
	public void print(Object obj) {
		checkDate();
		super.print(obj);
	}

    /**
       Print a string to the proper file, replacing the file every new
       day.
       @param str The string to print.
    */	
	public void print(String str) {
		checkDate();
		super.print(str);
	}
	
    /**
       Print a new line to the proper file, replacing the file every
       new day.
    */	
	public void println() {
		checkDate();
		super.println();
	}
	
    /**
       Print a Java object and a newline to the proper file, replacing
       the file every new day.
       @param obj The Java object to print.
    */	
	public void println(Object obj) {
		checkDate();
		super.println(obj);
	}
	
    /**
       Print a string and a new line to the proper file, replacing the
       file every new day.
       @param str THe string to print.
    */	
	public void println(String str) {
		checkDate();
		super.println(str);
	}
}
