/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.util;
 
//#APIDOC_EXCLUDE_FILE

/**
 * This class provides support for  
 * synchronizing threads acting on a generic resource in such a way that 
 * - If a thread is writing the resource no other thread can act on it 
 * in any way
 * - Several threads can read the resource at the same time
 * - If one or more threads are reading the resource no thread can write it
 * Original author Giovanni Caire - TILab
 */
public class RWLock {
	// The counter of threads currently reading the resource
	private int readersCnt = 0;
	
	// The Thread currently writing the resource (there can only be
	// one such a therad at a given time)
	private Thread currentWriter = null;
	
	// writeLock()/unlock() can be nested. This indicates the current 
	// depth
	private int writeLockDepth = 0;
	
	private Logger logger = Logger.getMyLogger(this.getClass().getName());

    /**
       Default constructor.
    */
    public RWLock() {
    }

    /**
       Acquire the protected resource with writing privileges. Only
       one writer at a time can access the protected resource, and no
       readers can access it at the same time. The locking is
       recursive (i.e. the same thread can acquire the lock multiple
       times, but must unlock it a matching number of times to
       actually free the protected resource).
    */
	public synchronized void writeLock() {
		Thread me = Thread.currentThread();
		while ((currentWriter != null && currentWriter != me) || readersCnt > 0) {
			// Someone (not me) is writing the resource OR
			// There are one or more Threads reading the resource
			// --> Go to sleep
			try {
				wait();
			}
			catch (InterruptedException ie) {
				if(logger.isLoggable(Logger.WARNING))
					logger.log(Logger.WARNING,"Unexpected interruption. "+ie.getMessage());
			}
		}
		writeLockDepth++;
		if (writeLockDepth == 1) {
			currentWriter = me;
			onWriteStart();
		}
	}
	
    /**
       Release the protected resource, previously acquired with
       writing privileges.
    */
	public synchronized void writeUnlock() {
		if (Thread.currentThread() == currentWriter) {
			writeLockDepth--;
			if (writeLockDepth == 0) {
				// I have finished writing the resource --> Wake up hanging threads
				currentWriter = null;
				notifyAll();
				onWriteEnd();
			}
		}
	}
	
    /**
       Acquire the protected resource with reading privileges. Many
       readers can access the protected resource at the same time, but
       no writer can access it while at least one reader is
       present. The locking is recursive (i.e. the same thread can
       acquire the lock multiple times, but must unlock it a matching
       number of times to actually free the protected resource).
    */
	public synchronized void readLock() {
		while (currentWriter != null) {
			// Someone is writing the resource --> Go to sleep
			try {
				wait();
			}
			catch (InterruptedException ie) {
				if(logger.isLoggable(Logger.WARNING))
					logger.log(Logger.WARNING,"Unexpected interruption. "+ie.getMessage());
			}
		}
		readersCnt++;
	}
	

    /**
       Release the protected resource, previously acquired with
       reading privileges.
    */
	public synchronized void readUnlock() {
		readersCnt--;
		if (readersCnt == 0) {
			// No one is reading the resource anymore --> Wake up threads
			// waiting to write it
			notifyAll();
		}
	}

    /**
       This placeholder method is called every time a thread actually
       acquires the protected resource with writing privileges (this
       means that, in case of multiple recursive locking by the same
       thread, this method is called only the first time). Subclasses
       can exploit this to transparently trigger a resource
       acquisition prolog.
    */	
	protected void onWriteStart() {
	}
	
    /**
       This placeholder method is called every time a thread actually
       releases the protected resource with writing privileges (this
       means that, in case of multiple recursive unlocking by the same
       thread, this method is called only the last time). Subclasses
       can exploit this to transparently trigger a resource release
       epilog.
    */	
	protected void onWriteEnd() {
	}
}
