/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.util;

//#APIDOC_EXCLUDE_FILE

import rca.util.leap.List;
import rca.util.leap.LinkedList;

/**
 * Implementation of a list of objects providing methods to 
 * synchronize threads acting on the list in such a way to prevent 
 * concurrent modifications (addition/remotion of elements) and 
 * concurrent scanning/modification. Concurrent scannings are allowed
 * instead.
 * Original author Giovanni Caire - TILab
 */
public class SynchList extends RWLock {
	// The actual list of objects
	private List innerList = null;


    /**
       Default constructor.
    */	
    public SynchList() {
    }

    /**
       This method grants writing privileges to the calling thread and
       grants access to the protected list.
       @return The inner, protected list.
    */
	public synchronized List startModifying() {
		writeLock();
		return innerList;
	}

    /**
       This method must be called when a writer thread has finished
       modifying the list, so that the associated readers-writer lock
       can be relinquished.
    */	
	public synchronized void stopModifying() {
		writeUnlock();
	}
	
    /**
       This method grants reading privileges to the calling thread and
       grants access to the protected list.
       @return The inner, protected list.
    */
	public synchronized List startScanning() {
		if (innerList != null) {
			readLock();
		}
		return innerList;
	}
	
    /**
       This method must be called when a reader thread has finished
       accessing the list, so that the associated readers-writer lock
       can be relinquished.
    */	
	public synchronized void stopScanning() {
		if (innerList != null) {
			readUnlock();
		}
	}

    //#APIDOC_EXCLUDE_BEGIN
	
	protected void onWriteStart() {
		if (innerList == null) {
			innerList = new LinkedList();
		}
	}
	
	protected void onWriteEnd() {
		if (innerList != null && innerList.size() == 0) {
			innerList = null;
		}
	}

    //#APIDOC_EXCLUDE_END

}
