/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.wrapper;

import rca.core.AID;
import rca.core.NotFoundException;
import rca.core.AgentManager;
import rca.core.AgentContainer;
import rca.core.MainContainer;

import rca.mtp.MTPException;

/**
   This class is a Proxy class, allowing access to a rca agent
   container. Invoking methods on instances of this class, it is
   possible to request services from <it>in-process</it> agent
   containers.
   This class must not be instantiated by applications. Instead, use
   the <code>createContainer()</code> method in class
   <code>Runtime</code>.
   <br>
   <b>NOT available in MIDP</b>
   <br>
   @see rca.core.Runtime#createAgentContainer(rca.core.Profile)

   Original author Giovanni Rimassa - Universita' di Parma
   Original author Giovanni Caire - TILAB
 */                                                      
public class ContainerController {

  private AgentContainer myImpl;
  private ContainerProxy myProxy;
  private String myPlatformName;

  protected PlatformController myPlatformController;

  /**
     This constructor requires a concrete
     implementation of a rca agent container, which cannot be
     instantiated by applications, so it cannot be meaningfully called
     from application code. The proper way to create an agent
     container from an application is to call the
     <code>Runtime.createContainer()</code> method.
     @see rca.core.Runtime#createAgentContainer(rca.core.Profile)
     @param impl A concrete implementation of a rca agent container.
     @param platformName the name of the platform
   */
  public ContainerController(ContainerProxy cp, AgentContainer impl, String platformName) {
    myProxy = cp;
      myImpl = impl;
    myPlatformName = platformName;
  }


  /**
   * Get agent proxy to local agent given its name.
   * @param localAgentName The short local name of the desired agent.
   * @throws ControllerException If any probelms occur obtaining this proxy.
   */
  public AgentController getAgent(String localAgentName) throws ControllerException {
    if(myImpl == null || myProxy == null) {
      throw new StaleProxyException();
    }

    AID agentID = new AID(localAgentName, AID.ISLOCALNAME);

    // Check that the agent exists
    rca.core.Agent instance = myImpl.acquireLocalAgent(agentID);
    if (instance == null) {
      throw new ControllerException("Agent " + localAgentName + " not found.");
    }
    myImpl.releaseLocalAgent(agentID);
    return new AgentController(agentID, myProxy, myImpl);
  }




  /**
     Creates a new rca agent, running within this container, 
     @param nickname A platform-unique nickname for the newly created
     agent. The agent will be given a FIPA compliant agent identifier
     using the nickname and the ID of the platform it is running on.
     @param className The fully qualified name of the class that
     implements the agent.
     @param args An object array, containing initialization parameters
     to pass to the new agent. 
     @return A proxy object, allowing to call state-transition forcing
     methods on the real agent instance.
   */
  public AgentController createNewAgent(String nickname, String className, Object[] args) throws StaleProxyException {
    if(myImpl == null || myProxy == null) {
      throw new StaleProxyException();
      }

    AID agentID = new AID(nickname, AID.ISLOCALNAME);

      try {
          myProxy.createAgent(agentID, className, args);
          return new AgentController(agentID, myProxy, myImpl);
      }
      catch (Throwable t) {
          throw new StaleProxyException(t);
      }
  }

    // HP Patch begin ----------------------------------------------------------------------------------
    /**
     * Add an Agent to this container. Typically Agent would be some class extending
     * Agent which was instantiated and configured.
     * @param nickname A platform-unique nickname for the newly created agent.
     * The agent will be given a FIPA compliant agent identifier using the nickname and
     * the ID of the platform it is running on.
     * @param anAgent The agent to be added to this agent container.
     * @return An AgentController, allowing to call state-transition forcing methods on the real agent instance.
     */
    public AgentController acceptNewAgent(String nickname, rca.core.Agent anAgent) throws StaleProxyException {
        if (myImpl == null || myProxy == null) {
          throw new StaleProxyException();
        }

      AID agentID = new AID(nickname, AID.ISLOCALNAME);
      // FIXME: This call skips the security checks on the local container
      try {
                    rca.core.NodeDescriptor nd = myImpl.getNodeDescriptor();
                    // The owner of the new agent is the owner of the local container.
                    // The new agent has NO initial credentials
          myImpl.initAgent(agentID, anAgent, nd.getOwnerPrincipal(), null);
      }
      catch(Exception e) {
          throw new StaleProxyException(e);
      }
      return new AgentController(agentID, myProxy, myImpl);
    }
    // HP Patch end ------------------------------------------------------------------------------------


  /**
     Shuts down this container, terminating all the agents running within it.
   */
  public void kill() throws StaleProxyException {
    if (myImpl == null || myProxy == null) {
      throw new StaleProxyException();
    }

    try {
        myProxy.killContainer();
        // release resources of this object
        myProxy = null;
        myImpl = null;
        myPlatformName = null;
    }
      catch (Throwable t) {
          throw new StaleProxyException(t);
      }
  }



  /**
     Installs a new message transport protocol, that will run within
     this container.

     @param address The transport address exported by the new MTP, in
     string format.
     @param className The fully qualified name of the Java class that
     implements the transport protocol.
     @exception MTPException If something goes wrong during transport
     protocol activation.
   */
  public void installMTP(String address, String className) throws MTPException, StaleProxyException {
    if (myImpl == null || myProxy == null) {
      throw new StaleProxyException();
    }

    try {
        myProxy.installMTP(address, className);
    }
      catch (Throwable t) {
          throw new StaleProxyException(t);
      }
  }

  /**
     Removes a message transport protocol, previously running within this
     container.

     @param address The transport address exported by the new MTP, in
     string format.
     @exception MTPException If something goes wrong during transport
     protocol activation.
     @exception NotFoundException If no protocol with the given
     address is currently installed on this container.
   */
  public void uninstallMTP(String address) throws MTPException, NotFoundException, StaleProxyException {
    if (myImpl == null || myProxy == null) {
      throw new StaleProxyException();
    }

    try {
        myProxy.uninstallMTP(address);
    }
      catch (Throwable t) {
          throw new StaleProxyException(t);
      }
  }

  /**
   * Retrieve the name of the platform the container wrapped by this
   * ContainerController belongs to.
   * @return the name (i.e. the HAP) of this platform.
   **/
  public String getPlatformName() {
    return myPlatformName;
  }

  /**
   * Retrieve the name of the wrapped container.
   * @return the name of this platform container.
   **/
  public String getContainerName() throws ControllerException {
    if(myImpl == null) {
      throw new ControllerException("Stale proxy.");
    }
    return myImpl.here().getName();
  }

  /**
     Retrieve a controller for the platform the container 
     wrapped by this ContainerController belongs to  and acts 
     as the Main Container.
     @return a <code>PlatfromController</code> for the platform
     the container wrapped by this <code>ContainerController</code>
     belongs to.
     @exception ControllerException If the container wrapped by this
     <code>ContainerController</code> is not valid or is not the 
     platform Main Container.
   */
  public PlatformController getPlatformController() throws ControllerException {
      initPlatformController();
      return myPlatformController;
  }

  protected void initPlatformController() throws ControllerException {
      if (myPlatformController == null) {
          if (myImpl == null) {
          throw new ControllerException("Stale proxy.");
          }
          MainContainer main = myImpl.getMain();
          if (main == null) {
          throw new ControllerException("Not a Main Container.");
          }
          if (main instanceof AgentManager) {
              myPlatformController = new PlatformControllerImpl(this, (AgentManager) main);
          }
          else {
          throw new ControllerException("Platform not accessible.");
          }
      }
  }
}
