/*---------------------------------------------------------------------------*
*  Copyright (c) 2006        Southeastern Universities Research Association, *
*                            Thomas Jefferson National Accelerator Facility  *
*                                                                            *
*    This software was developed under a United States Government license    *
*    described in the NOTICE file included as part of this distribution.     *
*                                                                            *
*    Author:  Vardan Gyurjyan                                                *
*             gurjyan@jlab.org                  Jefferson Lab, MS-6A         *
*             Phone: (757) 269-5879             12000 Jefferson Ave.         *
*             Fax:   (757) 269-5518             Newport News, VA 23606       *
*                                                                            *
*----------------------------------------------------------------------------*/

package rca.wrapper;

import rca.core.AgentManager;
import rca.core.ContainerID;
import rca.security.JADESecurityException;

import java.util.Hashtable;

/**
   Original author Giovanni Caire - TILAB
 */
class PlatformControllerImpl implements PlatformController {
	private ContainerController myContainer;
	private AgentManager myMain;
	private Hashtable listeners = new Hashtable();
  private State platformState = PlatformState.PLATFORM_STATE_VOID;
	
	PlatformControllerImpl(ContainerController c, AgentManager m) {
		myContainer = c;
		myMain = m;
    platformState = PlatformState.PLATFORM_STATE_READY;
    myMain.addListener(new AgentManagerListenerAdapter() {
		  public void removedContainer(rca.core.event.PlatformEvent ev) {
		  	ContainerID cid = ev.getContainer();
		  	try {
			  	if (cid.getName().equals(myContainer.getContainerName())) {
			  		// The local main container is terminating -->
			  		// The whole platform is terminating
	    			platformState = PlatformState.PLATFORM_STATE_KILLED;
			  	}
		  	}
		  	catch (ControllerException ce) {
		  		// Should never happen
		  		ce.printStackTrace();
		  	}
		  }
    } );    	
	}
	
  /**
   * Get the name of the platform.
   * @return String The platform name.
   */
  public String getName() {
  	return myContainer.getPlatformName();
  }
  
  /**
   * Start the platform after its been initialized.
   * @throws ControllerException If any probelms other than illegal state occur.
   * @throws IllegalStateException If state is illegal for this activity.
   */
  public void start() throws ControllerException {
  	// Just do nothing
  }

  /**
   * Suspend the agent platform. Next action may be resume or kill.
   * @throws ControllerException If any probelms other than illegal state occur.
   * @throws IllegalStateException If state is illegal for this activity.
   */
  public void suspend() throws ControllerException {
		throw new ControllerException("Not_Yet_Implemented");
  }

  /**
   * Activate the agent platform. Next action may be suspend or kill.
   * @throws ControllerException If any probelms other than illegal state occur.
   * @throws IllegalStateException If state is illegal for this activity.
   */
  public void resume() throws ControllerException {
		throw new ControllerException("Not_Yet_Implemented");
  }

  /**
   * Kill the agent platform. Kills all agents.
   * @throws ControllerException If any probelms other than illegal state occur.
   * @throws IllegalStateException If state is illegal for this activity.
   */
  public void kill() throws ControllerException {
  	try {
  		platformState = PlatformState.PLATFORM_STATE_KILLING;
  		// FIXME: Where do we get the principal and credentials
    	myMain.shutdownPlatform(null, null);
  	}
  	catch (JADESecurityException jse) {
  		throw new ControllerException("Security error. "+jse.getMessage());
  	}
  }

  /**
   * Get agent proxy to local agent given its name.
   * @param localAgentName The short local name of the desired agent.
   * @throws ControllerException If any probelms occur obtaining this proxy.
   */
  public AgentController getAgent(String localAgentName) throws ControllerException {
  	return myContainer.getAgent(localAgentName);
  }

  /**
   * Create a new agent.
   * @param nickName The name of the agent.
   * @param className The class implementing the agent.
   * @param args The agents parameters - typically String[] from a configuration file.
   * @return AgentController to enable control of the agent.
   * @throws IllegalStateException If state is illegal for this activity.
   * @throws ControllerException If any else goes wrong. All other exceptions are caught
   * and rethrown as a ControllerException.
   */
  public AgentController createNewAgent(String nickname, String className,
                                     Object[] args) throws ControllerException {
  	return myContainer.createNewAgent(nickname, className, args);
  }

  /**
   * Returns an instance of PlatformState.
   */
  public State getState() {
  	return platformState;
  }

  /**
   * Add a platform listener.
   * @param aListener The listener to be notified.
   */
  public synchronized void addPlatformListener(PlatformController.Listener aListener) throws ControllerException {
		ListenerWrapper wrapper = new ListenerWrapper(aListener);
  	myMain.addListener(wrapper);
		listeners.put(aListener, wrapper);
  }

  /**
   * Remove a platform listener.
   * @param aListener The listener to be notified.
   */
  public synchronized void removePlatformListener(PlatformController.Listener aListener) throws ControllerException {
  	ListenerWrapper wrapper = (ListenerWrapper) listeners.get(aListener);
  	if (wrapper != null) {
  		myMain.removeListener(wrapper);
  	}
  }

  /**
     Inner class AgentManagerListenerAdapter
     This utility class provides a dummy implementation of
     all the methods of the rca.core.AgentManager.Listener interface
   */
  class AgentManagerListenerAdapter implements AgentManager.Listener {
  	
	  public void addedContainer(rca.core.event.PlatformEvent ev) {
	  }
	  public void removedContainer(rca.core.event.PlatformEvent ev) {
	  }
	  public void bornAgent(rca.core.event.PlatformEvent ev) {
	  }
	  public void deadAgent(rca.core.event.PlatformEvent ev) {
	  }
	  public void movedAgent(rca.core.event.PlatformEvent ev) {
	  }
	  public void suspendedAgent(rca.core.event.PlatformEvent ev) {
	  }
	  public void resumedAgent(rca.core.event.PlatformEvent ev) {
	  }
	  public void frozenAgent(rca.core.event.PlatformEvent ev) {
	  }
	  public void thawedAgent(rca.core.event.PlatformEvent ev) {
	  }

	  public void addedMTP(rca.core.event.MTPEvent ev) {
	  }
	  public void removedMTP(rca.core.event.MTPEvent ev) {
	  }
	  public void messageIn(rca.core.event.MTPEvent ev) {
	  }
	  public void messageOut(rca.core.event.MTPEvent ev) {
	  }
  } // END of inner class AgentManagerListenerAdapter	  
  
  /**
     Inner class ListenerWrapper
     This utility class wraps a rca.wrapper.PlatformController.Listener
     into a rca.core.AgentManager.Listener
   */
  class ListenerWrapper extends AgentManagerListenerAdapter {
  	private PlatformController.Listener myListener;
  	
  	ListenerWrapper(PlatformController.Listener l) {
  		myListener = l;
  	}
  	
	  public void removedContainer(rca.core.event.PlatformEvent ev) {
	  	ContainerID cid = ev.getContainer();
	  	try {
		  	if (cid.getName().equals(myContainer.getContainerName())) {
		  		// The local main container is terminating -->
		  		// The whole platform is terminating
		  		myListener.killedPlatform(new rca.wrapper.PlatformEvent() {
				    public String getAgentGUID() {
				    	return null;
				    }
	
				    public String getPlatformName() {
				    	return myContainer.getPlatformName();
				    }
	
				    public int getEventType() {
				    	return PlatformEvent.KILLED_PLATFORM;
				    }
		  		} );
		  	}
	  	}
	  	catch (ControllerException ce) {
	  		// Should never happen 
	  		ce.printStackTrace();
	  	}
	  }
	  public void bornAgent(rca.core.event.PlatformEvent ev) {
	  	myListener.bornAgent(ev);
	  }
	  public void deadAgent(rca.core.event.PlatformEvent ev) {
	  	myListener.deadAgent(ev);
	  }
  } // END of inner class ListenerWrapper
}

